/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996-2000 Emmanuel CURIS
   Création : le 12 avril 2000

   Import de coordonnées cristallographiques sauvées en ASCII

   Essai de filtre d'import assez générique, pour faciliter un peu
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h> /* Pour isspace */

#include "systeme.h"

#include "atomes.h"
#include "cristallo.h"
#include "fichiers.h"
#include "science.h" /* Pour M_PI */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* #include "atm_litascii.txt" */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int CRIS_Import_Coordonnees(FILE *fichier, int format_choisi, struct infos_lecture *infos, const char *nom_fichier);
static int CRIS_Import_Initialiser(FILE *fichier, int format_choisi, struct infos_lecture *infos, const char *nom_fichier);
static int CRIS_Import_Terminer   (FILE *fichier, int format_choisi, struct infos_lecture *infos, const char *nom_fichier);

static int MOL_Import_Init_XYZ(FILE *fichier, struct infos_lecture *infos);

static double MOL_Import_AnalyseCoordonnee(char *chaine, double *erreur);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Chargement d'un fichier ASCII générique "simple" ~-~-~-~-~-~-~-~-~ */

/* -------- Ouverture et tout ça -------- */

int MOL_Import_XYZ(const char *nom_fichier, unsigned int format, struct infos_lecture *infos)
{
  FILE *fichier;
  int retour;

  /* 1. Ouverture du fichier */
  fichier = fopen( nom_fichier, "rb");
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir le fichier \"%s\" --- on passe au suivant\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier);
    (void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, NULL);
    return -1;
  }

  /* 2. Préparation en fonction du format */
  retour = CRIS_Import_Initialiser( fichier, format, infos, nom_fichier );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR à l'initialisation du fichier [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    (void) fclose( fichier );
    return -2;
  }

  /* 3. Ici, on est sur la partie intéressante : une ligne par atome à placer */
  retour = CRIS_Import_Coordonnees( fichier, format, infos, nom_fichier );
  /* On n'a plus besoin de la table des mailles */
  free( infos->maille ); infos->maille = NULL;

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans les coordonnées du fichier [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    (void) fclose( fichier );
    return -3;
  }

  /* 4. On charge la fin du fichier, si nécessaire */
  retour = CRIS_Import_Terminer( fichier, format, infos, nom_fichier );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR à la terminaison du fichier [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* 5. Chargement terminé sans erreur */
  (void) fclose( fichier );

  /* Chargement terminé, tout s'est bien passé */
  return 0;
}

/* -------- Initialiser la lecture, en fonction du format -------- */

static int CRIS_Import_Initialiser(FILE *fichier, int format_choisi, struct infos_lecture *infos, const char *nom_fichier)
{
  char *nom_molecule;
  int retour = -24;
  struct molecule *molecule = NULL;

  /* On se place au début de la partie intéressante */
  switch ( format_choisi )
  {
  case FORMAT3D_ASCII_XYZ:
    retour = MOL_Import_Init_XYZ( fichier, infos );
    break;

  case FORMAT3D_ASCII_CARINE:
    retour = MOL_Import_Init_Carine( fichier, infos );
    break;

  case FORMAT3D_ASCII_PDB:
  case FORMAT3D_ASCII_CIF:
    AVERTISSEMENT( "Format PDB/CIF ne devrait pas arriver ici [fichier %s]", nom_fichier );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR --- Format inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, format_choisi);
    return -1;
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le chargement du fichier %s [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_fichier, retour);
    return -5;
  }

  /* On crée la molécule qui contiendra le résultat */
  if ( NULL == infos->nom ) {
    nom_molecule = TXT_genere_chaine( "%F", nom_fichier );
  } else {
    nom_molecule = TXT_genere_chaine( "%s [%F]", infos->nom, nom_fichier );
  }
  molecule = MOL_creer( nom_molecule );
  if ( NULL == molecule ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la molécule\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -10;
  }

  /* La molécule sera utilisée pour le chargement */
  infos->molecule = molecule;

  if ( FAUX == ( infos->drapeaux & TROISD_COORD_FRAC ) ) {
    /* Coordonnées orthonormées : rien de spécial à faire en plus... */
    return 0;
  }

  /* Si la maille a été lue dans le fichier... */
  if ( infos->ligne_maille > 0 ) {
    OPTION_A_FAIRE( "Trouver le système cristallin !" );

    retour = CRIS_AdapteParametres( &(infos->systeme), &(infos->maille->a), &(infos->maille->b), &(infos->maille->c),
				    &(infos->maille->alpha), &(infos->maille->beta), &(infos->maille->gamma) );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Paramètres de maille incorrects.\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }

    /* On convertit les angles en radians */
    infos->maille->alpha *= M_PI / 180.0;
    infos->maille->beta  *= M_PI / 180.0;
    infos->maille->gamma *= M_PI / 180.0;
  }

  /* On crée une copie de la maille dans la molécule */
  molecule->maille = CRIS_CopierMaille( infos->maille );

  /* On prépare la liste des fragments pour les mailles, si demandé */
  if ( infos->drapeaux & TROISD_FRG_MAILLE ) {
    if ( infos->drapeaux & TROISD_MULTIMAILLES ) {
      infos->mailles = FRG_mailles( molecule, infos->i_min, infos->i_max, infos->j_min, infos->j_max, infos->k_min, infos->k_max);
    } else {
      infos->mailles = FRG_mailles( molecule, 0, 0, 0, 0, 0, 0 );
    }
    if ( NULL == infos->mailles ) {
      OPTION_A_FAIRE( "Demander si l'on doit continuer ou abandonner" );
      /* Pour l'instant, on continue, il n'y aura juste pas de fragment... */
    }
  }

  /* Chargement réussi */
  return 0;
}

/* -------- Init. d'un fichier XYZ quelconque : on saute les lignes de commentaire,... -------- */

static int MOL_Import_Init_XYZ(FILE *fichier, struct infos_lecture *infos)
{
  int c;
  unsigned long lgn_en_tete = 0, num_ligne = 0;
  boolean fin_en_tete = FAUX;

  c = fgetc( fichier );
  ungetc( c, fichier );
  do {
    /* On saute les lignes de commentaire */
    while ( c == infos->commentaire ) {
      c = ignore_ligne( fichier );
    }
    if ( EOF == c ) {
      (void) fprintf(stderr, "%s,%s l.%u : Fin de fichier atteinte\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }

    /* Ligne normale */
    if ( lgn_en_tete < infos->lg_en_tete ) {
      c = ignore_ligne( fichier ); /* On n'a pas encore passé l'en-tête */
      lgn_en_tete++;
    } else {
      if ( FAUX == ( infos->drapeaux & TROISD_COORD_FRAC ) ) {
	/* Coordonnées orthonormées : pas d'information de groupe d'espace ni de maille */
	return 0;
      }

      num_ligne++; /* Une ligne normale de plus */
      if ( ( num_ligne > infos->ligne_groupe ) && ( num_ligne > infos->ligne_maille ) ) {
	/* On a dépassé les lignes intéressantes */
	return 0;
      }

      if ( ( num_ligne == infos->ligne_groupe ) || ( num_ligne == infos->ligne_maille ) ) {
	if ( num_ligne == infos->ligne_groupe ) {
	  OPTION_A_FAIRE( "Analyser la ligne pour repérer le groupe d'espace" );
	}
	if ( num_ligne == infos->ligne_maille ) {
	  unsigned int nbr_colonnes;
	  int retour;
	  double *valeurs, *erreurs;

	  nbr_colonnes = compte_colonnes( fichier, infos->separateur );
	  if ( 0 == nbr_colonnes ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Ligne des paramètres de maille vide !\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    return -10;
	  }
	  valeurs = (double *) malloc( sizeof( double ) * nbr_colonnes );
	  if ( NULL == valeurs ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table des valeurs\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    return -11;
	  }
	  erreurs = (double *) malloc( sizeof( double ) * nbr_colonnes );
	  if ( NULL == valeurs ) {
	    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table des erreurs\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    free( valeurs ); valeurs = NULL;
	    return -12; /* On pourrait continuer, mais il est probable que la suite bloque aussi ! */
	  }

	  /* On lit les valeurs en chaque colonne */
	  retour = lit_valeurs( fichier, nbr_colonnes, infos->separateur, valeurs, erreurs );
	  if ( retour != 0 ) {
	    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la lecture des paramètres de maille [code %d]\n",
			   __FILE__, __FUNCTION__, __LINE__, retour);
	    free( erreurs ); erreurs = NULL;
	    free( valeurs ); valeurs = NULL;
	    return -13;
	  }

	  /* On place dans la maille */
	  if ( ( infos->col_a > 0 ) && ( infos->col_a <= nbr_colonnes ) ) {
	    infos->maille->a = valeurs[ infos->col_a - 1 ];
	    infos->maille->s_a = erreurs[ infos->col_a - 1 ];
	  }
	  if ( ( infos->col_b > 0 ) && ( infos->col_b <= nbr_colonnes ) ) {
	    infos->maille->b = valeurs[ infos->col_b - 1 ];
	    infos->maille->s_b = erreurs[ infos->col_b - 1 ];
	  }
	  if ( ( infos->col_c > 0 ) && ( infos->col_c <= nbr_colonnes ) ) {
	    infos->maille->c = valeurs[ infos->col_c - 1 ];
	    infos->maille->s_c = erreurs[ infos->col_c - 1 ];
	  }
	  if ( ( infos->col_alpha > 0 ) && ( infos->col_alpha <= nbr_colonnes ) ) {
	    infos->maille->alpha = valeurs[ infos->col_alpha - 1 ];
	    infos->maille->s_alpha = erreurs[ infos->col_alpha - 1 ];
	  }
	  if ( ( infos->col_beta > 0 ) && ( infos->col_beta <= nbr_colonnes ) ) {
	    infos->maille->beta = valeurs[ infos->col_beta - 1 ];
	    infos->maille->s_beta = erreurs[ infos->col_beta - 1 ];
	  }
	  if ( ( infos->col_gamma > 0 ) && ( infos->col_gamma <= nbr_colonnes ) ) {
	    infos->maille->gamma = valeurs[ infos->col_gamma - 1 ];
	    infos->maille->s_gamma = erreurs[ infos->col_gamma - 1 ];
	  }

	  /* Plus besoin des tables */
	  free( erreurs ); erreurs = NULL;
	  free( valeurs ); valeurs = NULL;

#ifdef _DEVELOPPE_
	  printf(" --- paramètres de maille lus dans le fichier --- \n");
	  printf("a = %f\tb = %f\tc = %f\talpha = %f\tbeta = %f\tgamma = %f\n",
		 infos->maille->a, infos->maille->b, infos->maille->c, infos->maille->alpha, infos->maille->beta, infos->maille->gamma);
#endif
	}
      } else {
	c = ignore_ligne( fichier ); /* Ligne ne correspondant ni au groupe, ni à la maille */
      }
    }
  } while ( FAUX == fin_en_tete );

  /* Si on est là, c'est que l'en-tête a été passé correctement */
  return 0;
}

/* -------- Lecture des coordonnées (un atome par ligne) -------- */

static int CRIS_Import_Coordonnees(FILE *fichier, int format_choisi __attribute__((__unused__)), struct infos_lecture *infos, const char *nom_fichier)
{
  int c;
  int retour;
  unsigned long num_ligne = 0;

  do {
    /* On lit le premier caractère sans le perdre... */
    c = fgetc( fichier );
    ungetc( c, fichier );

    if ( infos->commentaire == c ) {
      /* Ligne de commentaire au milieu : prévenir ? */
      c = ignore_ligne( fichier );
      continue;
    }
    if ( EOF == c ) break;

    num_ligne++; /* Une ligne utile de plus */
    retour = CRIS_LireLigneAscii( fichier, infos );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR en lisant la ligne %lu [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, num_ligne, nom_fichier);
      return -21;
    }
  } while ( c != EOF );

  /* Le chargement a réussi */
  return 0;
}

/* -------- Appel pour une ligne ASCII de coordonnées standard -------- */

int CRIS_LireLigneAscii(FILE *fichier, struct infos_lecture *infos)
{
  unsigned long i, nbr_colonnes, nombre_mailles = 0;
  Vecteur3D position = { 0.0, 0.0, 0.0 }, erreur = { 0.0, 0.0, 0.0 };
  double rayon = -1.0, charge = 0.0;
  int Z = HYDROGENE;
  char *ligne, **colonnes, *etiquette __attribute__((__unused__)) = NULL;

  ligne = lit_ligne( fichier, 0 );
  if ( NULL == ligne ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR, impossible de lire la ligne\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On la découpe en utilisant le séparateur */
  colonnes = TXT_DecomposerChaine( ligne, infos->separateur, &nbr_colonnes );
  if ( NULL == colonnes ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR, impossible de décomposer en colonnes [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, ligne);
    free( ligne ); ligne = NULL;
    return -2;
  }

  /* Pour chacun des éléments découpés */
  for ( i = 0 ; i < nbr_colonnes ; i++ ) {
    if ( infos->col_Z == ( i + 1 ) ) { /* Colonne des symboles */
      Z = ATM_num_atomique( colonnes[ i ] );
      if ( Z <= 0 ) { /* Symbole inconnu */
	Z = TXT_CNV_vNombre_L( colonnes[ i ], NULL ); /* Si codé numériquement */
      }
      if ( Z <= 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Atome inconnu [%s]\n",
		       __FILE__, __FUNCTION__, __LINE__, colonnes[ i ]);
      }
    } else if ( (i + 1) == infos->col_x ) {
      position.x = MOL_Import_AnalyseCoordonnee( colonnes[ i ], &(erreur.x) );
    } else if ( (i + 1) == infos->col_y ) {
      position.y = MOL_Import_AnalyseCoordonnee( colonnes[ i ], &(erreur.y) );
    } else if ( (i + 1) == infos->col_z ) {
      position.z = MOL_Import_AnalyseCoordonnee( colonnes[ i ], &(erreur.z) );
    } else if ( (i + 1) == infos->col_rayon ) {
      rayon = TXT_CNV_vNombre_D( colonnes[ i ], NULL, NULL );
    } else if ( (i + 1) == infos->col_charge ) {
      charge = TXT_CNV_vNombre_D( colonnes[ i ], NULL, NULL );
    } else if ( (i + 1) == infos->col_txt ) {
      etiquette = colonnes[ i ];
    }
  }

  /* Plus besoin des colonnes, mais encore de la ligne [étiquette éventuelle] */
  free( colonnes ); colonnes = NULL;

  /* Coordonnées orthonormées */
  if ( FAUX == ( infos->drapeaux & TROISD_COORD_FRAC ) ) {
    Atome *atome;

    atome = ATM_creer( infos->molecule, NULL, Z, &position, charge, rayon );
    if ( NULL == atome ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer %s en (%g, %g, %g)\n",
		     __FILE__, __FUNCTION__, __LINE__, ATM_Z_vers_symbole( Z ), position.x, position.y, position.z);
      /* Plus besoin de la ligne */
      free( ligne ); ligne = NULL;

      return +1; /* Erreur, mais mineure */
    }

    /* A-t-on besoin des erreurs ? */
    if ( infos->drapeaux & TROISD_LIRE_ERREURS ) {
      OPTION_A_FAIRE( "Assigner les erreurs à l'atome choisi" );
    }

    /* Terminé ! */
    return 0;
  }

  /* Si on est là : on veut des coordonnées fractionnaires */
  if ( infos->drapeaux & TROISD_MULTIMAILLES ) {
    /* Multi-mailles : pour éviter d'avoir des atomes en double, on s'oblige à être dans la maille (0,0,0) */
    CRIS_NormaliserCoordonneesFractionnaires( &position );
  }

  /* Combien de mailles ? */
  nombre_mailles  = (infos->i_max - infos->i_min) + 1;
  nombre_mailles *= (infos->j_max - infos->j_min) + 1;
  nombre_mailles *= (infos->k_max - infos->k_min) + 1;

  OPTION_A_FAIRE( "Mémoriser les opérations dans la maille de la molécule" );
  CRIS_GenereCristal( infos->molecule, &position, Z, 0.0, rayon, nombre_mailles, infos->mailles );

  /* Plus besoin de la ligne */
  free( ligne ); ligne = NULL;

  /* Chargement réussi */
  return 0;
}

/* -------- Analyse d'un champ contenant une coordonnée -------- */

static double MOL_Import_AnalyseCoordonnee(char *chaine, double *erreur)
{
  char *pos_fraction;
  double num = 0.0, den = 0.0;

  pos_fraction = strchr( chaine, '/' );
  if ( NULL == pos_fraction ) { /* Nombre classique */
    return TXT_CNV_vNombre_D( chaine, NULL, erreur );
  }

  /* On est de la forme a/b : pas d'erreur et on calcule la fraction */
  *erreur = 0.0;
  *pos_fraction = 0; /* Pour ne pas être gêné dans la lecture */
  num = TXT_CNV_vNombre_D( chaine, NULL, NULL ); /* Normalement, un entier suffirait, mais... */
  den = TXT_CNV_vNombre_D( pos_fraction + 1, NULL, NULL );
  if ( 0.0 == den ) {
    (void) fprintf(stderr, "%s,%s l.%u : FRACTION incorrecte, %g/%g [%s/%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, num, den, chaine, pos_fraction + 1);
    return 0.0;
  }

  return num / den;
}

/* -------- Travail à faire une fois que tous les atomes ont été chargés -------- */

static int CRIS_Import_Terminer(FILE *fichier __attribute__((__unused__)), int format_choisi __attribute__((__unused__)), struct infos_lecture *infos, const char *nom_fichier __attribute__((__unused__)))
{
  struct molecule *molecule;

  molecule = infos->molecule;

  OPTION_A_FAIRE( "Proposer de construire une table des liaisons" );

  MOL_cherche_barycentre( molecule, FAUX );
  MOL_initialise_vue( molecule );
  MOL_aller( molecule ); /* On force la reconstruction de la liste... */

  return 0;
}
