/* LASE - Logiciel d'Analyse des spectres EXAFS
   (C) 1997 - 2003, Emmanuel CURIS
   Cration : novembre 2003

   Gestion d'une maille cristalline
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include "systeme.h"

#include "cristallo.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Mmorisation et utilisation des informations cristallographiques ~-~-~-~-~-~-~-~-~ */

/* -------- Cration de la structure dcrivant une maille -------- */

struct maille *CRIS_CreerMaille(void)
{
  struct maille *maille;

  /* On rserve */
  maille = (struct maille *) malloc( sizeof( struct maille ) );
  if ( NULL == maille ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de rserver les informations de maille\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On initialise */
  maille->a = 0.0; maille->b = 0.0; maille->c = 0.0;
  maille->alpha = 0.0; maille->beta = 0.0; maille->gamma = 0.0;

  maille->s_a = 0.0; maille->s_b = 0.0; maille->s_c = 0.0;
  maille->s_alpha = 0.0; maille->s_beta = 0.0; maille->s_gamma = 0.0;

  maille->type = 'T';
  maille->Z = 1;

  maille->nbr_operations = 0;
  maille->nbr_translations = 0;

  maille->operations = NULL;
  maille->translations = NULL;

  return maille;
}

/* -------- Copie d'une maille dj existante -------- */

struct maille *CRIS_CopierMaille (struct maille *maille)
{
  struct maille *copie;

  if ( NULL == maille ) return NULL;

  /* On cre une nouvelle maille */
  copie = CRIS_CreerMaille();
  if ( NULL == copie ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la copie\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Copie de l'ensemble des infos... */
  *copie = *maille;

  /* Traitement des oprations & translations */
  if ( maille->nbr_operations > 0 ) {
    /* Pour ne pas avoir de problme  l'effacement, il faut une nouvelle table ! */
    copie->nbr_operations = 0; copie->operations = NULL;

    OPTION_A_FAIRE( "Copier les oprations de symtrie" );
  }
  if ( maille->nbr_translations > 0 ) {
    /* Pour ne pas avoir de problme  l'effacement, il faut une nouvelle table ! */
    copie->nbr_translations = 0; copie->translations = NULL;

    OPTION_A_FAIRE( "Copier les translations" );
  }

  return copie;
}

/* -------- Effacer une maille et les infos associes -------- */

int CRIS_EffacerMaille(struct maille *maille)
{
  if ( NULL == maille ) return -1;

  /* On efface les oprations */
  free( maille->operations ); maille->operations = NULL;
  free( maille->translations ); maille->translations = NULL;

  /* On efface la maille */
  free( maille );

  /* Fini sans erreur */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Conversions repre absolu / repre local de la maille ~-~-~-~-~-~-~-~-~ */

int CRIS_AbsoluVersFrac(struct maille *maille, Vecteur3D *absolu, Vecteur3D *fractionnaire)
{
  double c_a, c_b, c_g, s_a, s_b, s_g;

#ifdef _DEVELOPPE_
  if ( NULL == maille ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de maille indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -42;
  }
  if ( ( NULL == absolu ) || ( NULL == fractionnaire ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de point de dpart ni d'arrive\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -58;
  }
#endif

  switch ( maille->type )
  {
  case 'C': /* Seul a a un sens ; ne devrait jamais tre nul logiquement ! */
    fractionnaire->x = absolu->x / maille->a;
    fractionnaire->y = absolu->y / maille->a;
    fractionnaire->z = absolu->z / maille->a;
    return 0;

  case 'H':/* a = b, alpha = beta = 90, gamma = 120 */
  case 'r': /* Rhombodrique, axes hexagonaux [P] */
    fractionnaire->z = absolu->z / maille->c;
    fractionnaire->y = ( 2.0 / sqrt( 3.0 ) ) * absolu->y / maille->a ;
    fractionnaire->x = ( absolu->x - absolu->y / sqrt( 3.0 ) ) / maille->a ;
    return 0;

  case 'M': /* Monoclinique (axe b) : alpha = gamma = 90 */
    c_b = cos( maille->beta  ); s_b = sin( maille->beta );

    fractionnaire->z =  absolu->z / ( maille->c * s_b );
    fractionnaire->y =  absolu->y / maille->b;
    fractionnaire->x = ( absolu->x - c_b * absolu->z / s_b ) / maille->a;
    return 0;

  case 'm': /* Monoclinique (axe c) : alpha = beta = 90 */
    c_g = cos( maille->gamma ); s_g = sin( maille->gamma );

    fractionnaire->z =   absolu->z / maille->c ;
    fractionnaire->y =   absolu->y / ( maille->b * s_g );
    fractionnaire->x = ( absolu->x - c_g * absolu->y / s_g ) / maille->a;
    return 0;

  case 'O': /* Orthorhombique alpha = beta = gamma = 90 */
    fractionnaire->z = absolu->z / maille->c;
    fractionnaire->y = absolu->y / maille->b;
    fractionnaire->x = absolu->x / maille->a;
    return 0;

  case 'Q': /* Quadratique : a = b, alpha = beta = gamma = 90 */
    fractionnaire->z = absolu->z / maille->c;
    fractionnaire->y = absolu->y / maille->a;
    fractionnaire->x = absolu->x / maille->a;
    return 0;

  case 'R': /* Rhombodrique (axes rhombodriques) : a = b = c, alpha = beta = gamma */
    c_a = cos( maille->alpha ); s_a = sin( maille->alpha );

    fractionnaire->z =   absolu->z * s_a / ( maille->a * sqrt( fabs( 1.0 - 3.0 * c_a*c_a + 2.0 * c_a * c_a * c_a ) ) );
    fractionnaire->y = ( absolu->y - fractionnaire->z * ( maille->a / s_a) * ( c_a - c_a * c_a )    ) / ( maille->a * s_a );
    fractionnaire->x = ( absolu->x - maille->a * c_a * fractionnaire->y - maille->a * fractionnaire->z * c_a ) /   maille->a;
    return 0;
  }

  /* Ici : type inconnu ou 'T' ==> triclinique */
#ifdef _DEVELOPPE_
  if ( maille->type != 'T' ) {
    AVERTISSEMENT( "Type de maille inconnu (%c = %d)", maille->type, maille->type );
  }
#endif

  c_a = cos( maille->alpha );
  c_b = cos( maille->beta  );
  c_g = cos( maille->gamma ); s_g = sin( maille->gamma );

  fractionnaire->z =   absolu->z * s_g / ( maille->c * sqrt( fabs( 1.0 - c_a*c_a - c_b*c_b - c_g*c_g + 2 * c_a * c_b * c_g ) ) );
  fractionnaire->y = ( absolu->y - fractionnaire->z * ( maille->c / s_g) * ( c_a - c_b * c_g )    ) / ( maille->b * s_g );
  fractionnaire->x = ( absolu->x - maille->b * c_g * fractionnaire->y - maille->c * fractionnaire->z * c_b ) /   maille->a;

  return 0;
}

/* -------- Coordonnes dans le repre absolu des trois vecteurs de maille -------- */

int CRIS_VecteursMaille(struct maille *maille, Vecteur3D *a, Vecteur3D *b, Vecteur3D *c)
{
  double c_a, c_b, c_g, s_a, s_b, s_g;

#ifdef _DEVELOPPE_
  if ( NULL == maille ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de maille indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -42;
  }
  if ( ( NULL == a ) || ( NULL == b ) || ( NULL == c ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de vecteurs  calculer\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -58;
  }
#endif

  /* Le vecteur a est align avec l'axe des X */
  a->x = maille->a; a->y = 0.0; a->z = 0.0;

  /* Le vecteur b est dans le plan XOY */
  b->z = 0.0;

  /* Les autres coordonnes dpendent de la maille */
  switch ( maille->type )
  {
  case 'C': /* Seul a a un sens ; ne devrait jamais tre nul logiquement ! */
    b->x = 0.0; b->y = maille->a;
    c->x = 0.0; c->y = 0.0; c->z = maille->a;
    return 0;

  case 'H':/* a = b, alpha = beta = 90, gamma = 120 */
  case 'r': /* Rhombodrique, axes hexagonaux [P] */
    b->x = -0.5 * maille->a; b->y = maille->b * sqrt( 3.0 ) / 2.0;
    c->x = 0.0; c->y = 0.0; c->z = maille->c;
    return 0;

  case 'M': /* Monoclinique (axe b) : alpha = gamma = 90 */
    c_b = cos( maille->beta  ); s_b = sin( maille->beta );

    b->x = 0.0; b->y = maille->b;
    c->x = maille->c * c_b; c->y = 0.0; c->z = maille->c * s_b;
    return 0;

  case 'm': /* Monoclinique (axe c) : alpha = beta = 90 */
    c_g = cos( maille->gamma ); s_g = sin( maille->gamma );

    b->x = c_g * maille->b; b->y = s_g * maille->b;
    c->x = 0.0; c->y = 0.0; c->z = maille->c;
    return 0;

  case 'O': /* Orthorhombique alpha = beta = gamma = 90 */
    b->x = 0.0; b->y = maille->b;
    c->x = 0.0; c->y = 0.0; c->z = maille->c;
    return 0;

  case 'Q': /* Quadratique : a = b, alpha = beta = gamma = 90 */
    b->x = 0.0; b->y = maille->a;
    c->x = 0.0; c->y = 0.0; c->z = maille->c;
    return 0;

  case 'R': /* Rhombodrique (axes rhombodriques) : a = b = c, alpha = beta = gamma */
    c_a = cos( maille->alpha ); s_a = sin( maille->alpha );

    /* Vecteur b */
    b->x = c_a * maille->b; b->y = s_a * maille->b;

    /* Vecteur c */
    c->x = maille->c * c_a;
    c->y = maille->c * ( c_a - c_a * c_a ) / s_a;
    c->z = maille->c * sqrt( fabs(1 - 3.0 * c_a*c_a + 2 * c_a * c_a * c_a) ) / s_a;
    return 0;
  }

  /* Ici : type inconnu ou 'T' ==> triclinique */
#ifdef _DEVELOPPE_
  if ( maille->type != 'T' ) {
    AVERTISSEMENT( "Type de maille inconnu (%c = %d)", maille->type, maille->type );
  }
#endif

  c_a = cos( maille->alpha );
  c_b = cos( maille->beta  );
  c_g = cos( maille->gamma ); s_g = sin( maille->gamma );

  /* Vecteur b */
  b->x = c_g * maille->b;
  b->y = s_g * maille->b;

  /* Vecteur c */
  c->x = maille->c * c_b;
  c->y = maille->c * ( c_a - c_b * c_g ) / s_g;
  c->z = maille->c * sqrt( fabs(1 - c_a*c_a - c_b*c_b - c_g*c_g + 2 * c_a * c_b * c_g) ) / s_g;

  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Routines sur l'interface ~-~-~-~-~-~-~-~-~ */

int CRIS_DemanderIndicesMaille(BI_Boite_Demande *boite, int *i, int *j, int *k)
{
  struct infos_demande indices[ 3 ];
  const char *textes[ 3 ] = { "i =", "j =", "k =" };
  int reponse, idx;

  /* Contrle prliminaire */
  if ( ( NULL == boite ) || ( NULL == i ) || ( NULL == j ) || ( NULL == k ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Paramtres absurdes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Initialisation de la bote de demande des index de maille */
  indices[ 0 ].valeur_initiale.entier_S = *i;
  indices[ 1 ].valeur_initiale.entier_S = *j;
  indices[ 2 ].valeur_initiale.entier_S = *k;
  for ( idx = 0 ; idx < 3 ; idx++ ) {
    indices[ idx ].type = BI_CHAMP_DEMANDE_ENTIER_S;

    indices[ idx ].texte = textes[ idx ];
    indices[ idx ].unite = NULL; /* Sans unit */

    indices[ idx ].valeur_defaut.entier_S = 0;
    indices[ idx ].valeur_donnee = indices[ idx ].valeur_initiale;
  }

  /* On demande les valeurs */
  reponse = BI_DemandeMultiple( NULL, boite, 3, indices );
  if ( reponse < 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la demande des index\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }
  if ( reponse != (int) boite->bouton_annuler ) {
    *i = (int) indices[ 0 ].valeur_donnee.entier_S;
    *j = (int) indices[ 1 ].valeur_donnee.entier_S;
    *k = (int) indices[ 2 ].valeur_donnee.entier_S;
  }

  /* Termin... */
  return reponse;
}

