/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2002, Emmanuel Curis

   Interface graphique : routines générales indépendantes du système

   Travail sur le contenu des objets
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h> /* Pour isinf() ou équivalent */

#ifdef _UNIX_
# define _LIBRAIRIE_X_
#elif defined(_WINDOWS_)
# define _LIBRAIRIE_W_
#endif

#include "txt_utils.h"
#include "bib_interface.h"

#include "objets.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int OBJ_EstUnFils(OBJECT *arbre, unsigned int pere, unsigned int objet);

static int Tedinfo_MettreValeur  (TEDINFO *tedinfo, double valeur);
static int Tedinfo_PartieDecimale(TEDINFO *tedinfo, double p_decimale, unsigned int nbr_chiffres);
static int Tedinfo_PartieEntiere (TEDINFO *tedinfo, double p_entiere, unsigned int nbr_chiffres);

static int TXTInfini_MettreTexte (Texte_Infini *txt_infini, const char *texte);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Contrôle que les paramètres passés sont corrects -------- */

int OBJ_ControlerValeurs(unsigned int arbre, unsigned int objet)
{
  if ( arbre >= BI_nbr_arbres ) {
    (void) fprintf(stderr, "OBJ_ControlerValeurs : numéro d'arbre absurde [%u ; max. %u]\n", arbre, BI_nbr_arbres - 1);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }
  if ( NULL == arbres[ arbre ].adresse ) {
    (void) fprintf(stderr, "OBJ_ControlerValeurs : arbre non défini [%u]\n", arbre);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }
  if ( objet >= arbres[ arbre ].nbr_objets ) {
    (void) fprintf(stderr, "OBJ_ControlerValeurs : numéro d'objet absurde [%u ; max. %u]\n",
		   objet, arbres[ arbre ].nbr_objets - 1);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* Pas d'erreur détectée */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Arborescence ~-~-~-~-~-~-~-~- */

int OBJ_TrouverPere(unsigned int arbre, unsigned int objet)
{
  OBJECT *adr_arbre;
  int retour, pere = ROOT;

  /* Vérifications préliminaires */
  retour = OBJ_ControlerValeurs( arbre, objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "OBJ_TrouverPere : valeurs incorrectes... [arbre= %u, objet = %u]\n", arbre, objet);
    return -1;
  }
  if ( ROOT == objet ) {
    return -2;
  }

  adr_arbre = arbres[ arbre ].adresse;
  pere = OBJ_EstUnFils( adr_arbre, ROOT, objet );
  if ( pere < 0 ) {
    (void) fprintf(stderr, "OBJ_TrouverPere : impossible de trouver le père...");
  }

  /* On renvoie ce que l'on a pu trouver... */
  return pere;
}

/* Renvoie <0 si non trouvé, 0 si on est l'objet considéré, sinon numéro de l'objet père de niveau juste avant... */

static int OBJ_EstUnFils(OBJECT *adr_arbre, unsigned int pere, unsigned int objet)
{
  /* Il y a des fils : on les explore un par un... */
  if ( OB_HEAD( adr_arbre, pere ) > ROOT ) {
    int nv_pere, trouve;

    nv_pere = OB_HEAD( adr_arbre, pere );
    while ( nv_pere != (int) pere ) {
      if ( nv_pere == (int) objet ) { /* Ce fils est l'objet cherché */
	return pere;
      }
      trouve = OBJ_EstUnFils( adr_arbre, nv_pere, objet );
      if ( trouve > 0 ) return trouve; /* On a trouvé le fils ! */
      if ( 0 == trouve ) return nv_pere;

      nv_pere = OB_NEXT( adr_arbre, nv_pere );
    }

    /* Pas trouvé dans les fils */
    return -1;
  }

  /* Donc là, pas de fils : soit on est sur le bon, soit aucune chance... */
  if ( pere == objet )
    return 0;
  else
    return -1;
}

/* ~-~-~-~-~-~-~-~- Gestion du contenu ~-~-~-~-~-~-~-~- */

/* -------- Changer le type d'un objet -------- */

int OBJ_ChangerType(unsigned int n_arbre, unsigned int n_objet, int nv_type, ...)
{
  int type_avant;
  int retour;
  OBJECT *objet;

  /* Vérifications préliminaires */
  retour = OBJ_ControlerValeurs( n_arbre, n_objet );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : valeurs incorrectes [arbre = %u, objet = %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, n_arbre, n_objet);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }
  if ( ROOT == n_objet ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de changer le type de l'objet ROOT\n",
		    __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -2;
  }

  objet = arbres[ n_arbre ].adresse + n_objet;

  /* Type d'objet original */
  type_avant = objet->ob_type;
  if ( nv_type == type_avant ) {
    /* Même type : rien à faire ! */
    return 0;
  }

  switch ( nv_type )
  {
  case G_TXT_INFINI:
    if ( ( G_FTEXT == type_avant ) || ( G_FBOXTEXT == type_avant ) ||
	 ( G_TEXT  == type_avant ) || ( G_BOXTEXT  == type_avant ) ) {
      unsigned long lg_texte = 10; /* Arbitrairement, texte de 10 caractères en cas d'échec */
      char *texte = NULL;
      TEDINFO *tedinfo;
      Texte_Infini *txt_infini;

      tedinfo = objet->ob_spec.tedinfo;
      if ( tedinfo != NULL ) {
	lg_texte = tedinfo->te_txtlen;
	texte = tedinfo->te_ptext.editable;
	tedinfo->te_ptext.editable = NULL;
      }

      OBJ_LibererComplement( n_arbre, n_objet );
      objet->ob_type = nv_type;
      retour = OBJ_PreparerComplement( n_arbre, n_objet );
      if ( retour != 0 ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de créer le complément !\n",
			__FILE__, __FUNCTION__, __LINE__ );
	free( texte );
	return -2;
      }

      /* On réserve pour le texte */
      txt_infini = objet->ob_spec.txt_infini;
      txt_infini->texte = (char *) calloc( lg_texte + 1, sizeof( char ) );
      if ( NULL == txt_infini->texte ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de réserver le texte\n",
			__FILE__, __FUNCTION__, __LINE__ );
	return +5;
      }
      txt_infini->max_cars = lg_texte;

      /* On met le texte de la tedinfo */
      met_chaine( n_arbre, n_objet, 0, texte, TXT_COPIER );

      /* On n'a plus besoin du texte de la tedinfo */
      free( texte );
    } else {
      OBJ_LibererComplement( n_arbre, n_objet );
      objet->ob_type = nv_type;
      retour = OBJ_PreparerComplement( n_arbre, n_objet );
      if ( retour != 0 ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de créer le complément !\n",
			__FILE__, __FUNCTION__, __LINE__ );
	return -4;
      }
    }
    break;

  case G_NOMBRE: /* Pas forcément éditable, car peut servir à l'affichage uniquement ! */
    objet->ob_flags &= ~SELECTABLE; /* Mais il n'est pas sélectionable ! */

    {
      double valeur;

      /* Valeur numérique dans l'objet */
      valeur = OBJ_LireNombre( n_arbre, n_objet, NULL );

      /* On perd tout l'ancien [rien à conserver...] */
      OBJ_LibererComplement( n_arbre, n_objet );
      objet->ob_type = nv_type;
      retour = OBJ_PreparerComplement( n_arbre, n_objet );
      if ( retour != 0 ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de créer le complément !\n",
			__FILE__, __FUNCTION__, __LINE__ );
	return -4;
      }

      /* On remet la valeur numérique */
      OBJ_MettreNombre( n_arbre, n_objet, valeur, -1.0 );
    }
    break;

  case G_BUTTON:
    objet->ob_flags &= ~EDITABLE;  /* Un bouton n'est pas éditable */
    objet->ob_flags |= SELECTABLE; /* Mais il est sélectionable ! */
    if ( ( G_STRING == type_avant ) || ( G_SEPARATEUR == type_avant ) || ( G_TITLE == type_avant ) ) {
      /* Il suffit de changer le type, on garde le même texte */
      objet->ob_type = nv_type;
    } else {
      OBJ_LibererComplement( n_arbre, n_objet );
      objet->ob_type = nv_type;
      OBJ_PreparerComplement( n_arbre, n_objet );
    }
    break;

  case G_SEPARATEUR:
    objet->ob_flags &= ~(EDITABLE|SELECTABLE); /* Un séparateur n'est pas éditable ni sélectionnable */
    if ( ( G_STRING == type_avant ) || ( G_BUTTON == type_avant ) || ( G_TITLE == type_avant ) ) {
      /* Il suffit de changer le type, on garde le même texte */
      objet->ob_type = nv_type;
    } else {
      OBJ_LibererComplement( n_arbre, n_objet );
      objet->ob_type = nv_type;
      OBJ_PreparerComplement( n_arbre, n_objet );
    }
    break;

  case G_SSMENU:
    objet->ob_flags &= ~EDITABLE; /* Un sous-menu n'est pas éditable */
    objet->ob_flags |= SELECTABLE;

    OBJ_LibererComplement( n_arbre, n_objet );
    objet->ob_type = nv_type;
    OBJ_PreparerComplement( n_arbre, n_objet );
    break;

  case G_STYLELIGNE:
  case G_STYLEMARQUE:
    objet->ob_flags &= ~EDITABLE; /* Un sous-menu n'est pas éditable */    
    objet->ob_flags |= SELECTABLE;

    OBJ_LibererComplement( n_arbre, n_objet );
    objet->ob_type = nv_type;
    OBJ_PreparerComplement( n_arbre, n_objet );
    break;

  default:
    OPTION_A_FAIRE( "Conversion objet -> type %d" );
  }

  /* Doit-on redessiner ? */
  if ( VRAI == arbres[ n_arbre ].drapeaux.ouvert ) {
    BI_DessinerObjet( n_arbre, n_objet );
  }

  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Lire une valeur dans un objet ~-~-~-~-~-~-~-~-~ */

int OBJ_LireDate(unsigned int arbre, unsigned int objet, const char *format, struct tm *date)
{
  int retour;
  char *texte;

  /* Contrôles */
  retour = OBJ_ControlerValeurs( arbre, objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre ou objet incorrect [%d/arbre %u/objet %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, arbre, objet);
    return -1;
  }
  if ( NULL == date ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de mémoriser la date : date=NULL\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  /* Récupérer le texte de l'objet */
  texte = lit_chaine( arbre, objet, 0, NULL, TXT_COPIER );
  if ( NULL == texte ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'obtenir le texte de l'objet %u [arbre %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, objet, arbre);
    return -2;
  }

  retour = TXT_AnalyserDate( texte, format, date );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en décodant la date\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  return retour;
}

unsigned int OBJ_LireStyle(unsigned int arbre, unsigned int numero_objet)
{
  int retour;
  OBJECT *objet;

  /* Contrôles */
  retour = OBJ_ControlerValeurs( arbre, numero_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre ou objet incorrect [%d/arbre %u/objet %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, arbre, numero_objet);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_CONTINU;
  }

  objet = arbres[ arbre ].adresse + numero_objet;

  /* On renvoie la valeur... */
  switch ( objet->ob_type )
  {
  case G_STYLELIGNE:
    return objet->ob_spec.style_ligne;

  case G_STYLEMARQUE:
    return objet->ob_spec.style_marque;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Objet %u sans style associé, type incorrect !\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_objet);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_CONTINU;
  }
}

/* ~-~-~-~-~-~-~-~-~ Installer une valeur dans un objet ~-~-~-~-~-~-~-~-~ */

/* -------- Un caractère unique -------- */

int OBJ_MettreCaractere(unsigned int arbre, unsigned int num_objet, int caractere )
{
  int retour;
  char *chaine;
  OBJECT *objet;

  /* Contrôles */
  retour = OBJ_ControlerValeurs( arbre, num_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre ou objet incorrect [%d/arbre %u/objet %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, arbre, num_objet);
    return -1;
  }

  /* On récupère l'objet */
  objet = arbres[ arbre ].adresse + num_objet;

  switch ( objet->ob_type )
  {
  case G_TXT_INFINI:
    AVERTISSEMENT( "Mise en place d'un caractère unique dans un texte infini [arbre %u, objet %d]", arbre, num_objet );
    chaine = objet->ob_spec.txt_infini->texte;
    if ( NULL == chaine ) {
      chaine = (char *) calloc( 2, sizeof( char ) );
      if ( NULL == chaine ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver le texte\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -2;
      }
      objet->ob_spec.txt_infini->texte = chaine;
      objet->ob_spec.txt_infini->max_cars = 1;
    }
    *chaine = caractere;
    *(chaine + 1) = 0;
    break;

  case G_NOMBRE:
    AVERTISSEMENT( "Mise en place d'un caractère unique dans un champ numérique [arbre %u, objet %d]", arbre, num_objet );
    retour = OBJ_MettreNombre( arbre, num_objet, (double) caractere, 0.0 );
    break;

  case G_TEXT:
  case G_BOXTEXT: /* Texte, mais non éditable -> la valeur est dans le texte */
    (void) fprintf(stderr, "%s,%s l.%u : Tentative de modifier un caractère dans un texte constant\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return +10;

  case G_BUTTON:
  case G_STRING:
  case G_TITLE:
  case G_SEPARATEUR:
    (void) fprintf(stderr, "%s,%s l.%u : Tentative de modifier un caractère dans un texte constant\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return +11;

  case G_BOXCHAR:
    objet->ob_spec.gbox.caractere = caractere;
    break;

  case G_FTEXT:
  case G_FBOXTEXT:
    chaine = objet->ob_spec.tedinfo->te_ptext.editable;
    if ( NULL == chaine ) {
      AVERTISSEMENT( "Pas de texte éditable dans l'éditable [arbre %u, objet %u]", arbre, num_objet );
    }
    *chaine = caractere;
    *(chaine + 1) = 0;
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type d'objet non traité [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
    return -20;
  }

  /* On a mis le caractère : si l'arbre est ouvert */
  if ( VRAI == arbres[ arbre ].drapeaux.ouvert ) {
    if ( arbres[ arbre ].edit_courant == num_objet ) { /* On vient de modifier l'éditable courant... */
      /* On replace donc le curseur comme il faut */
      BI_PlacerCurseur( arbres + arbre, BI_InitialiserCurseur( objet ) );
    }

    /* De toute façon, on redessine */
    BI_DessinerObjet( arbre, num_objet );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Une date selon un format précis -------- */

int OBJ_MettreDate(unsigned int arbre, unsigned int objet, const char *format, struct tm *date)
{
  int retour;
  char *texte;

  /* Contrôles */
  retour = OBJ_ControlerValeurs( arbre, objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre ou objet incorrect [%d/arbre %u/objet %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, arbre, objet);
    return -1;
  }

  /* La date est-elle plausible ? */
  retour = 0;
  if ( date->tm_mon > DECEMBRE ) { date->tm_mon = DECEMBRE; retour = +1; }
  if ( FEVRIER == date->tm_mon ) { /* février */
    if ( date->tm_mday > 29 ) { date->tm_mday = 29; retour = +2; }
    if ( date->tm_mday == 29 ) {
      (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Contrôle des années bissextiles\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }
  } else if ( ( AVRIL == date->tm_mon ) || ( JUIN == date->tm_mon ) || ( SEPTEMBRE == date->tm_mon ) || ( NOVEMBRE == date->tm_mon ) ) {
    if ( date->tm_mday > 30 ) { date->tm_mday = 30; retour = +2; }
  } else {
    if ( date->tm_mday > 31 ) { date->tm_mday = 31; retour = +2; }
  }

  /* Texte contenant la date mise en forme */
  if ( NULL == format ) {
    format = "%d %B %Y";
  }

  /* On crée la date */
  texte = TXT_CreerDate( format, date );
  if ( NULL == texte ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la date avec TXT_CreerDate\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  /* On installe le texte dans l'objet */
  met_chaine( arbre, objet, 0, texte, TXT_COPIER );

  /* Plus besoin du texte */
  free( texte );

  /* Tout s'est bien passé ; >0 => date corrigée avant affichage */
  return retour;
}

/* -------- Valeur numérique entière (sans erreur !) -------- */

int OBJ_MettreEntier_S(unsigned int arbre, unsigned int objet, long valeur)
{
  int retour;
  char *nv_texte;

  /* Cas spécial : G_BOXCHAR */
  if ( G_BOXCHAR == (arbres[ arbre ].adresse + objet)->ob_type ) {
    int car = '0';

    if ( ( valeur > 0 ) && ( valeur < 10 ) ) {
      car = '0' + valeur;
    }
    return OBJ_MettreCaractere( arbre, objet, car );
  }

  retour = OBJ_MettreNombre( arbre, objet, (double) valeur, 0.0 );
  if( 0 == retour ) return 0; /* On a pu mettre la valeur */
  if ( retour < 0 ) { /* Erreur grave */
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans OBJ_MettreNombre() [%ld = %g], code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, valeur, (double) valeur, retour);
    return retour;
  }

  /* Erreur mineure : aurait-on essayé de mettre un texte dans un bouton ou une chaîne ?
     Si oui, on le fait [pour gérer les ascenseurs !] en _remplaçant_ la chaîne */
  nv_texte = TXT_CNV_L_vTexte( valeur, NULL, NULL );
  retour = OBJ_RemplacerTexte( arbre, objet, nv_texte, FAUX );
  if( 0 == retour ) return 0; /* On a pu mettre la valeur */
  if ( retour < 0 ) { /* Erreur grave */
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans OBJ_RemplacerTexte(), code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    free( nv_texte ); nv_texte = NULL;
    return retour;
  }

  /* On a échoué ! */

  return retour;
}

int OBJ_MettreEntier_U(unsigned int arbre, unsigned int objet, unsigned long valeur)
{
  int retour;

  retour = OBJ_MettreNombre( arbre, objet, (double) valeur, 0.0 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans OBJ_MettreNombre() [%lu = %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, valeur, (double) valeur);
  }

  return retour;
}

/* -------- Valeur numérique réelle quelconque, avec erreur si erreur > 0.0 -------- */

int OBJ_MettreNombre(unsigned int arbre, unsigned int num_objet, double valeur, double erreur)
{
  int retour;
  OBJECT *objet;

  /* Faux objet : appels vides... */
  if ( ROOT == num_objet ) return 0;

  /* Contrôles */
  retour = OBJ_ControlerValeurs( arbre, num_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre ou objet incorrect [%d/arbre %u/objet %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, arbre, num_objet);
    return -1;
  }
#ifdef _DEVELOPPE_
# ifdef __linux__ /* LINUX : On a, a priori, les fonctions isinf() et isnan() */
  if ( ( isinf( valeur ) ) || ( isnan( valeur ) ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Valeur incorrecte [%g, erreur %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, valeur, erreur);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -2;
  }
# endif
#endif

  /* Quel est l'objet modifié ? */
  objet = arbres[ arbre ].adresse + num_objet;

  switch ( objet->ob_type )
  {
  case G_BUTTON:
  case G_STRING:
  case G_TITLE:
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : Valeur dans un bouton, une chaîne : dangereux -> ignoré\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return +1; /* Erreur, mineure */

  case G_TXT_INFINI: /* Longueur illimitée -> rien de spécial ! */
    {
      char *chaine_valeur;

      chaine_valeur = TXT_CNV_D_vTexte( valeur, erreur, 7, 7, 3, NULL, NULL );
      met_chaine( arbre, num_objet, 0, chaine_valeur, TXT_COPIER );
      free( chaine_valeur );
    }
    return 0; /* Redessin éventuel déjà fait dans met_chaine() */

  case G_NOMBRE:
    {
      char *chaine_valeur, *chaine;

      chaine_valeur = TXT_CNV_D_vTexte( valeur, erreur, 7, 7, 3, NULL, NULL );
      if ( objet->ob_spec.nombre->proprietes & BI_EDITABLE_UNITE ) {
	chaine = TXT_genere_chaine( "%s %s", chaine_valeur, objet->ob_spec.nombre->unite );
	free( chaine_valeur ); chaine_valeur = NULL;
      } else if ( objet->ob_spec.nombre->proprietes & BI_EDITABLE_ANGLE ) {
	chaine = TXT_genere_chaine( "%s"DEGRE, chaine_valeur );
	free( chaine_valeur ); chaine_valeur = NULL;
      } else {
	chaine = chaine_valeur;
      }
      met_chaine( arbre, num_objet, 0, chaine, TXT_COPIER );
      free( chaine );
    }
    return 0;

  case G_TEXT:
  case G_BOXTEXT:
    OPTION_A_FAIRE( "Mettre une valeur dans G_TEXT ou G_BOXTEXT" );
    return +5;

  case G_FTEXT:
  case G_FBOXTEXT: /* On ne tient pas compte de l'erreur */
    retour = Tedinfo_MettreValeur( objet->ob_spec.tedinfo, valeur );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant mettre %g dans la tedinfo [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, valeur, retour);
      return -60;
    }
    break;

  default:
    AVERTISSEMENT( "Type d'objet inconnu [%d]\n", objet->ob_type );
    return +45;
  }

  /* Si l'arbre est ouvert, on redessine l'objet pour afficher le changement */
  if ( VRAI == arbres[ arbre ].drapeaux.ouvert ) {
    BI_DessinerObjet( arbre, num_objet );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Cas particulier : installer une vaeur numérique dans une TEDINFO -------- */

static int Tedinfo_MettreValeur(TEDINFO *tedinfo, double valeur)
{
  const char *masque; /* Adresse du masque  */
  const char *pos_fin, *pos_debut, *pos_point __attribute__((__unused__));
  char *chaine;       /* Adresse de la chaîne affichée */
  double p_decimale;
  int retour;
  unsigned int nbr_chiffres_avant = 0, nbr_chiffres_apres = 0;

  /* Init. */
  if ( NULL == tedinfo ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de TEDINFO !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -50;
  }
  chaine = tedinfo->te_ptext.editable;
  masque = tedinfo->te_ptmplt;
  if ( ( NULL == chaine ) || ( NULL == masque ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Chaîne NULL dans la TEDINFO\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -51;
  }

  /* On initialise la chaîne éditable */
  memset( chaine, 0, tedinfo->te_txtlen );

  /* On cherche, dans le masque, le premier caractère éditable */
  pos_debut = strchr( masque, '_' );
  if ( pos_debut == NULL ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucun caractère éditable dans l'objet\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -52;
  }

  /* On cherche le premier caractère autre que '_' après ce premier éditable */
  pos_fin = pos_debut;
  while ( '_' == *pos_fin ) {
    nbr_chiffres_avant++; /* Un chiffre de plus avant la virgule */
    pos_fin++;
  }

  /* On peut commencer par mettre la partie entière ! */
  retour = Tedinfo_PartieEntiere( tedinfo, valeur, nbr_chiffres_avant );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le codage de la partie entière [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return +1;
  }

  /* De quel type de champ s'agit-il ? */
  if ( 0 == *pos_fin ) {
    return 0; /* Champ entier... */
  }
  if ( ( *pos_fin != '.' ) && ( *pos_fin != ',' ) ) {
    /* Autre chose qu'un séparateur décimal -> on ignore la suite ! */
    (void) fprintf(stderr, "%s, %s l.%u : Type de champ inconnu [masque = %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, pos_debut);
    return +5;
  }
  pos_point = pos_fin;
  pos_fin++;

  /* On repère le nombre de chiffres après la virgule */
  while ( '_' == *pos_fin ) {
    nbr_chiffres_apres++;
    pos_fin++;
  }
  if ( 0 == nbr_chiffres_apres ) {
    /* Aucune place réservée pour ce qu'il y a après la virgule ! */
    (void) fprintf(stderr, "%s, %s l.%u : Type de champ inconnu [masque = %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, pos_debut);
    return +6;
  }

  /* Ici, il reste donc à placer la partie décimale */
  valeur = fabs( valeur ); /* Le signe a déjà été placé -> valeurs positives */
  p_decimale = valeur - (double) (int) valeur;

  /* On essaye de garder la précision maximale -> on évite TXT_CNV_D_vTexte qui fait sauter les 9 multiples finaux ! */
  retour = Tedinfo_PartieDecimale( tedinfo, p_decimale, nbr_chiffres_apres );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant mettre la partie décimale\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +20;
  }

  /* Installation réussie */
  return 0;
}

/* -------- Mettre la partie entière -------- */

static int Tedinfo_PartieEntiere(TEDINFO *tedinfo, double valeur, unsigned int nbr_chiffres)
{
  boolean negatif = FAUX;
  char *car_chiffre;
  double min, max;
  unsigned int chiffre, nbr_chiffres_reel;

  /* Cas spécial : un seul chiffre */
  if ( 1 == nbr_chiffres ) {
    if ( valeur <   0.0 ) return -11;
    if ( valeur >= 10.0 ) return -12;

    *(tedinfo->te_ptext.editable) = (int) valeur;
    return 0;
  }

  /* Quelles sont les valeurs limites, compte-tenu du nombre de chiffres indiqué ? */
  max = pow( 10.0, (double) nbr_chiffres );
  min = -max / 10.0;

  /* La valeur est-elle dans ces bornes ? */
  if ( valeur <= min ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : Valeur trop petite [%g, min = %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, valeur, min);
#endif
    return -1;
  } else if ( valeur >= max ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : Valeur trop grande [%g, max = %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, valeur, max);
#endif
    return -2;
  }

  if ( valeur < 0.0 ) {
    negatif = VRAI;
    valeur  = -valeur;
  }
  if ( valeur < 1.0 ) { /* 0,... sera de toute façon avec un 0 ! */
    valeur = 0.0;
    nbr_chiffres_reel = 1;
  } else {
    nbr_chiffres_reel = 1 + (unsigned int) log10( valeur );
  }

  car_chiffre = tedinfo->te_ptext.editable;

  /* Tant que le nombre de chiffres n'est pas égal au nombre de chiffres réel */
  while ( nbr_chiffres > nbr_chiffres_reel + 1 ) { /* Espaces - on garde la place pour le signe éventuel */
    *car_chiffre = ' ';
    car_chiffre++;
    nbr_chiffres--;
  }
  if ( nbr_chiffres > nbr_chiffres_reel ) { /* On peut avoir égalité pour un nombre >0 avec exactement le bon nombre ! */
    if ( VRAI == negatif ) {
      *car_chiffre = '-';
    } else {
      *car_chiffre = ' ';
    }
    car_chiffre++;
    nbr_chiffres--;
  }

  /* Ici, on a donc autant de cases libres qu'il y a de chiffres */
  car_chiffre += nbr_chiffres - 1;
  while ( nbr_chiffres > 0 ) {
    /* On stocke le chiffre des unités */
    chiffre = (unsigned int) fmod( valeur, 10.0 );
    *car_chiffre = '0' + chiffre;

    /* On divise : chiffre des dizaines devient le chiffre des unités */
    valeur /= 10.0;

    /* On passe au chiffre suivant */
    car_chiffre--;
    nbr_chiffres--;
  }

  /* Installation terminée */
  return 0;
}

/* -------- Placer la partie décimale -------- */

static int Tedinfo_PartieDecimale(TEDINFO *tedinfo, double p_decimale, unsigned int nbr_chiffres)
{
  char *chaine, *pos_virgule;
  unsigned int i = 0;
  int chiffre;

  if ( 0.0 == p_decimale ) {
    /* Valeur entière ! */
    return 0;
  }

  /* On se place à la suite de la partie entière */
  chaine  = tedinfo->te_ptext.editable;
  chaine += strlen( chaine );
  pos_virgule = chaine;

  /* On écrit, chiffre par chiffre */
  while ( i < nbr_chiffres ) {
    p_decimale *= 10.0;
    chiffre = (int) p_decimale;
    p_decimale -= (double) chiffre;

    *chaine = '0' + chiffre;
    chaine++;
    i++;
  }

  chaine--;
  if ( '0' == *chaine ) {
    /* On fait sauter les 0 finaux */
    while ( *chaine == '0' ) {
      *chaine = 0;
      chaine--;
    }
  } else if ( '9' == *chaine ) {
    char *debut;

    /* On fait sauter les 9 finaux */
    /* a. On vérifie que l'on n'a pas la valeur maximale, [-]999...99,9999 */
    debut = chaine;
    while ( *chaine == '9' ) {
      chaine--;
      if ( chaine == tedinfo->te_ptext.editable ) {
	return 0; /* Rien d'autre à faire */
      }
      if ( '-' == *chaine ) {
	return 0; /* Rien d'autre à faire */
      }
    }

    /* b. On peut donc faire la modif */
    chaine = debut;
    while ( *chaine == '9' ) {
      if ( chaine > pos_virgule )
	*chaine = 0;
      else
	*chaine = '0';
      chaine--;
    }
    if ( ' ' == *chaine ) { /* 99,9999 -> 100,0000 */
      *chaine = '1';
    } else {
      (*chaine)++;
    }
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Indiquer une unité pour une valeur -------- */

int OBJ_MettreUnite(unsigned int numero_arbre, unsigned int numero_objet, const char *unite)
{
  OBJECT *objet;
  int retour;
  struct infos_editable *infos_nombre;

  /* Validité */
  retour = OBJ_ControlerValeurs( numero_arbre, numero_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : valeurs incorrectes... [arbre= %u, objet = %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_arbre, numero_objet);
    return -1;
  }

  /* Quel objet nous intéresse ? */
  objet = arbres[ numero_arbre ].adresse + numero_objet;
  if ( objet->ob_type != G_NOMBRE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Type d'objet incorrect [%d] --- seul les objets de type %d [G_NOMBRE] peuvent avoir une unité associée\n",
		   __FILE__, __FUNCTION__, __LINE__, objet->ob_type, G_NOMBRE );
    return -2;
  }

  /* On récupère ses infos */
  infos_nombre = objet->ob_spec.nombre;
  if ( NULL == infos_nombre ) {
    (void) fprintf( stderr, "%s,%s l.%u : Pas d'infos étendues G_NOMBRE [arbre : %u, objet : %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, numero_arbre, numero_objet );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -3;
  }

  /* On adapte les unités... */
  return OBJ_OptionsEditable( numero_arbre, numero_objet,
			      infos_nombre->proprietes & BI_EDITABLE_UNITE,
			      infos_nombre->min, infos_nombre->max, unite );

}

/* -------- Associer un sous-menu à un objet -------- */

int OBJ_MettreMenu(unsigned int arbre, unsigned int num_objet,
		   unsigned long nombre_options, const char **options, const char *titre,
		   void (*gestion)(unsigned int arbre, unsigned int objet, unsigned long option_choisie) )
{
  OBJECT *objet;
  int retour;

  /* Validité */
  retour = OBJ_ControlerValeurs( arbre, num_objet );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : valeurs incorrectes... [arbre= %u, objet = %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, arbre, num_objet );
    return -1;
  }

  /* Quel objet nous intéresse ? */
  objet = arbres[ arbre ].adresse + num_objet;
  if ( objet->ob_type != G_SSMENU ) {
    (void) fprintf( stderr, "%s, %s l.%u : Type d'objet incorrect [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
    return -2;
  }

  /* Y a-t-il déjà un menu associé à l'objet ? */
  if ( ( objet->ob_spec.menu->options != NULL ) && ( objet->ob_spec.menu->options != options ) ) {
    /* On ne cherche pas à libérer, car sous-menu peuvent être statiques, constants dans le programme */
    objet->ob_spec.menu->nombre_options = 0;
    objet->ob_spec.menu->options = NULL;
    objet->ob_spec.menu->gestion = ( void (*)(unsigned int,unsigned int,unsigned long) ) NULL;

    /* Même chose pour le titre */
    objet->ob_spec.menu->titre   = NULL;
  }

  /* On associe la nouvelle table de menus */
  objet->ob_spec.menu->options = options;
  objet->ob_spec.menu->gestion = gestion;
  objet->ob_spec.menu->titre   = titre;
  objet->ob_spec.menu->nombre_options = nombre_options;
  objet->ob_spec.menu->option_choisie = 0; /* Par défaut, on affiche le premier choix possible ! */

  /* Pas de souci */
  return 0;
}

/* -------- Demander ou indiquer quelle est l'option choisie dans un sous-menu -------- */

unsigned long OBJ_OptionMenu(unsigned int numero_arbre, unsigned int numero_objet, unsigned long option, int mode)
{
  OBJECT *objet;
  int retour;

  /* Validation de l'objet */
  retour = OBJ_ControlerValeurs( numero_arbre, numero_objet );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : valeurs incorrectes... [arbre= %u, objet = %u]\n",
		    __FILE__, __FUNCTION__, __LINE__, numero_arbre, numero_objet );
    return 0;
  }

  /* Quel objet nous intéresse ? */
  objet = arbres[ numero_arbre ].adresse + numero_objet;
  if ( objet->ob_type != G_SSMENU ) {
#ifdef _DEVELOPPE_
    /* Peut être normal dans certaines conditions [cf. xan_matrice.c] */
    (void) fprintf( stderr, "%s, %s l.%u : Type d'objet incorrect [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
#endif
    return 0;
  }

  if ( METTRE == mode ) {
    if ( option >= objet->ob_spec.menu->nombre_options ) {
      AVERTISSEMENT( "Option inexistante [%lu, 0 à %lu autorisés]", option, objet->ob_spec.menu->nombre_options );
    } else {
      /* On appelle, si nécessaire, la fonction associée à l'objet */
      if ( objet->ob_spec.menu->gestion != ( void (*)(unsigned int, unsigned int, unsigned long) ) NULL ) {
	objet->ob_spec.menu->gestion( numero_arbre, numero_objet, option );
      }

      /* On mémorise l'option */
      objet->ob_spec.menu->option_choisie = option;

      /* On redessine l'objet, mais seulement si on affiche le texte de l'option ! */
      if ( NULL == objet->ob_spec.menu->texte ) {
	/* Et seulement si l'arbre est ouvert ! */
	if ( VRAI == arbres[ numero_arbre ].drapeaux.ouvert ) {
	  BI_DessinerObjet( numero_arbre, numero_objet );
	}
      }
    }
  }

  /* On renvoie l'option actuellement choisie */
  return objet->ob_spec.menu->option_choisie;
}

/* -------- Mettre un style de ligne/marque dans un objet -------- */

int OBJ_MettreStyle(unsigned int numero_arbre, unsigned int numero_objet, unsigned int style)
{
  int retour;
  OBJECT *objet;

  /* Contrôles */
  retour = OBJ_ControlerValeurs( numero_arbre, numero_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Arbre ou objet incorrect [%d/arbre %u/objet %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour, numero_arbre, numero_objet);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  objet = arbres[ numero_arbre ].adresse + numero_objet;

  /* On renvoie la valeur... */
  switch ( objet->ob_type )
  {
  case G_STYLELIGNE:
    objet->ob_spec.style_ligne = style;
    break;

  case G_STYLEMARQUE:
    objet->ob_spec.style_marque = style;
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Objet %u sans style associé, type incorrect !\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_objet);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_CONTINU;
  }

  /* On force le redessin, si l'arbre est ouvert ! */
  if ( VRAI == arbres[ numero_arbre ].drapeaux.ouvert ) {
    BI_DessinerObjet( numero_arbre, numero_objet );
  }

  /* Mise en place réussie ! */
  return 0;
}

/* -------- Mettre un texte dans un objet --------
    On ne touche pas à la chaîne existante, on y copie le texte passé

    Remarque : si texte infini, on peut étendre la chaîne existante !
 */

int OBJ_CopierTexte(unsigned int arbre, unsigned int num_objet, const char *texte, unsigned long lg_max)
{
  int retour;
  OBJECT *objet;

  /* Validité */
  retour = OBJ_ControlerValeurs( arbre, num_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : valeurs incorrectes... [arbre= %u, objet = %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, arbre, num_objet);
    return -1;
  }

  /* Quel objet nous intéresse ? */
  objet = arbres[ arbre ].adresse + num_objet;

  switch ( objet->ob_type )
  {
  case G_BOXCHAR:
    AVERTISSEMENT( "Utiliser OBJ_MettreCaractere pour G_BOXCHAR [arbre %u, objet %u]", arbre, num_objet );
    return +1;

  case G_BUTTON:
  case G_STRING:
    if ( NULL == texte ) {
      /* On rend vide le texte */
      if ( objet->ob_spec.txt_malloc != NULL ) {
	objet->ob_spec.txt_malloc[ 0 ] = 0;
      }
    } else {
      AVERTISSEMENT( "Pas de réel contrôle de débordement pour G_BUTTON/G_STRING [arbre %u, objet %u ; %s]", arbre, num_objet, texte );
      if ( 0 == lg_max ) {
	(void) fprintf( stderr, "%s, %s l.%u : Pas de longueur maximale indiquée, contrôle impossible\n",
			__FILE__, __FUNCTION__, __LINE__ );
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	return -10;
      }
      /* On copie le texte dans l'existant, en tenant compte de la limite */
      strncpy( objet->ob_spec.txt_malloc, texte, lg_max );
    }
    break;

  case G_FTEXT:
    OPTION_A_FAIRE( "Copier un texte dans un FTEXT" );
    break;

  case G_TXT_INFINI:
    TXTInfini_MettreTexte( objet->ob_spec.txt_infini, texte );
    break;

#ifdef _DEVELOPPE_
  case G_SSMENU:
    if ( NULL == objet->ob_spec.menu ) {
      (void) fprintf(stderr, "%s, %s l.%u : Pas d'information de menu !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return -11;
    }

    (void) fprintf(stderr, "%s, %s l.%u : Impossible de copier un texte dans G_SSMENU --- On ne peut que le remplacer !",
		   __FILE__, __FUNCTION__, __LINE__);
    return -19;
#endif

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Type d'objet non-traité ou inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -20;
  }

  /* Si nécessaire, on redessine */
  if ( VRAI == arbres[ arbre ].drapeaux.ouvert ) {
    BI_DessinerObjet( arbre, num_objet );

    /* En s'occupant aussi du curseur, s'il est dans cet objet ! */
    if ( num_objet == arbres[ arbre ].edit_courant ) {
      BI_PlacerCurseur( arbres + arbre, BI_InitialiserCurseur( objet ) );
    }
  }

  /* Terminé, sans erreur */
  return 0;
}

/* ---- Cas particulier d'un texte infini ---- */

static int TXTInfini_MettreTexte(Texte_Infini *txt_infini, const char *texte)
{
  unsigned long lg_nouveau_texte;

  if ( NULL == txt_infini ) return -1;

  if ( NULL == texte ) {
    /* On fait donc un texte vide ! */
    if ( NULL != txt_infini->texte ) {
      txt_infini->texte[ 0 ] = 0;
      txt_infini->prem_car = 0;
    }
    return 0;
  }

  /* Quelle est la taille du nouveau texte ? */
  lg_nouveau_texte = strlen( texte );

  /* Est-ce qu'il tient dans l'objet actuel ?
     [attention, max_cars = 0 et lg_nouveau_texte_0 = 0 peut se produire, si le texte est NULL : en tenir compte !] */
  if ( ( lg_nouveau_texte > txt_infini->max_cars ) || ( NULL == txt_infini->texte ) ) {
    char *nv_texte;

    /* NON ==> On augmente la taille du tampon */
    nv_texte = (char *) realloc( txt_infini->texte, sizeof( char ) * ( lg_nouveau_texte + 1 ) ); /* NPO le 0 final ! */
    if ( NULL == nv_texte ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de réserver un texte plus long...\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
    txt_infini->texte = nv_texte;

    /* On mémorise la nouvelle taille */
    txt_infini->max_cars = lg_nouveau_texte; /* 0 final non compté... */
  }

  /* On peut maintenant copier en toute sécurité */
  strcpy( txt_infini->texte, texte );

  /* On réaffiche depuis le début du texte */
  txt_infini->prem_car = 0;

  /* Terminé sans anicroche */
  return 0;
}

/* -------- Remplacer le texte d'un objet -------- */

int OBJ_RemplacerTexte(unsigned int arbre, unsigned int num_objet, const char *texte, boolean copier_original)
{
  int retour;
  OBJECT *objet;
  char *vrai_texte = NULL;

  /* Validité */
  retour = OBJ_ControlerValeurs( arbre, num_objet );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : valeurs incorrectes... [arbre= %u, objet = %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, arbre, num_objet);
    return -1;
  }
  if ( NULL == texte ) return -2;

  if ( VRAI == copier_original ) {
    vrai_texte = strdup( texte );
    if ( NULL == vrai_texte ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de copier le texte %s !\n",
		     __FILE__, __FUNCTION__, __LINE__, texte);
      return -5;
    }
  }

  objet = arbres[ arbre ].adresse + num_objet;
  switch ( objet->ob_type )
  {
  case G_BOXCHAR:
    AVERTISSEMENT( "Utiliser OBJ_MettreCaractere pour G_BOXCHAR [arbre %u, objet %u]", arbre, num_objet );
    return +1;

  case G_BUTTON:
  case G_STRING:
  case G_TITLE:
    /* On libère l'ancien texte */
    free( objet->ob_spec.txt_malloc );

    /* On met le nouveau */
    if ( VRAI == copier_original ) {
      objet->ob_spec.txt_malloc = vrai_texte;
      vrai_texte = NULL;
    } else
      objet->ob_spec.texte = texte;
    break;

  case G_SSMENU:
    if ( VRAI == copier_original ) {
      (void) fprintf( stderr, "%s, %s l.%u : Impossible de mettre un texte réservé avec malloc() dans un G_SSMENU !\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      (void) fprintf( stderr, "%s, %s l.%u : Il ne serait jamais libéré...\n",
		      __FILE__, __FUNCTION__, __LINE__ );
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    } else {
      objet->ob_spec.menu->texte = texte;
    }
    break;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : Type d'objet non-traité ou inconnu [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, objet->ob_type );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -20;
  }

  /* On efface le texte vrai s'il existe et n'a pas été utilisé */
  free( vrai_texte ); vrai_texte = NULL;

  /* Si nécessaire, on redessine */
  if ( VRAI == arbres[ arbre ].drapeaux.ouvert ) {
    BI_DessinerObjet( arbre, num_objet );
  }

  /* Terminé, sans erreur */
  return 0;
}

