/* LASE - Logiciel d'analyse des spectres EXAFS
   (C) Emmanuel CURIS, 2004-1998

   Sauvegarde au format postscript
   L'unit de base est le 1/100 de mm = taille d'un point donc 
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "math_utils.h" /* Pour Puissance() */

#include "ps.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define HAUTEUR_TEXTE (12.0 / 72.0 * 2540.0 ) /* 12 p, 1 p = 1/72 pouce, 1 pouce = 2.54 cm, on travaille en 1/100 mm */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int  PS_GrilleDecimale(FILE *fichier, struct rect_ps *cadre, InfosAxe *axe);
static int  PS_GrilleLog     (FILE *fichier, struct rect_ps *cadre, InfosAxe *axe);
static void PS_Grille_Log    (FILE *fichier, struct rect_ps *cadre, InfosAxe *axe, double *points, double echelle, int p10, double val_ref);

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* -------- Trac d'un cadre pour graphes, avec graduation des bords bas & gauche */

int PS_Cadre(FILE *fichier, struct infos_ps *infos, struct rect_ps *cadre)
{
  int retour;
  double delta_min;
  char *legende;

  if ( NULL == cadre ) return -1;
  if ( ( cadre->haut <= 0.0 ) || ( cadre->larg <= 0.0 ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Cadre rduit  1 point [larg = %g, haut = %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, cadre->larg, cadre->haut);
    return -25;
  }
  (void) fprintf( fichier, "%% Cadre = rgion du plan affiche\n");

  /* Le rectangle entourant : noir, fond blanc */
  retour = PS_Rectangle( fichier, infos->proprietes & PS_NIVEAU_2, infos->epaisseur, cadre->cadre.couleur_bord, cadre->cadre.couleur_fond,
			 cadre->x_BG, cadre->y_BG, cadre->larg, cadre->haut );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant faire le cadre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }

  /* Quel intervalle minimal entre deux graduations ? */
  delta_min = 3.0 * infos->epaisseur * 100.0; /* 3 paisseurs de traits, en 1/100 mm */
  if ( delta_min < 50.0 ) delta_min = 50.0; /* Avec au minimum 0,5 mm */

  /* On s'occupe de l'axe des X */
  /* 0. vrification de la validit des bornes */
  retour = AXE_ValiderBornes( &(cadre->cadre.axe_X) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Bornes de l'axe X incorrectes [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -2;
  }

  /* 1. Grille ventuelle */
  retour = PS_Grille( fichier, cadre, &(cadre->cadre.axe_X) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de tracer la grille [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* 2. Graduation en bas, toujours prsente */
  retour = PS_AxeGradue( fichier, cadre->x_BG, cadre->y_BG, cadre->x_BG + cadre->larg, cadre->y_BG,
			 &(cadre->cadre.axe_X), delta_min, infos->haut_texte );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_AxeGradue_H [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -2;
  }

  /* 3. La lgende */
  legende = AXE_LireLegende( &(cadre->cadre.axe_X) );
  if ( legende != NULL ) {
    PS_Texte( fichier, cadre->x_BG + cadre->larg / 2.0, cadre->y_BG - HAUTEUR_TEXTE - 200.0, legende,
	      BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_HAUT, 0.0, cadre->cadre.axe_X.couleur_legende );
    free( legende ); legende = NULL;
  }

  /* On s'occupe de l'axe des Y */
  /* 0. vrification de la validit des bornes */
  retour = AXE_ValiderBornes( &(cadre->cadre.axe_Y) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Bornes de l'axe Y incorrectes [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -2;
  }

  /* 1. Grille ventuelle */
  retour = PS_Grille( fichier, cadre, &(cadre->cadre.axe_Y) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de tracer la grille [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* 2. Graduation  gauche, toujours prsente */
  retour = PS_AxeGradue( fichier, cadre->x_BG, cadre->y_BG, cadre->x_BG, cadre->y_BG + cadre->haut,
			 &(cadre->cadre.axe_Y), delta_min, infos->haut_texte );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_AxeGradue_V [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -2;
  }

  /* 3. La lgende */
  legende = AXE_LireLegende( &(cadre->cadre.axe_Y) );
  if ( legende != NULL ) {
    PS_Texte( fichier, cadre->x_BG - 4.0 * HAUTEUR_TEXTE, cadre->y_BG + cadre->haut/2.0, legende,
	      BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_BAS, 90.0, cadre->cadre.axe_Y.couleur_legende );
    free( legende ); legende = NULL;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Grille en fond de cadre ~-~-~-~-~-~-~-~-~ */

/* -------- Fonction d'appel globale -------- */

int PS_Grille(FILE *fichier, struct rect_ps *cadre, InfosAxe *axe)
{
  int retour = 0;

  if ( ( NULL == fichier ) || ( NULL == cadre ) ) return -1;
  if ( FAUX == ( axe->proprietes & AXE_AVEC_GRILLE ) ) return 0; /* Pas de grille */

  (void) fprintf( fichier, "%% Grille\n" );

  /* Couleur, style de la grille */
  PS_Couleur( fichier, axe->couleur_grille );
  PS_StyleLigne( fichier, axe->style_grille );

  switch ( axe->type_graduation )
  {
  case GRADUATION_AUCUNE:
    break;

  case GRADUATION_DECIMALE:
    retour = PS_GrilleDecimale( fichier, cadre, axe );
    break;

  case GRADUATION_LOGARITHMIQUE:
    retour = PS_GrilleLog( fichier, cadre, axe );
    break;

  case GRADUATION_PARABOLIQUE:
  case GRADUATION_GAUSSIENNE:
  default:
    (void) fprintf(stderr, "%s, %s l.%u : Mthode de graduation inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, axe->type_graduation);

  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le trac de la grille [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -20;
  }

  return 0;
}

/* -------- Grille dcimale -------- */

static int PS_GrilleDecimale(FILE *fichier, struct rect_ps *cadre, InfosAxe *axe)
{
  double points[ 4 ] = { 0, 0, 0, 0 };  /* Init pour viter un avertissement ind avec gcc 4.0 */
  double valeur, pas, depart, echelle;
  int chiffre;

  /* Coordonnes fixes */
  if ( axe->proprietes & AXE_VERTICAL ) {
    points[ 0 ] = cadre->x_BG;
    points[ 2 ] = cadre->x_BG + cadre->larg;
    echelle = ( axe->max - axe->min ) / cadre->haut;
  } else {
    points[ 1 ] = cadre->y_BG;
    points[ 3 ] = cadre->y_BG + cadre->haut;
    echelle = ( axe->max - axe->min ) / cadre->larg;
  }

  /* Quel pas utiliser ? (la puissance de 10 utilise pour le texte, a priori) */
  pas = axe->pas_graduation;
  if ( pas <= 0.0 ) {
    /* Si automatique : un cran en-dessous de la puissance optimale ! */
    pas = Puissance( 10.0, AXE_GraduationMajeure( axe->min, axe->max ) - 1 );
  }

  /* O dmarre-t-on ? */
  if ( axe->min > 0.0 ) {
    double x_p;

    x_p = axe->min / pas;
    chiffre = (int) ( x_p - floor( x_p / 10.0 ) * 10.0);
    depart  = ( floor( x_p ) - ( double ) chiffre ) * pas;
  } else {
    double x_p;

    x_p = axe->min / pas;
    chiffre = (int) ( x_p - ceil( x_p / 10.0 ) * 10.0);
    depart  = ( ceil( x_p ) - ( double ) chiffre ) * pas;
  }

  /* On saute toutes les graduations avant le dbut de l'axe */
  valeur = depart + pas * (double) chiffre;
  while ( valeur < axe->min ) {
    chiffre++;
    valeur = depart + pas * (double) chiffre;
  }

  /* Tant que l'on est dans la partie visible, on fait les tics */
  while ( valeur  <= axe->max ) {
    /* Position des points */
    if ( axe->proprietes & AXE_VERTICAL ) {
      points[ 1 ] = cadre->y_BG + ( valeur - axe->min ) / echelle;
      points[ 3 ] = points[ 1 ];
    } else {
      points[ 0 ] = cadre->x_BG + ( valeur - axe->min ) / echelle;
      points[ 2 ] = points[ 0 ];
    }

    /* Trac */
    (void) fprintf( fichier, "%g %g moveto %g %g lineto ",
		    points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );

    /* Passage au tic suivant */
    chiffre++;
    valeur = depart + pas * (double) chiffre; /* Recalcul pour viter de rester sur place si pas trop faible ! */
  }

  /* Termin */
  (void) fprintf( fichier, "stroke\n" );

  return 0;
}

/* -------- Grille logarithmique -------- */

static int PS_GrilleLog(FILE *fichier, struct rect_ps *cadre, InfosAxe *axe)
{
  int p10_min, p10_max, p10;
  double points[ 4 ] = { 0, 0, 0, 0 };  /* Init pour viter un avertissement ind avec gcc 4.0 */
  double pos_depart, pos_fin, place_puissance, hauteur, echelle;

  /* Coordonnes fixes */
  if ( axe->proprietes & AXE_VERTICAL ) {
    points[ 0 ] = cadre->x_BG;
    points[ 2 ] = cadre->x_BG + cadre->larg;
    hauteur = cadre->haut;
  } else {
    points[ 1 ] = cadre->y_BG;
    points[ 3 ] = cadre->y_BG + cadre->haut;
    hauteur = cadre->larg;
  }
  echelle = ( axe->max - axe->min ) / hauteur;

  /* Quelle place occupe une puissance complte sur l'axe [= un intervalle de 1 en log] */
  place_puissance = 1.0 / echelle;
  if ( place_puissance < 1000.0 ) { /* Moins de 1 cm : seul certaines puissances seront places => sans grille */
    return +1;
  }

  /* Quelles sont les puissances de 10 extrmes impliques ? */
  p10_min = floor( axe->min );
  p10_max = ceil ( axe->max );

  /* Pour chacune de ces puissances, on gradue */
  for ( p10 = p10_min ; p10 < p10_max ; p10++ ) {
    pos_depart  = p10 - axe->min; /* log10( 10^n ) = n : 1er tic */
    pos_depart /= echelle;

    pos_fin = (p10 + 1) - axe->min; /* log10( 10^n+1 ) = n + 1 : dernier tic, 1er suivant */
    pos_fin /= echelle;

    /* On place le tic de fin, s'il est dans la rgion */
    if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
      points[ 1 ] = cadre->y_BG + pos_fin;
      if ( ( points[ 1 ] > cadre->y_BG ) && ( points[ 1 ] < ( cadre->y_BG + cadre->haut - 1 ) ) ) {
	points[ 3 ] = points[ 1 ];
	pos_depart += cadre->y_BG;

	/* Trac */
	(void) fprintf( fichier, "%g %g moveto %g %g lineto ",
			points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
/*       } else { */
/* 	(void) fprintf(stderr, "%s, %s l.%u : Refus, y=%g [%g -> %g] (p10 = %d)\n", */
/* 		       __FILE__, __FUNCTION__, __LINE__, points[ 1 ], cadre->y_BG, cadre->y_BG + cadre->haut - 1, p10 ); */
      }
    } else {
      points[ 0 ] = cadre->x_BG + pos_fin;
      if ( points[ 0 ] < ( cadre->x_BG + cadre->larg - 1 ) ) {
	pos_depart += cadre->x_BG;

	points[ 2 ] = points[ 0 ];
	/* Trac */
	(void) fprintf( fichier, "%g %g moveto %g %g lineto ",
			points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
      }
    }

    /* On place le niveau en-dessous avec une taille infrieure */
    PS_Grille_Log( fichier, cadre, axe, points, echelle, p10, Puissance( 10.0, p10 ) );
  }

  /* Termin sans problme */
  (void) fprintf( fichier, "stroke\n" );
  return 0;
}

/* -------- Placer les tics logarithmiques entre deux positions --------

   p10        : la puissance de 10 que l'on reprsente
   pos_depart : la position _cran_ du tic pour le tic "1"
 */

static void PS_Grille_Log(FILE *fichier, struct rect_ps *cadre, InfosAxe *axe, double *points, double echelle,
			   int p10, double val_ref)
{
  double positions[ 10 ], point_avant;
  unsigned int i;

  /* Position des tics  l'cran */
  for ( i = 0 ; i < 10 ; i++ ) {
    positions[ i ] = val_ref +  Puissance( 10.0, p10 ) * (double) i; /* Valeur dcimale */
    positions[ i ] = log10( positions[ i ] ); /* Correspondance en log */
  }

  /* Prparation */
  if ( axe->proprietes & AXE_VERTICAL ) { /* Axe vertical */
    point_avant = cadre->y_BG + ( positions[ 0 ] - axe->min ) / echelle;
  } else { /* Axe horizontal */
    point_avant = cadre->x_BG + ( positions[ 0 ] - axe->min ) / echelle;
  }

  /* On place les traits pour "2"  "9" [1 plac avant comme "x" d'un tour suprieur, "10" = "1" du suivant]  */
  for ( i = 1 ; i < 10 ; i++ ) {
    /* On trace les traits */
    if ( axe->proprietes & AXE_VERTICAL ) {
      points[ 1 ] = cadre->y_BG + ( positions[ i ] - axe->min ) / echelle;
/*       printf("%s : points[ 1 ] = %g, avant = %g (delta = %g)\n", __FUNCTION__, points[ 1 ], point_avant, point_avant - points[ 1 ] ); */
      if ( points[ 1 ] <= cadre->y_BG ) { /* On est en-dessous */
	continue;
      }

      if ( points[ 1 ] >= ( cadre->y_BG + cadre->haut - 1 ) ) { /* On dborde par en haut */
	return; /* La suite sera hors-rgion */
      }

      if ( ( points[ 1 ] - point_avant ) < 200.0 ) return; /* Trop serrs : la suite aussi ! */
      point_avant = points[ 1 ];

      points[ 3 ] = points[ 1 ];
    } else {
      points[ 0 ] = cadre->x_BG + ( positions[ i ] - axe->min ) / echelle;
      if ( points[ 0 ] <= cadre->x_BG ) { /* On est en-dessous */
	continue;
      }
      if ( points[ 0 ] >= ( cadre->x_BG + cadre->larg - 1 ) ) {
	return; /* On dborde par en haut : la suite sera hors-rgion */
      }

      if ( ( points[ 0 ] - point_avant ) < 200.0 ) return; /* Trop serrs : la suite aussi ! */
      point_avant = points[ 0 ];

      points[ 2 ] = points[ 0 ];
    }

    /* Trac */
    (void) fprintf( fichier, "%g %g moveto %g %g lineto ",
		    points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );
  }
}
