/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Cration : 18 mars 1999

   Export d'une structure 3d (une molcule) vers POV 3.0
   L'utilisation du fichier pov cr suppose que chimie.inc soit accessible par POV
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "exafs.h"
#include "atomes.h"
#include "feff.h"
#include "fichiers.h"
#include "modeles.h"
#include "tracer.h" /* Pour les modes de dessin de la molcule */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

static char *chemin_export_pov = NULL;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void choix_texture (FILE *fichier, unsigned int Z);

static int POV_Atome     (FILE *fichier, Atome *atome, int mode);
static int POV_ModeleFEFF(FILE *fichier, Modele_FEFF *modele, int mode);
static int POV_Molecule  (FILE *fichier, Molecule *molecule, int mode);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

void export_pov(Molecule *molecule, int mode)
{
  char *nom_fichier;
  unsigned int i;
  Fragment *fragment;
  FILE *fich_pov = NULL;
  Liaison *liaison;
  BI_InfosSelecteur infos_selecteur = {
    NULL,
    BI_SELECTEUR_FILTRE_UTILISATEUR | BI_SELECTEUR_AVEC_SAUVEGARDE | BI_SELECTEUR_MEMORISER_CHEMIN,
    NULL,
    1, NULL,
    0, NULL,
    (void (*)(unsigned int, unsigned int, const char *) ) NULL
  };
  const char *masque = "*.pov";

  if ( NULL == molecule ) return;

  /* Prparation du slecteur de fichiers */
  infos_selecteur.titre  = (char *) "Cr"e_AIGU"er un script POV-Ray";
  infos_selecteur.filtres = &masque;
  infos_selecteur.chemin = chemin_export_pov;

  nom_fichier = BI_SelecteurSauvegarde( &infos_selecteur, NULL, fen_vue_3d );
  /* On mmorise les chemins */
  chemin_export_pov = infos_selecteur.chemin;

  if ( NULL == nom_fichier ) return;

  fich_pov = fopen(nom_fichier, "w");
  if (fich_pov == NULL) {
    (void) SYS_ERREUR(ERREUR_ECRITURE, nom_fichier, NULL, NULL);
    return;
  }
  free(nom_fichier);

  /* On commence par les dclarations gnrales */
  fprintf(fich_pov, "// Fichier POV g"e_AIGU"n"e_AIGU"r"e_AIGU" par LASE v. %f\n", VERSION / 10.0);
  fprintf(fich_pov, "// LASE - Logiciel d'Analyse des Spectres EXAFS, "COPYRIGHT" Emmanuel CURIS 2002\n");
  fprintf(fich_pov, "//\n// Mol"e_AIGU"cule : %s\n\n", molecule->nom);
  fprintf(fich_pov, "#include \"colors.inc\"\n");
  fprintf(fich_pov, "#include \"glass.inc\"\n");
  fprintf(fich_pov, "#include \"chimie.inc\"\n\n");

  /* On positionne l'oeil */
  fprintf(fich_pov, "camera { location <%f, %f, %f>\n", molecule->pos_oeil.x, molecule->pos_oeil.y, molecule->pos_oeil.z );
  fprintf(fich_pov, "         sky <%f, %f, %f>\n", molecule->x_h, molecule->y_h, molecule->z_h);
  fprintf(fich_pov, "         angle %f\n", molecule->angle1);
  fprintf(fich_pov, "         look_at <%f, %f, %f> }\n\n", molecule->point_vise.x, molecule->point_vise.y, molecule->point_vise.z );

  /* On positionne une source de lumire, pour la forme */
  fprintf(fich_pov, "light_source { <%f, %f, %f> color White }\n", 
	  molecule->pos_oeil.x, molecule->pos_oeil.y, molecule->pos_oeil.z);

  /* On dessine la molcule */
  POV_Molecule(fich_pov, molecule, mode);

  /* On dessine les liaisons ventuelles, par des cylindres */
  /* Premire tape : liaisons dans les fragments */
  for (fragment = molecule->fragments, i = 0 ; i < molecule->nbr_fragments ;
       i++, fragment = fragment->suivant)
  {
    if (fragment->liaisons == NULL) /* Aucune liaison dans le fragment */
      continue;

    fprintf(fich_pov, "\n// Liaisons du fragment %s\n", fragment->nom);
    for (liaison = fragment->liaisons ; liaison != NULL ; liaison = liaison->suivante)
    {
      float x_m, y_m, z_m;

      x_m = (liaison->atome_1->position.x + liaison->atome_2->position.x) / 2.0;
      y_m = (liaison->atome_1->position.y + liaison->atome_2->position.y) / 2.0;
      z_m = (liaison->atome_1->position.z + liaison->atome_2->position.z) / 2.0;
      fprintf(fich_pov, "cylinder { <%f, %f, %f> <%f, %f, %f> 0.03 ",
	      liaison->atome_1->position.x, liaison->atome_1->position.y, liaison->atome_1->position.z, x_m, y_m, z_m);
      choix_texture(fich_pov, liaison->atome_1->num_atomique);
      fprintf(fich_pov, " }\n");
      fprintf(fich_pov, "cylinder { <%f, %f, %f> <%f, %f, %f> 0.03 ",
	      liaison->atome_2->position.x, liaison->atome_2->position.y, liaison->atome_2->position.z, x_m, y_m, z_m);
      choix_texture(fich_pov, liaison->atome_2->num_atomique);
      fprintf(fich_pov, " }\n");
    }
  }

  /* Deuxime tape : liaisons entre fragments */
  if (molecule->liaisons != NULL) {
    fprintf(fich_pov, "\n// Liaisons intramolculaires, entre fragments\n");
    for (liaison = molecule->liaisons ; liaison != NULL ; liaison = liaison->suivante) {
      float x_m, y_m, z_m;

      x_m = (liaison->atome_1->position.x + liaison->atome_2->position.x) / 2.0;
      y_m = (liaison->atome_1->position.y + liaison->atome_2->position.y) / 2.0;
      z_m = (liaison->atome_1->position.z + liaison->atome_2->position.z) / 2.0;
      fprintf(fich_pov, "cylinder { <%f, %f, %f> <%f, %f, %f> 0.03 ",
	      liaison->atome_1->position.x, liaison->atome_1->position.y, liaison->atome_1->position.z, x_m, y_m, z_m);
      choix_texture(fich_pov, liaison->atome_1->num_atomique);
      fprintf(fich_pov, " }\n");
      fprintf(fich_pov, "cylinder { <%f, %f, %f> <%f, %f, %f> 0.03 ",
	      liaison->atome_2->position.x, liaison->atome_2->position.y, liaison->atome_2->position.z, x_m, y_m, z_m);
      choix_texture(fich_pov, liaison->atome_2->num_atomique);
      fprintf(fich_pov, " }\n");
    }
  }

#ifdef sans_modele
  /* On trace le chemin courant */
  if ( OBJ_choisit(AFF_STRUCTURE, AST_CHEMINS, QUESTION) == VRAI )
  {
    Absorbeur *absorbeur;
    Chemin *chemin;
    unsigned int i = 0;
    double x_deb, y_deb, z_deb;

    absorbeur = &(mol_courante->absorbeurs[num_absorbeur - 1]);
    chemin = absorbeur->chemins;
    while (i < num_chemin)
    {
      i++;
      chemin = chemin->suivant;
      if (chemin == NULL)
      {
#ifdef _DEVELOPPE_
	fprintf(stderr, "export_pov : pas de chemin %d pour cet absorbeur !\n", num_chemin);
#endif
	return;
      }
    }

    /* On a le chemin : plus qu' tracer des cylindres vert de diamtre 0.1 pour le reprsenter... */
    x_deb = absorbeur->absorbeur->x; /* On dmarre de l'absorbeur */
    y_deb = absorbeur->absorbeur->y;
    z_deb = absorbeur->absorbeur->z;

    fprintf(fich_pov, "\n// Le chemin de diffusion choisi\n");
    if (chemin->nbr_atomes == 2) { /* Rflexion simple : un seul cylindre suffit */
      Atome *atome = chemin->atomes[0];

      fprintf(fich_pov, "cylinder { <%f, %f, %f> <%f, %f, %f> 0.05 texture { pigment {\n",
	      x_deb, y_deb, z_deb, atome->x, atome->y, atome->z);
      fprintf(fich_pov, "\tgradient <%f, %f, %f>\n",
	      atome->x - x_deb, atome->y - y_deb, atome->z - z_deb);
      fprintf(fich_pov, "\tcolor_map { [0.0,0.4 color Green color Green] [0.4,1.0 color White filter 1 color White filter 1 ] } scale 0.2");
      fprintf(fich_pov, "\n} } }\n");
    } else {
      for (i = 0 ; i < chemin->nbr_atomes ; i++)
      {
	Atome *atome = chemin->atomes[i];

	fprintf(fich_pov, "cylinder { <%f, %f, %f> <%f, %f, %f> 0.05 texture { pigment {\n",
		x_deb, y_deb, z_deb, atome->x, atome->y, atome->z);
	fprintf(fich_pov, "\tgradient <%f, %f, %f>\n",
		atome->x - x_deb, atome->y - y_deb, atome->z - z_deb);
	fprintf(fich_pov, "\tcolor_map { [0.0,0.4 color Green color Green] [0.4,1.0 color White filter 1 color White filter 1 ] } scale 0.2");
	fprintf(fich_pov, "\n} } }\n");

	x_deb = atome->x; y_deb = atome->y; z_deb = atome->z;
      }
    }
  }
#endif /* sans_modele */

  /* On travaille sur un fond blanc */

  fprintf(fich_pov, "\nbackground { color White }\n");

  (void) fclose(fich_pov);
}

/* -------- Dessin d'un atome -------- */

static int POV_Atome(FILE *fichier, Atome *atome, int mode)
{
  if ( NULL == fichier ) return FAUX;
  if ( NULL == atome ) return FAUX;

  (void) fprintf(fichier, "sphere { <%f, %f, %f>", atome->position.x, atome->position.y, atome->position.z );
  if ( mode & V3D_ATOME_REDUIT ) /* Rduction au tiers */
    (void) fprintf(fichier, " %f ", atome->rayon / 300.0);
  else /* Taille relle */
    (void) fprintf(fichier, " %f ", atome->rayon / 100.0); /* Rayons en pm mais coords en ansgtroms */

  choix_texture(fichier, atome->num_atomique);
  (void) fprintf(fichier, " }\n");

  return VRAI;
}

/* -------- Dessin des atomes d'un modle FEFF -------- */

static int POV_ModeleFEFF(FILE *fichier, Modele_FEFF *modele, int mode)
{
  FEFF_atome *atome;

  if ( NULL == fichier ) return FAUX;
  if ( NULL == modele ) return FAUX;

  for ( atome = modele->parametres.prem_atome ; atome != NULL ; atome = atome->suivant ) {
    if ( (FAUX == (mode & V3D_VOIR_SELECTION)) || ( VRAI == atome->atome->marques.visible) ) {
      if ( FAUX == POV_Atome(fichier, atome->atome, mode) )
	return FAUX;
    }
  }

  return VRAI;
}

/* -------- Ensemble des atomes  dessiner -------- */

static int POV_Molecule(FILE *fichier, Molecule *molecule, int mode)
{
  if ( NULL == fichier ) return FAUX;
  if ( NULL == molecule ) return FAUX;
  if ( FAUX == (mode & V3D_VOIR_ATOMES) ) return VRAI;

  if ( NULL == molecule->premier ) return FAUX;

  if ( mode & V3D_VOIR_MODELE ) {
    Absorbeur *absorbeur;

    if ( NULL == molecule->absorbeurs ) {
      return VRAI;
    }
    absorbeur = molecule->absorbeurs + molecule->absorbeur_courant;

    /* On dessine l'absorbeur, qui appartient toujours au modle de toute faon */
    if ( (FAUX == (mode & V3D_VOIR_SELECTION)) || ( VRAI == absorbeur->absorbeur->marques.visible) ) {
      if ( FAUX == POV_Atome(fichier, absorbeur->absorbeur, mode) )
	return FAUX;
    }

    if (NULL == absorbeur->modeles) {
      return VRAI;
    }

    /* Reste  dessiner le modle... */
    switch (absorbeur->modele_actif->type)
    {
    case MODELE_EXAFS_CHEMINS:
      POV_ModeleFEFF(fichier, absorbeur->modele_actif->infos_modele.feff, mode);
      break;
    }
  } else {
    Atome *atome;

    for ( atome = molecule->premier ; atome != NULL ; atome = atome->suivant ) {
      if ( (FAUX == (mode & V3D_VOIR_SELECTION)) || ( VRAI == atome->marques.visible) ) {
	POV_Atome(fichier, atome, mode);
      }
    }
  }

  return VRAI;
}

/* -------- Indique la texture  utiliser pour les atomes, en fonction du numro atomique -------- */

static void choix_texture(FILE *fichier, unsigned int Z)
{
  switch (Z)
  {
  case AZOTE:
    fprintf(fichier, "%s", "pigment { Azote }");
    break;
      
  case CARBONE:
    fprintf(fichier, "%s", "pigment { Carbone }");
    break;

  case CALCIUM:
    fprintf(fichier, "%s", "texture { Calcium }");
    break;

  case CHLORE:
    fprintf(fichier, "%s", "texture { Chlore }");
    break;

  case HYDROGENE:
    fprintf(fichier, "%s", "texture { Hydrogene }");
    break;

  case OXYGENE:
    fprintf(fichier, "%s", "pigment { Oxygene }");
    break;

  case PLATINE:
    fprintf(fichier, "%s", "texture { Platine }");
    break;

  case SOUFRE:
    fprintf(fichier, "%s", "pigment { Soufre }");
    break;
    
  case ZINC:
    fprintf(fichier, "%s", "texture { Zinc }");
    break;

  default:
    fprintf(fichier, "%s", "pigment { color non_fait }");
#ifdef _DEVELOPPE_
    printf("Atome pas encore prvu sous POV : Z = %d\n", Z);
#endif
  }
}
