/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Module de calculs thoriques divers
   Dfinitions pour les sections efficaces (surtout d'absorption)
*/

#ifndef ABSORPTION_H_
# define ABSORPTION_H_ /* Protection contre l'inclusion multiple */

struct graphe; /* Dans graphes.h */

/* ~~~~~~~~ Section efficace : photoabsorption / effet photolectrique ~~~~~~~~ */

/* -------- Les seuils d'absorption -------- */

#define ABSORPTION_TOTALE    -10 /* Absorption due  tous les seuils */

#define APRES_SEUIL_K -1 /* Si e > e_k, donc aprs le seuil K */
#define SEUIL_K        0 
#define SEUIL_LI       1
#define SEUIL_LII      2
#define SEUIL_LIII     3
#define SEUIL_MI       4
#define SEUIL_MII      5
#define SEUIL_MIII     6
#define SEUIL_MIV      7
#define SEUIL_MV       8
#define SEUIL_NI       9
#define SEUIL_NII     10
#define SEUIL_NIII    11
#define SEUIL_NIV     12
#define SEUIL_NV      13
#define SEUIL_NVI     14
#define SEUIL_NVII    15
#define SEUIL_OI      16
#define SEUIL_OII     17
#define SEUIL_OIII    18
#define SEUIL_OIV     19
#define SEUIL_OV      20
#define SEUIL_OVI     21
#define SEUIL_OVII    22
#define SEUIL_PI      23
#define SEUIL_PII     24
#define SEUIL_PIII    25
#define SEUIL_QI      26

#ifdef AVEC_TRANSURANIENS_ /* Seuils n'existant que pour les transuraniens lourds (>= Lr) */
# define SEUIL_PIV  27
# define SEUIL_PV   28
# define SEUIL_QII  29
# define SEUIL_QIII 30

# define NOMBRE_SEUILS 31 /* On ne compte pas les pseudo-seuils */
#else
# define NOMBRE_SEUILS 27 /* On ne compte pas les pseudo-seuils */
#endif

/* -- Charger les informations sur les seuils d'absorption -- */

int          SEUIL_InitialiserTables(void);
int          SEUIL_LibererTables    (void);
unsigned int SEUIL_PremierElement   (unsigned int seuil);

/* -- Manipuler un seuil -- */

double TH_DonneesSeuil (int Z, int seuil, double *hauteur, double *largeur); /* Donne les informations sur un seuil */
double TH_QuelSeuil     (int Z, double e_min, double e_max, int *seuil);
int    TH_QuelleRegion   (int Z, double energie);
int    TH_IdentifierSeuil(double energie, double marge, int *seuil, boolean historique);

const char *TH_TexteSeuil(int seuil); /* Nom du seuil... */

/* -- Modliser un seuil -- */

#define SEUIL_MODELE_ESCALIER 0 /* Seuil idal : marche d'escalier */
#define SEUIL_MODELE_LORENTZ  1 /* Donc Arc tg en intgrant */
#define SEUIL_MODELE_GAUSS    2 /* Donc erf en intgrant */

int   TH_SimulerSeuil(struct graphe *graphe, double position, double hauteur, double largeur, int methode);

/* -- Modliser l'absorption intrinsque aprs seuil -- */

#define MODELE_MU0_CONSTANT  0
#define MODELE_MU0_LENGELER  1
#define MODELE_MU0_VICTOREEN 2
#define MODELE_MU0_MCMASTER  3
#define MODELE_MU0_EBEL      4
#define MODELE_MU0_HEITLER   5

#ifdef _DEVELOPPE_
# define NOMBRE_MODELES_MU0 6
#else
# define NOMBRE_MODELES_MU0 5 /* Heitler ne fonctionne pas encore */
#endif

#define TH_CALCUL_MU0_AVEC_EXAFS (1 << 1) /* Bit mis : on indique des oscillations EXAFS */
#define TH_CALCUL_MU0_AVEC_MU0   (1 << 2) /* Bit mis : on veut un mu0 qui servira en post-seuil */
#define TH_CALCUL_MU0_MU0_EXP    (1 << 3) /* Bit mis : on indique ce mu0 a partir d'un graphe  */
#define TH_CALCUL_MU0_AVEC_SEUIL (1 << 4) /* Bit mis : on veut le seuil proprement dit */
#define TH_CALCUL_MU0_AVEC_RAIES (1 << 5) /* Bit mis : on veut les raies du preseuil aussi */

struct infos_construction_mu0 {
  unsigned int options;		/* Masque de bits, cf. ci-dessus */
  unsigned int Z;		/* Numero atomique de l'element */
  struct infos_exafs {
    struct graphe *signal_exafs; /* Les oscillations EXAFS [NULL = aucune] */
    unsigned int ponderation;	/* du signal EXAFS */
  } exafs;
  struct infos_mu0 { /* Les parametres pour le mu0 post-seuil */
    struct graphe *mu0_exp;	/* le mu0 experimental a utiliser */
    unsigned int ponderation;	/* du mu0 experimental indique */
    unsigned int modele_theorique; /* Le modele theorique a utiliser pour simuler mu0 */
  } mu0;
  struct infos_seuil {
    double energie;		/* E0 (eV) */
    unsigned int type;		/* K, L... */
    double hauteur;		/* hauteur de seuil attendue */
    double largeur;		/* largeur du seuil (eV) */
    unsigned int methode;	/* de calcul theorique du seuil */
  } seuil;
};

struct graphe *TH_ConstruireMu0(double e_min, double e_max, double pas, struct infos_construction_mu0 *parametres);

double TH_CalculerMu0           (int Z, double energie, int methode, int seuil, double e_seuil);
int    TH_CalculerMu0Intrinseque(struct graphe *graphe, int modele, double e0, int Z, int seuil);
int    TH_CalculerMu0Total      (struct graphe *graphe, struct graphe *mu0_intrinseque);

const char *TH_TexteModeleMu0(int modele);

/* -- Modliser l'absorption par les autres atomes, les autres seuils [(matrice)] -- */

#define MUM_SELON_VICTOREEN 0 /* Modle pour estimer le fond continu */
#define MUM_SELON_MCMASTER  1
#define MUM_SELON_EBEL      2

struct graphe *TH_fond_continu(struct graphe *grf); /* Dans Theorie/matrice.c */

void   TH_AjouterMum(struct graphe *grf, int Z, float proportion, int methode, boolean avec_seuils);
double TH_CalculerMu(int Z, double energie, int methode, int seuil);

/* Ebel     : (E) = exp( A0 + A1ln(E) + A2ln(E) + A3ln(E) + A4ln^4(E) + A5 ln^5(E)) par seuil */

void   Ebel_EffacerTables(void);
int    Ebel_seuil      (struct graphe *graphe, int Z, float proportion, int seuil, double e_seuil);
int    TH_coef_Ebel    (int Z, int seuil, boolean total, double c[ 6 ], double *h_seuil); /* Lire les coefficients */
double TH_Ebel         (double energie, int Z, double c[ 6 ]); /* Calcul */

/* McMaster : (E) = exp( A0 + A1ln(E) + A2ln(E) + A3ln(E) ) */

int    McMaster_seuil  (struct graphe *graphe, int Z, float proportion, int seuil, double e_seuil);
int    TH_coef_McMaster(int Z, int seuil, double *c1, double *c2, double *c3, double *c4); /* Lire les coefficients */
double TH_McMaster     (double energie, int Z, double c1, double c2, double c3, double c4); /* Calcul */

/* Victoreen : (E) = C/E + D/E^4 */

int    victorine_seuil  (struct graphe *graphe, int Z, float proportion, int seuil, double e_seuil);
int    TH_coef_Victoreen(int Z, int seuil, double *c , double *d); /* Lire les coefficients */
double TH_Victoreen     (double energie, int Z, double c1, double c2); /* Calcul */

/* ~~~~~~~~ Section efficace : diffusion ~~~~~~~~ */

#define DIFFUSION_INCOHERENTE -3 /* Diffusion Compton, inlastique */
#define DIFFUSION_COHERENTE   -2 /* Diffusion Thompson, lastique */

#define DIFFUSION_COEF_McMaster 0 /* Utilisation des tables de McMaster */

int TH_DiffusionElastique  (int Z, struct graphe *graphe, int methode, double f_massique);
int TH_DiffusionInelastique(int Z, struct graphe *graphe, int methode, double f_massique);

#endif /* ABSORPTION_H_ */
