/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS
   Création : mai 2006

   Interface graphique avec l'utilisateur

   Analyse de variance à un facteur (fixe ou aléatoire)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "serie.h"
#include "interface.h"
#include "proba.h"
#include "statistiques.h"
#include "tracer.h"

/* -------- Variables locales -------- */

#define I_A1F_OPT_CHARGER_MODALITE  0  /* Charger un graphe comme valeurs pour la modalité */
#define I_A1F_OPT_IMPORTER_MODALITE 1  /* Importer des valeurs pour cette modalité */
#define I_A1F_OPT_SAUVER_MODALITE   2  /* Sauver le graphe des valeurs de cette modalité */
#define I_A1F_OPT_EXPORTER_MODALITE 3  /* Exporter en ASCII les valeurs de cette modalité */
#define I_A1F_OPT_CHANGER_VALEURS   5  /* [4 : séparateur] Changer les valeurs */
#define I_A1F_OPT_VALEUR_EN_X       6  /* Assigner une valeur à la variable X */
#define I_A1F_OPT_CONTROLE          7  /* Marquer cette modalité comme témoin */
#define I_A1F_OPT_RENOMMER          8  /* Changer le nom de cette modalité */
#define I_A1F_OPT_ESTIMER_EFFET    10  /* [9 : séparateur] Estimer l'effet pour cette modalité */

#define I_A1F_NOMBRE_OPTIONS_MODALITES 11

#define I_A1F_VAR_BARTLETT          0 /* Test de Bartlett pour contrôler l'égalité des variances */
#define I_A1F_VAR_LEVENE_MOYENNE    1 /* Test de Levene, en utilisant la moyenne [test original] */
#define I_A1F_VAR_LEVENE_MEDIANE    2 /* Test de Levene, en utilisant la médiane */
#define I_A1F_VAR_LEVENE_M_TRONQUEE 3 /* Test de Levene, en utilisant la moyenne tronquée */

#define I_A1F_NOMBRE_OPTIONS_VARIANCES 4

#include "int_anva1F.txt"

/* -------- Fonctions locales -------- */

static int  I_A1F_GererArbre      (unsigned int numero_objet, int x, int y);
static int  I_A1F_PreparerArbre   (unsigned int numero_arbre);
static void I_A1F_ActiverQuanti   (boolean activer);
static void I_A1F_InstallerANVA   (struct infos_ANVA_1F *infos);
static void I_A1F_InstallerFacteur(struct anva_facteur *facteur);

static void I_A1F_OptionsTemoin  (struct infos_ANVA_1F *infos);
static void I_A1F_OptionsVariance(unsigned int arbre __attribute__((__unused__)), unsigned int objet __attribute__((__unused__)), unsigned long action);

static void I_A1F_AjouterModalite(Fenetre *fenetre, struct infos_ANVA_1F *infos, unsigned int nombre_modalites, unsigned int modalite_avant);

static int  I_A1F_Clic        (Fenetre *fenetre, int, int, int, int, int);
static int  I_A1F_ClicModalite(Fenetre *fenetre, struct infos_ANVA_1F *infos, unsigned int option, int x, int y);

static int  I_A1F_Messages(Fenetre *fenetre, Message *message);

static unsigned int I_A1F_QuelleModalite(Fenetre *fenetre, struct infos_ANVA_1F *infos, int x, int y);

static void I_A1F_Redessin      (Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);
static void I_A1F_Redessin_Quali(Fenetre *fenetre, struct infos_ANVA_1F *infos);
static void I_A1F_Redessin_Quanti(Fenetre *fenetre, struct infos_ANVA_1F *infos);
static void I_A1F_Redessin_BornesY(struct infos_ANVA_1F *infos, double *y_min, double *y_max);

static void I_A1F_RedessinModalite_Quali(Fenetre *fenetre, struct modalite *modalite, InfosAxe *axe_Y, double echelle_Y,
					 int x_HG, int y_HG, unsigned int largeur, unsigned int hauteur);
static void I_A1F_RedessinModalite_Quanti(Fenetre *fenetre, struct modalite *modalite, int y_BG, InfosAxe *axe_Y, double echelle_Y, int x_valeurs);
static void I_A1F_RedessinModalite_Texte(Fenetre *fenetre, struct modalite *modalite,
					 int x_HG, int y_HG, unsigned int largeur, unsigned int hauteur);

#define IA1_DEMANDE_NOM_FACTEUR      0 /* Numéro des champs dans la boîte de demande */
#define IA1_DEMANDE_NOMBRE_MODALITES 1
#define IA1_DEMANDE_ALEATOIRE        2

int I_ANVA_1facteur(struct graphe *reference)
{
  int retour;
  struct infos_ANVA_1F *infos = NULL;
  unsigned long i;
  Fenetre *fenetre = NULL;
  BI_Champ_Demande questions[ 3 ] =
    { { BI_CHAMP_DEMANDE_CHAINE  , txt_nom_facteur     , NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
      { BI_CHAMP_DEMANDE_ENTIER_U, txt_nombre_modalites, NULL, BI_EDITABLE_MIN, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 2 }, { 0 } },
      { BI_CHAMP_DEMANDE_BINAIRE , txt_effet_aleatoire , NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } } };

  if ( VRAI == ARB_EstOuvert( ANVA_1FACTEUR ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s, %s l.%u : Appel avec l'arbre déjà ouvert [arbre %u en fenêtre %p] !\n",
		   __FILE__, __FUNCTION__, __LINE__, ANVA_1FACTEUR,  ARB_QuelleFenetre( arbres + ANVA_1FACTEUR, FAUX ) );
#endif
    /* Arbre déjà ouvert : terminer l'analyse en cours d'abord ! => on le fait passer au premier plan */
    FEN_premierplan( ARB_QuelleFenetre( arbres + ANVA_1FACTEUR, FAUX ) );
    return -2;
  }

  /* On réserve les informations pour l'ANVA 1 facteur */
  infos = (struct infos_ANVA_1F *) calloc( 1, sizeof( struct infos_ANVA_1F ) );
  if ( NULL == infos ) {
    ERR_MemoireInsuffisante( __FILE__, __FUNCTION__, __LINE__, sizeof( struct infos_ANVA_1F ),
			     "Impossible de r"e_AIGU"server les informations pour l'ANVA" );
    return -1;
  }
  infos->facteur.nom = NULL;
  infos->facteur.modalites = NULL;
  infos->moyennes = NULL;   infos->variances = NULL;
  infos->QT =  0.0;   infos->QF =  0.0;   infos->QR =  0.0;
  infos->ST = -1.0;   infos->SF = -1.0;   infos->SR = -1.0;
  infos->rapport = -2.0;   infos->proba_critique = -2.0; /* Avec les valeurs ci-dessus, correspond à une erreur ! */
  infos->type_modele = ANVA_MODELE_DROITE;  /* Régression linéaire par défaut */
  infos->QE =  0.0; infos->QM =  0.0;
  infos->SE = -1.0; infos->SM = -1.0;
  infos->modele.rl = NULL;
  infos->resultat = NULL;

  /* On commence par demander le nombre de modalités du facteur */
  /* a. On utilise la référence, si elle existe, pour faire une proposition sur le nombre de modalités... */
  if ( reference != NULL ) {
    Serie_Graphes *serie;

    serie = GRF_QuelleSerie( reference );
    if ( serie != NULL ) {
      Graphe *graphe;

      for ( graphe = serie->premier_graphe, i = 0 ;
	    graphe != NULL ;
	    graphe = graphe->suivant ) {
	if ( ( GRF_SERIE == graphe->genre_donnees ) && ( NULL == graphe->x ) ) {
	  i++;
	}
      }

      /* On suppose que chaque graphe de la série correspond aux valeurs pour une modalité ! */
      questions[ IA1_DEMANDE_NOMBRE_MODALITES ].valeur_initiale.entier_U = i;
      questions[ IA1_DEMANDE_NOMBRE_MODALITES ].valeur_defaut.entier_U = i;
    }
  }

  /* b. On pose réellement la question  */
  questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_initiale.chaine = "Facteur 1";

  retour = BI_DemandeMultiple( principale, &Q_NombreModalites, 2, questions );
  if ( retour != 1 ) {
    /* Annulation ou problème */
    /*     printf("%s : Annulation, code %d\n", __FUNCTION__, retour ); */
    free( infos ); infos = NULL;
    return -1;
  }

  /* c. On contrôle les valeurs indiquées */
  if ( questions[ IA1_DEMANDE_NOMBRE_MODALITES ].valeur_donnee.entier_U < 2 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Nombre de modalités absurde [%lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, questions[ IA1_DEMANDE_NOMBRE_MODALITES ].valeur_donnee.entier_U );
    free( questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine );
    questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine = NULL;
    free( infos ); infos = NULL;
    return -2;
  }

  /* On initialise le facteur */
  retour = ANVA_InitialiserFacteur( &(infos->facteur), questions[ IA1_DEMANDE_NOMBRE_MODALITES ].valeur_donnee.entier_U );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : Erreur en voulant préparer le facteur ! [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    free( questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine );
    questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine = NULL;
    free( infos ); infos = NULL;
    return -20;
  }
  ANVA_FACTEUR_Renommer( &(infos->facteur), questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine );
  free( questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine );
  questions[ IA1_DEMANDE_NOM_FACTEUR ].valeur_donnee.chaine = NULL;

  /* On initialise les modalités */
  infos->modalite_controle = -1;  /* Pas de modalité contrôle par défaut */
  for ( i = 0 ; i < infos->facteur.nbr_modalites ; i++ ) {
    char *nom_modalite;

    nom_modalite = TXT_genere_chaine( "Traitement %c", 'A' + i );
    ANVA_MODALITE_Renommer( infos->facteur.modalites + i, strdup( nom_modalite ) );
    free( nom_modalite ); nom_modalite = NULL;
  }
  if ( reference != NULL ) {
    /* On place les graphes utilisables dans les modalités gardées ! */
    Serie_Graphes *serie;
    unsigned long i_modalite;

    serie = GRF_QuelleSerie( reference );
    if ( serie != NULL ) {
      Graphe *graphe;

      /* On suppose que chaque graphe utilisable de la série correspond aux valeurs pour une modalité ! */
      for ( graphe = serie->premier_graphe, i_modalite = 0 ;
	    graphe != NULL ;
	    graphe = graphe->suivant ) {
	if ( ( GRF_SERIE == graphe->genre_donnees ) && ( NULL == graphe->x ) ) {
	  infos->facteur.modalites[ i_modalite ].valeurs = graphe;
	  i_modalite++;
	  if ( i_modalite >= infos->facteur.nbr_modalites ) break;  /* On a initialisé toutes les modalités */
	}
      }
    }

    /* On initialise les modalités... et les valeurs de l'ANVA à un facteur */
    ANVA_1F_Calculs( infos );

    /* On en affiche les résultats */
    I_A1F_InstallerANVA( infos );
  }

  /* On ouvre la fenêtre d'affichage des résultats & d'interface */
  fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, titre_fenetre, NULL,
				ANVA_1FACTEUR, BI_FEN_ARBRE_DROITE, I_A1F_PreparerArbre,
				I_A1F_Redessin, NULL, I_A1F_Clic, I_A1F_Messages, ( void (*)(Fenetre *, int, int, int, int) ) NULL, ( int (*)(Fenetre *) )NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'ouvrir la fenêtre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( infos ); infos = NULL;
    return -5;
  }
  FEN_Ouvrir( fenetre );

  /* On associe les infos à la fenêtre */
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, infos );

  /* On met les valeurs initiales dans l'arbre */
  I_A1F_InstallerANVA( infos );

  /* Calcul sans problème [on ne libère pas les infos : le sera en fermant la fenêtre !] */
  return 0;
}

/* ~-~-~-~-~-~-~-~ Clics sur l'affichage [à gauche dans la fenêtre] ~-~-~-~-~-~-~-~ */

static int I_A1F_Clic(Fenetre *fenetre, int bouton, int nombre_clics __attribute__((__unused__)), int etat_clavier __attribute__((__unused__)), int x, int y)
{
  struct infos_ANVA_1F *infos;
  int option_choisie;

  /* On récupère les infos */
  infos = (struct infos_ANVA_1F *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return BI_EVENEMENT_TRAITE; /* Aucune information actuellement disponible ! */

  /* On fait apparaître le bon menu */
  if ( 1 == bouton ) {
    if ( infos->facteur.proprietes & ANVA_FACTEUR_QUANTITATIF ) {
      /* Le clic ne sert à rien sur cet affichage [Message ?]*/
    } else {
      option_choisie = SS_menu_texte( options_modalite, I_A1F_NOMBRE_OPTIONS_MODALITES, titre_options_modalite, 0, ROOT );
      if ( option_choisie >= 0 ) {
	int retour __attribute__((__unused__));

	retour = I_A1F_ClicModalite( fenetre, infos, option_choisie, x, y );
      }
    }

    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Gestion des options associées à une modalité -------- */

static int I_A1F_ClicModalite(Fenetre *fenetre, struct infos_ANVA_1F *infos, unsigned int option, int x, int y)
{
  char *nv_nom;
  int reponse;
  unsigned int numero_modalite;
  struct infos_demande valeur_X = { BI_CHAMP_DEMANDE_REEL, "X", NULL, BI_EDITABLE_FORMULE | BI_EDITABLE_UNITE, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };

  /* On récupère la modalité qui nous intéresse : celle sur laquelle on a cliqué */
  numero_modalite = I_A1F_QuelleModalite( fenetre, infos, x, y );

  /* On choisit l'option */
  switch ( option )
  {
  case I_A1F_OPT_CHARGER_MODALITE:
    OPTION_A_FAIRE( "Charger les valeurs..." );
    break;

  case I_A1F_OPT_IMPORTER_MODALITE:
    OPTION_A_FAIRE( "Importer les valeurs..." );
    break;

  case I_A1F_OPT_SAUVER_MODALITE:
    OPTION_A_FAIRE( "Sauver les valeurs" );
    break;

  case I_A1F_OPT_EXPORTER_MODALITE:
    OPTION_A_FAIRE( "Exporter les valeurs..." );
    break;

  case I_A1F_OPT_CHANGER_VALEURS:
    if ( NULL == infos->facteur.modalites[ numero_modalite ].valeurs ) {
      BI_Champ_Demande question =
	{ BI_CHAMP_DEMANDE_CHAINE, txt_valeurs_modalite, NULL, BI_EDITABLE_MIN, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };
      int reponse;

      /* On demande les valeurs à mettre */
      reponse = BI_DemandeMultiple( fenetre, &Q_ValeursModalite, 1, &question );
      if ( 2 == reponse ) break;	/* On annule l'ajout de cette modalité */

      /* On crée le graphe des valeurs */
      infos->facteur.modalites[ numero_modalite ].valeurs = GRF_nv_ListeValeurs( GR_SSTYPE, question.valeur_donnee.chaine );

      /* Plus besoin de cette chaîne */
      free( question.valeur_donnee.chaine ); question.valeur_donnee.chaine = NULL;
    } else {
      OPTION_A_FAIRE( "Changer les valeurs" );
    }
    break;

  case I_A1F_OPT_VALEUR_EN_X:
    valeur_X.valeur_initiale.reel = infos->facteur.modalites[ numero_modalite ].valeur;
    reponse = BI_DemandeValeur( fenetre, &Q_ValeurX, &valeur_X );
    if ( reponse != 1 ) {
      /* Annulation -> pas besoin de redessiner */
      return BI_EVENEMENT_TRAITE;
    }
    infos->facteur.modalites[ numero_modalite ].valeur = valeur_X.valeur_donnee.reel;

    /* Peut changer les valeurs calculées, si en mode quantitatif -> recalcul au cas où ! */
    break;

  case I_A1F_OPT_CONTROLE:
    if ( numero_modalite == (unsigned) infos->modalite_controle ) {
      /* On redemande ==> on annule le marquage comme contrôle */
      infos->modalite_controle = -1;
    } else {
      infos->modalite_controle = numero_modalite;
    }

    /* Redessin seulement : ne change pas les calculs ! */
    FEN_redessin( fenetre, BI_RDS_UTIL );
    return BI_EVENEMENT_TRAITE;

  case I_A1F_OPT_RENOMMER:
    nv_nom = BI_Demande_Texte( fenetre, &Q_NomModalite, infos->facteur.modalites[ numero_modalite ].nom );
    if ( NULL == nv_nom ) {
      /* Annulation -> pas besoin de redessiner */
      return BI_EVENEMENT_TRAITE;
    }
    ANVA_MODALITE_Renommer( infos->facteur.modalites + numero_modalite, nv_nom );
    free( nv_nom ); nv_nom = NULL;

    /* Redessin seulement : ne change pas les calculs ! */
    FEN_redessin( fenetre, BI_RDS_UTIL );
    return BI_EVENEMENT_TRAITE;

  case I_A1F_OPT_ESTIMER_EFFET:
    OPTION_A_FAIRE( "Estimer l'effet" );
    break;

  default:
    AVERTISSEMENT( "Option inconnue [%u]", option );
    return BI_EVENEMENT_TRAITE;
  }

  /* On redessine et on recalcule, car en général quelque chose a changé ! */
  ANVA_1F_Calculs( infos );
  I_A1F_InstallerANVA( infos );

  /* On a fini et tout s'est bien passé */
  return BI_EVENEMENT_TRAITE;
}

/* ~-~-~-~-~-~-~-~ Gestion de l'affichage [à gauche dans la fenêtre] ~-~-~-~-~-~-~-~ */

/* -------- Indique quelle modalité se trouve à la position indiquée -------- */

static unsigned int I_A1F_QuelleModalite(Fenetre *fenetre, struct infos_ANVA_1F *infos, int x, int y __attribute__((__unused__)))
{
  unsigned int numero_modalite;

  if ( infos->facteur.proprietes & ANVA_FACTEUR_QUANTITATIF ) {
    OPTION_A_FAIRE( "Détection de la modalité en quantitatif" );
    numero_modalite = 0;
  } else {
    unsigned int larg_modalite;

    /* Deux cas faciles */
    if ( ( 0 == infos->facteur.nbr_modalites ) || ( NULL == infos->facteur.modalites ) ) return 0;
    if ( 1 == infos->facteur.nbr_modalites ) return 0;

    /* Quelle place est occupée par une modalité à l'écran ? */
    larg_modalite = fenetre->w_larg / infos->facteur.nbr_modalites;

    /* On en déduit la modalité */
    numero_modalite = x / larg_modalite;
  }

  /* On retourne ce que l'on a trouvé */
  (void) fprintf(stderr, "%s, %s l.%u : On trouve la modalité %u [X=%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, numero_modalite, x );

  return numero_modalite;
}

/* -------- Routines de redessin -------- */

static void I_A1F_Redessin(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  struct infos_ANVA_1F *infos;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut);

  /* On récupère les infos */
  infos = (struct infos_ANVA_1F *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return;	/* Aucune information actuellement disponible ! */

  if ( infos->facteur.proprietes & ANVA_FACTEUR_QUANTITATIF ) {
    I_A1F_Redessin_Quanti( fenetre, infos );
  } else {
    I_A1F_Redessin_Quali( fenetre, infos );
  }
}

static void I_A1F_Redessin_BornesY(struct infos_ANVA_1F *infos, double *y_min, double *y_max)
{
  unsigned long i = 0;

  /* On se place sur la première modalité avec des valeurs */
  while ( NULL == infos->facteur.modalites[ i ].valeurs ) {
    i++;

    if ( i >= infos->facteur.nbr_modalites ) {
      /* Aucune modalité ne contient de valeur : pas de tracé ! */
      *y_min = -1.0;
      *y_max =  1.0;
      return;
    }
  }

  /* Ici, i est l'index de la première modalité contenant des valeurs */
  *y_min = infos->facteur.modalites[ i ].valeurs->y_min;
  *y_max = infos->facteur.modalites[ i ].valeurs->y_max;

  for (  ; i < infos->facteur.nbr_modalites ; i++ ) {
    double min, max;

    if ( NULL == infos->facteur.modalites[ i ].valeurs ) continue;

    /* La modalité a des valeurs : on adapte */
    min = infos->facteur.modalites[ i ].valeurs->y_min;
    if ( min < *y_min ) *y_min = min;

    max = infos->facteur.modalites[ i ].valeurs->y_max;
    if ( max > *y_max ) *y_max = max;
  }

  /* Contrôle d'intervalle non nul */
  if ( *y_min == *y_max ) {
    if ( 0.0 == *y_min ) {
      *y_min = -0.1;
      *y_max =  0.1;
    } else {
      double valeur;

      valeur = *y_min;

      *y_min = valeur * 0.9;
      *y_max = valeur * 1.1;
    }
  }
}

static void I_A1F_Redessin_Quali(Fenetre *fenetre, struct infos_ANVA_1F *infos)
{
  InfosAxe axe_Y;
  unsigned int lg_modalite, ht_modalite;
  unsigned int i;		/* Compteur sur les modalités */
  double echelle_Y;

  /* Hauteur réservée aux valeurs : 4/5 de la fenêtre */
  ht_modalite = ( fenetre->w_haut * 4 ) / 5;

  /* On place donc le séparateur */
  TRC_Separateur( fenetre, HORIZONTAL, 0, ht_modalite, fenetre->w_larg );

  if ( 0 == infos->facteur.nbr_modalites ) {
    TRC_Z_Message( fenetre, 0, 0, fenetre->w_larg, ht_modalite, "Aucune modalité définie" );
    return;
  }

  /* Largeur occupée par une modalité */
  lg_modalite = fenetre->w_larg / infos->facteur.nbr_modalites;
  if ( lg_modalite < 10 ) {
    lg_modalite = 10;		/* Au moins 10 pixels de large pour voir quelque chose ! */
  }

  /* On construit l'axe des Y [bornes] */
  AXE_Initialiser( &axe_Y, VRAI );
  axe_Y.proprietes &= ~AXE_VISIBLE; /* L'axe Y=0 n'a aucun intérêt particulier ici */
  I_A1F_Redessin_BornesY( infos, &(axe_Y.min), &(axe_Y.max) );

  /* On calcule l'échelle [dans le sens positif] */
  echelle_Y = AXE_Echelle( &axe_Y, 2, (ht_modalite - 1) - 2 );

  /* On dessine chaque modalité, avec son axe */
  for ( i = 0 ; i < infos->facteur.nbr_modalites ; i++ ) {
    /* On redessine la modalité */
    I_A1F_RedessinModalite_Quali( fenetre, infos->facteur.modalites + i, &axe_Y, echelle_Y,
				  i * lg_modalite, 0, lg_modalite, ht_modalite );
    if ( i == (unsigned int) infos->modalite_controle ) {
      /* Modalité de contrôle : on met un cadre rouge autour */
      TRC_Couleur( ROUGE );
      TRC_Style( BI_CONTINU );
      TRC_Epaisseur( 3 );
      TRC_Rectangle( fenetre, i * lg_modalite + 2, 1, lg_modalite - 4, ht_modalite - 3 );
      TRC_Epaisseur( 1 );
    }

    /* On affiche les informations textuelles associées */
    I_A1F_RedessinModalite_Texte( fenetre, infos->facteur.modalites + i,
				  i * lg_modalite, ht_modalite, lg_modalite, fenetre->w_haut - ht_modalite );

    /* Fin : On met le séparateur */
    TRC_Separateur( fenetre, VERTICAL, lg_modalite * i, 0, fenetre->w_haut );
  }

  /* On dessine la moyenne globale */
  TRC_Style( BI_TIRET_COURT );
  TRC_Couleur( BLEU );

  TRC_LigneH( fenetre, 0, TRC_ConvertitY_PvE( infos->moyenne, axe_Y.min, ht_modalite - 3, echelle_Y ), fenetre->w_larg );
}

static void I_A1F_RedessinModalite_Quali(Fenetre *fenetre, struct modalite *modalite, InfosAxe *axe_Y, double echelle_Y,
					 int x_HG, int y_HG, unsigned int largeur, unsigned int hauteur)
{
  int x_axe, x_valeurs, y_BG;

  /* Position des objets */
  /* 1. Position de l'axe : au milieu */
  x_axe = x_HG + largeur / 2;

  /* 2. Position des valeurs individuelles : au milieu de la 2e moitié = au 3/4 ! */
  x_valeurs = x_HG + ( 3 * largeur ) / 4;

  /* 3. Position du coin inférieur gauche [y_min] */
  y_BG = y_HG + (hauteur - 1) - 2;

  /* Préparation du dessin */
  TRC_Style( BI_CONTINU );
  TRC_Couleur( NOIR );		/* Mettre en rouge l'active [?] */
  TRC_Epaisseur( 1 );

  /* On trace l'axe */
  TRC_LigneV( fenetre, x_axe, y_HG + 2, hauteur - 4 );
  TRC_AxeGradue( fenetre, x_axe, y_BG, x_axe, y_HG + 2, axe_Y, 2, 10 );

  /* On trace les valeurs */
  if ( NULL == modalite->valeurs ) {
    /* Comme on aura souvent plus de place en hauteur, le texte d'avertissement est tourné ! */
    TRC_Texte( fenetre, x_HG + largeur / 2, y_HG + hauteur / 2, BI_TEXTE_CENTRER_H, BI_TEXTE_CENTRER_V, 90, "Aucune valeur pour cette modalit"e_AIGU );
  } else {
    I_A1F_RedessinModalite_Quanti( fenetre, modalite, y_BG, axe_Y, echelle_Y, x_valeurs );
  }
}

static void I_A1F_RedessinModalite_Quanti(Fenetre *fenetre, struct modalite *modalite, int y_BG, InfosAxe *axe_Y, double echelle_Y, int x_valeurs)
{
  int y_valeur;
  unsigned long i;
  Graphe *graphe;

  /* On récupère le graphe */
  graphe = modalite->valeurs;
  if ( NULL == graphe ) {
    return;
  }

  /* On trace les valeurs */
  for ( i = 0 ; i < graphe->nombre_point ; i++ ) {
    y_valeur = TRC_ConvertitY_PvE( graphe->y[ i ], axe_Y->min, y_BG, echelle_Y );
    TRC_Marque( fenetre, x_valeurs, y_valeur, BI_MARQUE_ASTERISQUE, 6 );
  }

  /* On place la moyenne */
  TRC_Couleur( ROUGE );
  y_valeur = TRC_ConvertitY_PvE( modalite->moyenne, axe_Y->min, y_BG, echelle_Y );
  TRC_Marque( fenetre, x_valeurs, y_valeur, BI_MARQUE_ASTERISQUE, 6 );

  /* On place l'erreur sur la moyenne */
  if ( modalite->valeurs->nombre_point > 1 ) {
    double sigma_moyenne;
    int y_debut, y_fin;

    sigma_moyenne  = modalite->SCR / ( modalite->valeurs->nombre_point - 1 );
    sigma_moyenne /= modalite->valeurs->nombre_point;
    sigma_moyenne  = sqrt( sigma_moyenne );

    y_debut = TRC_ConvertitY_PvE( modalite->moyenne - sigma_moyenne, axe_Y->min, y_BG, echelle_Y );
    y_fin   = TRC_ConvertitY_PvE( modalite->moyenne + sigma_moyenne, axe_Y->min, y_BG, echelle_Y );

    TRC_Ligne( fenetre, x_valeurs, y_debut, x_valeurs, y_fin );
  }
}

static void I_A1F_RedessinModalite_Texte(Fenetre *fenetre, struct modalite *modalite,
					 int x_HG, int y_HG, unsigned int largeur, unsigned int hauteur)
{
  int x_texte, y_texte;
  unsigned int larg_texte __attribute__((__unused__)), haut_texte;
  int attributs_texte[ 10 ];
  char *texte;

  /* Dimensions d'un caractère */
  vqt_attributes( _PRG_station->handle, attributs_texte);
  larg_texte = attributs_texte[ 6 ];
  haut_texte = attributs_texte[ 7 ] + 2; /* Interligne de 2 pixels entre les textes */

  /* Le nom de la modalité, en vertical sur deux fois la hauteur du texte */
  TRC_Couleur( BLEU );
  x_texte = x_HG + haut_texte;
  y_texte = y_HG + hauteur / 2;
  TRC_Texte( fenetre, x_texte, y_texte, BI_TEXTE_CENTRER_H, BI_TEXTE_CENTRER_V, 90, modalite->nom );

  /* Les informations numériques */
  TRC_Couleur( NOIR );
  x_texte = x_HG + 2 * haut_texte;
  y_texte = y_HG + 2;

  /* 1re ligne : le nombre de valeurs */
  if ( NULL == modalite->valeurs ) {
    TRC_Couleur( ROUGE );
    TRC_Texte( fenetre, x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, "Aucune valeur" );
    return;
  }
  texte = TXT_genere_chaine( "%u valeurs", modalite->valeurs->nombre_point );
  TRC_Z_Texte( fenetre, x_HG, y_HG, largeur, hauteur,
	       x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  free( texte ); texte = NULL;
  y_texte += haut_texte + 2;

  /* 2e ligne : valeur moyenne */
  texte = TXT_genere_chaine( "m  = %*f", modalite->moyenne, GRF_QuelleUnite( modalite->valeurs, AXE_Y ) );
  TRC_Z_Texte( fenetre, x_HG, y_HG, largeur, hauteur,
	       x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  free( texte ); texte = NULL;
  y_texte += haut_texte + 2;

  /* 3e ligne : écart-type */
  if ( modalite->valeurs->nombre_point > 1 ) {
    texte = TXT_genere_chaine( "s* = %*f", sqrt( modalite->SCR / (double) (modalite->valeurs->nombre_point - 1) ),
			       GRF_QuelleUnite( modalite->valeurs, AXE_Y ) );
  } else {
    texte = TXT_genere_chaine( "s* = %*f", 0.0, GRF_QuelleUnite( modalite->valeurs, AXE_Y ) );
  }
  TRC_Z_Texte( fenetre, x_HG, y_HG, largeur, hauteur,
	       x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  free( texte ); texte = NULL;
  y_texte += haut_texte + 2;

  /* 4e ligne : minimum */
  texte = TXT_genere_chaine( "min = %*f", modalite->valeurs->y_min,
			     GRF_QuelleUnite( modalite->valeurs, AXE_Y ) );
  TRC_Z_Texte( fenetre, x_HG, y_HG, largeur, hauteur,
	       x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  free( texte ); texte = NULL;
  y_texte += haut_texte + 2;

  /* 5e ligne : maximum */
  texte = TXT_genere_chaine( "max = %*f", modalite->valeurs->y_max,
			     GRF_QuelleUnite( modalite->valeurs, AXE_Y ) );
  TRC_Z_Texte( fenetre, x_HG, y_HG, largeur, hauteur,
	       x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  free( texte ); texte = NULL;
  y_texte += haut_texte + 2;

  /* 6e ligne : estimation de l'effet */
  texte = TXT_genere_chaine( ALPHA" = %*f", modalite->effet,
			     GRF_QuelleUnite( modalite->valeurs, AXE_Y ) );
  TRC_Z_Texte( fenetre, x_HG, y_HG, largeur, hauteur,
	       x_texte, y_texte, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, texte );
  free( texte ); texte = NULL;
  y_texte += haut_texte + 2;
}

/* -------- Redessin dans le cas d'un facteur quantitatif : modèle Y=f(X) utilisable ! -------- */

static void I_A1F_Redessin_Quanti(Fenetre *fenetre, struct infos_ANVA_1F *infos)
{
  InfosRegion2D region;
  double echelle_X, echelle_Y;
  int retour, y_barre, y_residus;
  unsigned int i, ht_modele __attribute__((__unused__));

  /* On sépare la zone en 2 : 2/3 pour le modèle, 1/3 pour les résidus */
  y_residus = ( fenetre->w_haut * 2 ) / 3;
  ht_modele = y_residus - 1;
  TRC_Separateur( fenetre, HORIZONTAL, 0, y_residus, fenetre->w_larg );

  /* On initialise la zone 2D */
  region.proprietes = 0;
  retour = TRC_Z2D_Initialiser( &region, fenetre, NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant initialiser la zone de dessin\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }
  region.proprietes |= CADRE2D_AVEC_CADRE;
  AXE_Initialiser( &(region.axe_X), FAUX );
  region.axe_X.proprietes &= ~AXE_AVEC_GRILLE;
  AXE_Initialiser( &(region.axe_Y), VRAI );
  region.axe_Y.proprietes &= ~AXE_AVEC_GRILLE;

  /* On adapte pour prévoir la place des légendes */
  region.x_BG += 50;
  region.y_BG  = y_residus - 20;
  region.largeur -= 52;		/* On garde deux pixels de marge à droite ! */
  region.hauteur = y_residus - 20;

  /* On cherche les bornes en X */
  region.axe_X.min = infos->facteur.modalites[ 0 ].valeur;
  region.axe_X.max = region.axe_X.min;
  for ( i = 1 ; i < infos->facteur.nbr_modalites ; i++ ) {
    /* Adaptation en X */
    if ( infos->facteur.modalites[ i ].valeur < region.axe_X.min ) region.axe_X.min = infos->facteur.modalites[ i ].valeur;
    if ( infos->facteur.modalites[ i ].valeur > region.axe_X.max ) region.axe_X.max = infos->facteur.modalites[ i ].valeur;
  }

  /* On adapte les X pour que les valeurs ne soient pas sur les bords : 1% de la différence totale, de chaque côté */
  region.axe_X.min -= ( region.axe_X.max - region.axe_X.min ) * 0.01;
  region.axe_X.max += ( region.axe_X.max - region.axe_X.min ) * 0.01;

  /* On adapte les Y pour être sûr que le modèle soit partout visible */
  I_A1F_Redessin_BornesY( infos, &(region.axe_Y.min), &(region.axe_Y.max) );
  if ( infos->resultat != NULL ) {
    if ( infos->resultat->y_min < region.axe_Y.min ) region.axe_Y.min = infos->resultat->y_min;
    if ( infos->resultat->y_max > region.axe_Y.max ) region.axe_Y.max = infos->resultat->y_max;
  }

  /* On trace le cadre & tout ça */
  region.couleur_fond = -1;	/* Transparent */
  region.couleur_bord = NOIR;	/* Mais bord noir */
  TRC_DessineRegion2D( fenetre, &region );

  /* On calcule les échelles */
  retour = TRC_Z2D_Echelles( &region, 0, &echelle_X, &echelle_Y );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur dans le calcul des échelles, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* On place les modalités */
  for ( i = 0 ; i < infos->facteur.nbr_modalites ; i++ ) {
    int x_modalite;

    TRC_Style( BI_CONTINU );
    TRC_Epaisseur( 1 );
    TRC_Couleur( NOIR );		/* Mettre en rouge l'active [?] */

    x_modalite = TRC_ConvertitX_PvE( infos->facteur.modalites[ i ].valeur, region.axe_X.min, region.x_BG, echelle_X );
    I_A1F_RedessinModalite_Quanti( fenetre, infos->facteur.modalites + i, region.y_BG, &(region.axe_Y), echelle_Y, x_modalite );
  }

  /* On place la moyenne globale, horizontale : comparaison sur la nécessité d'un modèle... */
  TRC_Style( BI_TIRET_COURT );
  TRC_Couleur( BLEU );

  y_barre = TRC_ConvertitY_PvE( infos->moyenne, region.axe_Y.min, region.y_BG, echelle_Y );
  TRC_LigneH( fenetre, region.x_BG, y_barre, region.largeur );

  /* On trace le modèle */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->resultat, BLEU, BI_CONTINU, BI_MARQUE_ASTERISQUE, AXE_Y, NULL, NULL );

  /* B. On trace les résidus */
  region.y_BG    = fenetre->w_haut - 20;
  region.hauteur = (fenetre->w_haut - y_residus) - 20;

  /* On garde la même échelle, mais centrée sur la moyenne [à améliorer ensuite !] */
  region.axe_Y.min -= infos->moyenne;
  region.axe_Y.max -= infos->moyenne;

  /* On trace le cadre & tout ça */
  TRC_DessineRegion2D( fenetre, &region );

  /* On calcule les échelles */
  retour = TRC_Z2D_Echelles( &region, 0, &echelle_X, &echelle_Y );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur dans le calcul des échelles, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

}

/* ~-~-~-~-~-~-~-~ Options de l'interface [arbre ou menu contextuel] ~-~-~-~-~-~-~-~ */

static void I_A1F_AjouterModalite(Fenetre *fenetre, struct infos_ANVA_1F *infos, unsigned int nombre_modalites, unsigned int modalite_avant)
{
  BI_Champ_Demande questions[ 3 ] =
    { { BI_CHAMP_DEMANDE_CHAINE, txt_nom_modalite    , NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
      { BI_CHAMP_DEMANDE_CHAINE, txt_valeurs_modalite, NULL, BI_EDITABLE_MIN, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
      { BI_CHAMP_DEMANDE_REEL  , txt_X_modalite      , NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } } };
  int reponse, retour;
  unsigned int i = 1;
  struct modalite *modalite;

  if ( 0 == nombre_modalites ) return;
  if ( modalite_avant >= infos->facteur.nbr_modalites ) {
    if ( ( infos->facteur.nbr_modalites > 0 ) || ( modalite_avant > 0 ) ) {
      AVERTISSEMENT( "Modalité avant [%u] absurde, max. %u", modalite_avant, nombre_modalites - 1 );
    }
    modalite_avant = 0;
  }
  do {
    /* On libère les réponses de la modalité précédente */
    free( questions[ 0 ].valeur_donnee.chaine ); questions[ 0 ].valeur_donnee.chaine = NULL;
    free( questions[ 1 ].valeur_donnee.chaine ); questions[ 1 ].valeur_donnee.chaine = NULL;

    /* On demande le nom & les valeurs de la modalité */
    reponse = BI_DemandeMultiple( fenetre, &Q_ValeursModalite, 3, questions );
    if ( 2 == reponse ) break;	/* On annule l'ajout de cette modalité */

    /* On crée la modalité correspondante [après la modalité indiquée, pour l'instant] */
    retour = ANVA_FACTEUR_AjouterModalite( &(infos->facteur), modalite_avant );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en voulant créer la modalité, code %d\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      break;
    }
    if ( 1 == infos->facteur.nbr_modalites ) {
      modalite = infos->facteur.modalites;
    } else {
      modalite = infos->facteur.modalites + modalite_avant + 1;
    }

    /* On lui donne le nom indiqué */
    ANVA_MODALITE_Renommer( modalite, questions[ 0 ].valeur_donnee.chaine );

    /* On crée le graphe des valeurs */
    modalite->valeurs = GRF_nv_ListeValeurs( GR_SSTYPE, questions[ 1 ].valeur_donnee.chaine );

    /* On mémorise la valeur en X */
    modalite->valeur = questions[ 2 ].valeur_donnee.reel;

    /* On passe à la modalité suivante */
    i++;
  } while ( i < nombre_modalites );

  /* On a terminé : on libère les dernières réponses */
  free( questions[ 0 ].valeur_donnee.chaine ); questions[ 0 ].valeur_donnee.chaine = NULL;
  free( questions[ 1 ].valeur_donnee.chaine ); questions[ 1 ].valeur_donnee.chaine = NULL;

  /* Et on force un redessin & les recalculs */
  ANVA_1F_Calculs( infos );
  I_A1F_InstallerANVA( infos );
/*   FEN_redessin( fenetre, BI_RDS_UTIL ); */
}

/* -------- Appeler une méthode de comparaison à un témoin -------- */

static void I_A1F_OptionsTemoin(struct infos_ANVA_1F *infos)
{
  int sens = 0;	/* 0 : bilatéral ; > 0 : unilatéral > témoin ; < 0 : unilatéral < témoin */
  int retour;
  double alpha_global;
  unsigned int methode, approximation;
  unsigned long i;

  /* Quelques contrôles */
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'informations associées !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  if ( infos->facteur.nbr_modalites < 2 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas assez de modalités, test sans signification\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  if ( infos->modalite_controle < 0 ) {
    (void) BI_ERR_alerte( NULL, &ERR_AucunTemoin, VRAI );
    (void) fprintf(stderr, "%s, %s l.%u : Aucun contrôle n'a été indiqué !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* On vérifie qu'une analyse n'a pas déjà été faite */
  if ( infos->comparaisons != NULL ) {
    int reponse;

    reponse = BI_ERR_alerte( NULL, &Q_AnalyseDejaFaite, VRAI, infos->nombre_comparaisons );
    if ( 2 == reponse ) return;

    /* On efface les comparaisons déjà faites ! */
    STAT_EffacerComparaisons( infos->nombre_comparaisons, infos->comparaisons );
    infos->nombre_comparaisons = 0;
    infos->comparaisons = NULL;
  }

  /* On lit le risque global souhaité */
  alpha_global = OBJ_LireNombre( ANVA_1FACTEUR, A1F_ALPHA_GLOBAL, NULL ) / 100.0;  /* Entré en % ! */

  /* On demande les options et paramètres pour faire le test */
  retour = I_STAT_CMP_OptionsTemoin( &methode, &approximation, &sens, &alpha_global );
  if ( retour != 0 ) {
    AVERTISSEMENT( "Annulation [code %d]", retour );
    return;
  }

  /* On fait l'étude statistique suivant les critères demandés */
  retour = ANVA_1F_ComparerAuTemoin( infos, methode, approximation, sens, alpha_global );
  if ( retour != 0 ) {
    AVERTISSEMENT( "Problème dans ANVA_1F_ComparerAuTemoin [code %d]", retour );
    return;
  }

  /* On affiche les résultats [en attendant mieux !] */
  for ( i = 0 ; i < infos->nombre_comparaisons ; i++ ) {
    printf("IC pour %s - %s : [%g, %g]\n", infos->comparaisons[ i ].modalites[ 0 ]->nom, infos->comparaisons[ i ].modalites[ 1 ]->nom, 
	   infos->comparaisons[ i ].ic_bas, infos->comparaisons[ i ].ic_haut);
  }

}

static void I_A1F_OptionsVariance(unsigned int arbre __attribute__((__unused__)), unsigned int objet __attribute__((__unused__)), unsigned long action)
{
  switch ( action )
  {
  case I_A1F_VAR_BARTLETT:
    OPTION_A_FAIRE( "Test de Bartlett" );
    break;

  case I_A1F_VAR_LEVENE_MOYENNE:
    OPTION_A_FAIRE( "Test de Levène / moyenne" );
    break;

  case I_A1F_VAR_LEVENE_MEDIANE:
    OPTION_A_FAIRE( "Test de Levène / médiane" );
    break;

  case I_A1F_VAR_LEVENE_M_TRONQUEE:
    OPTION_A_FAIRE( "Test de Levène / moyenne tronquée" );
    break;

  default:
    AVERTISSEMENT( "Option inconnue [%lu]", action );
  }
}

/* ~-~-~-~-~-~-~-~ Gestion de l'arbre [à droite dans la fenêtre] ~-~-~-~-~-~-~-~ */

static int I_A1F_GererArbre(unsigned int numero_objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int nouvel_etat, retour;
  struct infos_ANVA_1F *infos;

  /* On récupère les informations */
  infos = (struct infos_ANVA_1F *) ARB_LireComplement( ANVA_1FACTEUR );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'informations associées à l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_TRAITE;
  }

  switch ( numero_objet )
  {
  case A1F_AJOUTER:
    I_A1F_AjouterModalite( ARB_QuelleFenetre( arbres + ANVA_1FACTEUR, FAUX ), infos, 1, infos->facteur.nbr_modalites - 1 );
    return BI_EVENEMENT_TRAITE;

  case A1F_OTER:
    OPTION_A_FAIRE( "Ôter une modalité" );
    return BI_EVENEMENT_TRAITE;

  case A1F_EXPORTER:
    OPTION_A_FAIRE( "Exporter les résultats [tableau, figure]" );
    return BI_EVENEMENT_TRAITE;

  case A1F_AVEC_RL:
    nouvel_etat = OBJ_choisit( ANVA_1FACTEUR, A1F_AVEC_RL, ECHANGE );
    if ( VRAI == nouvel_etat ) {
      /* X est une variable quantitative */
      infos->facteur.proprietes |= ANVA_FACTEUR_QUANTITATIF;
      I_A1F_ActiverQuanti( VRAI );
      retour = ANVA_1F_PreparerModele( infos, infos->type_modele );
      if ( retour < 0 ) {	/* Erreur grave */
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans la préparation du modèle, code %d\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	/* On repasse donc en mode qualitatif */
	infos->facteur.proprietes &= ~ANVA_FACTEUR_QUANTITATIF; 
	I_A1F_ActiverQuanti( FAUX );

	/* Et on a fini, sans nécessité de redessin */
	return BI_EVENEMENT_TRAITE;
      }
      retour = ANVA_1F_CalculerModele( infos );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur dans le calcul du modèle, code %d\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	/* Peut être normal tant que le modèle n'a pas été configuré correctement ! */
      }
      /* On affiche les résultats ! */
      I_A1F_InstallerANVA( infos );
    } else {
      /* X est qualitative */
      infos->facteur.proprietes &= ~ANVA_FACTEUR_QUANTITATIF; 
      I_A1F_ActiverQuanti( FAUX );
    }
    FEN_redessin( ARB_QuelleFenetre( arbres + ANVA_1FACTEUR, FAUX ), BI_RDS_UTIL );
    return BI_EVENEMENT_TRAITE;

  case A1F_TEST_TEMOIN:
    OBJ_choisit( ANVA_1FACTEUR, numero_objet, FORCE_OUI );
    I_A1F_OptionsTemoin( infos );
    OBJ_choisit( ANVA_1FACTEUR, numero_objet, FORCE_NON );
    FEN_redessin( ARB_QuelleFenetre( arbres + ANVA_1FACTEUR, FAUX ), BI_RDS_UTIL );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Activer/désactiver les options pour le quantitatif -------- */

static void I_A1F_ActiverQuanti(boolean activer)
{
  OBJ_active( ANVA_1FACTEUR, A1F_MODELE_QUANT, activer );

  OBJ_active( ANVA_1FACTEUR, A1F_Q_MODELE  , activer );
  OBJ_active( ANVA_1FACTEUR, A1F_DDL_MODELE, activer );

  OBJ_active( ANVA_1FACTEUR, A1F_Q_ECART  , activer );
  OBJ_active( ANVA_1FACTEUR, A1F_DDL_ECART, activer );

  OBJ_active( ANVA_1FACTEUR, A1F_SM, activer );
  OBJ_active( ANVA_1FACTEUR, A1F_F_MODELE, activer );
  OBJ_active( ANVA_1FACTEUR, A1F_P_MODELE, activer );

  OBJ_active( ANVA_1FACTEUR, A1F_S2_ECART, activer );
  OBJ_active( ANVA_1FACTEUR, A1F_F_ECART , activer );
  OBJ_active( ANVA_1FACTEUR, A1F_P_ECART , activer );

  OBJ_active( ANVA_1FACTEUR, A1F_PARAMETRES, activer );
}

/* -------- Installer tous les paramètres dans l'arbre -------- */

static void I_A1F_InstallerANVA(struct infos_ANVA_1F *infos)
{
  /* On commence par installer le facteur */
  I_A1F_InstallerFacteur( &(infos->facteur) );

  /* On installe les valeurs numériques */
  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_QT, infos->QT, 0.0 );
  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_QR, infos->QR, 0.0 );
  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_QFACT, infos->QF, 0.0 );

  OBJ_MettreEntier_U( ANVA_1FACTEUR, A1F_DDL_FACT, infos->ddl_F );
  OBJ_MettreEntier_U( ANVA_1FACTEUR, A1F_DDL_RES , infos->ddl_R );
  OBJ_MettreEntier_U( ANVA_1FACTEUR, A1F_DDL_TOT , infos->ddl_T );

  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_ST, infos->ST, 0.0 );
  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_SR, infos->SR, 0.0 );
  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_SF, infos->SF, 0.0 );

  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_RAPPORT   , infos->rapport, 0.0 );
  OBJ_MettreNombre( ANVA_1FACTEUR, A1F_P_CRITIQUE, infos->proba_critique, 0.0 );

  if ( infos->facteur.proprietes & ANVA_FACTEUR_QUANTITATIF ) {
    /* On installe les résultats de la modélisation */
    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_Q_MODELE, infos->QM, 0.0 );
    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_Q_ECART , infos->QE, 0.0 );

    OBJ_MettreEntier_U( ANVA_1FACTEUR, A1F_DDL_MODELE, infos->ddl_M );
    OBJ_MettreEntier_U( ANVA_1FACTEUR, A1F_DDL_ECART , infos->ddl_E );

    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_SM, infos->SM, 0.0 );
    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_F_MODELE, infos->RM, 0.0 );
    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_P_MODELE, infos->pc_M, 0.0 );

    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_S2_ECART, infos->SE, 0.0 );
    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_F_ECART , infos->RE, 0.0 );
    OBJ_MettreNombre( ANVA_1FACTEUR, A1F_P_ECART , infos->pc_E, 0.0 );
  }
}

static void I_A1F_InstallerFacteur(struct anva_facteur *facteur)
{
  /* Le nombre de modalités */
  OBJ_MettreEntier_U( ANVA_1FACTEUR, A1F_NBR_MODALITE, facteur->nbr_modalites );

  /* Le facteur est-il à effets fixes ou aléatoires ? */
  OBJ_choisit( ANVA_1FACTEUR, A1F_FIXE, ( 0 == ( facteur->proprietes & ANVA_FACTEUR_ALEATOIRE ) ) );

  /* Le facteur est-il quantitatif ? */
  if ( facteur->proprietes & ANVA_FACTEUR_QUANTITATIF ) {
    OBJ_choisit( ANVA_1FACTEUR, A1F_AVEC_RL, FORCE_OUI );
  } else {
    OBJ_choisit( ANVA_1FACTEUR, A1F_AVEC_RL, FORCE_NON );
  }
}

/* -------- Préparation de l'arbre -------- */

static int I_A1F_PreparerArbre(unsigned int numero_arbre)
{
  int retour;

  /* On installe les sous-menus */
  OBJ_ChangerType( numero_arbre, A1F_OPT_VARIANCE, G_SSMENU );
  OBJ_MettreMenu ( numero_arbre, A1F_OPT_VARIANCE, I_A1F_NOMBRE_OPTIONS_VARIANCES, options_variance, titre_options_variance, I_A1F_OptionsVariance );
  OBJ_RemplacerTexte( numero_arbre, A1F_OPT_VARIANCE, titre_variance, FAUX );

/*   OBJ_ChangerType( numero_arbre, A1F_TEST_TEMOIN, G_SSMENU ); */
/*   OBJ_MettreMenu ( numero_arbre, A1F_TEST_TEMOIN, I_A1F_NOMBRE_OPTIONS_TEMOIN, options_temoin, titre_options_temoin, I_A1F_OptionsTemoin ); */
/*   OBJ_RemplacerTexte( numero_arbre, A1F_TEST_TEMOIN, titre_temoin, FAUX ); */

  OPTION_A_FAIRE( "Installer les sous-menus" );

  /* On commence par initialiser l'arbre [pas de contrôle, car normalement appelé uniquement s'il faut le faire !] */
  retour = ARB_Initialiser( numero_arbre, NULL, I_A1F_GererArbre, NULL );
  if ( retour != 0 ) return retour;

  /* On déplace les objets pour enjoliver */
  OBJ_deplace( numero_arbre, A1F_TXT_QR   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_QR       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_DDL_RES  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_TXT_DDL_R, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( numero_arbre, A1F_TXT_QT   , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_QT, 0    , +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_DDL_TOT  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_TXT_DDL_T, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( numero_arbre, A1F_TXT_SR    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_SR        , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_TXT_P     , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_P_CRITIQUE, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( numero_arbre, A1F_TXT_ST, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_ST    , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( numero_arbre, A1F_TXT_VARIANCE, 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_OPT_VARIANCE, 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_TXT_DISTRIB , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( numero_arbre, A1F_OPT_DISTRIB , 0, -1, DEPLACE_PIXELS );

#ifdef AVEC_TRADUCTION_
  OPTION_A_FAIRE( "Traduction de l'arbre" );
#endif

  return 0;
}

/* ~-~-~-~-~-~-~-~ Gestion de la fenêtre elle-même ~-~-~-~-~-~-~-~ */

/* -------- Gestion du message de fermeture -------- */

#define I_A1F_QUITTER_ISOLE_PERDRE      1  /* On veut perdre cette modalité isolée */
#define I_A1F_QUITTER_ISOLE_GARDER      2  /* On veut garder cette modalité : placer dans une nouvelle série */
#define I_A1F_QUITTER_ISOLE_TOUT_PERDRE 3  /* On veut tout perdre de cette modalité */
#define I_A1F_QUITTER_ISOLE_TOUT_GARDER 4  /* On veut tout garder */

static int I_A1F_Messages(Fenetre *fenetre, Message *message)
{
  int reponse = -1;
  unsigned int i;
  struct infos_ANVA_1F *infos;
  struct serie *serie_modalites = NULL;

  if ( NULL == message ) return BI_EVENEMENT_ERREUR;
  if ( message->type != WM_CLOSED ) return BI_EVENEMENT_NON_TRAITE;

  /* On récupère les infos de la fenêtre */
  infos = (struct infos_ANVA_1F *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) {
    /* Aucune information actuellement ==> fermeture directe */
    FEN_Detruire( fenetre );
    return BI_EVENEMENT_FEN_FERMEE;
  }
  
  /* On regarde si certaines modalités sont en graphes isolés */
  for ( i = 0 ; i < infos->facteur.nbr_modalites ; i++ ) {
    Graphe *valeurs;

    valeurs = infos->facteur.modalites[ i ].valeurs;
    if ( NULL == valeurs ) continue;  /* Modalité sans valeurs ! */
    if ( VRAI == GRF_EstIsole( valeurs ) ) {
      if ( ( reponse != I_A1F_QUITTER_ISOLE_TOUT_PERDRE ) && ( reponse != I_A1F_QUITTER_ISOLE_TOUT_GARDER ) ) {
	reponse = BI_ERR_alerte( fenetre, &Q_ModaliteIsolee, VRAI, valeurs->nom );
      }

      if ( ( I_A1F_QUITTER_ISOLE_TOUT_PERDRE == reponse ) || ( I_A1F_QUITTER_ISOLE_PERDRE == reponse ) ) {
	GRF_eff_isole( valeurs ); valeurs = NULL;
      } else if ( ( I_A1F_QUITTER_ISOLE_GARDER == reponse ) || ( I_A1F_QUITTER_ISOLE_TOUT_GARDER == reponse ) ) {
	if ( NULL == serie_modalites ) {
	  OPTION_A_FAIRE( "Demander le nom de la série pour sa création" );
	}

	/* On insère le graphe dans cette série */
	GRF_InsererDansSerie( valeurs, serie_modalites, NULL );
      }
    }

    /* On dissocie les valeurs de la modalité */
    infos->facteur.modalites[ i ].valeurs = NULL;
  }

  OPTION_A_FAIRE( "Libération propre des modalités et des infos !" );

  /* On dissocie les infos de la fenêtre */
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

  /* On ferme la fenêtre */
  FEN_Detruire( fenetre );

  /* Et on a donc traité correctement le message, en fermant une fenêtre */
  return BI_EVENEMENT_FEN_FERMEE;
}
