/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 2007, Emmanuel CURIS
   Cration : 4 juillet 2007

   Interface graphique

   Crer "ex-nihilo" d'un atome dans une molcule
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
/* #include <string.h> */
/* #include <ctype.h> */

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"
#include "cristallo.h"
#include "interface.h"

/* -------- Variables locales -------- */

struct infos_arbre {
  unsigned long premier_fragment_affiche;
};

/* -------- Les textes [isols pour traduction] -------- */

#include "int_creer_atome.txt"

/* -------- Fonctions locales [déclarations] -------- */

static int  I_ATM_GererArbre   (unsigned int numero_objet, int x, int y);
static void I_ATM_PreparerArbre(void);

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Crer un atome, en coordonnes orthonormes absolues --------
   Renvoie l'atome cr ; NULL en cas de problme
 */

Atome *I_ATM_CreerAtome_Absolu(Molecule *molecule, boolean multiple, Fenetre *mere)
{
  boolean sortie = FAUX;
  int reponse, numero_atomique;
  double charge, rayon;
  Atome *atome = NULL;
  Vecteur3D coordonnees = { 0.0, 0.0, 0.0 };
  Vecteur3D_Erreurs erreurs_coordonnees = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

  if ( NULL == molecule ) return NULL;  /* Pas de molécule ! */

  /* Prparer l'arbre */
  I_ATM_PreparerArbre( );

  /* On ne peut pas demander les symtriques */
  OBJ_active( CREER_ATOME, CA_AVEC_SYMETRIK, FORCE_NON );

  /* Les coordonnes absolues sont en Å */

  /* On demande les coordonnes */
  do {
    reponse = ARB_Bloquant( CREER_ATOME, titre_creer_atome, mere, CREER_ATOME );
    if ( CRA_ANNULER == reponse ) break;	/* On annule */
    if ( ( FAUX == multiple ) || ( CA_DERNIER == reponse ) ) {
      sortie = VRAI;
    }

    /* On cre l'atome dans la molcule */
    /* 1. Lecture du numro atomique */
    numero_atomique = OBJ_NumeroAtomique( CREER_ATOME, CA_NUM_ATOMIQUE );

    /* 2. Lecture des coordonnes */
    OBJ_LireCoordonnees( CREER_ATOME, CA_POS_X, CA_POS_Y, CA_POS_Z, &coordonnees, &erreurs_coordonnees );

    /* 3. Lecture des paramtres facultatifs */
    rayon  = OBJ_LireNombre( CREER_ATOME, CA_RAYON , NULL );
    charge = OBJ_LireNombre( CREER_ATOME, CA_CHARGE, NULL );

    /* 4. Cration de l'atome */
    atome = ATM_creer( molecule, NULL, numero_atomique, &coordonnees, charge, rayon );
    if ( NULL == atome ) {
      /* Problme en crant l'atome */
      (void) BI_ERR_alerte( principale, &ERR_CreationAtome_Absolu, VRAI );
    }

    /* On remet la bote  0 [pas d'atome, coordonnes nulles, rayon et charge nuls,...] */
    OBJ_CopierTexte( CREER_ATOME, CA_NUM_ATOMIQUE, "", 0 );
    OBJ_MettreCoordonnees( CREER_ATOME, CA_POS_X, CA_POS_Y, CA_POS_Z, NULL, NULL );
    OBJ_MettreNombre( CREER_ATOME, CA_RAYON , 0.0, 0.0 );
    OBJ_MettreNombre( CREER_ATOME, CA_CHARGE, 0.0, 0.0 );
  } while ( sortie == FAUX );

  /* On renvoie le dernier atome créé */
  return atome;
}

/* -------- Créer un atome, en coordonnées fractionnaires --------
   Renvoie l'atome créé ; NULL en cas de problème

   Si une table de mailles est fournie, sera créé dans toutes ces mailles.
   Sinon, sera créé dans le(s) fragment(s) indiqués par l'utilisateur
 */

Atome *I_ATM_CreerAtome_Fractionnaire(Molecule *molecule, boolean multiple, unsigned long nombre_mailles, Fragment **mailles, Fenetre *mere)
{
  boolean sortie = FAUX;
  int reponse, numero_atomique;
  double charge, rayon;
  Atome *atome = NULL;
  Vecteur3D coordonnees = { 0.0, 0.0, 0.0 };
  Vecteur3D_Erreurs erreurs_coordonnees = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

  if ( NULL == molecule ) return NULL;  /* Pas de molécule ! */
  if ( NULL == molecule->maille ) {
    return NULL;
  }

  /* Préparer l'arbre */
  I_ATM_PreparerArbre( );

  /* On peut demander les symétriques, si des opérations de symétrie existent */
  if ( ( molecule->maille->operations != NULL ) || ( molecule->maille->translations != NULL ) ) {
    OBJ_active( CREER_ATOME, CA_AVEC_SYMETRIK, FORCE_OUI );
  } else {
    OBJ_active ( CREER_ATOME, CA_AVEC_SYMETRIK, FORCE_NON );
    OBJ_choisit( CREER_ATOME, CA_AVEC_SYMETRIK, FORCE_NON );
  }

  /* Les coordonnées fractionnaires sont sans unité */

  /* On demande les coordonnées */
  do {
    reponse = ARB_Bloquant( CREER_ATOME, titre_creer_atome, mere, CREER_ATOME );
    if ( CRA_ANNULER == reponse ) break;	/* On annule */
    if ( ( FAUX == multiple ) || ( CA_DERNIER == reponse ) ) {
      sortie = VRAI;
    }

    /* On crée l'atome dans la molécule */
    /* 1. Lecture du numéro atomique */
    numero_atomique = OBJ_NumeroAtomique( CREER_ATOME, CA_NUM_ATOMIQUE );

    /* 2. Lecture des coordonnées */
    OBJ_LireCoordonnees( CREER_ATOME, CA_POS_X, CA_POS_Y, CA_POS_Z, &coordonnees, &erreurs_coordonnees );

    /* 3. Lecture des paramètres facultatifs */
    rayon  = OBJ_LireNombre( CREER_ATOME, CA_RAYON , NULL );
    charge = OBJ_LireNombre( CREER_ATOME, CA_CHARGE, NULL );

    /* 4. Création de l'atome */
    reponse = CRIS_GenereCristal( molecule, &(coordonnees), numero_atomique, charge, rayon,
				  nombre_mailles, mailles );
    if ( reponse != 0 ) {
      /* Problème en créant l'atome */
      (void) BI_ERR_alerte( principale, &ERR_CreationAtome_Frac, VRAI );
    }

    /* On remet la boîte à 0 [pas d'atome, coordonnées nulles, rayon et charge nuls,...] */
    OBJ_CopierTexte( CREER_ATOME, CA_NUM_ATOMIQUE, "", 0 );
    OBJ_MettreCoordonnees( CREER_ATOME, CA_POS_X, CA_POS_Y, CA_POS_Z, NULL, NULL );
    OBJ_MettreNombre( CREER_ATOME, CA_RAYON , 0.0, 0.0 );
    OBJ_MettreNombre( CREER_ATOME, CA_CHARGE, 0.0, 0.0 );
  } while ( sortie == FAUX );

  /* On renvoie le dernier atome créé */
  return atome;
}

/* ~-~-~-~-~-~-~-~-~ Gestion de la boîte de dialogue ~-~-~-~-~-~-~-~-~ */

static int I_ATM_GererArbre(unsigned int numero_objet, int x __attribute__((__unused__)), int y)
{

  if ( ( ( numero_objet >= CA_FRAGMENT_1 ) && ( numero_objet <= CA_FRAGMENT_N ) ) ||
       ( CA_FRG_AVANT == numero_objet ) || ( CA_FRG_FOND_ASC == numero_objet ) || ( CRA_BOUTON_ASC == numero_objet ) || ( CA_FRG_APRES == numero_objet ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Clic en Y = %d\n",
		   __FILE__, __FUNCTION__, __LINE__, y);
    OPTION_A_FAIRE( "Gérer le défileur de fragments" );

    return BI_EVENEMENT_TRAITE;
  }

  switch ( numero_objet )
  {
  case CA_NV_FRAGMENT:
    OPTION_A_FAIRE( "Créer un nouveau fragment" );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

static void I_ATM_PreparerArbre(void)
{
  int retour;

  if ( VRAI == arbres[ CREER_ATOME ].drapeaux.prepare ) return;

  /* Initialisation de l'arbre */
  retour = ARB_Initialiser( CREER_ATOME, NULL, I_ATM_GererArbre, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%sl.%u : Impossible d'initialiser l'arbre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* Enjolivements visuels */

  /* Traductions */
#ifdef AVEC_TRADUCTION_
#endif

  /* Préparation terminée */
  arbres[ CREER_ATOME ].drapeaux.prepare = VRAI;
}
