/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 2007, Emmanuel CURIS
   Création : 3 juillet 2007

   Interface graphique

   Créer "ex-nihilo" une molécule
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
/* #include <string.h> */
/* #include <ctype.h> */

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"
#include "cristallo.h"
#include "interface.h"

/* -------- Variables locales -------- */

struct infos_arbre {
  struct systeme_cristallin *systemes; /* Table des réseaux définis */
  unsigned long nbr_systemes; /* Nombre de systèmes */
  int systeme_choisi;
  int groupe_choisi;
  unsigned long nombre_mailles;
  Fragment **mailles;
};

#define LGMAX_NOM_RESEAU 20
#define LGMAX_NOM_GROUPE 20

/* -------- Les textes [isolés pour traduction] -------- */

#include "int_creer_molecule.txt"

/* -------- Fonctions locales -------- */

static void I_MOL_PreparerArbre(void);
static int  I_MOL_GererArbre   (unsigned int objet, int x, int y);

static void I_MOL_ActiverCristallo(boolean sens);

static int I_MOL_GenererMaille(Molecule *molecule, struct infos_arbre *infos);

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Création d'une molécule -------- */

Molecule *I_MOL_CreerMolecule(Molecule *avant __attribute__((__unused__)))
{
  int retour;
  char *nom_molecule;
  struct infos_arbre infos = { NULL, 0, 0, 0, 0, NULL };
  Molecule *molecule = NULL;

  /* On lit la liste des systèmes cristallins. */
  infos.systemes = CRIS_CreerListeSystemes( PREFIXE_REP "/Lase/Cristallo/", &(infos.nbr_systemes) );
  if ( NULL == infos.systemes ) { /* Pas de coordonnées fractionnaires possibles */
    OBJ_active ( NVLE_MOLECULE, NM_AVEC_MAILLE, FORCE_NON);
    OBJ_choisit( NVLE_MOLECULE, NM_AVEC_MAILLE, FORCE_NON);
    I_MOL_ActiverCristallo( FAUX );
  } else {
    OBJ_active( NVLE_MOLECULE, NM_AVEC_MAILLE, FORCE_OUI );

    /* On installe le système : met à jour les champs éditables ; quand on désactive tout ensuite au cas où, pas de surprise ! */
    OBJ_ChoisirSystemeCr( NVLE_MOLECULE, NM_BRAVAIS, LGMAX_NOM_RESEAU, 0, METTRE, infos.systemes, infos.nbr_systemes,
			  NM_VALEUR_A, NM_VALEUR_B, NM_VALEUR_C, NM_VALEUR_ALPHA, NM_VALEUR_BETA, NM_VALEUR_GAMMA );

    /* On active ce qu'il faut pour les coordonnées fractionnaires */
    I_MOL_ActiverCristallo( OBJ_choisit( NVLE_MOLECULE, NM_AVEC_MAILLE, QUESTION ) );
  }

  /* On prépare la boîte de dialogue */
  I_MOL_PreparerArbre();

  /* On associe les informations à l'arbre */
  ARB_MettreComplement( NVLE_MOLECULE, &infos );

  /* On demande les paramètres de la molécule */
  retour = ARB_Bloquant( NVLE_MOLECULE, titre_creer_molecule, fenetre_onglets, 0 );
  ARB_MettreComplement( NVLE_MOLECULE, NULL );  /* Plus besoin de lui */
  if ( retour != NM_OK ) {
    /* Annulation, ou erreur */
    return NULL;
  }

  /* On cre la molcule */
  nom_molecule = lit_chaine( NVLE_MOLECULE, NM_NOM_MOLECULE, 0, NULL, TXT_REMPLACE );
  molecule = MOL_creer( nom_molecule );
  if ( NULL == molecule ) {
    (void) BI_ERR_alerte( principale, &ERR_CreationMolecule, VRAI );
    return NULL;
  }

  if ( VRAI == OBJ_choisit( NVLE_MOLECULE, NM_AVEC_MAILLE, QUESTION ) ) {
    /* Il faut crer la maille ! */
    retour = I_MOL_GenererMaille( molecule, &infos );
    if ( retour != 0 ) {
      /* Problme lors de la cration */
      retour = BI_ERR_alerte( principale, &Q_MailleNonCreee, VRAI );
      if ( 2 == retour ) {
	/* On dcide d'annuler la cration de la molcule */
	CRIS_EffacerListeSystemes( infos.systemes, infos.nbr_systemes );
	MOL_effacer( molecule, VRAI, VRAI );
	return NULL;
      }
    }
  }

  /* Plus besoin de la liste des systmes cristallins */
  CRIS_EffacerListeSystemes( infos.systemes, infos.nbr_systemes );

  /* On demande la cration des atomes */
  if ( molecule->maille != NULL ) {
    /* Comme on a demand la maille, on demande les coordonnes fractionnaires */
    I_ATM_CreerAtome_Fractionnaire( molecule, VRAI, infos.nombre_mailles, infos.mailles, fenetre_onglets );
  } else {
    /* Pas de maille : on demande les coordonnes absolues */
    I_ATM_CreerAtome_Absolu( molecule, VRAI, fenetre_onglets );
  }

  /* Plus besoin de la table des mailles */
  infos.nombre_mailles = 0;
  free( infos.mailles ); infos.mailles = NULL;

  /* On prpare la molcule */
  MOL_cherche_barycentre( molecule, FAUX );
  MOL_initialise_vue( molecule );

  /* On adapte la liste des molcules */
  MOL_PrepareListe();

  /* On a termin : on renvoie la molcule */
  return molecule;
}

/* ~-~-~-~-~-~-~-~-~ Fonctions locales ~-~-~-~-~-~-~-~-~ */

static int I_MOL_GenererMaille(Molecule *molecule, struct infos_arbre *infos)
{
  int retour;
  int i_min, j_min, k_min, i_max, j_max, k_max;
  struct maille *maille = NULL;

  if ( ( NULL == molecule ) || ( NULL == infos ) ) return -1;

  /* On créer la maille de la molécule */
  maille = CRIS_CreerMaille();
  if ( NULL == maille ) {
    return +1;		      /* Erreur n'empêchant pas la création */
  }

  /* On lit les paramètres de maille */
  maille->a = OBJ_LireNombre( NVLE_MOLECULE, NM_VALEUR_A, &(maille->s_a) );
  maille->b = OBJ_LireNombre( NVLE_MOLECULE, NM_VALEUR_B, &(maille->s_b) );
  maille->c = OBJ_LireNombre( NVLE_MOLECULE, NM_VALEUR_C, &(maille->s_c) );

  /* Pour les angles, on ne peut pas utiliser OBJ_LireAngle car on veut mémoriser en degrés... */
  maille->alpha = OBJ_LireNombre( NVLE_MOLECULE, NM_VALEUR_ALPHA, &(maille->s_alpha) );
  maille->beta  = OBJ_LireNombre( NVLE_MOLECULE, NM_VALEUR_BETA , &(maille->s_beta ) );
  maille->gamma = OBJ_LireNombre( NVLE_MOLECULE, NM_VALEUR_GAMMA, &(maille->s_gamma) );

  /* On adapte ces paramètres en fonction du groupe d'espace */
  retour = CRIS_AdapteParametres( infos->systemes + infos->systeme_choisi,
				  &(maille->a), &maille->b, &maille->c, &maille->alpha, &maille->beta, &maille->gamma );
  if ( retour < 0 ) { /* Erreur grave */
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%sl.%u : ERREUR dans les paramètres de maille [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
#endif
    CRIS_EffacerMaille( maille ); maille = NULL;
    return +2;
  }

  /* On adapte les angles en radians */
  maille->alpha *= M_PI / 180.0;
  maille->beta  *= M_PI / 180.0;
  maille->gamma *= M_PI / 180.0;

  /* On mémorise le type de système cristallin */
  maille->type = infos->systemes[ infos->systeme_choisi ].type;

  /* On lit les opérations du groupe d'espace */
  retour = CRIS_LireOperationsGroupes( infos->systemes[ infos->systeme_choisi ].groupes + infos->groupe_choisi,
				       PREFIXE_REP "/Lase/Cristallo/", infos->systemes[ infos->systeme_choisi ].dossier );
  if ( retour != 0 ) {
    /* Impossible de lire les opérations du groupe d'espace */
    CRIS_EffacerMaille( maille ); maille = NULL;
    return +3;
  }

  /* On mémorise ces opérations */
  maille->nbr_operations   = infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].nbr_operations;
  maille->nbr_translations = infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].nbr_translations;
  maille->operations   = infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].operations;
  maille->translations = infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].translations;

  /* Et on les élimine du système pour ne pas les libérer par erreur ! */
  infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].nbr_operations = 0;
  infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].nbr_translations = 0;
  infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].operations = NULL;
  infos->systemes[ infos->systeme_choisi ].groupes[ infos->groupe_choisi ].translations = NULL;

  /* On associe tout cela à la molécule */
  MOL_AssocierCristallo( molecule, maille );

  /* On crée les fragments des mailles */
  i_min = OBJ_LireEntier_S( NVLE_MOLECULE, NM_I_MIN ); /* Nombre de mailles à générer */
  i_max = OBJ_LireEntier_S( NVLE_MOLECULE, NM_I_MAX );
  if ( i_max < i_min ) { int i = i_min ; i_min = i_max ; i_max = i; }

  j_min = OBJ_LireEntier_S( NVLE_MOLECULE, NM_J_MIN );
  j_max = OBJ_LireEntier_S( NVLE_MOLECULE, NM_J_MAX );
  if ( j_max < j_min ) { int j = j_min ; j_min = j_max ; j_max = j; }

  k_min = OBJ_LireEntier_S( NVLE_MOLECULE, NM_K_MIN );
  k_max = OBJ_LireEntier_S( NVLE_MOLECULE, NM_K_MAX );
  if ( k_max < k_min ) { int k = k_min ; k_min = k_max ; k_max = k; }

  infos->mailles = FRG_mailles( molecule, i_min, i_max, j_min, j_max, k_min, k_max );
  if ( infos->mailles != NULL ) {
    infos->nombre_mailles  = (i_max - i_min) + 1;
    infos->nombre_mailles *= (j_max - j_min) + 1;
    infos->nombre_mailles *= (k_max - k_min) + 1;
  }

  /* Terminé sans souci */
  return 0;
}

/* -------- Gestion de l'arbre ------------ */

static int I_MOL_GererArbre(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  int etat, nv_choix;
  struct infos_arbre *infos;

  /* On récupère les infos de l'arbre */
  infos = (struct infos_arbre *) ARB_LireComplement( NVLE_MOLECULE );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas d'informations associées à l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }

  switch ( objet )
  {
  case NM_AVEC_MAILLE:
    etat = OBJ_choisit( NVLE_MOLECULE, NM_AVEC_MAILLE, ECHANGE );

    /* On active les boutons correspondants */
    I_MOL_ActiverCristallo( etat == FORCE_OUI );

    if ( VRAI == etat ) { /* Coordonnées fractionnaires actives -> on installe le système cristallin actif */
      OBJ_ChoisirSystemeCr( NVLE_MOLECULE, NM_BRAVAIS, LGMAX_NOM_RESEAU, infos->systeme_choisi, METTRE, infos->systemes, infos->nbr_systemes,
			    NM_VALEUR_A, NM_VALEUR_B, NM_VALEUR_C, NM_VALEUR_ALPHA, NM_VALEUR_BETA, NM_VALEUR_GAMMA );
      OBJ_ChoisirGroupeEsp( NVLE_MOLECULE, NM_GROUPE_ESPACE, LGMAX_NOM_GROUPE, infos->groupe_choisi, METTRE, infos->systemes + infos->systeme_choisi );
    }
    return BI_EVENEMENT_TRAITE;

  case NM_BRAVAIS:
    nv_choix = OBJ_ChoisirSystemeCr( NVLE_MOLECULE, NM_BRAVAIS, LGMAX_NOM_RESEAU, infos->systeme_choisi, QUESTION, infos->systemes, infos->nbr_systemes,
				     NM_VALEUR_A, NM_VALEUR_B, NM_VALEUR_C, NM_VALEUR_ALPHA, NM_VALEUR_BETA, NM_VALEUR_GAMMA );
    if ( (nv_choix >= 0) && (nv_choix != (int) infos->systeme_choisi) ) {
      infos->systeme_choisi = (unsigned int) nv_choix;
      OBJ_active( NVLE_MOLECULE, NM_GROUPE_ESPACE, VRAI ); /* A pu être désactivé si aucun groupe pour le précédent choisi */

      /* On installe le premier groupe de la liste */
      OBJ_ChoisirGroupeEsp( NVLE_MOLECULE, NM_GROUPE_ESPACE, LGMAX_NOM_GROUPE, 0, METTRE, infos->systemes + infos->systeme_choisi );

      /* On demande le groupe d'espace associé, s'il y en a */
      infos->groupe_choisi = OBJ_ChoisirGroupeEsp( NVLE_MOLECULE, NM_GROUPE_ESPACE, LGMAX_NOM_GROUPE, infos->groupe_choisi, QUESTION,
						   infos->systemes + infos->systeme_choisi );
    }
    return BI_EVENEMENT_TRAITE;

  case NM_GROUPE_ESPACE:
    infos->groupe_choisi = OBJ_ChoisirGroupeEsp( NVLE_MOLECULE, NM_GROUPE_ESPACE, LGMAX_NOM_GROUPE, infos->groupe_choisi, QUESTION,
						 infos->systemes + infos->systeme_choisi );
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Macros de gestion -------- */

static void I_MOL_ActiverCristallo(boolean sens)
{
  int que_faire;

  /* On choisit le sens */
  que_faire = ( sens == VRAI ) ? ( FORCE_OUI ) : ( FORCE_NON );

  /* On s'occupe des boutons */
  OBJ_active( NVLE_MOLECULE, NM_BRAVAIS      , que_faire );
  OBJ_active( NVLE_MOLECULE, NM_GROUPE_ESPACE, que_faire );

  OBJ_active( NVLE_MOLECULE, NM_I_MIN, que_faire );
  OBJ_active( NVLE_MOLECULE, NM_I_MAX, que_faire );
  OBJ_active( NVLE_MOLECULE, NM_J_MIN, que_faire );
  OBJ_active( NVLE_MOLECULE, NM_J_MAX, que_faire );
  OBJ_active( NVLE_MOLECULE, NM_K_MIN, que_faire );
  OBJ_active( NVLE_MOLECULE, NM_K_MAX, que_faire );

  OBJ_active( NVLE_MOLECULE, NM_VALEUR_A, que_faire );
  if ( FAUX == sens ) {
    /* On est sûr que les paramètres de maille sont coupés */
    OBJ_active( NVLE_MOLECULE, NM_VALEUR_B, FORCE_NON );
    OBJ_active( NVLE_MOLECULE, NM_VALEUR_C, FORCE_NON );

    OBJ_active( NVLE_MOLECULE, NM_VALEUR_ALPHA, FORCE_NON );
    OBJ_active( NVLE_MOLECULE, NM_VALEUR_BETA , FORCE_NON );
    OBJ_active( NVLE_MOLECULE, NM_VALEUR_GAMMA, FORCE_NON );
  } else {
    /* Le choix des paramètres de maille dépend du système cristallin retenu */
  }
}

/* -------- Préparation de l'arbre -------- */

static void I_MOL_PreparerArbre(void)
{
  if ( VRAI == arbres[ NVLE_MOLECULE ].drapeaux.prepare ) return;

  /* On prépare les sous-menus */

  /* On initialise */
  if ( FAUX == arbres[ NVLE_MOLECULE ].drapeaux.init ) {
    int retour;

    OPTION_A_FAIRE( "Documentation, liste des molécules disponibles" );
    retour = ARB_Initialiser( NVLE_MOLECULE, NULL, I_MOL_GererArbre, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  /* Améliorations esthétiques */
  OBJ_deplace( NVLE_MOLECULE, NM_TXT_TYPE     , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_TYPE_MOLECULE, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( NVLE_MOLECULE, NM_BRAVAIS      , 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( NVLE_MOLECULE, NM_GROUPE_ESPACE, 0, 500, DEPLACE_RELATIF );

  OBJ_deplace( NVLE_MOLECULE, NM_TXT_B   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_VALEUR_B, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_TXT_C   , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_VALEUR_C, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( NVLE_MOLECULE, NM_TXT_BETA    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_VALEUR_BETA , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_TXT_GAMMA   , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_VALEUR_GAMMA, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( NVLE_MOLECULE, NM_TXT_J, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_J_MIN, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_J_MAX, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_TXT_K, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_K_MIN, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( NVLE_MOLECULE, NM_K_MAX, 0, +2, DEPLACE_PIXELS );

#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( NVLE_MOLECULE, NM_TXT_NOM, txt_nle_mol_nom );
  OBJ_TraduireTexte( NVLE_MOLECULE, NM_NOM_MOLECULE, txt_nle_mol_type );

  OBJ_TraduireTexte( NVLE_MOLECULE, NM_TXT_AV_MAILLE, txt_nle_mol_parametres );

  OBJ_TraduireTexte( NVLE_MOLECULE, NM_OK, txt_nle_mol_creer );
  OBJ_TraduireTexte( NVLE_MOLECULE, NM_ANNULER, txt_nle_mol_annuler );
#endif

  /* L'arbre est maintenant près... */
  arbres[ NVLE_MOLECULE ].drapeaux.prepare = VRAI;
}
