/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1996 - 2006

   Interface pour la gestion de base des graphes (copie, effacement,...)
*/

#include <stdio.h>
#include <stdlib.h>

#include "exafs.h"
#include "systeme.h"
#include "txt_utils.h"

#include "interface.h"
#include "serie.h"
#include "statistiques.h"

#include "int_graphes.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int I_Deplacement(Graphe *graphe, Serie_Graphes *serie_arrivee, Graphe *avant);

/* -------- Déplacer un graphe dans une autre série -------- */

int I_GRF_Deplacer(struct graphe *graphe, boolean multiple)
{
  Graphe **graphes = NULL;
  Serie_Graphes *serie_arrivee;
  unsigned long nbr_graphes = 0;

  if ( VRAI == multiple ) {
    /* Choix des graphes à copier */
    if ( NULL == graphe ) {
      graphes = GRF_mchoisir( GR_TOUS, 1, 0, &nbr_graphes );
    } else {
      graphes = GRF_mchoisir( graphe->type, 1, 0, &nbr_graphes );
    }
    if ( NULL == graphes ) return -1;
  } else if ( NULL == graphe ) {
    return -20;
  }

  /* Choix de la série d'arrivée */
  serie_arrivee = I_SERIE_GRF_Choisir( -1, ROOT, 0 );
  if (NULL == serie_arrivee) {
    free( graphes ); graphes = NULL;
    return -2;
  }

  if ( VRAI == serie_arrivee->drapeaux.protegee ) {
    (void) BI_ERR_alerte( principale, &ERR_SerieProtegee, VRAI );
    free( graphes ); graphes = NULL;
    return -3;
  }

  if ( serie_arrivee == serie_courante ) {
    (void) BI_ERR_alerte( principale, &MSG_MemeSerie, VRAI );
    return 0;
  }

  /* Déplacement */
  if ( VRAI == multiple ) {
    unsigned long i;

    for ( i = 0 ; i < nbr_graphes ; i++ ) {
      I_Deplacement( graphes[ i ], serie_arrivee, (i > 0) ? (graphes[ i - 1 ]) : (NULL) );
    }
  } else {
    I_Deplacement( graphe, serie_arrivee, NULL );
  }

  /* On a fini... */
  free( graphes ); graphes = NULL;
  return 0;
}

/* -------- Isoler un graphe, en créant une nouvelle série -------- */

int I_GRF_Isoler(struct graphe *graphe, boolean multiple)
{
  char *titre;
  Graphe **graphes = NULL;
  Serie_Graphes *serie_arrivee;
  unsigned long nbr_graphes = 1;

  if ( VRAI == multiple ) {
    /* Choix des graphes à copier */
    if ( NULL == graphe ) {
      graphes = GRF_mchoisir( GR_TOUS, 1, 0, &nbr_graphes );
    } else {
      graphes = GRF_mchoisir( graphe->type, 1, 0, &nbr_graphes );
    }
    if ( NULL == graphes ) return -1;

    if ( nbr_graphes == serie_courante->nbr_graphes ) {
      (void) BI_ERR_alerte( principale, &MSG_TousChoisis, VRAI );
      free( graphes ); graphes = NULL;
      return -19;
    }
  } else if ( NULL == graphe ) {
    return -20;
  }

  /* Création de la série d'arrivée */
  titre = BI_Demande_Texte( principale, &Q_NomNouvelleSerie, serie_courante->nom, (unsigned int) (nbr_graphes) );
  serie_arrivee = SERIE_GRF_CreerSerie( serie_courante, titre );
  free( titre ); titre = NULL;
  if (NULL == serie_arrivee) {
    (void) fprintf(stderr, "%s,%s l.%u : impossible de créer une nouvelle série\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( graphes ); graphes = NULL;
    return -21;
  }

  if ( VRAI == multiple ) {
    unsigned long i;

    for ( i = 0 ; i < nbr_graphes ; i++ ) {
      I_Deplacement( graphes[ i ], serie_arrivee, (i > 0) ? (graphes[ i - 1 ]) : (NULL) );
    }
  } else {
    I_Deplacement( graphe, serie_arrivee, NULL );
  }

  /* On a fini... */
  free( graphes ); graphes = NULL;
  return 0;
}

/* -------- Effacer un graphe, avec demandes --------
 |  Renvoie 0 si l'effacement a eu lieu sans problème
 |    (cf. interface.h)
 |
 | graphe : le graphe à effacer
 | demander : VRAI = on demande confirmation avant l'effacement (même pour les sauvés)
 |            FAUX = on ne demande rien si sauvé
 | demander_si_pas_sauve : VRAI = on demande s'il faut sauver, quoi qu'il arrive
 |                         FAUX = on ne demande rien et on efface sans sauvegarder
 */

int I_GRF_Effacer(Graphe *graphe, boolean demander, boolean demander_si_pas_sauve)
{
  int reponse;

  /* Préalable... */
  if ( NULL == graphe ) return +1;

  if ( GRF_EstProtege( graphe ) ) {
    (void) BI_ERR_alerte( principale, &ERR_GrapheProtege, VRAI, graphe->nom );
    return -5;
  }

  /* Contrôles */
  if ( ( VRAI == demander_si_pas_sauve ) && ( FAUX == graphe->drapeaux.sauve ) ) {
    /* Le graphe n'a pas été sauvé : on demande ce que l'on fait avec cela */
    reponse = BI_ERR_alerte( principale, &Q_EffacerGraphe_NonSauve_Unique, VRAI, graphe->nom );
    if ( I_EFFACER_GRF_NS_U_ANNULER == reponse ) {
      /* Effacement annulé */
      return +1;
    }
    if ( I_EFFACER_GRF_NS_U_SAUVER == reponse ) {
      int retour;

      retour = I_GRF_Sauver( graphe );
      if ( retour != 0 ) {
	/* Pb dans la sauvegarde : effacement annulé... */
	reponse = BI_ERR_alerte( principale, &ERR_Sauvegarde, VRAI, graphe->nom );
	if ( 2 == reponse ) {
	  /* Effacement annulé... à cause d'un problème ! */
	  return -1;
	}
      }
    }

    /* Plus besoin de demander */
    demander = FAUX;
  }

  if ( VRAI == demander ) {
    /* On demande confirmation de l'effacement... */
    reponse = BI_ERR_alerte( principale, &Q_EffacerGraphe_Unique, VRAI, graphe->nom );
    if ( I_EFFACER_GRF_U_ANNULER == reponse ) {
      return +2;
    }
  }

  /* On efface */
  GRF_efface( graphe, VRAI, FAUX );

  /* On met à jour l'ascenseur de la fenêtre principale, au besoin */
  I_PrincipaleGRF_RefaireAscenseurs( FAUX, VRAI );

  /* On a terminé */
  return 0;
}

/* -------- Déplacement d'un seul graphe avec contrôle -------- */

static int I_Deplacement(Graphe *graphe, Serie_Graphes *serie_arrivee, Graphe *avant)
{
  if ( ( GRF_SERIE == graphe->genre_donnees ) && ( graphe->x != NULL ) ) {
    int reponse;

    reponse = BI_ERR_alerte( principale, &Q_PerdreXSerie, VRAI, graphe->nom );
    if ( 2 == reponse ) {
      return -1;
    }
  }

  GRF_ChangeSerie( graphe, serie_arrivee, avant );
  return 0;
}

/* ~-~-~-~-~-~-~-~ Interfaces pour les créations simples de graphes ~-~-~-~-~-~-~-~ */

#define I_GRF_DEMANDE_NOM     0	/* Toujours le nom demandé en premier ! */

#define I_GRF_DEMANDE_DERNIERE 1  /* Les réponses sont toujours dans cet ordre, par homogénéité */
#define I_GRF_DEMANDE_ENCORE   2
#define I_GRF_DEMANDE_ANNULE   3

/* -------- Création d'un graphe contenant une série de valeurs aléatoires -------- */

Graphe *I_GRF_CreerAleatoire(Graphe *avant)
{
  Graphe *nouveau = NULL;
  int reponse;
  struct processus_aleatoire conditions = { PA_BRUIT_BLANC, { { 0 } }, { DP_GAUSSIENNE, 0, { { 0, 0 } } } };
  struct serie *serie_cible = NULL;
  unsigned long nombre_valeurs;

  /* On appelle l'arbre considéré */
  reponse = ARB_Bloquant( TIRAGE_ALEATOIRE, "Paramètres des tirages", principale, 0 );
  if ( TA_ANNULER == reponse ) return NULL;

  /* On mémorise les paramètres pour la génération */
  nombre_valeurs = OBJ_LireEntier_U( TIRAGE_ALEATOIRE, TA_NOMBRE );
  conditions.loi.parametres.Gauss.mu    = OBJ_LireNombre( TIRAGE_ALEATOIRE, TA_PARAMETRE_1, NULL );
  conditions.loi.parametres.Gauss.sigma = OBJ_LireNombre( TIRAGE_ALEATOIRE, TA_PARAMETRE_2, NULL );

  /* On génère les valeurs */
  nouveau = PA_GenererStationnaire( &conditions, nombre_valeurs );

  /* On insère le résultat après le graphe indiqué */
  serie_cible = I_GRF_InsererDansSerie( serie_courante, avant, nouveau );
  if ( NULL == serie_cible ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'insérer le graphe dans une quelconque série !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( nouveau ); nouveau = NULL;
    return NULL;
  }

  /* On met le nom */
/*   GRF_titre( nouveau,  */

  /* On a terminé, sans encombre */
  return nouveau;
}

/* -------- Création d'un graphe contenant des couples de valeurs (X,Y) [sans nécessairement de lien fonctionnel] -------- */

#define I_GRF_DEMANDE_VALEURS_X 1
#define I_GRF_DEMANDE_LEGENDE_X 2
#define I_GRF_DEMANDE_UNITE_X   3
#define I_GRF_DEMANDE_VALEURS_Y 4
#define I_GRF_DEMANDE_LEGENDE_Y 5
#define I_GRF_DEMANDE_UNITE_Y   6

Graphe *I_GRF_CreerCouples(Graphe *avant)
{
  BI_Champ_Demande demandes[ 7 ] = {
    { BI_CHAMP_DEMANDE_CHAINE, "Nom          :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Valeurs en X :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Légende en X :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Unité en X   :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Valeurs en Y :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Légende en Y :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Unité en Y   :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } }
  };
  Graphe *nouveau = NULL;
  struct serie *serie_cible = NULL;
  int reponse;
  unsigned int couleur = NOIR, style = BI_AUCUNE_LIGNE, marque = BI_MARQUE_CERCLE;

  do {
    /* On demande la série de valeurs */
    reponse = BI_DemandeMultiple( principale, &Q_ValeursSerie, 7, demandes );
    if ( I_GRF_DEMANDE_ANNULE == reponse) break;

    /* On construit le graphe isolé correspondant */
    nouveau = GRF_nv_DoubleListe( GR_SSTYPE, demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_donnee.chaine, demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_donnee.chaine );
    if ( nouveau != NULL ) {
      /* On a pu créer le graphe : on l'insère dans la série et on adapte légendes, titre,... */
      /* On n'a plus besoin des valeurs en X et en Y */
      free( demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_donnee.chaine ); demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_donnee.chaine = NULL;
      free( demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_donnee.chaine ); demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_donnee.chaine = NULL;

      /* On donne un nom au graphe */
      GRF_titre( nouveau, demandes[ I_GRF_DEMANDE_NOM ].valeur_donnee.chaine );

      /* On insère le graphe dans la série des résultats */
      serie_cible = I_GRF_InsererDansSerie( serie_cible, avant, nouveau );

      /* On colore ce graphe */
      GRF_couleur( nouveau, couleur, style, marque );

      /* On lui associe légende et unité */
      GRF_ChangerLegende( nouveau, AXE_X, demandes[ I_GRF_DEMANDE_LEGENDE_X ].valeur_donnee.chaine, VRAI );
      GRF_ChangerUnite( nouveau, AXE_X, demandes[ I_GRF_DEMANDE_UNITE_X ].valeur_donnee.chaine );

      GRF_ChangerLegende( nouveau, AXE_Y, demandes[ I_GRF_DEMANDE_LEGENDE_Y ].valeur_donnee.chaine, VRAI );
      GRF_ChangerUnite( nouveau, AXE_Y, demandes[ I_GRF_DEMANDE_UNITE_Y ].valeur_donnee.chaine );

      /* On insèrera la suivant après, avec la couleur suivante */
      avant = nouveau;
      TRC_CouleurSuivante( &couleur, NULL, &marque );
    } else {
      OPTION_A_FAIRE( "Afficher un message" );

      /* On repropose même X, même Y pour que l'utilisateur puisse corriger */
      free( demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_initiale.chaine );
      demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_initiale = demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_donnee;
      demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_donnee.chaine = NULL;

      free( demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_initiale.chaine );
      demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_initiale = demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_donnee;
      demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_donnee.chaine = NULL;

      /* On ne quitte pas, sauf si l'on annule */
      if ( I_GRF_DEMANDE_DERNIERE == reponse ) reponse = I_GRF_DEMANDE_ENCORE;
    }

    /* On prépare pour le tour suivant : on propose même nom, même légende, même unité par défaut */
    free( demandes[ I_GRF_DEMANDE_NOM ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_NOM ].valeur_initiale = demandes[ I_GRF_DEMANDE_NOM ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_NOM ].valeur_donnee.chaine = NULL;

    free( demandes[ I_GRF_DEMANDE_LEGENDE_X ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_LEGENDE_X ].valeur_initiale = demandes[ I_GRF_DEMANDE_LEGENDE_X ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_LEGENDE_X ].valeur_donnee.chaine = NULL;

    free( demandes[ I_GRF_DEMANDE_UNITE_X ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_UNITE_X ].valeur_initiale = demandes[ I_GRF_DEMANDE_UNITE_X ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_UNITE_X ].valeur_donnee.chaine = NULL;

    free( demandes[ I_GRF_DEMANDE_LEGENDE_Y ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_LEGENDE_Y ].valeur_initiale = demandes[ I_GRF_DEMANDE_LEGENDE_Y ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_LEGENDE_Y ].valeur_donnee.chaine = NULL;

    free( demandes[ I_GRF_DEMANDE_UNITE_Y ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_UNITE_Y ].valeur_initiale = demandes[ I_GRF_DEMANDE_UNITE_Y ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_UNITE_Y ].valeur_donnee.chaine = NULL;
  } while ( I_GRF_DEMANDE_ENCORE == reponse );

  /* On libère les noms, légendes, unités, valeurs qui peuvent rester */
  free( demandes[ I_GRF_DEMANDE_NOM ].valeur_initiale.chaine );
  free( demandes[ I_GRF_DEMANDE_LEGENDE_X ].valeur_initiale.chaine );
  free( demandes[ I_GRF_DEMANDE_UNITE_X ].valeur_initiale.chaine );
  free( demandes[ I_GRF_DEMANDE_VALEURS_X ].valeur_initiale.chaine );
  free( demandes[ I_GRF_DEMANDE_VALEURS_Y ].valeur_initiale.chaine );

  /* On renvoie le dernier graphe créé */
  return nouveau;
}

/* -------- Création d'un graphe contenant une simple série de valeurs -------- */

#define I_GRF_DEMANDE_VALEURS 1
#define I_GRF_DEMANDE_LEGENDE 2
#define I_GRF_DEMANDE_UNITE   3

Graphe *I_GRF_CreerSerie(Graphe *avant)
{
  BI_Champ_Demande demandes[ 4 ] = {
    { BI_CHAMP_DEMANDE_CHAINE, "Nom     :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Valeurs :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Légende :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_CHAINE, "Unité   :", NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } }
  };
  Graphe *nouveau = NULL;
  struct serie *serie_cible = NULL;
  int reponse;
  unsigned int couleur = NOIR, style = BI_AUCUNE_LIGNE, marque = BI_MARQUE_PLUS;

  do {
    /* On demande la série de valeurs */
    reponse = BI_DemandeMultiple( principale, &Q_ValeursSerie, 4, demandes );
    if ( I_GRF_DEMANDE_ANNULE == reponse) break;

    /* On construit le graphe isolé correspondant */
    nouveau = GRF_nv_ListeValeurs( GR_SSTYPE, demandes[ I_GRF_DEMANDE_VALEURS ].valeur_donnee.chaine );
    free( demandes[ I_GRF_DEMANDE_VALEURS ].valeur_donnee.chaine ); demandes[ I_GRF_DEMANDE_VALEURS ].valeur_donnee.chaine = NULL;
    if ( NULL == nouveau ) {
      /* Échec en créant le graphe : on recommence... */
      free( demandes[ I_GRF_DEMANDE_NOM     ].valeur_donnee.chaine ); demandes[ I_GRF_DEMANDE_NOM     ].valeur_donnee.chaine = NULL;
      free( demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_donnee.chaine ); demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_donnee.chaine = NULL;
      free( demandes[ I_GRF_DEMANDE_UNITE   ].valeur_donnee.chaine ); demandes[ I_GRF_DEMANDE_UNITE   ].valeur_donnee.chaine = NULL;
      continue;
    }

    /* On donne un nom au graphe */
    GRF_titre( nouveau, demandes[ I_GRF_DEMANDE_NOM ].valeur_donnee.chaine );

    /* On insère le graphe dans la série des résultats */
    serie_cible = I_GRF_InsererDansSerie( serie_cible, avant, nouveau );

    /* On colore ce graphe */
    GRF_couleur( nouveau, couleur, style, marque );

    /* On lui associe légende et unité */
    GRF_ChangerLegende( nouveau, AXE_Y, demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_donnee.chaine, VRAI );
    GRF_ChangerUnite( nouveau, AXE_Y, demandes[ I_GRF_DEMANDE_UNITE ].valeur_donnee.chaine );

    /* On prépare pour le tour suivant : on propose même nom, même légende, même unité par défaut */
    free( demandes[ I_GRF_DEMANDE_NOM ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_NOM ].valeur_initiale = demandes[ I_GRF_DEMANDE_NOM ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_NOM ].valeur_donnee.chaine = NULL;

    free( demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_initiale = demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_donnee.chaine = NULL;

    free( demandes[ I_GRF_DEMANDE_UNITE ].valeur_initiale.chaine );
    demandes[ I_GRF_DEMANDE_UNITE ].valeur_initiale = demandes[ I_GRF_DEMANDE_UNITE ].valeur_donnee;
    demandes[ I_GRF_DEMANDE_UNITE ].valeur_donnee.chaine = NULL;

    avant = nouveau;
    TRC_CouleurSuivante( &couleur, NULL, &marque );
  } while ( I_GRF_DEMANDE_ENCORE == reponse );

  /* On libère les noms, légendes, unités qui peuvent rester */
  free( demandes[ I_GRF_DEMANDE_NOM ].valeur_initiale.chaine );
  free( demandes[ I_GRF_DEMANDE_LEGENDE ].valeur_initiale.chaine );
  free( demandes[ I_GRF_DEMANDE_UNITE ].valeur_initiale.chaine );

  /* On renvoie le dernier graphe ainsi créé */
  return nouveau;
}
