/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS
   Création : mars 2006

   Interface graphique avec l'utilisateur

   Liste des modèles disponibles
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"		/* Pour avoir accès aux modèles des molécules */
#include "graphes.h"
#include "couches.h"		/* Pour éditer les modèles FEFF ! */
#include "interface.h"
#include "modeles.h"

/* -------- Variables locales -------- */

static struct infos_liste {
  unsigned int famille;
  ModeleAjustement *premier_modele;
  unsigned long nombre_modeles;
  unsigned long premier_affiche;
  unsigned long modele_actif;
} infos;

#define LGMAX_FAMILLE_MODELES 45
#define LGMAX_TYPE_MODELE     40
#define LGMAX_NOM_MODELE      32
#define LGMAX_EXPERIENCE      27
#define LGMAX_ALGORITHME      27
#define LGMAX_ESTIMATEUR      27

/* -------- Les sous-menus déroulants ------- */

#define LM_OPTIONS_CREER   0	/* Créer un nouveau modèle */
#define LM_OPTIONS_EFFACER 1	/* Effacer le modèle actif */
#define LM_OPTIONS_COPIER  2	/* Copier le modèle actif */
#define LM_NOMBRE_OPTIONS 3

#define LM_RESIDUS_ABSOLUS  1	/* [0 : separateur] Calculer les residus absolus, moindres carres */
#define LM_RESIDUS_RELATIFS 2	/* Calculer les residus relatifs, moindres carres */
#define LM_RESIDUS_PONDERES 3	/* Calculer les residus ponderes, moindres carres */
#define LM_NOMBRE_OPTIONS_RESIDUS 4

#include "int_lst_modeles.txt"

/* -------- Fonctions locales -------- */

static int LMOD_GererArbre(unsigned int objet, int x, int y);

static void LMOD_ChoisirExperience(ModeleAjustement *modele);
static void LMOD_ParametresLibres (ModeleAjustement *modele);

static void LMOD_AfficherModele (ModeleAjustement *modele);
static void LMOD_CouperArret    (boolean couper);
static void LMOD_CouperGeneraux (boolean couper);
static void LMOD_CouperResultats(boolean couper);
static void LMOD_ToutCouper     (boolean couper);

static void LMOD_GererMenuOptions (unsigned int num_arbre, unsigned int num_objet, unsigned long action);
static void LMOD_GererMenuResidus (unsigned int num_arbre, unsigned int num_objet, unsigned long action);
static void LMOD_GererMenuValidite(unsigned int num_arbre, unsigned int num_objet, unsigned long action);

static ModeleAjustement *LMOD_ModeleEdite(void);

static const char *LMOD_NVTexte(unsigned int arbre, unsigned long numero_objet);

/* ~-~-~-~-~-~-~-~- Gestion de l'arbre ~-~-~-~-~-~-~-~- */

static int LMOD_GererArbre(unsigned int objet, int x __attribute__((__unused__)), int y)
{
  int reponse;
  ModeleAjustement *modele = NULL;

  if ( LM_FAMILLE == objet ) {
    reponse = SS_menu_texte( noms_familles_modeles, LM_NOMBRE_FAMILLES, NULL, LISTE_MODELES, LM_FAMILLE );
    if ( reponse < 0 ) {
      return BI_EVENEMENT_TRAITE;
    }

    LMOD_ChangerFamille( reponse );
    return BI_EVENEMENT_TRAITE;
  }

  if ( NULL == infos.premier_modele ) {
    if ( LM_OPTIONS == objet ) {
      /* Si aucun modèle, la seule option raisonnable est de créer un modèle : on l'applique directement */
      LMOD_GererMenuOptions( LISTE_MODELES, LM_OPTIONS, LM_OPTIONS_CREER );
      return BI_EVENEMENT_TRAITE; /* A un sens si aucun modèle... */
    }
    return BI_EVENEMENT_TRAITE; /* Aucune modèle : aucun intérêt ! */
  }

  if ( ( ( objet >= LM_PREMIER_MODEL ) && ( objet <= LM_DERNIER_MODEL ) ) ||
       ( LM_MODELE_PREV == objet ) || ( LM_FOND_ASC == objet ) || ( LM_BOUTON_ASC == objet ) || ( LM_MODELE_NEXT == objet ) ) {
    /* On enregistre le modèle actuel --- avant, car ensuite change infos.modele_actif ->
       on enregistrerait les paramètres de l'ancien modèle dans le nouveau ! [en particulier le nom !] */
    LMOD_EnregistrerModele( NULL );

    reponse = ARB_GererDefileur_O( LISTE_MODELES, LM_PREMIER_MODEL, LM_DERNIER_MODEL, LGMAX_NOM_MODELE,
				   LM_MODELE_PREV, LM_FOND_ASC, LM_BOUTON_ASC, LM_MODELE_NEXT,
				   infos.nombre_modeles, NULL, &(infos.modele_actif), &(infos.premier_affiche),
				   LMOD_NVTexte, objet, y );
    if ( reponse != 0 ) {
      /* Problème, annulation,... */
      AVERTISSEMENT( "Problème dans ARB_GererDefileur_O, code %d", reponse );
      return BI_EVENEMENT_TRAITE;
    }

    if ( ( LM_MODELE_PREV == objet ) || ( LM_FOND_ASC == objet ) || ( LM_BOUTON_ASC == objet ) || ( LM_MODELE_NEXT == objet ) ) {
      /* On se contente de faire défiler la liste -> pas besoin de changer l'affichage */
      return BI_EVENEMENT_TRAITE;
    }

    /* On récupère le nouveau modèle édité */
    modele = LMOD_ModeleEdite();

    /* On installe le modèle actif */
    LMOD_AfficherModele( modele );

    /* Activations dépendant du type de modèle */
    switch ( infos.famille )
    {
    case LM_FAMILLE_GENERIQUE_R:
      break;

    case LM_FAMILLE_MOLECULE:
      break;

    case LM_FAMILLE_ABSORBEUR:
      mol_courante->absorbeurs[ mol_courante->absorbeur_courant ].modele_actif = modele;
      ABS_AllerModele( modele, VRAI );
      break;

    case LM_FAMILLE_DECOMPOSER:
      OPTION_A_FAIRE( "Activer ce modèle pour la décomposition de spectres" );
      break;

    default:
      AVERTISSEMENT( "Famille de modèle non-reconnue [%d]", infos.famille );
      break;
    }

    return BI_EVENEMENT_TRAITE;
  }

  /* On récupère le modèle édité */
  modele = LMOD_ModeleEdite();
  if ( NULL == modele ) return BI_EVENEMENT_TRAITE;

  switch ( objet )
  {
  case LM_GRAPHE_EXP:
    OBJ_choisit( LISTE_MODELES, objet, FORCE_OUI );
    LMOD_ChoisirExperience( modele );
    if ( modele->experience != NULL ) {
      OBJ_CopierTexte( LISTE_MODELES, LM_GRAPHE_EXP, modele->experience->nom, LGMAX_EXPERIENCE );
      OBJ_active( LISTE_MODELES, LM_PROPOSER, FORCE_OUI );
    } else {
      OBJ_CopierTexte( LISTE_MODELES, LM_GRAPHE_EXP, txt_aucun_graphe_modele, LGMAX_EXPERIENCE );
      OBJ_active( LISTE_MODELES, LM_PROPOSER, FORCE_NON );
    }
    OBJ_choisit( LISTE_MODELES, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case LM_PROPOSER:
    OBJ_choisit( LISTE_MODELES, objet, FORCE_OUI );
    OBJ_MettreNombre( LISTE_MODELES, LM_X_MIN, modele->experience->x_min, 0.0 );
    OBJ_MettreNombre( LISTE_MODELES, LM_X_MAX, modele->experience->x_max, 0.0 );
    OBJ_choisit( LISTE_MODELES, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case LM_ALGORITHME:
    OPTION_A_FAIRE( "Choix de l'algorithme" );
    break;

  case LM_ESTIMATEUR:
    OPTION_A_FAIRE( "Choix de l'estimateur" );
    break;

  case LM_PARAMS_SPECIF:
    OBJ_choisit( LISTE_MODELES, objet, FORCE_OUI );
    LMOD_ParametresLibres( modele );
    OBJ_choisit( LISTE_MODELES, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case LM_ITERATIONS:
    reponse = OBJ_choisit( LISTE_MODELES, objet, ECHANGE );
    OBJ_active( LISTE_MODELES, LM_NOMBRE_ITERAT, reponse );
    return BI_EVENEMENT_TRAITE;

  case LM_ERREUR:
    reponse = OBJ_choisit( LISTE_MODELES, objet, ECHANGE );
    OBJ_active( LISTE_MODELES, LM_ERREUR_MIN, reponse );
    return BI_EVENEMENT_TRAITE;

  case LM_DELTA_ERREUR:
    reponse = OBJ_choisit( LISTE_MODELES, objet, ECHANGE );
    OBJ_active( LISTE_MODELES, LM_DELTA_ERR, reponse );
    return BI_EVENEMENT_TRAITE;

  case LM_AJUSTER:
    OBJ_choisit( LISTE_MODELES, objet, FORCE_OUI );
    LMOD_EnregistrerModele( NULL );
    MODELE_AjustementComplet( modele, NULL, VRAI, VRAI, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );
    if ( modele->residu != NULL ) {
      /* Il y a des résultats d'ajustement ! */
      OBJ_active( LISTE_MODELES, LM_OPT_RESIDUS , FORCE_OUI );
      OBJ_active( LISTE_MODELES, LM_OPT_PARAMS  , FORCE_OUI );
      OBJ_active( LISTE_MODELES, LM_OPT_VALIDITE, FORCE_OUI );
    }
    OBJ_choisit( LISTE_MODELES, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;
  }

  /* Autre bouton : fonction par défaut... */
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Mettre à jour la liste des modèles, si la famille est celle affichée -------- */

void LMOD_MettreAJour(unsigned int numero_famille)
{
  if ( numero_famille != infos.famille ) return; /* Autre famille que celle affichée : pas à s'en préoccuper ! */

  /* On fait croire qu'une autre famille était affichée --- il n'y aura jamais 80000 familles ! */
  infos.famille = 80000;

  /* Et on demande de passer à cette famille */
  LMOD_ChangerFamille( numero_famille );
}

void LMOD_ChangerFamille(unsigned int nv_famille)
{
  if ( nv_famille >= LM_NOMBRE_FAMILLES ) {
    (void) fprintf( stderr, "%s,%s l.%u : Numéro de famille absurde [nv_famille=%u]\n",
		    __FILE__, __FUNCTION__, __LINE__, nv_famille );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  if ( nv_famille == infos.famille ) return; /* On reste sur la même famille */

  /* On mémorise la famille en cours d'édition ! */
  infos.famille = nv_famille;

  /* On met le nom de la famille dans l'objet */
  OBJ_CopierTexte( LISTE_MODELES, LM_FAMILLE, noms_familles_modeles[ nv_famille ], LGMAX_FAMILLE_MODELES );

  /* On récupère le premier modèle de la série */
  switch ( nv_famille )
  {
  case LM_FAMILLE_GENERIQUE_R:
    infos.premier_modele = I_ModeleGenerique_Premier();
    break;

  case LM_FAMILLE_MOLECULE:
    if ( NULL == mol_courante ) {
      infos.premier_modele = NULL;
    } else {
      infos.premier_modele = mol_courante->modele;
    }
    break;

  case LM_FAMILLE_ABSORBEUR:
    if ( ( NULL == mol_courante ) || ( NULL == mol_courante->absorbeurs ) ) {
      infos.premier_modele = NULL;
    } else {
      infos.premier_modele = mol_courante->absorbeurs[ mol_courante->absorbeur_courant ].modeles;
    }
    break;

  case LM_FAMILLE_DECOMPOSER:
    infos.premier_modele = I_ModeleDecomposition_Premier();
    break;

  default:
    OPTION_A_FAIRE( "Détection du premier modèle : famille ??" );
    break;
  }

  /* Initialisation en fonction */
  if ( NULL == infos.premier_modele ) {
    infos.nombre_modeles = 0;
    LMOD_ToutCouper( VRAI );
  } else {
    ModeleAjustement *modele;
    unsigned long n = 0;

    LMOD_ToutCouper( FAUX );

    modele = infos.premier_modele;
    while ( modele != NULL ) {
      n++;
      modele = modele->suivant;
    }
    infos.nombre_modeles = n;
  }
  infos.premier_affiche = 0;
  infos.modele_actif = 0;

  /* On prépare le défileur en fonction de cela */
  ARB_PreparerDefileur_O( LISTE_MODELES, LM_PREMIER_MODEL, LM_DERNIER_MODEL, LGMAX_NOM_MODELE,
			  LM_MODELE_PREV, LM_FOND_ASC, LM_BOUTON_ASC, LM_MODELE_NEXT,
			  infos.nombre_modeles, NULL, infos.modele_actif, infos.premier_affiche,
			  LMOD_NVTexte );

  /* On affiche pour le modèle choisi */
  LMOD_AfficherModele( infos.premier_modele );
}

static void LMOD_AfficherModele(ModeleAjustement *modele)
{
  if ( NULL == modele ) {
    return;
  }

  if ( VRAI == modele->drapeaux.en_cours ) {
    /* Ajustement en cours : pas mal de choses sont inaccessibles */
    LMOD_CouperGeneraux( VRAI );
    LMOD_CouperArret   ( VRAI );

    OBJ_active( LISTE_MODELES, LM_PARAMS_SPECIF, FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_TYPE_DU_MODEL, FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_AJUSTER, FORCE_NON );
  } else {
    LMOD_CouperGeneraux( FAUX );
    LMOD_CouperArret   ( FAUX );

    OBJ_active( LISTE_MODELES, LM_PARAMS_SPECIF, FORCE_OUI );
    OBJ_active( LISTE_MODELES, LM_TYPE_DU_MODEL, FORCE_OUI );
    OBJ_active( LISTE_MODELES, LM_AJUSTER, FORCE_OUI );
  }

  /* Le graphe expérimental associé */
  if ( NULL == modele->experience ) {
    OBJ_CopierTexte( LISTE_MODELES, LM_GRAPHE_EXP, txt_aucun_graphe_modele, LGMAX_EXPERIENCE );
    OBJ_active( LISTE_MODELES, LM_PROPOSER, FORCE_NON );
  } else {
    OBJ_CopierTexte( LISTE_MODELES, LM_GRAPHE_EXP, modele->experience->nom, LGMAX_EXPERIENCE );
    OBJ_active( LISTE_MODELES, LM_PROPOSER, FORCE_OUI );
  }

  /* Les bornes d'ajustement */
  OBJ_MettreIntervalle( LISTE_MODELES, LM_X_MIN, LM_X_MAX, ROOT, ROOT, modele->x_min, modele->x_max, 0, 0 );

  /* La méthode de minimisation */
  OBJ_CopierTexte( LISTE_MODELES, LM_ALGORITHME, "Pseudo-gradient", LGMAX_ALGORITHME );
  OBJ_CopierTexte( LISTE_MODELES, LM_ESTIMATEUR, "Moindres carr"e_AIGU"s", LGMAX_ESTIMATEUR );
  OBJ_choisit( LISTE_MODELES, LM_AVEC_ERREURS , modele->drapeaux.pondere);
  OBJ_choisit( LISTE_MODELES, LM_AV_CONTRAINTE, modele->drapeaux.contraintes);

  /* Que faire en fin d'ajustement ? */
  if ( NULL == modele->residu ) {
    /* Ajustement jamais fait ou perdu */
    OBJ_active( LISTE_MODELES, LM_OPT_RESIDUS , FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_OPT_PARAMS  , FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_OPT_VALIDITE, FORCE_NON );
  } else {
    OBJ_active( LISTE_MODELES, LM_OPT_RESIDUS , FORCE_OUI );
    OBJ_active( LISTE_MODELES, LM_OPT_PARAMS  , FORCE_OUI );
    OBJ_active( LISTE_MODELES, LM_OPT_VALIDITE, FORCE_OUI );
  }
  if ( modele->condition_arret & ARRET_AJUSTEMENT_TESTS ) {
    OBJ_choisit( LISTE_MODELES, LM_FIN_VALIDITE, FORCE_OUI );
  } else {
    OBJ_choisit( LISTE_MODELES, LM_FIN_VALIDITE, FORCE_NON );
  }
  if ( modele->condition_arret & ARRET_AJUSTEMENT_COVARIANCE ) {
    OBJ_choisit( LISTE_MODELES, LM_FIN_MATRICE, FORCE_OUI );
  } else {
    OBJ_choisit( LISTE_MODELES, LM_FIN_MATRICE, FORCE_NON );
  }

  /* Les critères d'arrêt de l'ajustement */
  if ( modele->condition_arret & ARRET_AJUSTEMENT_ITERATIONS ) {
    OBJ_choisit( LISTE_MODELES, LM_ITERATIONS, FORCE_OUI );
    OBJ_MettreEntier_U( LISTE_MODELES, LM_NOMBRE_ITERAT, modele->max_iteration );
  } else {
    OBJ_choisit( LISTE_MODELES, LM_ITERATIONS, FORCE_NON );
  }

  if ( modele->condition_arret & ARRET_AJUSTEMENT_ERREUR ) {
    OBJ_choisit( LISTE_MODELES, LM_ERREUR, FORCE_OUI );
    OBJ_active ( LISTE_MODELES, LM_ERREUR_MIN, FORCE_OUI );
    OBJ_MettreEntier_U( LISTE_MODELES, LM_ERREUR_MIN, modele->err_min );
  } else {
    OBJ_choisit( LISTE_MODELES, LM_ITERATIONS, FORCE_NON );
    OBJ_active ( LISTE_MODELES, LM_ERREUR_MIN, FORCE_NON );
  }

  if ( modele->condition_arret & ARRET_AJUSTEMENT_DELTA_ERREUR ) {
    OBJ_choisit( LISTE_MODELES, LM_DELTA_ERREUR, FORCE_OUI );
    OBJ_active ( LISTE_MODELES, LM_DELTA_ERR, FORCE_OUI );
    OBJ_MettreEntier_U( LISTE_MODELES, LM_DELTA_ERR, modele->derr_min );
  } else {
    OBJ_choisit( LISTE_MODELES, LM_DELTA_ERREUR, FORCE_NON );
    OBJ_active ( LISTE_MODELES, LM_DELTA_ERR, FORCE_NON );
  }

  if ( modele->condition_arret & ARRET_AJUSTEMENT_DELTA_VARIABLE ) {
    OBJ_choisit( LISTE_MODELES, LM_DELTA_PARAM, FORCE_OUI );
  } else {
    OBJ_choisit( LISTE_MODELES, LM_DELTA_PARAM, FORCE_NON );
  }

  /* Le nom du modèle */
  OBJ_CopierTexte( LISTE_MODELES, LM_NOM_MODELE, modele->nom_modele, 0 );
}

/* -------- Mémoriser les informations du modèle --------
   On peut l'appeler de l'extérieur : on mémorise quand on lance un ajustement...
 */

void LMOD_EnregistrerModele(ModeleAjustement *modele_demande)
{
  ModeleAjustement *modele;
  char *chaine;

  /* On cherche quel est le modèle actuellement modifié */
  modele = LMOD_ModeleEdite();
  if ( NULL == modele ) return;	/* Aucun modèle en cours de modification */

  /* On vérifie que c'est bien celui demandé */
  if ( ( modele_demande != NULL ) && ( modele_demande != modele ) ) {
    /* Le modèle indiqué n'est pas celui édité : inutile d'enregistrer */
    return;
  }

  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) BI_ERR_alerte( principale, &MSG_AjustementEnCours, VRAI );
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Modification d'un modèle en cours d'ajustement : interdit !\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return;
  }

  /* -- Paramètres généraux -- */
  /* Le graphe expérimental est changé "en temps réel" */
  /* Bornes d'ajustement */
  OBJ_LireIntervalle( LISTE_MODELES, LM_X_MIN, LM_X_MAX, ROOT, ROOT, 
		      &(modele->x_min), &(modele->x_max), NULL, NULL );
  /* Algorithme */
  modele->algorithme = OBJ_OptionMenu( LISTE_MODELES, LM_ALGORITHME, 0, QUESTION );
  /* Estimateur utilise */
/*   modele->estimateur = OBJ_OptionMenu( LISTE_MODELES, LM_ESTIMATEUR, 0, QUESTION ); */
  /* Autres conditions d'ajustement */
  modele->drapeaux.pondere     = OBJ_choisit( LISTE_MODELES, LM_AVEC_ERREURS , QUESTION );
  modele->drapeaux.contraintes = OBJ_choisit( LISTE_MODELES, LM_AV_CONTRAINTE, QUESTION );

  /* -- Résultats -- */
  /* Que faire en fin d'ajustement ? */
  if ( OBJ_choisit( LISTE_MODELES, LM_FIN_VALIDITE, QUESTION ) == VRAI ) {
    modele->condition_arret |= ARRET_AJUSTEMENT_TESTS;
  } else {
    modele->condition_arret &= ~ARRET_AJUSTEMENT_TESTS;
  }
  if ( OBJ_choisit( LISTE_MODELES, LM_FIN_MATRICE, QUESTION ) == VRAI ) {
    modele->condition_arret |= ARRET_AJUSTEMENT_COVARIANCE;
  } else {
    modele->condition_arret &= ~ARRET_AJUSTEMENT_COVARIANCE;
  }

  /* -- Critères d'arrêt -- */
  if ( OBJ_choisit( LISTE_MODELES, LM_ITERATIONS, QUESTION ) == VRAI ) {
    modele->condition_arret |= ARRET_AJUSTEMENT_ITERATIONS;
    modele->max_iteration = OBJ_LireEntier_U( LISTE_MODELES, LM_NOMBRE_ITERAT );
  } else {
    modele->condition_arret &= ~ARRET_AJUSTEMENT_ITERATIONS;
    modele->max_iteration = 100000;
  }
  if ( OBJ_choisit( LISTE_MODELES, LM_ERREUR, QUESTION ) == VRAI ) {
    modele->condition_arret |= ARRET_AJUSTEMENT_ERREUR;
    modele->err_min = OBJ_LireEntier_S( LISTE_MODELES, LM_ERREUR_MIN );
  } else {
    modele->condition_arret &= ~ARRET_AJUSTEMENT_ERREUR;
  }
  if ( OBJ_choisit( LISTE_MODELES, LM_DELTA_ERREUR, QUESTION ) == VRAI ) {
    modele->condition_arret |= ARRET_AJUSTEMENT_DELTA_ERREUR;
    modele->derr_min = OBJ_LireEntier_S( LISTE_MODELES, LM_DELTA_ERR );
  } else {
    modele->condition_arret &= ~ARRET_AJUSTEMENT_DELTA_ERREUR;
  }
  if ( OBJ_choisit( LISTE_MODELES, LM_DELTA_PARAM, QUESTION ) == VRAI ) {
    modele->condition_arret |= ARRET_AJUSTEMENT_DELTA_VARIABLE;
  } else {
    modele->condition_arret &= ~ARRET_AJUSTEMENT_DELTA_VARIABLE;
  }

  /* Le nom du modèle */
  chaine = lit_chaine( LISTE_MODELES, LM_NOM_MODELE, 0, NULL, TXT_COPIER );
  MODELE_Nommer( modele, chaine );
  free( chaine ); chaine = NULL;
}

/* -------- Macros d'activation / désactivation -------- */

static void LMOD_ToutCouper(boolean couper)
{
  LMOD_CouperGeneraux ( couper );
  LMOD_CouperResultats( couper );
  LMOD_CouperArret    ( couper );

  if ( VRAI == couper ) {
    OBJ_active( LISTE_MODELES, LM_PARAMS_SPECIF, FORCE_NON );

    OBJ_active( LISTE_MODELES, LM_NOM_MODELE, FORCE_NON );

    OBJ_active( LISTE_MODELES, LM_TYPE_DU_MODEL, FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_AJUSTER, FORCE_NON );
  } else {
    OBJ_active( LISTE_MODELES, LM_NOM_MODELE, FORCE_OUI );
    /* Autres : dépendent de l'état => fait ailleurs */
  }
}

static void LMOD_CouperArret(boolean couper)
{
  int nouvel_etat;

  if ( VRAI == couper ) nouvel_etat = FORCE_NON;
  else                  nouvel_etat = FORCE_OUI;

  OBJ_active( LISTE_MODELES, LM_ITERATIONS, nouvel_etat );
  if ( VRAI == couper ) {
    OBJ_active( LISTE_MODELES, LM_NOMBRE_ITERAT, FORCE_NON );
  } /* Sinon : dépend de l'état => fait ailleurs */

  OBJ_active( LISTE_MODELES, LM_DELTA_PARAM, nouvel_etat );

  OBJ_active( LISTE_MODELES, LM_ERREUR, nouvel_etat );
  if ( VRAI == couper ) {
    OBJ_active( LISTE_MODELES, LM_ERREUR_MIN, FORCE_NON );
  } /* Sinon : dépend de l'état => fait ailleurs */

  OBJ_active( LISTE_MODELES, LM_DELTA_ERREUR, nouvel_etat );
  if ( VRAI == couper ) {
    OBJ_active( LISTE_MODELES, LM_DELTA_ERR, FORCE_NON );
  } /* Sinon : dépend de l'état => fait ailleurs */
}

static void LMOD_CouperGeneraux(boolean couper)
{
  int nouvel_etat;

  if ( VRAI == couper ) nouvel_etat = FORCE_NON;
  else                  nouvel_etat = FORCE_OUI;

  OBJ_active( LISTE_MODELES, LM_GRAPHE_EXP, nouvel_etat );

  OBJ_active( LISTE_MODELES, LM_X_MIN, nouvel_etat );
  OBJ_active( LISTE_MODELES, LM_X_MAX, nouvel_etat );
  if ( VRAI == couper ) {
    OBJ_active( LISTE_MODELES, LM_PROPOSER, FORCE_NON );
  } else {
    /* Activation dépend de la présence d'un graphe expérimental : fait ailleurs ! */
  }

  OBJ_active( LISTE_MODELES, LM_ALGORITHME, nouvel_etat );
  OBJ_active( LISTE_MODELES, LM_ESTIMATEUR, nouvel_etat );

  OBJ_active( LISTE_MODELES, LM_AVEC_ERREURS , nouvel_etat );
  OBJ_active( LISTE_MODELES, LM_AV_CONTRAINTE, nouvel_etat );
}

static void LMOD_CouperResultats(boolean couper)
{
  int nouvel_etat;

  if ( VRAI == couper ) nouvel_etat = FORCE_NON;
  else                  nouvel_etat = FORCE_OUI;

  if ( VRAI == couper ) {
    OBJ_active( LISTE_MODELES, LM_OPT_RESIDUS , FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_OPT_PARAMS  , FORCE_NON );
    OBJ_active( LISTE_MODELES, LM_OPT_VALIDITE, FORCE_NON );
  } else {
    /* Dépend de la présence de résultats : ailleurs ! */
  }

  OBJ_active( LISTE_MODELES, LM_FIN_VALIDITE, nouvel_etat );
  OBJ_active( LISTE_MODELES, LM_FIN_MATRICE , nouvel_etat );
}

/* -------- Aides à la gestion -------- */

static ModeleAjustement *LMOD_ModeleEdite(void)
{
  ModeleAjustement *modele;
  unsigned long i;

  if ( NULL == infos.premier_modele ) return NULL;	/* Aucun modèle en cours de modification */

  /* On avance dans la liste jusqu'à avoir le bon modèle */
  for ( i = 0, modele = infos.premier_modele ;
	i < infos.modele_actif ;
	i++, modele = modele->suivant ) {
    if ( NULL == modele ) {
      AVERTISSEMENT( "Modèle %lu actif, mais seulement %lu dans la liste", infos.modele_actif, i );
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return NULL;
    }
  }

  /* On renvoie le modèle en cours d'édition */
  return modele;
}

static const char *LMOD_NVTexte(unsigned int arbre __attribute__((__unused__)), unsigned long numero_objet)
{
  ModeleAjustement *modele;
  unsigned long i;

  for ( i = 0, modele = infos.premier_modele ; i < numero_objet ; modele = modele->suivant, i++ ) {
    if ( NULL == modele ) {
      AVERTISSEMENT( "Objet %lu demandé, mais NULL", i );
      return " ";		/* Pas de modèle associé ! */
    }
  }
  /* Là, on est sur le modèle demandé */
  if ( modele != NULL ) {
    return modele->nom_modele;
  }

  /* Pas de modèle ! */
  AVERTISSEMENT( "Objet %lu demandé, mais NULL", i );
  return " ";
}

/* ~-~-~-~-~-~-~-~-~ Options avancées ~-~-~-~-~-~-~-~-~ */

/* -------- Choisir un graphe expérimental à modéliser -------- */

static void LMOD_ChoisirExperience(ModeleAjustement *modele)
{
  Graphe *choisi = NULL;

  if ( NULL == modele ) return;	/* Aucun modèle en cours d'ajustement */

  /* En fonction de la famille de modèle, on choisit les graphes adaptés */
  switch ( infos.famille )
  {
  case LM_FAMILLE_GENERIQUE_R:
    choisi = GRF_choisir( GR_TOUS, FAUX, VRAI, NULL );
    break;

  case LM_FAMILLE_MOLECULE:
    OPTION_A_FAIRE( "Demander une liste des graphes potentiels / modèle molécule" );
    break;

  case LM_FAMILLE_ABSORBEUR:
    switch ( modele->type )
    {
    case MODELE_EXAFS_CHEMINS:	/* Type FEFF */
      OPTION_A_FAIRE( "Choisir dans une liste de graphes plus souple (donc y compris TF, par exemple)" );
      choisi = GRF_choisir( GR_TT_EXAFS, FAUX, VRAI, NULL );
      break;

    default:
      OPTION_A_FAIRE( "Modèle de type autre que FEFF" );
    }
    break;

  default:
    AVERTISSEMENT( "Famille de modèles inconnue [%d]", infos.famille );
    return;
  }

  if ( NULL == choisi ) return;	/* Rien à faire ! */

  /* On change le graphe expérimental */
  MODELE_MettreReference( modele, choisi, QUESTION );
}

/* -------- Paramétrer le modèle, paramètres spécifiques -------- */

static void LMOD_ParametresLibres(ModeleAjustement *modele)
{
  if ( NULL == modele ) return;

  switch ( modele->type )
  {
  case MODELE_EXAFS_CHEMINS:
    COUCHE_OuvrirEditeur( modele->infos_modele.feff );
    break;

  case MODELE_GENERIQUE:
    I_ModeleGenerique( &(modele->infos_modele.generique) );
    break;

  default:
    AVERTISSEMENT( "Type de modèle inconnu [%d]", modele->type );
    return;
  }
}

/* -------- Sous-menu du bouton "options" [options générales...] -------- */

static void LMOD_GererMenuOptions(unsigned int num_arbre __attribute__((__unused__)), unsigned int num_objet __attribute__((__unused__)), unsigned long action)
{
  ModeleAjustement *modele __attribute__((__unused__));

  /* On mémorise le modèle en cours d'édition */
  LMOD_EnregistrerModele( NULL );

  /* L'option "nouveau" fonctionne toujours */
  if ( LM_OPTIONS_CREER == action ) {
    int retour;

    switch ( infos.famille )
    {
    case LM_FAMILLE_GENERIQUE_R:
      retour = I_ModeleGenerique_Nouveau( infos.modele_actif );
      if ( retour != 0 ) {
	AVERTISSEMENT( "Problème dans I_ModeleGenerique_Nouveau, code %d", retour );
	return;
      }
      break;

    case LM_FAMILLE_MOLECULE:
      OPTION_A_FAIRE( "Créer un nouveau modèle dans la molécule courante" );
      break;

    case LM_FAMILLE_ABSORBEUR:
      retour = I_ModeleAbsorbeur_Nouveau( NULL, infos.modele_actif );
      if ( retour != 0 ) {
	AVERTISSEMENT( "Problème dans I_ModeleAbsorbeur_Nouveau, code %d", retour );
	return;
      }
      break;

    case LM_FAMILLE_DECOMPOSER:
      (void) I_ModeleDecomposition_Nouveau( infos.modele_actif, NULL );
      break;

    default:
      AVERTISSEMENT( "Famille de modèle inconnue ou absurde : %d", infos.famille );
      break;
    }

    /* On remet à jour la liste des modèles */
    LMOD_MettreAJour( infos.famille );
    return;
  }

  /* Quel est le modèle en cours de modification ? */
  modele = LMOD_ModeleEdite();

  switch ( action )
  {
  case LM_OPTIONS_EFFACER:
    OPTION_A_FAIRE( "Effacer le modèle actif" );
    break;

  case LM_OPTIONS_COPIER:
    OPTION_A_FAIRE( "Copier le modèle actif" );
    break;

  default:
    AVERTISSEMENT( "Option inconnue [%lu]", action );
    return;
  }

  /* On met à jour la liste des modèles ! */
  LMOD_MettreAJour( infos.famille );
}

/* -------- Appeler les options de calculs sur les residus -------- */

static void LMOD_GererMenuResidus(unsigned int num_arbre __attribute__((__unused__)), unsigned int num_objet __attribute__((__unused__)), unsigned long action)
{
  ModeleAjustement *modele;

  /* On récupère le modèle actuellement affiché */
  modele = LMOD_ModeleEdite();
  if ( NULL == modele ) return;
  if ( NULL == modele->residu ) {
    (void) BI_ERR_alerte( NULL, &ERR_AucunResidu, VRAI );
    return;
  }
  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) BI_ERR_alerte( NULL, &MSG_AjustementEnCours, VRAI );
    return;
  }

  switch ( action )
  {
  case LM_RESIDUS_ABSOLUS:	/* Calculer les residus absolus */
    I_MOD_AfficherResidusAbsolus( modele );
    return;

  case LM_RESIDUS_RELATIFS:	/* Calculer les residus relatifs */
    I_MOD_AfficherResidusRelatifs( modele );
    return;

  case LM_RESIDUS_PONDERES:	/* Calculer les residus ponderes par les incertitudes */
    I_MOD_AfficherResidusPonderes( modele );
    return;

  default:
    AVERTISSEMENT( "Option inconnue [action = %lu]", action )
  }
}

/* -------- Appeler les options de validité [simple appel à MODELE_ControlerValidite() !] -------- */

static void LMOD_GererMenuValidite(unsigned int num_arbre __attribute__((__unused__)), unsigned int num_objet __attribute__((__unused__)), unsigned long action)
{
  ModeleAjustement *modele;

  /* On récupère le modèle actuellement affiché */
  modele = LMOD_ModeleEdite();

  /* Et on en vérifie la validité avec l'opération choisie ! */
  I_MOD_VALIDITE_ControlerValidite( modele, action );
}

/* ~-~-~-~-~-~-~-~- Initialisation de l'arbre ~-~-~-~-~-~-~-~- */

void LMOD_PreparerArbre(void)
{
  /* Restent en bouton, pour simplicité :
      - le sous-menu de choix des familles
      - celui de choix du graphe expérimental associé */

  /* Les sélecteurs d'options peuvent être convertis en sous-menus */
  OBJ_ChangerType( LISTE_MODELES, LM_OPTIONS, G_SSMENU );
  OBJ_MettreMenu ( LISTE_MODELES, LM_OPTIONS, LM_NOMBRE_OPTIONS, options_modeles, NULL, LMOD_GererMenuOptions );

  OBJ_ChangerType( LISTE_MODELES, LM_OPT_VALIDITE, G_SSMENU );
  OBJ_MettreMenu ( LISTE_MODELES, LM_OPT_VALIDITE, MOD_NOMBRE_METHODES_VALIDATION, options_validite, titre_ssmenu_validite, LMOD_GererMenuValidite );

  OBJ_ChangerType( LISTE_MODELES, LM_ALGORITHME, G_SSMENU );
  OBJ_MettreMenu ( LISTE_MODELES, LM_ALGORITHME, MOD_NOMBRE_ALGORITHMES, noms_algorithmes, titre_ssmenu_algorithme, NULL );

  OBJ_ChangerType( LISTE_MODELES, LM_ESTIMATEUR, G_SSMENU );
  OBJ_MettreMenu ( LISTE_MODELES, LM_ESTIMATEUR, MOD_NOMBRE_ALGORITHMES, noms_estimateurs, titre_ssmenu_estimateur, NULL );

  OBJ_ChangerType( LISTE_MODELES, LM_OPT_RESIDUS, G_SSMENU );
  OBJ_MettreMenu ( LISTE_MODELES, LM_OPT_RESIDUS, LM_NOMBRE_OPTIONS_RESIDUS, options_residus, titre_ssmenu_residus, LMOD_GererMenuResidus );

  /* Initialisation */
  infos.famille = 82500; /* Pour que le 1er appel avec changement ait lieu !
			    [il n'y aura jamais autant de familles de modèle, j'espère...] */
  infos.premier_modele  = NULL;
  infos.nombre_modeles  = 0;
  infos.premier_affiche = 0;
  infos.modele_actif    = 0;

  if ( FAUX == arbres[ LISTE_MODELES ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( LISTE_MODELES, "", LMOD_GererArbre, ( int (*)(unsigned int, int, int *) ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR à l'initialisation de l'arbre [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return;
    }
  }

  /* Meilleure disposition des objets */
  ARB_EnjoliverDefileur( LISTE_MODELES, LM_FND_MODELES, LM_PREMIER_MODEL, LM_DERNIER_MODEL,
			 LM_MODELE_PREV, LM_FOND_ASC, LM_BOUTON_ASC, LM_MODELE_NEXT, ARB_ASCENSEUR_DROITE );

  OBJ_deplace( LISTE_MODELES, LM_TXT_PARAMETRE, 0, +500, DEPLACE_RELATIF );

  OBJ_deplace( LISTE_MODELES, LM_TXT_XMAX, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_X_MAX   , 0, +1, DEPLACE_PIXELS);

  OBJ_deplace( LISTE_MODELES, LM_TXT_ESTIMATEU, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_ESTIMATEUR   , 0, +1, DEPLACE_PIXELS);

  OBJ_deplace( LISTE_MODELES, LM_TXT_RESULTATS, 0, +500, DEPLACE_RELATIF );

  OBJ_deplace( LISTE_MODELES, LM_TXT_PARAMS, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_OPT_PARAMS, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_TXT_VALIDITE, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_OPT_VALIDITE, 0, +2, DEPLACE_PIXELS);

  OBJ_deplace( LISTE_MODELES, LM_TXT_FIN_AJUST, 0, -1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_FIN_MATRICE, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_TXT_MATRICE, 0, +2, DEPLACE_PIXELS);

  OBJ_deplace( LISTE_MODELES, LM_TXT_ARRET, 0, +500, DEPLACE_RELATIF );

  OBJ_deplace( LISTE_MODELES, LM_DELTA_PARAM, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_TXT_SI_PARAM, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_ERREUR, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_TXT_SI_ERR_IN, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_ERREUR_MIN, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_DELTA_ERREUR, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_TXT_SI_ERR_VA, 0, +3, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_DELTA_ERR, 0, +3, DEPLACE_PIXELS);

  OBJ_deplace( LISTE_MODELES, LM_TXT_NOM, 0, -3, DEPLACE_PIXELS);
  OBJ_deplace( LISTE_MODELES, LM_NOM_MODELE, 0, -3, DEPLACE_PIXELS);

#ifdef AVEC_TRADUCTION_
#endif

  arbres[ LISTE_MODELES ].drapeaux.prepare = VRAI;

  /* On commence par afficher les modèles génériques... */
  LMOD_ChangerFamille( LM_FAMILLE_GENERIQUE_R );
}
