/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2003, Emmanuel CURIS

   Généralités du programme LASE (interface graphique)

   Gestion et préparation du menu des calculs & modèles théoriques
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "systeme.h"

#include "exafs.h"

#include "interface.h"
#include "modeles.h"
#include "serie.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#include "int_modeles_affichage.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Lancer l'ajustement et, le cas échéant, afficher les messages indiquant les échecs de l'ajustement ------ */

int I_MOD_LancerAjustement(struct modele_ajuste *modele)
{
  int retour;
  /* Fenetre *fenetre = NULL; */

  if ( NULL == modele ) return +1;

  /* On lance l'ajustement en interne */
  retour = MODELE_AjustementComplet( modele, NULL, FAUX, FAUX, Q_MODELE_ANCIENS_RESULTATS_DEMANDER );

  /* On vérifie le résultat */
  if ( retour != 0 ) {
    switch( retour ) {
    case ERREUR_MODELE_AJUSTEMENT_EN_COURS:
      BI_ERR_annuler( texte_erreur_ajustement_en_cours );
      break;

    default:
      BI_ERR_annuler( texte_erreur_indefinie, retour );
    }
  }

  /* On a fini... */
  return retour;
}

/* -------- Affichage des valeurs des residus "absolus", au sens des moindres carres -------- */

int I_MOD_AfficherResidusAbsolus(struct modele_ajuste *modele)
{
  double scr, cmr, variance = -1;
  unsigned long nombre_points, nombre_parametres;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->residu ) {
    (void) fprintf( stderr, "%s : Pas de résidu pour le modele !\n", __FUNCTION__ );
    return -1;
  }

  /* Nombre de points ajustés */
  nombre_points = modele->residu->nombre_point;

  /* Nombre de variables ajustées */
  nombre_parametres = modele->nbr_variables;

  /* Erreur fondamentale : somme des carrés des résidus */
  scr = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, FAUX, FAUX );

  /* Valeurs dérivées */
  cmr = scr / nombre_points;
  if ( nombre_points > nombre_parametres ) {
    variance = scr / ( nombre_points - nombre_parametres );
  } else {
    OPTION_A_FAIRE( "Modèle surparamétré : afficher un message d'erreur !" )
  }

  /* Affichage */
  (void) BI_ERR_alerte( NULL, &MSG_ResidusAbsolus, VRAI,
			nombre_parametres, nombre_points, scr, cmr, variance );
  return 0;
}

/* -------- Affichage des valeurs des residus "ponderes" -------- */

int I_MOD_AfficherResidusPonderes(struct modele_ajuste *modele)
{
  double scr, cmr, variance = -1.0;
  unsigned long nombre_points, nombre_parametres;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->residu ) {
    (void) fprintf( stderr, "%s : Pas de residu pour le modele !\n", __FUNCTION__ );
    return -2;
  }
  if ( ( NULL == modele->reference ) || ( NULL == modele->reference->erreur_y ) ) {
    (void) BI_ERR_alerte( NULL, &ERR_PasDIncertitude, VRAI );
    return -3;
  }

  /* Nombre de points ajustes */
  nombre_points = modele->residu->nombre_point;

  /* Nombre de variables ajustees */
  nombre_parametres = modele->nbr_variables;

  /* Erreur fondamentale : somme des carres des residus */
  scr = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, VRAI, FAUX );

  /* Valeurs derivees */
  cmr = scr / nombre_points;
  if ( nombre_points > nombre_parametres ) {
    variance = scr / ( nombre_points - nombre_parametres );
  } else {
    OPTION_A_FAIRE( "Modele surparametre : afficher un message d'erreur !" )
  }

  /* Affichage */
  (void) BI_ERR_alerte( NULL, &MSG_ResidusPonderes, VRAI,
			nombre_parametres, nombre_points,
			scr, cmr, variance );
  return 0;
}

/* -------- Affichage des valeurs des residus "relatifs" -------- */

int I_MOD_AfficherResidusRelatifs(struct modele_ajuste *modele)
{
  double scr, scr_relatif, sy2, cmr_relatif, variance_relatif = -1.0;
  unsigned long nombre_points, nombre_parametres;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->residu ) {
    (void) fprintf( stderr, "%s : Pas de residu pour le modele !\n", __FUNCTION__ );
    return -1;
  }

  /* Nombre de points ajustes */
  nombre_points = modele->residu->nombre_point;

  /* Nombre de variables ajustees */
  nombre_parametres = modele->nbr_variables;

  /* Erreur fondamentale : somme des carres des residus */
  scr         = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, FAUX, FAUX );
  scr_relatif = MODELE_CalculerErreur( modele, MODELE_ERREUR_CARRES, FAUX, VRAI );
  sy2         = MODELE_CalculerErreur( modele, MODELE_ERREUR_NORME2, FAUX, FAUX );

  /* Valeurs derivees */
  cmr_relatif = scr_relatif / nombre_points;
  if ( nombre_points > nombre_parametres ) {
    variance_relatif = scr_relatif / ( nombre_points - nombre_parametres );
  } else {
    OPTION_A_FAIRE( "Modele surparametre : afficher un message d'erreur !" )
  }

  /* Affichage */
  (void) BI_ERR_alerte( NULL, &MSG_ResidusRelatifs, VRAI,
			nombre_parametres, nombre_points,
			scr_relatif, cmr_relatif, variance_relatif,
			scr / sy2 );
  return 0;
}
