/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2003, Emmanuel Curis

   Interface utilisateur : routines gnrales indpendantes du systme

   Fonctions avances pour la gestion de certains objets
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "systeme.h"

#include "chimie_utils.h"
#include "txt_utils.h"
#include "interface.h"
#include "absorption.h"
#include "cristallo.h"		/* Pour les systmes cristallins */
#include "graphes.h"		/* Pour les types de graphes */
#include "images.h"		/* Pour les types et genres d'images */
#include "science.h"		/* Pour M_PI */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#include "int_objets.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void InstallerSysteme(struct systeme_cristallin *systeme, unsigned int arbre, unsigned int objet, unsigned long lgmax_texte,
			     unsigned int obj_a, unsigned int obj_b, unsigned int obj_c, unsigned int obj_alpha, unsigned int obj_beta, unsigned int obj_gamma);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Choisir le modle semi-empirique de coefficient d'absorption atomique pour un seuil d'absorption -------- */

int OBJ_ChoisirModeleMu0(unsigned int arbre, unsigned int objet, unsigned long lgmax_texte, int modele, int mode)
{
  int nv_modele, i;
  const char *noms_modeles[ NOMBRE_MODELES_MU0 ];

  if ( METTRE == mode ) {
    met_chaine( arbre, objet, lgmax_texte, TH_TexteModeleMu0( modele ), TXT_COPIER );
    return modele;
  }

  /* Si on est l, c'est que l'on souhaite changer le modle */
  for ( i = 0 ; i < NOMBRE_MODELES_MU0 ; i++ )
    noms_modeles[ i ] = TH_TexteModeleMu0( i );

  nv_modele = SS_menu_texte( noms_modeles , NOMBRE_MODELES_MU0, "Mod"e_GRAVE"le de "MU"0", arbre, objet);
  if ( nv_modele < 0 ) return modele; /* Modele inchang */
  if ( nv_modele == modele ) return modele;

  met_chaine( arbre, objet, lgmax_texte, noms_modeles[ nv_modele ], TXT_COPIER );
  return nv_modele;
}

/* -------- Choisir le seuil d'absorption -------- */

int OBJ_ChoisirSeuil(unsigned int arbre, unsigned int objet, unsigned long lgmax_texte, int seuil, int mode)
{
  int nv_seuil, i;
  const char *noms_seuils[ NOMBRE_SEUILS ];

  if ( METTRE == mode ) {
    met_chaine( arbre, objet, lgmax_texte, TH_TexteSeuil( seuil ), TXT_COPIER );
    return seuil;
  }

  /* Si on est l, c'est que l'on souhaite changer le seuil */
  for ( i = 0 ; i < NOMBRE_SEUILS ; i++ )
    noms_seuils[ i ] = TH_TexteSeuil( i );

  nv_seuil = SS_menu_texte( noms_seuils , NOMBRE_SEUILS, "Choix du seuil", arbre, objet);
  if ( nv_seuil < 0 ) return seuil; /* Seuil inchang */
  if ( nv_seuil == seuil ) return seuil;

  met_chaine( arbre, objet, lgmax_texte, noms_seuils[ nv_seuil ], TXT_COPIER );
  return nv_seuil;
}

/* -------- Choisir le type de graphe  utiliser -------- */

int OBJ_ChoisirTypeGraphe(unsigned int arbre, unsigned int objet, unsigned long lgmax_texte __attribute__((__unused__)), int type, int mode, boolean avec_internes)
{
  unsigned long i, nbr_types;
  const char **noms_types;
  int nv_type;

  if ( METTRE == mode ) {
    OBJ_RemplacerTexte( arbre, objet, GRF_TexteTypeGraphe( type ), VRAI );
    return type;
  }

  OBJ_choisit( arbre, objet, FORCE_OUI );

  nbr_types = NOMBRE_TYPES_GRAPHES;
  if ( VRAI == avec_internes ) {
    nbr_types += 3; /* Autre type & sans type et sparateur */
  }

  noms_types = (const char **) malloc( sizeof( char *) * nbr_types );
  if ( NULL == noms_types ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la table des types / mmoire insuffisante \n",
		   __FILE__, __FUNCTION__, __LINE__);
    OBJ_choisit( arbre, objet, FORCE_NON );
    return type;
  }

  /* On remplit : les types classiques en premier */
  for ( i = 0 ; i < NOMBRE_TYPES_GRAPHES ; i++ ) {
    noms_types[ i ] = GRF_TexteTypeGraphe( i );
  }

  /* Si ncessaire, le sparateur et la suite */
  if ( VRAI == avec_internes ) {
    noms_types[ i++ ] = "----";
    noms_types[ i++ ] = GRF_TexteTypeGraphe( GR_SSTYPE );
    noms_types[ i++ ] = GRF_TexteTypeGraphe( GR_AUTRETYPE );
  }

  /* On appelle ce sous-menu */
  nv_type = SS_menu_texte( noms_types, nbr_types, "Type des donn"e_AIGU"es du graphe", arbre, objet );
  free( noms_types );
  if ( nv_type < 0 ) {
    OBJ_choisit( arbre, objet, FORCE_NON );
    return type;
  }

  /* On tient compte des types spciaux */
  if ( nv_type == NOMBRE_TYPES_GRAPHES + 1 ) nv_type = GR_SSTYPE;
  if ( nv_type == NOMBRE_TYPES_GRAPHES + 2 ) nv_type = GR_AUTRETYPE;

  if ( nv_type == type ) {
    OBJ_choisit( arbre, objet, FORCE_NON );
    return nv_type;
  }

  OBJ_RemplacerTexte( arbre, objet, GRF_TexteTypeGraphe( nv_type ), VRAI );
  OBJ_choisit( arbre, objet, FORCE_NON );
  return nv_type;
}

/* -------- Choisir le genre d'une image -------- */

int OBJ_ChoisirGenreImage(unsigned int arbre, unsigned int objet, int genre, int mode)
{
  const char **noms_genres;
  int nv_genre;
  unsigned long i, nombre_genres;

  if ( objet > ROOT ) {
    OBJ_choisit( arbre, objet, FORCE_OUI );
  }

  if ( METTRE == mode ) {
    OBJ_RemplacerTexte( arbre, objet, IMG_TXT_Genre( genre ), VRAI );
    return genre;
  }

  /* Par dfaut, tous les genres utilisateurs */
  nombre_genres = NOMBRE_GENRE_IMAGES;

  /* On rserve la table */
  noms_genres = (const char **) malloc( sizeof( char *) * nombre_genres );
  if ( NULL == noms_genres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la table des genres / mmoire insuffisante \n",
		   __FILE__, __FUNCTION__, __LINE__);
    OBJ_choisit( arbre, objet, FORCE_NON );
    return genre;
  }

  /* On remplit : les types classiques en premier */
  for ( i = 0 ; i < nombre_genres ; i++ ) {
    noms_genres[ i ] = IMG_TXT_Genre( i );
  }

  /* On appelle le sous-menu */
  nv_genre = SS_menu_texte( noms_genres, nombre_genres, titre_menu_genres_image, arbre, objet );
  free( noms_genres ); noms_genres = NULL;
  if ( nv_genre < 0 ) {
    OBJ_choisit( arbre, objet, FORCE_NON );
    return genre;
  }

  if ( nv_genre == genre ) {
    OBJ_choisit( arbre, objet, FORCE_NON );
    return nv_genre;
  }

  OBJ_RemplacerTexte( arbre, objet, IMG_TXT_Genre( nv_genre ), VRAI );
  OBJ_choisit( arbre, objet, FORCE_NON );
  return nv_genre;
}

/* -------- Lire un numro atomique : Z ou symbole -------- */

int OBJ_NumeroAtomique(unsigned int arbre, unsigned int objet)
{
  char *texte;
  int Z;

  texte = lit_chaine( arbre, objet, 0, NULL, TXT_REMPLACE );
  if ( NULL == texte ) {
    (void) fprintf(stderr, "OBJ_NumeroAtomique : ERREUR dans lit_chaine() [arbre %u, objet %u]\n", arbre, objet);
    return SYMBOLE_CHIMIQUE_INCONNU;
  }

  /* Plus frquent : un symbole -> on essaye de le lire */
  Z = ATM_num_atomique( texte );
  if ( SYMBOLE_CHIMIQUE_INCONNU == Z ) {
    /* On essaye de lire une valeur numrique */
    Z = TXT_CNV_vNombre_L( texte, NULL );
    if ( ( Z < HYDROGENE ) || ( Z > MEITNERIUM ) ) {
      (void) fprintf(stderr, "OBJ_NumeroAtomique : Impossible de trouver Z [%s]\n", texte);
      Z = 0;
    }
  }

  return Z;
}

/* --------- Choisir le groupe d'espace ----------- */

int OBJ_ChoisirGroupeEsp(unsigned int arbre, unsigned int objet, unsigned long lgmax_texte, int actuel, int mode, struct systeme_cristallin *systeme)
{
  union {
    const char **cte;
    char **var;
  } noms;
  char *nom_groupe;
  unsigned long i;
  int choix;

  /* Pas de systme, ou pas de groupes dans le systme -> rien  faire, tant en demande qu'en installation ! */
  if ( ( NULL == systeme ) || ( NULL == systeme->groupes ) || ( 0 == systeme->nbr_groupes ) ) {
    if ( systeme != NULL ) {
      (void) fprintf(stderr, "%s,%sl.%u : Systme %s, %u groupes, en %p\n",
		     __FILE__, __FUNCTION__, __LINE__, systeme->nom, systeme->nbr_groupes, (void *) systeme->groupes );
    }
    OBJ_active( arbre, objet, FORCE_NON );
    met_chaine( arbre, objet, lgmax_texte, "Aucun groupe trouv"e_AIGU, TXT_COPIER );
    return 0;
  }

  if ( METTRE == mode ) {
#ifdef _DEVELOPPE_
    if ( ( actuel < 0 ) || ( actuel >= (int) systeme->nbr_groupes ) ) {
      (void) fprintf(stderr, "%s,%sl.%u : Index de groupe absurde [%d, doit tre entre 0 et %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, actuel, systeme->nbr_groupes - 1);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return 0;
    }
#endif

    nom_groupe = TXT_genere_chaine("%u : %s", systeme->groupes[ actuel ].numero, systeme->groupes[ actuel ].symbole );
    met_chaine( arbre, objet, lgmax_texte, nom_groupe, TXT_COPIER );
    free( nom_groupe );
    return actuel;
  }

  /* Si on est l : on demande -> bouton activ ! */
  OBJ_choisit( arbre, objet, FORCE_OUI );

  /* On rserve la table des noms */
  noms.cte = (const char **) malloc( sizeof( char * ) * systeme->nbr_groupes );
  if ( NULL == noms.cte ) {
    (void) fprintf(stderr, "%s,%sl.%u : Mmoire insuffisante pour la liste des noms\n",
		   __FILE__, __FUNCTION__, __LINE__);
    OBJ_choisit( arbre, objet, FORCE_NON );
    return 0;
  }

  /* On l'initialise */
  for ( i = 0 ; i < systeme->nbr_groupes ; i++ ) {
    noms.var[ i ] = TXT_genere_chaine("%u : %s", systeme->groupes[ i ].numero, systeme->groupes[ i ].symbole );
  }

  /* On appelle le sous-menu associ */
  choix = SS_menu_texte( noms.cte, systeme->nbr_groupes, "Groupe d'espace", arbre, objet );
  for (i = 0 ; i < systeme->nbr_groupes ; i++ ) {
    free( noms.var[ i ] ); noms.var[ i ] = NULL;
  }
  free( noms.var ); noms.var = NULL;
  if ( ( choix < 0 ) || ( choix == actuel ) ) { /* Slection annule, ou on garde le mme */
    OBJ_choisit( arbre, objet, FORCE_NON );
    return actuel;
  }

  /* Slection modifie : on installe le nouveau groupe slectionn */
  nom_groupe = TXT_genere_chaine("%u : %s", systeme->groupes[ choix ].numero, systeme->groupes[ choix ].symbole );
  met_chaine( arbre, objet, lgmax_texte, nom_groupe, TXT_COPIER );
  free( nom_groupe );

  /* Termin */
  OBJ_choisit( arbre, objet, FORCE_NON );
  return choix;
}

/* --------- Choisir le systme cristallin -------- */

int OBJ_ChoisirSystemeCr(unsigned int arbre, unsigned int objet, unsigned long lgmax_texte, int actuel, int mode,
			 struct systeme_cristallin *systemes, unsigned long nbr_systemes,
			 unsigned int obj_a, unsigned int obj_b, unsigned int obj_c, unsigned int obj_alpha, unsigned int obj_beta, unsigned int obj_gamma)
{
  const char **noms;
  unsigned long i;
  int choix;

  /* Pas de systme -> rien  faire, tant en demande qu'en installation ! */
  if ( ( NULL == systemes ) || ( 0 == nbr_systemes ) ) {
    OBJ_active( arbre, objet, FORCE_NON );
    met_chaine( arbre, objet, lgmax_texte, "Aucun syst"e_GRAVE"me trouv"e_AIGU, TXT_COPIER );
    return 0;
  }

  if ( METTRE == mode ) {
#ifdef _DEVELOPPE_
    if ( ( actuel < 0 ) || ( actuel >= (int) nbr_systemes ) ) {
      (void) fprintf(stderr, "%s,%sl.%u : Index de systme absurde [%d, doit tre entre 0 et %lu]\n",
		     __FILE__, __FUNCTION__, __LINE__, actuel, nbr_systemes - 1);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return 0;
    }
#endif

    InstallerSysteme( systemes + actuel, arbre, objet, lgmax_texte, obj_a, obj_b, obj_c, obj_alpha, obj_beta, obj_gamma );
    return actuel;
  }

  /* Si on est l : on demande -> bouton activ ! */
  OBJ_choisit( arbre, objet, FORCE_OUI );

  /* On rserve la table des noms */
  noms = (const char **) malloc( sizeof( char * ) * nbr_systemes );
  if ( NULL == noms ) {
    (void) fprintf(stderr, "%s,%sl.%u : Mmoire insuffisante pour la liste des noms\n",
		   __FILE__, __FUNCTION__, __LINE__);
    OBJ_choisit( arbre, objet, FORCE_NON );
    return 0;
  }

  /* On l'initialise */
  for ( i = 0 ; i < nbr_systemes ; i++ ) {
    noms[ i ] = systemes[ i ].nom;
  }

  /* On appelle le sous-menu associ */
  choix = SS_menu_texte( noms, nbr_systemes, "Syst"e_GRAVE"me cristallin", arbre, objet );
  free( noms ); noms = NULL;
  if ( ( choix < 0 ) || ( choix == actuel ) ) { /* Slection annule, ou on garde le mme */
    OBJ_choisit( arbre, objet, FORCE_NON );
    return actuel;
  }

  /* Slection modifie : on installe le nouveau systme slectionn
     L'installation des groupes d'espace est  la charge de l'appelant */
  InstallerSysteme( systemes + choix, arbre, objet, lgmax_texte, obj_a, obj_b, obj_c, obj_alpha, obj_beta, obj_gamma );

  /* Termin */
  OBJ_choisit( arbre, objet, FORCE_NON );
  return choix;
}

static void InstallerSysteme(struct systeme_cristallin *systeme, unsigned int arbre, unsigned int objet, unsigned long lgmax_texte,
			     unsigned int obj_a, unsigned int obj_b, unsigned int obj_c, unsigned int obj_alpha, unsigned int obj_beta, unsigned int obj_gamma)
{
  /* Le nom du systme */
  met_chaine( arbre, objet, lgmax_texte, systeme->nom, TXT_COPIER );

  /* Quels paramtres de maille sont pertinents ? */
  if ( 'C' == systeme->type ) { /* Cubique */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_NON ); /* a = b = c */
    OBJ_active( arbre, obj_c, FORCE_NON );

    /* Angles imposs  90 */
    OBJ_active( arbre, obj_alpha, FORCE_NON ); OBJ_MettreNombre( arbre, obj_alpha, 90.0, 0.0 );
    OBJ_active( arbre, obj_beta , FORCE_NON ); OBJ_MettreNombre( arbre, obj_beta , 90.0, 0.0 );
    OBJ_active( arbre, obj_gamma, FORCE_NON ); OBJ_MettreNombre( arbre, obj_gamma, 90.0, 0.0 );
  } else if ( ( 'H' == systeme->type ) || ( 'r' == systeme->type ) ) { /* Hexagonal ; rhombodrique avec axes hexagonaux */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_NON ); /* a = b */
    OBJ_active( arbre, obj_c, FORCE_OUI );

    /* Angles imposs  90, 90 et 120 */
    OBJ_active( arbre, obj_alpha, FORCE_NON ); OBJ_MettreNombre( arbre, obj_alpha, 90.0, 0.0 );
    OBJ_active( arbre, obj_beta , FORCE_NON ); OBJ_MettreNombre( arbre, obj_beta , 90.0, 0.0 );
    OBJ_active( arbre, obj_gamma, FORCE_NON ); OBJ_MettreNombre( arbre, obj_gamma, 120.0, 0.0 );
  } else if ( 'M' == systeme->type ) { /* Monoclinique (axe b) */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_OUI );
    OBJ_active( arbre, obj_c, FORCE_OUI );

    /* Angles imposs  90, quelconque (>90) et 90 */
    OBJ_active( arbre, obj_alpha, FORCE_NON ); OBJ_MettreNombre( arbre, obj_alpha, 90.0, 0.0 );
    OBJ_active( arbre, obj_beta , FORCE_OUI );
    OBJ_active( arbre, obj_gamma, FORCE_NON ); OBJ_MettreNombre( arbre, obj_gamma, 90.0, 0.0 );
  } else if ( 'm' == systeme->type ) { /* Monoclinique (axe c) */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_OUI );
    OBJ_active( arbre, obj_c, FORCE_OUI );

    /* Angles imposs  90, 90 et quelconque (>90) */
    OBJ_active( arbre, obj_alpha, FORCE_NON ); OBJ_MettreNombre( arbre, obj_alpha, 90.0, 0.0 );
    OBJ_active( arbre, obj_beta , FORCE_NON ); OBJ_MettreNombre( arbre, obj_beta , 90.0, 0.0 );
    OBJ_active( arbre, obj_gamma, FORCE_OUI );
  } else if ( 'O' == systeme->type ) { /* Orthorhombique */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_OUI );
    OBJ_active( arbre, obj_c, FORCE_OUI );

    /* Angles imposs  90, 90 et 90 */
    OBJ_active( arbre, obj_alpha, FORCE_NON ); OBJ_MettreNombre( arbre, obj_alpha, 90.0, 0.0 );
    OBJ_active( arbre, obj_beta , FORCE_NON ); OBJ_MettreNombre( arbre, obj_beta , 90.0, 0.0 );
    OBJ_active( arbre, obj_gamma, FORCE_NON ); OBJ_MettreNombre( arbre, obj_gamma, 90.0, 0.0 );
  } else if ( 'Q' == systeme->type ) { /* Quadratique */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_NON ); /* a = b */
    OBJ_active( arbre, obj_c, FORCE_OUI );

    /* Angles imposs  90, 90 et 90 */
    OBJ_active( arbre, obj_alpha, FORCE_NON ); OBJ_MettreNombre( arbre, obj_alpha, 90.0, 0.0 );
    OBJ_active( arbre, obj_beta , FORCE_NON ); OBJ_MettreNombre( arbre, obj_beta , 90.0, 0.0 );
    OBJ_active( arbre, obj_gamma, FORCE_NON ); OBJ_MettreNombre( arbre, obj_gamma, 90.0, 0.0 );
  } else if ( 'R' == systeme->type ) { /* Rhombodrique */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_NON ); /* a = b = c */
    OBJ_active( arbre, obj_c, FORCE_NON );

    OBJ_active( arbre, obj_alpha, FORCE_OUI );
    OBJ_active( arbre, obj_beta , FORCE_NON ); /* alpha = beta = gamma */
    OBJ_active( arbre, obj_gamma, FORCE_NON );
#ifdef _DEVELOPPE_
  } else if ( 'T' != systeme->type ) { /* Autre qu'un type connu */
    AVERTISSEMENT( "Groupe de type inconnu [%d]", (int) systeme->type );
#endif
  } else { /* T = triclinique & autres systmes inconnus */
    OBJ_active( arbre, obj_a, FORCE_OUI );
    OBJ_active( arbre, obj_b, FORCE_OUI );
    OBJ_active( arbre, obj_c, FORCE_OUI );

    /* Angles quelconques */
    OBJ_active( arbre, obj_alpha, FORCE_OUI );
    OBJ_active( arbre, obj_beta , FORCE_OUI );
    OBJ_active( arbre, obj_gamma, FORCE_OUI );
  }
}

/* -------- Lire un angle (en radians, affich en dgrs) dans un champ ditable -------- */

double OBJ_LireAngle(unsigned int arbre, unsigned int objet, double *erreur)
{
  double angle, err_angle;

  /* On lit la valeur */
  angle = OBJ_LireNombre( arbre, objet, &err_angle );

  /* On la passe entre -360 et 360 */
  while ( angle < -360.0 ) {
    angle += 360.0;
  }
  while ( angle > 360.0 ) {
    angle -= 360.0;
  }

  /* On convertit en radians */
  angle *= M_PI / 180.0;
  err_angle *= M_PI / 180.0;

  /* On renvoit les rsultats */
  if ( erreur != NULL ) *erreur = err_angle;
  return angle;
}

/* -------- Mettre un angle (en radians, affich en degrs) dans un champ ditable -------- */

int OBJ_MettreAngle(unsigned int arbre, unsigned int objet, double angle, double erreur)
{
  int retour;

  /* On convertit l'angle (et son erreur aussi !) en degrs */
  angle  *= 180.0 / M_PI;
  erreur *= 180.0 / M_PI;

  /* On le place dans l'intervalle [-360, 360] */
  while ( angle < -360.0 ) {
    angle += 360.0;
  }
  while ( angle > 360.0 ) {
    angle -= 360.0;
  }

  /* On s'assure que l'on a la bonne unit affiche */
  retour = OBJ_OptionsEditable( arbre, objet,
				BI_EDITABLE_MIN|BI_EDITABLE_MAX|BI_EDITABLE_ANGLE|BI_EDITABLE_UNITE,
				-360.0, 360.0, DEGRE );
#ifdef _DEVELOPPE_
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : ERREUR en prparant l'objet ditable\n",
		    __FILE__, __FUNCTION__, __LINE__ );
  }
#endif

  /* On installe la valeur corrige dans l'objet ditable */
  retour = OBJ_MettreNombre( arbre, objet, angle, erreur );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : ERREUR en installant la valeur numrique\n",
		    __FILE__, __FUNCTION__, __LINE__ );
  }

  return retour;
}

/* -------- Lire les coordonnes d'un point dans l'espace [le plan] -------- */

int OBJ_LireCoordonnees(unsigned int arbre, unsigned int obj_x, unsigned int obj_y, unsigned int obj_z,
			struct vect_3d *point, struct err_coord_3d *erreurs)
{
  double val_erreur;

  if ( NULL == point ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de vecteur pour mmoriser le rsultat\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -5;
  }

  point->x = OBJ_LireNombre( arbre, obj_x, &val_erreur );
  if ( erreurs != NULL ) erreurs->var_x = val_erreur * val_erreur; /* On a lu l'cart-type, mais on mmorise la variance ! */

  point->y = OBJ_LireNombre( arbre, obj_y, &val_erreur );
  if ( erreurs != NULL ) erreurs->var_y = val_erreur * val_erreur; /* On a lu l'cart-type, mais on mmorise la variance ! */

  point->z = OBJ_LireNombre( arbre, obj_z, &val_erreur );
  if ( erreurs != NULL ) erreurs->var_z = val_erreur * val_erreur; /* On a lu l'cart-type, mais on mmorise la variance ! */
  /* Remarque : on n'a rien lu sur les covariances : que faire ? */

  return 0;
}

/* -------- Mettre les coordonnes d'un point dans l'espace [le plan] -------- */

int OBJ_MettreCoordonnees(unsigned int arbre, unsigned int obj_x, unsigned int obj_y, unsigned int obj_z,
			  struct vect_3d *point, struct err_coord_3d *erreurs)
{
  if ( NULL == point ) { /* NULL -> vecteur (0 0 0) */
    OBJ_MettreNombre( arbre, obj_x, 0.0, 0.0 );
    OBJ_MettreNombre( arbre, obj_y, 0.0, 0.0 );
    OBJ_MettreNombre( arbre, obj_z, 0.0, 0.0 );
    return 0;
  }

  /* Si une erreur est prsente, c'est la variance qui est mmorise, mais on affiche l'cart-type */
  /* Actuellement, rien pour afficher les covariances */
  OBJ_MettreNombre( arbre, obj_x, point->x, ( erreurs != NULL ) ? ( sqrt( erreurs->var_x ) ) : ( 0.0 ) );
  OBJ_MettreNombre( arbre, obj_y, point->y, ( erreurs != NULL ) ? ( sqrt( erreurs->var_y ) ) : ( 0.0 ) );
  OBJ_MettreNombre( arbre, obj_z, point->z, ( erreurs != NULL ) ? ( sqrt( erreurs->var_z ) ) : ( 0.0 ) );

  return 0;
}

/* -------- Lire les paramtres pour un intervalle de calcul -------- */

int OBJ_LireIntervalle(unsigned int arbre, unsigned int obj_min, unsigned int obj_max, unsigned int obj_pas,
		       unsigned int obj_nbr_points, double *min, double *max, double *pas, unsigned long *nbr_points)
{
  double val_min, val_max, val_pas;
  unsigned long nombre_points;

  /* On lit les valeurs ; si objet manquant [ROOT], contient 0 */
  val_min = OBJ_LireNombre( arbre, obj_min, NULL );
  val_max = OBJ_LireNombre( arbre, obj_max, NULL );
  val_pas = OBJ_LireNombre( arbre, obj_pas, NULL );
  nombre_points = OBJ_LireEntier_U( arbre, obj_nbr_points );

  /* On complte pour la valeur manquante */
  if ( ROOT == obj_min ) { /* On ne connat pas k_min */
    val_min = val_max - val_pas * (double) nombre_points;
  } else if ( ROOT == obj_max ) { /* On ne connat pas k_max */
    val_max = val_min + val_pas * (double) nombre_points;
  } else if ( ROOT == obj_pas ) { /* On ne connat pas le pas */
    if ( ROOT == obj_nbr_points ) { /* Ni pas, ni nombre de points : on veut juste un intervalle */
      if ( val_min > val_max ) {
	double x;

	x = val_min; val_min = val_max; val_max = x;
      }
      if ( min != NULL ) *min = val_min;
      if ( max != NULL ) *max = val_max;
      return 0;
    }
    if ( 0 == nombre_points ) {
      (void) fprintf(stderr, "%s,%s l.%u : Nombre de points absurde\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
    val_pas = ( val_max - val_min ) / (double) (nombre_points);
  } else if ( ROOT == obj_nbr_points ) {
    double v_max;

    if ( val_pas == 0.0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Valeur du pas absurde [%g]\n",
		     __FILE__, __FUNCTION__, __LINE__, val_pas);
      return -2;
    }
    nombre_points = 1 + (unsigned long) fabs( (val_max - val_min) / val_pas );

    /* On contrle que l'on inclut bien le dernier point, mme si le pas n'est pas exact */
    v_max = val_min + val_pas * (double) nombre_points;
    if ( v_max < val_max )
      nombre_points++;
  }

  /* On contrle la cohrence de tout cela */
  if ( ( 0.0 == val_pas ) || ( fabs( val_pas ) > fabs( val_max - val_min ) ) ) { /* Pas nul ou suprieur  l'intervalle */
    /* Rq : inclus le cas des deux bornes confondues ! */
    (void) fprintf(stderr, "%s,%s l.%u : Pas nul, absurde\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }
  if( 0 == nombre_points ) {
    (void) fprintf(stderr, "%s,%s l.%u : Nombre de points absurde [0]\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }
  if ( ( val_max < val_min ) && ( val_pas > 0.0 ) ) {
    double i;
    i = val_max ; val_max = val_min ; val_min = i;
  } else if ( ( val_min < val_max ) && ( val_pas < 0.0 ) ) {
    val_pas = -val_pas;
  }

  /* On renvoit les valeurs */
  if ( min != NULL ) *min = val_min;
  if ( max != NULL ) *max = val_max;
  if ( pas != NULL ) *pas = val_pas;
  if ( nbr_points != NULL ) *nbr_points = nombre_points;
  return 0;
}

/* -------- Mettre les paramtres pour un intervalle de calcul -------- */

int OBJ_MettreIntervalle(unsigned int arbre, unsigned int obj_min, unsigned int obj_max, unsigned int obj_pas,
		       unsigned int obj_nbr_points, double min, double max, double pas, unsigned long nbr_points)
{
  int retour = 0;

  if ( obj_min > ROOT ) {
    retour = OBJ_MettreNombre( arbre, obj_min, min, 0.0 );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR survenue dans OBJ_MettreNombre(%u, %u, %g, 0) [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, arbre, obj_min, min, retour);
    }
  }
  if ( obj_max > ROOT ) {
    retour = OBJ_MettreNombre( arbre, obj_max, max, 0.0 );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR survenue dans OBJ_MettreNombre(%u, %u, %g, 0) [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, arbre, obj_max, max, retour);
    }
  }
  if ( obj_pas > ROOT ) {
    retour = OBJ_MettreNombre( arbre, obj_pas, pas, 0.0 );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR survenue dans OBJ_MettreNombre(%u, %u, %g, 0) [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, arbre, obj_pas, pas, retour);
    }
  }
  if ( obj_nbr_points > ROOT ) {
    retour = OBJ_MettreEntier_U( arbre, obj_nbr_points, nbr_points );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR survenue dans OBJ_MettreEntier_U(%u, %u, %lu) [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, arbre, obj_nbr_points, nbr_points, retour);
    }
  }

  /* On renvoie ce que l'on a pu faire... */
  return retour;
}
