/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2003, Emmanuel CURIS

   Généralités du programme LASE (interface graphique)

   Gestion et préparation du menu des calculs & modèles théoriques
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "systeme.h"

#include "exafs.h"

#include "absorption.h"
#include "chimie_utils.h"	/* Pour les noms des elements... */
#include "graphes.h"		/* Pour GRF_aller() */
#include "interface.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_RAIE       15
#define LGMAX_NOM_SEUIL      13
#define LGMAX_NOM_MODELE     30
#define LGMAX_NOM_GRAPHE_MU0 30
#define LGMAX_NOM_GRAPHE_KHI 30

static int seuil_affiche = SEUIL_K; /* Mémorisation d'un appel à l'autre */
static int modele_mu0_affiche = MODELE_MU0_MCMASTER;

struct infos_arbre {
  Graphe *mu0;
  Graphe *signal_exafs;
};

#include "int_th_construire_mu0.txt" /* Les textes [traductions] */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void I_TH_XAS_ConstruireMu0_PreparerArbre(void);
static int  I_TH_XAS_ConstruireMu0_GererArbre   (unsigned int objet, int x, int y);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

struct graphe *I_TH_XAS_ConstruireMu0(void) {
  boolean coherent;
  struct infos_arbre infos;
  struct infos_construction_mu0 parametres;
  double e_min, e_max, pas;
  Graphe *mu0_total = NULL, *mu0 = NULL;
  char *titre;
  int retour;

  /* Initialisation de l'arbre */
  if ( FAUX == arbres[ SIMULER_MU ].drapeaux.prepare ) {
    I_TH_XAS_ConstruireMu0_PreparerArbre();
  }
  infos.mu0 = NULL;
  infos.signal_exafs = NULL;

  ARB_MettreComplement( SIMULER_MU, &infos );

  /* Préparation... */
  OBJ_ChoisirSeuil( SIMULER_MU, SM_SEUIL_ELEMENT, LGMAX_NOM_SEUIL, seuil_affiche, METTRE );

  OBJ_ChoisirModeleMu0( SIMULER_MU, SM_MODELE_MU0, LGMAX_NOM_MODELE, modele_mu0_affiche, METTRE );
  (void) I_TH_XAS_ConstruireMu0_GererArbre( SM_MU0_THEORIQUE, 0, 0 );

  OBJ_choisit( SIMULER_MU, SM_AV_EXAFS, FORCE_NON );
  OBJ_active( SIMULER_MU, SM_EXAFS, FORCE_NON );
  met_chaine( SIMULER_MU, SM_EXAFS, LGMAX_NOM_GRAPHE_KHI, ERR_PasDeSignalEXAFS, TXT_COPIER );
  OBJ_active( SIMULER_MU, SM_PONDERATION, FORCE_NON );

  /* Paramétrage par l'utilisateur */
  do {
    coherent = VRAI;		/* On part du principe que tout ira bien... */

    /* Appel de l'arbre */
    retour = ARB_Bloquant( SIMULER_MU, titre_boite_construire_mu, principale, 0 );
    if ( SM_ANNULER == retour ) {
      /* On enleve les infos de l'arbre... */
      ARB_MettreComplement( SIMULER_MU, NULL );
      return NULL;
    }

    /* Intervalle de construction */
    retour = OBJ_LireIntervalle( SIMULER_MU, SM_E_DEBUT, SM_E_FIN, SM_PAS, 0,
				 &e_min, &e_max, &pas, NULL );
    if ( retour != 0 ) {
      AVERTISSEMENT( "Parametrage de l'intervalle absurde : recommencez [%g, %g, %g]",
		     e_min, e_max, pas );
      coherent = FAUX;
    }

    /* Lecture des parametres dans l'arbre */
    parametres.options = 0;
    parametres.Z = OBJ_NumeroAtomique( SIMULER_MU, SM_ELEMENT );

    /* Oscillations EXAFS eventuelles */
    if ( VRAI == OBJ_choisit( SIMULER_MU, SM_AV_EXAFS, QUESTION ) ) {
      if ( NULL == infos.signal_exafs ) {
	(void) BI_ERR_alerte( NULL, &ERR_PasDeSpectreEXAFS, VRAI );
	coherent = FAUX;
      } else {
	parametres.options |= TH_CALCUL_MU0_AVEC_EXAFS;
      }
      parametres.exafs.signal_exafs = infos.signal_exafs;
      parametres.exafs.ponderation  = OBJ_LireEntier_S( SIMULER_MU, SM_PONDERATION );
    } else {
      parametres.exafs.signal_exafs = NULL;
      parametres.exafs.ponderation  = 0;
    }

    /* Modele du mu0 apres seuil eventuel */
    if ( VRAI == OBJ_choisit( SIMULER_MU, SM_MU0, QUESTION ) ) {
      parametres.options |= TH_CALCUL_MU0_AVEC_MU0;

      if ( VRAI == OBJ_choisit( SIMULER_MU, SM_MU0_THEORIQUE, QUESTION ) ) {
	parametres.mu0.modele_theorique = modele_mu0_affiche;
      } else {
	/* Utilisation d'un graphe experimental */
	parametres.options |= TH_CALCUL_MU0_MU0_EXP;
	parametres.mu0.mu0_exp = infos.mu0;

	/* Si on donne EXAFS _et_ mu0, on peut supposer qu'ils ont la même pondération... */
	if ( parametres.options & TH_CALCUL_MU0_AVEC_EXAFS ) {
	  parametres.mu0.ponderation = parametres.exafs.ponderation;
	} else {
	  parametres.mu0.ponderation = 0;
	  OPTION_A_FAIRE( "Demander la ponderation pour mu0..." );
	}
      }
    }

    /* Modele du seuil */
    parametres.seuil.type    = seuil_affiche;
    parametres.seuil.energie = OBJ_LireNombre( SIMULER_MU, SM_E0, NULL );
    parametres.seuil.hauteur = OBJ_LireNombre( SIMULER_MU, SM_HAUTEUR_SEUIL, NULL );
    parametres.seuil.methode = SEUIL_MODELE_ESCALIER;
    if ( VRAI == OBJ_choisit( SIMULER_MU, SM_SEUIL, QUESTION ) ) {
      parametres.options |= TH_CALCUL_MU0_AVEC_SEUIL;
      /* Comment modéliser le seuil [mu(seuil)] ? */
      if ( VRAI == OBJ_choisit( SIMULER_MU, SM_SEUIL_LORENTZ, QUESTION ) ) {
	parametres.seuil.methode = SEUIL_MODELE_LORENTZ;
      } else if ( VRAI == OBJ_choisit( SIMULER_MU, SM_SEUIL_GAUSS, QUESTION ) ) {
	parametres.seuil.methode = SEUIL_MODELE_GAUSS;
      }
      parametres.seuil.largeur = OBJ_LireNombre( SIMULER_MU, SM_LARGEUR_SEUIL, NULL );
    }

    /* Modele du preseuil */
    if ( VRAI == OBJ_choisit( SIMULER_MU, SM_PRESEUIL, QUESTION ) ) {
      parametres.options |= TH_CALCUL_MU0_AVEC_RAIES;
    }
  } while ( coherent != VRAI );

  /* Plus besoin de l'arbre et de ses infos */
  ARB_MettreComplement( SIMULER_MU, NULL );

  /* On construit le spectre */
  mu0_total = TH_ConstruireMu0( e_min, e_max, pas, &parametres );
  if ( NULL == mu0_total ) {
    (void) BI_ERR_alerte( NULL, &ERR_CreationDuGraphe, VRAI );
    return NULL;
  }

  /* Insertion dans la liste des graphes accessibles a l'utilisateur :
     Par defaut, dans la serie courante, apres le graphe courant */
  I_GRF_InsererDansSerie( NULL, NULL, mu0_total );

  /* Ameliorations pour l'affichage */
  titre = TXT_genere_chaine( "Section efficace de photo-absorption : %s, seuil %s",
			     ATM_Z_vers_symbole( parametres.Z ), TH_TexteSeuil( seuil_affiche ) );
  GRF_titre( mu0_total, titre );
  free( titre ); titre = NULL;

  /* Termine */
  return mu0;
}

/* -------- Gestion de la boite de dialogue -------- */

static int I_TH_XAS_ConstruireMu0_GererArbre(unsigned int objet, int x, int y)
{
  int etat, Z;
  struct infos_arbre *infos;
  Graphe *choix;

  infos = (struct infos_arbre *) ARB_LireComplement( SIMULER_MU );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'infos !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }

  if ( ( objet >= SM_NOM_RAIE_1 ) && ( objet <= SM_NOM_RAIE_N ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Affichage de la raie choisie\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  switch ( objet )
  {
    /* Raies du préseuil */
  case SM_RAIE_LORENTZ:
  case SM_RAIE_GAUSS:
    OBJ_choisit( SIMULER_MU, objet, FORCE_OUI );
    return BI_EVENEMENT_TRAITE;

    /* Seuil proprement dit */
  case SM_SEUIL:
    etat = OBJ_choisit( SIMULER_MU, SM_SEUIL, ECHANGE );
    OBJ_active( SIMULER_MU, SM_RAIE_LORENTZ, etat );
    OBJ_active( SIMULER_MU, SM_RAIE_GAUSS, etat );
    OBJ_active( SIMULER_MU, SM_E0, etat );
    OBJ_active( SIMULER_MU, SM_HAUTEUR_SEUIL, etat );
    OBJ_active( SIMULER_MU, SM_LARGEUR_SEUIL, etat );
    return BI_EVENEMENT_TRAITE;

  case SM_SEUIL_LORENTZ:
  case SM_SEUIL_GAUSS:
    OBJ_choisit( SIMULER_MU, objet, FORCE_OUI );
    return BI_EVENEMENT_TRAITE;

  case SM_SEUIL_ELEMENT:
    seuil_affiche = OBJ_ChoisirSeuil( SIMULER_MU, SM_SEUIL_ELEMENT, LGMAX_NOM_SEUIL, seuil_affiche, QUESTION );
    return BI_EVENEMENT_TRAITE;

  case SM_PROPOSER:
    Z = OBJ_NumeroAtomique( SIMULER_MU, SM_ELEMENT );
    if ( Z >= HYDROGENE ) {
      double energie, hauteur = 1.0, largeur = 2.0;

      energie = (double) TH_DonneesSeuil( Z, seuil_affiche, &hauteur, &largeur );
      if ( hauteur <= 0.0 ) { /* On utilise les valeurs de McMaster */
	double A_apres[ 4 ], A_avant[ 4 ];

	TH_coef_McMaster( Z, seuil_affiche, A_avant, A_avant + 1, A_avant + 2, A_avant + 3 );
	TH_coef_McMaster( Z, seuil_affiche - 1, A_apres, A_apres + 1, A_apres + 2, A_apres + 3 );

	hauteur  = TH_McMaster( energie, Z, A_apres[ 0 ], A_apres[ 1 ], A_apres[ 2 ], A_apres[ 3 ] );
	hauteur -= TH_McMaster( energie, Z, A_avant[ 0 ], A_avant[ 1 ], A_avant[ 2 ], A_avant[ 3 ] );
      }
      if ( largeur < 0.0 ) largeur = 0.0;

      OBJ_MettreNombre( SIMULER_MU, SM_E0, energie, 0.0 );
      OBJ_MettreNombre( SIMULER_MU, SM_HAUTEUR_SEUIL, hauteur, 0.0 );
      OBJ_MettreNombre( SIMULER_MU, SM_LARGEUR_SEUIL, largeur, 0.0 );

      OBJ_MettreNombre( SIMULER_MU, SM_E_DEBUT, rint( energie - 100.0 ), 0.0 ); /* Différence rint()/round() ? */
      OBJ_MettreNombre( SIMULER_MU, SM_E_FIN  , rint( energie + 1000.0 ), 0.0 );
      OBJ_MettreNombre( SIMULER_MU, SM_PAS, 2.0, 0.0 );
    }
    return BI_EVENEMENT_TRAITE;

    /* Absorption intrinsèque après seuil */
  case SM_MU0:
    etat = OBJ_choisit( SIMULER_MU, SM_MU0, ECHANGE );
    if ( FAUX == etat ) {
      OBJ_active( SIMULER_MU, SM_MU0_THEORIQUE, FORCE_NON );
      OBJ_active( SIMULER_MU, SM_MU0_EXP, FORCE_NON );
      OBJ_active( SIMULER_MU, SM_MODELE_MU0, FORCE_NON );
      OBJ_active( SIMULER_MU, SM_GRF_MU0, FORCE_NON );
    } else {
      OBJ_active( SIMULER_MU, SM_MU0_THEORIQUE, FORCE_OUI );
      OBJ_active( SIMULER_MU, SM_MU0_EXP, FORCE_OUI );
      if ( VRAI == OBJ_choisit( SIMULER_MU, SM_MU0_THEORIQUE, QUESTION ) ) {
	OBJ_active( SIMULER_MU, SM_MODELE_MU0, FORCE_OUI );
	OBJ_active( SIMULER_MU, SM_GRF_MU0, FORCE_NON );
      } else {
	OBJ_active( SIMULER_MU, SM_MODELE_MU0, FORCE_NON );
	OBJ_active( SIMULER_MU, SM_GRF_MU0, FORCE_OUI );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case SM_MU0_THEORIQUE:
    OBJ_choisit( SIMULER_MU, objet, FORCE_OUI );
    OBJ_active( SIMULER_MU, SM_MODELE_MU0, FORCE_OUI );
    OBJ_active( SIMULER_MU, SM_GRF_MU0, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case SM_MU0_EXP:
    OBJ_choisit( SIMULER_MU, objet, FORCE_OUI );
    OBJ_active( SIMULER_MU, SM_MODELE_MU0, FORCE_NON );
    OBJ_active( SIMULER_MU, SM_GRF_MU0, FORCE_OUI );
    return BI_EVENEMENT_TRAITE;

  case SM_MODELE_MU0:
    modele_mu0_affiche = OBJ_ChoisirModeleMu0( SIMULER_MU, SM_MODELE_MU0, LGMAX_NOM_MODELE, modele_mu0_affiche, QUESTION );
    return BI_EVENEMENT_TRAITE;

  case SM_GRF_MU0:
    choix = GRF_choisir( GR_MU0, FAUX, VRAI, NULL );
    if ( choix != NULL ) {
      infos->mu0 = choix;
      met_chaine( SIMULER_MU, SM_EXAFS, LGMAX_NOM_GRAPHE_KHI, choix->nom, TXT_COPIER );
    }
    return BI_EVENEMENT_TRAITE;

    /* Oscillations EXAFS */
  case SM_AV_EXAFS:
    etat = OBJ_choisit( SIMULER_MU, objet, ECHANGE );
    OBJ_active( SIMULER_MU, SM_EXAFS, etat );
    OBJ_active( SIMULER_MU, SM_PONDERATION, etat );
    return BI_EVENEMENT_TRAITE;

  case SM_EXAFS:
    choix = GRF_choisir( GR_EXAFS, FAUX, VRAI, NULL );
    if ( choix != NULL ) {
      infos->signal_exafs = choix;
      met_chaine( SIMULER_MU, SM_EXAFS, LGMAX_NOM_GRAPHE_KHI, choix->nom, TXT_COPIER );
    }
    return BI_EVENEMENT_TRAITE;
  }

#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "%s,%s l.%u : Objet %u non géré [X=%d, Y=%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, objet, x, y);
#endif
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Preparation de l'arbre/boite de dialogue -------- */

static void I_TH_XAS_ConstruireMu0_PreparerArbre(void)
{
  int retour;

  /* Initialisation */
  retour = ARB_Initialiser( SIMULER_MU, NULL, I_TH_XAS_ConstruireMu0_GererArbre, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en initialisant l'arbre [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* Enjolivation */
  OBJ_deplace( SIMULER_MU, SM_TXT_PAS, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_PAS    , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( SIMULER_MU, SM_AJOUTER_RAIE, 0, -1, DEPLACE_PIXELS );

  ARB_EnjoliverDefileur( SIMULER_MU, SM_FND_RAIES, SM_NOM_RAIE_1, SM_NOM_RAIE_N,
			 SM_RAIE_AVANT, SM_FND_ASC, SM_BOUTON_ASC, SM_RAIE_APRES, ARB_ASCENSEUR_DROITE );

  OBJ_deplace( SIMULER_MU, SM_TXT_POSITION, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_POSITION    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_TXT_LARGEUR , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_LARGEUR     , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( SIMULER_MU, SM_TXT_HAUTEUR  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_HAUTEUR_SEUIL, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_TXT_LG_SEUIL , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_LARGEUR_SEUIL, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( SIMULER_MU, SM_TXT_MODELE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_MODELE_MU0, 0, +1, DEPLACE_PIXELS );
  met_chaine ( SIMULER_MU, SM_GRF_MU0, LGMAX_NOM_GRAPHE_MU0, "Pas de coefficient d'absorption", TXT_COPIER );
  OBJ_deplace( SIMULER_MU, SM_TXT_GRF_MU0, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_GRF_MU0, 0, +2, DEPLACE_PIXELS );

  met_chaine ( SIMULER_MU, SM_EXAFS, LGMAX_NOM_GRAPHE_KHI, "Pas de signal EXAFS", TXT_COPIER );
  OBJ_deplace( SIMULER_MU, SM_TXT_PONDERAT, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( SIMULER_MU, SM_PONDERATION , 0, +1, DEPLACE_PIXELS );

  arbres[ SIMULER_MU ].drapeaux.prepare = VRAI;
}
