/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1996 - 2004, Emmanuel CURIS
   Cration de ce fichier : 5 aout 2004

   Interface pour le tri, la gestion des doublons d'un graphe
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"

#include "interface.h"
#include "graphes.h"
#include "tri.h"

#include "interface_tri.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_TriParBlocs(Graphe *graphe, unsigned int nbr_blocs);

static int  TRI_GererArbre   (unsigned int num_objet, int x, int y);
static void TRI_PreparerArbre(int genre);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~- Gestion de l'interface utilisateur pour le tri des graphes ~-~-~-~-~-~- */

/* -------- Appeler de l'extrieur --------
    Renvoie 0    si on a pu trier le graphe
            != 0 en cas de problme (<0 si problme grave)
 */

int I_GRF_Trier(Graphe *graphe)
{
  int retour;
  unsigned long nombre_doublons = 0;
  boolean moyenner = VRAI;

  /* Controles sur le graphe indiqu en paramtres */
  if ( NULL == graphe ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de graphe indiqu !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( ( FAUX == GRF_EstReel( graphe ) ) && ( FAUX == GRF_EstComplexe( graphe ) ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Genre de graphe non triable [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->genre_donnees );
#endif
    (void) BI_ERR_alerte( principale, &ERR_GenreIncorrect, VRAI );
    return +1;			/* Pas bien grave, juste rien  faire... */
  } else if ( ( GRF_SERIE == graphe->genre_donnees ) && ( graphe->x != NULL ) ) {
    OPTION_A_FAIRE( "Controler que les X ne sont pas en commun pour toute la srie" );
    (void) BI_ERR_alerte( principale, &ERR_SerieAvecX, VRAI );
    return +2;
  }

  /* On prpare l'arbre */
  TRI_PreparerArbre( graphe->genre_donnees );

  /* On appelle l'arbre */
  retour = ARB_Bloquant( ORDONNE_GRAPHE, titre_arbre_tri, principale, 0 );
  if ( TRI_ANNULE == retour ) {
    return +2; /* On annule : erreur mineure */
  }

  /* On fait ce qu'il faut faire */

  /* ~~~~~~~~~~~~~~~ Tri par X croissants ~~~~~~~~~~~~~~~~~~~~~~~~~~ */

  if ( OBJ_choisit( ORDONNE_GRAPHE, TRI_TRI_OK, QUESTION ) == VRAI ) { /* On veut trier */
    if ( VRAI == OBJ_choisit( ORDONNE_GRAPHE, TRI_ALEATOIRE, QUESTION ) ) { /* Tri complet */
      /* Graphe compltement dsordonn */
      retour = GRF_Trier( graphe, GRF_TRI_RAPIDE ); /* On essaye un tri rapide par dfaut ! */
    } else {
      /* Graphe tri par blocs, supposs de meme taille */
      retour = GRF_TriParBlocs( graphe, OBJ_LireEntier_U( ORDONNE_GRAPHE, TRI_NBR_BLOCS ) );
    }

    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le tri [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return -5;
    }
  }

  /* ~~~~~~~~~~~~~~~ limination des doublons [on peut en faire plusieurs successivement ! ~~~~~~~~~~~~~~~~~~~~~~ */

  moyenner = OBJ_choisit( ORDONNE_GRAPHE, TRI_MOY_DOUBLON, QUESTION );

  /* Cas des regroupements par paquet de n X */
  if ( OBJ_choisit( ORDONNE_GRAPHE, TRI_GROUPER_OK, QUESTION ) == VRAI ) {
    unsigned int nbr_x;

    nbr_x = OBJ_LireEntier_U( ORDONNE_GRAPHE, TRI_GROUPE_PARAM );
    nombre_doublons = GRF_TraiterDoublons( graphe, GRF_DOUBLONS_GROUPE, (double) nbr_x, moyenner, VRAI );
  }

  /* Cas des doublons dfinis par Delta_x < qq chose */
  if ( OBJ_choisit( ORDONNE_GRAPHE, TRI_DOUBLES_OK, QUESTION ) ) {
    double difference;

    difference = OBJ_LireNombre( ORDONNE_GRAPHE, TRI_PARAM_DOUBLE, NULL );

    nombre_doublons = GRF_TraiterDoublons( graphe, GRF_DOUBLONS_SOUPLE, difference, moyenner, VRAI );
    if ( nombre_doublons > 0 ) { /* Des doublons => les bornes ne sont plus bonnes */
      (void) BI_ERR_alerte( NULL, &MSG_PresqueDoublons, VRAI, nombre_doublons, difference );
    }
  }

  /* Cas des doublons simples */
  if ( OBJ_choisit(ORDONNE_GRAPHE, TRI_DOUBLON_OK, QUESTION) == VRAI) {
    nombre_doublons = GRF_TraiterDoublons( graphe, GRF_DOUBLONS_SIMPLE, 0.0, moyenner, VRAI );
    if ( nombre_doublons > 0 ) { /* Des doublons => les bornes ne sont plus bonnes */
      (void) BI_ERR_alerte( NULL, &MSG_Doublons, VRAI, nombre_doublons );

      GRF_bornes(graphe, DESORDONNE, DESORDONNE); /* Bornes ont pu changer... */
    }
  }

  if ( 0 == nombre_doublons ) {
    (void) BI_ERR_alerte( NULL, &MSG_AucunDoublon, VRAI);
  } else {
    /* On doit peut-tre mettre  jour l'affichage : il y a eu des changements */
    GRF_MiseAJour( graphe, VRAI );
  }

  return VRAI; /* Fini */
}

/* -------- Tri par bloc d'un graphe : prparation des blocs et tout a -------- */

static int GRF_TriParBlocs(Graphe *graphe, unsigned int nbr_blocs)
{
  TRI_infos *blocs, resultat = { NULL, NULL, 0, 0, 0, NULL, NULL };
  double *tmp;
  unsigned int i;
  unsigned long nbr_pts; /* Nombre de points par bloc */

  /* Controle de la faisabilit */
  if ( nbr_blocs < 2 ) {
    (void) fprintf(stderr, "%s,%s l.%u : %u blocs, rien  faire\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_blocs);
    return +1;
  }

  if ( graphe->nombre_point % nbr_blocs ) {
    /* % = reste de la division; non-nul => pas assez de points pour avoir tous ces blocs */
    (void) BI_ERR_alerte(NULL, &ERR_NombreBlocs, VRAI, nbr_blocs, graphe->nombre_point);
    return -1;
  }

  /* Combien de points par bloc [la division tombe forcment juste !] */
  nbr_pts = graphe->nombre_point / nbr_blocs;

  /* On rserve pour le rsultat */
  resultat.x = (double *) calloc( graphe->nombre_point, sizeof(double) );
  if ( resultat.x == NULL ) {
    (void) SYS_ERREUR( MANQUE_MEMOIRE, __FILE__, __FUNCTION__, "resultat.x");
    return -2;
  }

  resultat.y = (double *) calloc( graphe->nombre_point, sizeof(double) );
  if ( resultat.y == NULL ) {
    (void) SYS_ERREUR( MANQUE_MEMOIRE, __FILE__, __FUNCTION__, "resultat.y");
    free( resultat.x ); resultat.x = NULL;
    return -2;
  }

  resultat.nbr_points = graphe->nombre_point;

  /* On rserve les blocs */
  blocs = (TRI_infos *) calloc( nbr_blocs, sizeof(TRI_infos) );
  if ( NULL == blocs ) {
    (void) SYS_ERREUR( MANQUE_MEMOIRE, __FILE__, __FUNCTION__, "blocs");
    free( resultat.x ); resultat.x = NULL;
    free( resultat.y ); resultat.y = NULL;
    return -3;
  }

  /* Prparation des blocs */
  for ( i = 0 ; i < nbr_blocs ; i++ ) {
    blocs[ i ].x = graphe->x + i * nbr_pts;
    blocs[ i ].y = graphe->y + i * nbr_pts;
    blocs[ i ].position = 0;
    blocs[ i ].nbr_points = nbr_pts;
  }

  /* Fusion des blocs */
  TRI_Blocs( &resultat, blocs, nbr_blocs);

  /* On libre ce que l'on a utilis    */
  free( blocs ); blocs = NULL;

  /* On assigne les nouvelles valeurs au graphe
     (avec cette version, aucun risque de  race condition en multi-thread) */
  tmp = graphe->x; graphe->x = resultat.x; free( tmp ); tmp= NULL;
  tmp = graphe->y; graphe->y = resultat.y; free( tmp ); tmp= NULL;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Gestion effective de l'arbre pour paramtrer le tri (et les doublons) -------- */

static int TRI_GererArbre(unsigned int num_objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  switch ( num_objet )
  {
  case TRI_ALEATOIRE:
    OBJ_active(ORDONNE_GRAPHE, TRI_NBR_BLOCS, FORCE_NON);
    OBJ_choisit(ORDONNE_GRAPHE, TRI_ALEATOIRE, FORCE_OUI);
    return BI_EVENEMENT_TRAITE;

  case TRI_BLOCS:
    OBJ_active(ORDONNE_GRAPHE, TRI_NBR_BLOCS, FORCE_OUI);
    OBJ_choisit(ORDONNE_GRAPHE, TRI_BLOCS, FORCE_OUI);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Prparation de l'arbre d'interface -------- */

static void TRI_PreparerArbre(int genre)
{
  int retour;

  if ( FAUX == arbres[ ORDONNE_GRAPHE ].drapeaux.prepare ) {
    /* Mise en forme de l'arbre & initialisation */
    retour = ARB_Initialiser( ORDONNE_GRAPHE, "menu_graphes.html#tri", TRI_GererArbre, ( int (*)(unsigned int, int, int *) ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return;
    }

    /* Positionnement des objets */
    OBJ_deplace( ORDONNE_GRAPHE, TRI_TXT_BLOCS, 0, +500, DEPLACE_RELATIF );
    OBJ_deplace( ORDONNE_GRAPHE, TRI_NBR_BLOCS, 0, +500, DEPLACE_RELATIF );

#ifdef AVEC_TRADUCTION_
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_TXT_GRAPHE, texte_arbre_tri_le_graphe  );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_ALEATOIRE , texte_arbre_tri_desordonne );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_BLOCS     , texte_arbre_tri_par_blocs  );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_TXT_BLOCS , texte_arbre_tri_nbr_blocs );

    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_TXT_DOUBLONS, texte_arbre_tri_doublons );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_ELIM_DOUBLON, texte_arbre_tri_eliminer );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_MOY_DOUBLON , texte_arbre_tri_moyenner );

    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_TXT_PAQUETS, texte_arbre_tri_par_paquets );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_TXT_FAU_DBL, texte_arbre_tri_quasi_doublons );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_TXT_SIMPLE , texte_arbre_tri_simples );

    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_LANCER, texte_arbre_tri_ok );
    OBJ_TraduireTexte( ORDONNE_GRAPHE, TRI_ANNULE, texte_arbre_tri_annuler );
#endif

    /* L'arbre est maintenant pret */
    arbres[ ORDONNE_GRAPHE ].drapeaux.prepare = VRAI;
  }

  /* Qu'est-ce qui est possible avec tel ou tel type de graphe ? */
  switch ( genre )
  {
  case GRF_FCT_REELLE:
    OBJ_active( ORDONNE_GRAPHE, TRI_BLOCS, FORCE_OUI );
    break;

  case GRF_FCT_COMPLEXE:
    OBJ_active( ORDONNE_GRAPHE, TRI_BLOCS, FORCE_OUI );	/* Impossible de trier par blocs */
    ARB_SimulerClic( ORDONNE_GRAPHE, TRI_ALEATOIRE, 1 );
    break;
  }
}
