/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998-2002, Emmanuel CURIS - Cration de ce fichier: 9 juin 1998

   Calcul des coefficients d'une spline approchant le graphe indiqu
   Cas d'un graphe rel
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "graphes.h"
#include "lissage.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int preparer_tableaux(unsigned long nbr_pts, double *x, double *y,
			     double *b, double *c, double *d);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Partie essentielle : construction des tableaux de coefficients en chaque point -\
 |   Renvoie 0 si tout s'est bien pass, <0 en cas d'erreur                                 |
 |                                                                                          |
 | nbr_pts : nombre de points dans le tableau  lisser                                      |
 | x, y    : tableaux des couples (x, y)                                                    |
 | b, c, d : tableaux des coefficients supposs dj rservs                               |
 \----------------------------------------------------------------------------------------- */

static int preparer_tableaux(unsigned long nbr_pts, double *x, double *y,
			     double *b, double *c, double *d)
{
  unsigned long i;

  if ( (x == NULL) || (y == NULL) || (b == NULL) || (c == NULL) || (d == NULL) )
    return -1;

  if ( x[1] == x[0] ) {
    return ERR_LIS_DOUBLON;
  }
  d[0] =   x[1] - x[0];
  c[1] = ( y[1] - y[0] ) / d[0];
  for (i = 1 ; i < nbr_pts - 1 ; i++) {
    d[i]     = x[i + 1] - x[i];
    if ( d[i] == 0.0 ) {
      return ERR_LIS_DOUBLON;
    }
    b[i]     = 2.0 * ( d[i - 1] + d[i] );
    c[i + 1] = ( y[i + 1] - y[i] ) / d[i];
    c[i]     = c[i+1] - c[i];
  }
  b[0] = -d[0];
  b[nbr_pts - 1] = -d[nbr_pts - 2];
		
  if ( (x[1] == x[3]) || (x[2] == x[0]) || (x[3] == x[0]) ) {
    return ERR_LIS_DOUBLON;
  }
  c[0]  = c[2] / (x[3] - x[1])
        - c[1] / (x[2] - x[0]);
  c[0] *= d[0] * d[0] / (x[3] - x[0]);

  if ( (x[nbr_pts - 1] == x[nbr_pts - 3]) ||
       (x[nbr_pts - 2] == x[nbr_pts - 4]) ||
       (x[nbr_pts - 4] == x[nbr_pts - 1]) ) {
    return ERR_LIS_DOUBLON;
  }
  c[nbr_pts - 1]  = c[nbr_pts - 2] / (x[nbr_pts - 1] - x[nbr_pts - 3])
	          - c[nbr_pts - 3] / (x[nbr_pts - 2] - x[nbr_pts - 4]);
  c[nbr_pts - 1] *= - d[nbr_pts - 2] * d[nbr_pts - 2] / (x[nbr_pts - 1] - x[nbr_pts - 4]);
		
  for (i = 1 ; i < nbr_pts ; i ++) {
    double pas;

    if ( b[i - 1] == 0.0 ) { /* b[i] = 2 * (x[i + 1] - x[i - 1]) */
      return ERR_LIS_DOUBLON;
    }
    pas = d[i - 1] / b[i - 1];
    b[i] -= pas * d[i - 1];
    c[i] -= pas * c[i - 1];
  }

  if ( b[nbr_pts - 1] == 0.0 ) {
    return ERR_LIS_NON_TRIE;
  }
  c[nbr_pts - 1] /= b[nbr_pts - 1];
  for (i = nbr_pts - 2 ; i > 0 ; i--) {
    c[i] -= d[i] * c[i + 1];
    if (b[i] == 0.0) { /* b ne peut tre nul que si les X ne sont pas tris */
      return ERR_LIS_NON_TRIE;
    }
    c[i] /= b[i];
  }
  c[0] -= d[0] * c[1];
  c[0] /= b[0]; /* b[0] a dj t vrifi non-nul */

  /* Les d ont dj t vrifis non-nuls */
  b[nbr_pts - 1]  = (y[nbr_pts - 1] - y[nbr_pts - 2] ) / d[nbr_pts - 2];
  b[nbr_pts - 1] += d[nbr_pts - 2] * (c[nbr_pts - 2] + 2.0 * c[nbr_pts - 1] );
  for (i = 0 ; i < nbr_pts - 1 ; i++) {
    b[i]  = (y[i + 1] - y[i]) / d[i];
    b[i] -= d[i] * ( c[i + 1] + 2.0 * c[i] );
    d[i]  = ( c[i + 1] - c[i] ) / d[i];
    c[i] *= 3.0;
  }
  c[nbr_pts - 1] *= 3.0;
  d[nbr_pts - 1]  = d[nbr_pts - 2];

  return 0;
}

/*********** Lissage par une "spline", la plus adapte possible ********************************
 *  Suppose que le graphe  lisser soit tri, par ordre croissant des X                        *
 *          que la liste de zones soit trie (mme ordre) et disjointe                         *
 ***********************************************************************************************/

/* --------- Prpare les tableaux de coefficients utiliss pour les splines ---------------------\
 |  Renvoie pas grand chose, -1 si erreur                                                        |
 |  Version pour une srie de portions du graphe seulement                                       |
 |  Les quatre tableaux crs dans la procdure ne sont PAS librs : ne pas oublier de le faire |
 |                                                                                               |
 | graphe     : graphe  lisser
 | nbr_zones  : nombre d'intervalles de points valables
 | zones      : bornes de ces intervalles ([0] = min1, [1] = max1, [2] = min2, ...)
 | b, c, d    : tableaux des coefficients (rservs DANS la procdure, d'o leur type...)
 | tableau    : tableau de la portion de graphe rellement utilise, cr sur place
 | nbr_points : nombre de points dans le tableau ci-dessus
 \---------------------------------------------------------------------------------------------- */
 
int SPL_prepare(const Graphe *graphe, char nbr_zones, double *zones, 
		double **b, double **c, double **d, double **tableau, unsigned long *nbr_points)
{
  short int a;
  int retour;
  double x_min, x_max; /* Intervalle total de lissage */
  register double *zone_x, *zone_y, x;
  unsigned long i, nbr_pts = 0;

  if ( (nbr_zones == 0) || (zones == NULL) ) return -1; /* Aucun intervalle spcifi */

  x_min = zones[0]; x_max = zones[1];
  for (a = 2 ; a <= 2*(nbr_zones - 1) ; a += 2) {
    if (zones[a]     < x_min) x_min = zones[a];
    if (zones[a + 1] > x_max) x_max = zones[a + 1];
  }

  /* Calcul du nombre de points */
  zone_x = graphe->x;
  for (i = 0 ; i < graphe->nombre_point ; i++) {
    x = *zone_x;
    for (a = 0 ; a < 2*nbr_zones ; a += 2) {
      if ( (x >= zones[a]) && (x <= zones[a + 1]) ) {
	nbr_pts++;
	break;
      }
    }
    zone_x++;
  }

  if (nbr_pts < 4) { /* Moins de quatre -> sans intrt */
    return ERR_LIS_TROP_PEU_POINTS;
  }

  *tableau = (double *) calloc( 2 * nbr_pts, sizeof(double) ); /* X et Y... */
  if (*tableau == NULL) {
    return ERR_LIS_MEMOIRE;
  }

  *b = (double *) calloc( nbr_pts, sizeof(double) );
  *c = (double *) calloc( nbr_pts, sizeof(double) );
  *d = (double *) calloc( nbr_pts, sizeof(double) );
  if ( (*d == NULL) || (*c == NULL) || (*b == NULL) ) {
    free(*tableau);
    free(*b); free(*c); free(*d); /* On libre la mmoire ventuellement prise */
    return ERR_LIS_MEMOIRE;
  }

  /* On recopie la partie intressante du graphe dans tableau */
  zone_x = *tableau;
  zone_y = *tableau + nbr_pts;
  for (i = 0 ; i < graphe->nombre_point ; i++) {
    x = *(graphe->x + i);
    for (a = 0 ; a < 2*nbr_zones ; a += 2) {
      if ( (x >= zones[a]) && (x <= zones[a + 1]) ) {
	*zone_x = x;
	*zone_y = *(graphe->y + i);
	zone_x++;
	zone_y++;
	break;
      }
    }
  }

  /* Prparation des tableaux de coefficients */
  retour = preparer_tableaux(nbr_pts, *tableau, (*tableau) + nbr_pts, *b, *c, *d);
  if (retour < 0) { /* Problme */
    free(*b); free(*c); free(*d); /* On libre la mmoire ventuellement prise */
    free(*tableau);
    *tableau = NULL;
    *b = NULL; *c = NULL; *d = NULL;
    return retour;
  }
    
  *nbr_points = nbr_pts;
  return 1; /* Pas de problme */
}

/* --------- Prpare les tableaux de coefficients utiliss pour les splines ----------------\
 |  Renvoie 0 si tout va bien, <0 en cas d'erreur                                           |
 |  Version pour _tout_ le graphe pass en paramtre                                        |
 |  Les 3 tableaux crs dans la procdure ne sont PAS librs : ne pas oublier de le faire |
 |                                                                                          |
 | graphe     : graphe  lisser                                                             |
 | b, c, d    : tableaux des coefficients (rservs DANS la procdure, d'o leur type...)   |
 \----------------------------------------------------------------------------------------- */
 
int SPL_tt_prep(const Graphe *graphe, double **b, double **c, double **d)
{
  int retour;
  unsigned long nbr_pts;

  if (graphe == NULL) {
#ifdef _DEVELOPPE_
    fprintf(stderr, "SPL_tt_prep appele avec un graphe nul !\n");
#endif
    return -1000;
  }

  if (graphe->nombre_point < 4) { /* Moins de quatre -> sans intrt */
    (void) fprintf(stderr, "SPL_tt_prep : %s a trop peu de points !\n", graphe->nom);    
    return ERR_LIS_TROP_PEU_POINTS;
  }

  nbr_pts = graphe->nombre_point; /* Pour viter de se prendre la tte aprs... */

  *b = (double *) calloc( nbr_pts, sizeof(double) );
  *c = (double *) calloc( nbr_pts, sizeof(double) );
  *d = (double *) calloc( nbr_pts, sizeof(double) );
  if ( (*d == NULL) || (*c == NULL) || (*b == NULL) ) {
    free(*b); free(*c); free(*d); /* On libre la mmoire ventuellement prise */
    *b = NULL; *c = NULL; *d = NULL;
#ifdef _DEVELOPPE_
    fprintf(stderr, "SPL_tt_prep : mmoire insuffisante (taille demande : 3 * %lu)\n", sizeof(double) * nbr_pts);
    fprintf(stderr, "   tableaux de la spline\n");
#endif
    return ERR_LIS_MEMOIRE;
  }

  /* Prparation des tableaux de coefficients */
  retour = preparer_tableaux(nbr_pts, graphe->x, graphe->y, *b, *c, *d);
  if (retour < 0) { /* Problme */
    free(*b); free(*c); free(*d); /* On libre la mmoire ventuellement prise */
    *b = NULL; *c = NULL; *d = NULL;
    (void) fprintf(stderr, "SPL_tt_prep : erreur en prparant les tableaux...\n");    
    return retour;
  }

  return 0; /* Pas de problme */
}

/* --------- Calcule un point de la spline obtenue par la procdure prcdente -------------\
 |   Renvoie la valeur en ce point (0 en cas d'erreur...)                                   |
 |   Version pour une spline calcule sur plusieurs graphes                                 |
 |   Suppose de plus que tableau est class par X croissant                                 |
 |                                                                                          |
 | x       : point en lequel calculer                                                       |
 | nbr_pts : nombre de points utiliss pour le calcul de la spline                          |
 | tableau : tableaux des X et Y (tous les X puis tous les Y), suppos tri par X croissant |
 | b, c, d : coefficients de la spline calculs par la fonction au-dessus                   |
 \----------------------------------------------------------------------------------------- */

double SPL_calcule(double x, unsigned long nbr_pts, double *tableau,
		   double *b, double *c, double *d)
{
  double renvoi = 0.0;
  double delta_x;
  unsigned long i;

  if (tableau == NULL) return 0.0;
  if (b == NULL) return 0.0;
  if (c == NULL) return 0.0;
  if (d == NULL) return 0.0;

  /* On recherche le point en x le plus proche, par dichotomie */
  if ( x <= tableau[0]) {
    delta_x = x - tableau[0];
    i = 0;
  } else if (x >= tableau[nbr_pts - 1]) {
    delta_x = x - tableau[nbr_pts - 1];
    i = nbr_pts - 1;
  } else {
    /* Recherche du point le plus proche */

    i = 1;
    while (x > tableau[i])
      i++; /* lment qui suit juste x */
    i--; /* On prend l'lment juste avant */
    
    delta_x = x - tableau[i];
  }

  if (fabs(delta_x) > 0.0) {
    renvoi  = delta_x * d[i];
    renvoi += c[i];
    renvoi *= delta_x;
    renvoi += b[i];
    renvoi *= delta_x;
  }
  renvoi += tableau[nbr_pts + i]; /* R = Y + b * X + c * X^2 + d * X^3 */
  
  return renvoi;
}

/* --------- Calcule un point de la spline obtenue par la procdure prcdente -------------\
 |   Renvoie la valeur en ce point (0 en cas d'erreur, mais avec une alerte...)             |
 |   Version pour une spline sur tout un graphe                                             |
 |                                                                                          |
 | graphe  : graphe qui a servi pour le calcul de la spline                                 |
 | x       : point en lequel calculer                                                       |
 | b, c, d : coefficients de la spline calculs par la fonction au-dessus                   |
 \----------------------------------------------------------------------------------------- */

double SPL_tt_calc(const Graphe *graphe, double x, double *b, double *c, double *d)
{
  double renvoi = 0;
  double delta_x;
  unsigned long i;

  if (graphe == NULL) {
#ifdef _DEVELOPPE_
    fprintf(stderr, "SPL_tt_calc appele avec un graphe nul !\n");
#endif
    return 0.0;
  }

  if ( (b == NULL) || (c == NULL) || (d == NULL) ) { 
#ifdef _DEVELOPPE_
    fprintf(stderr, "SPL_tt_calc appele avec des tableaux nuls (graphe : %s) !\n", graphe->nom);
#endif
    return 0.0; /* Problme -> donner une alerte */
  }
  
  /* On recherche le point en x le plus proche, par dichotomie */
  if ( x <= graphe->x_min)
  {
    delta_x = x - graphe->x_min;
    i = 0;
  }
  else if (x >= graphe->x_max)
  {
    delta_x = x - graphe->x_max;
    i = graphe->nombre_point - 1;
  }
  else
  {
    /* Recherche du point le plus proche */

    i = 1;
    while (x > (graphe->x)[i])
      i++; /* lment qui suit juste x */
    i--; /* On prend l'lment juste avant */
    
    delta_x = x - (graphe->x)[i];
  }

  if (fabs(delta_x) > 0)
  {
    renvoi  = delta_x * d[i];
    renvoi += c[i];
    renvoi *= delta_x;
    renvoi += b[i];
    renvoi *= delta_x;
  }
  renvoi += (graphe->y)[i]; /* R = Y + b * X + c * X^2 + d * X^3 */

  return renvoi;
}

/* --------- Calcule un point de la spline obtenue par la procdure prcdente -------------\
 |           ainsi que la drive en ce point                                               |
 |   Renvoie la valeur en ce point (0 en cas d'erreur, mais avec une alerte...)             |
 |           (et, comme paramtre, la valeur de la drive)                                 |
 |   Version pour une spline sur tout un graphe                                             |
 |                                                                                          |
 | graphe  : graphe qui a servi pour le calcul de la spline                                 |
 | x       : point en lequel calculer                                                       |
 | b, c, d : coefficients de la spline calculs par la fonction au-dessus                   |
 | derivee : valeur de la drive, calcule d'aprs la spline, en ce point                  |
 \----------------------------------------------------------------------------------------- */

double SPL_tt_derive(const Graphe *graphe, double x, double *b, double *c, double *d, double *derivee)
{
  double renvoi = 0;
  double delta_x;
  unsigned long i;

  if (b == NULL) return 0; /* Problme -> donner une alerte */
  if (c == NULL) return 0;
  if (d == NULL) return 0;
  
  /* On recherche le point en x le plus proche, par dichotomie */
  if ( x <= graphe->x_min)
  {
    delta_x = x - graphe->x_min;
    i = 0;
  }
  else if (x >= graphe->x_max)
  {
    delta_x = x - graphe->x_max;
    i = graphe->nombre_point - 1;
  }
  else
  {
    /* Recherche du point le plus proche */

    i = 1;
    while (x > (graphe->x)[i])
      i++; /* lment qui suit juste x */
    i--; /* On prend l'lment juste avant */
    
    delta_x = x - (graphe->x)[i];
  }

  if (fabs(delta_x) > 0) {
    renvoi  = delta_x * d[i];
    renvoi += c[i];
    renvoi *= delta_x;
    renvoi += b[i];
    renvoi *= delta_x;

    /* Drive = b + 2 * c * X + 3 * d * X^2 */
    *derivee = b[ i ] + 2.0 * c[ i ] * delta_x + 3.0 * d[ i ] * delta_x * delta_x;
  } else {
    *derivee = b[ i ];
  }
  renvoi += (graphe->y)[i]; /* R = Y + b * X + c * X^2 + d * X^3 */

  return renvoi;
}
