/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998-2003, Emmanuel CURIS

   Rsolution d'un systme par la mthode du pivot de Gauss
   (Merci  Bruno Blanchet pour le source en pascal)

   Suppose que le systme  une solution
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "matrices.h"

/*#define DEBUG_GAUSS_ */

/* -------- Rsolution d'un systme de la forme AX = B, X : routine gnrale
    Renvoie 0 si tout s'est bien pass
            code d'erreur sinon

   A : matrice [carre] des coefficients
   X : vecteur [colonne] des inconnues
   B : vecteur [colonne] des coefficients constants
   methode : la mthode  utiliser pour la rsolution, cf. matrice.h pour les codes
   -------- */

int MAT_ResoudSysteme(Matrice *A, Matrice *X, Matrice *B, unsigned int methode)
{
  int retour = -1974;

  /* Contrles prliminaires */
  if ( ( NULL == A ) || ( NULL == X ) || ( NULL == B ) ) {
    AVERTISSEMENT( "Une matrice au moins NULL : A=%p, X = %p, B = %p", (void *) A, (void *) X, (void *) B );
    return -1;
  }

  /* Appel des fonctions en fonction de la mthode indique */
  switch ( methode )
  {
  case MATRICE_LINEAIRE_GAUSS:
    /* Mthode du pivot de Gauss */
    retour = MAT_ResoudGauss( A, X, B );
    break;

  default:
    AVERTISSEMENT( "Code de mthode inconnu : %u", methode );
  }

  /* On renvoie le code de retour */
  return retour;
}

/* ~-~-~-~-~-~-~-~-~ Rsolution effective ~-~-~-~-~-~-~-~-~ */

/* -------- Mthode du pivot de Gauss -------- */

int MAT_ResoudGauss(Matrice *mat_A, Matrice *mat_X, Matrice *mat_B)
{
  int retour;
  int i, j, k, l_pivot = 0, c_pivot = 0; /* i : colonne, j : ligne */
  int *idx_sol;
  unsigned int taille_systeme;

  /* Contrles */
  if ( ( NULL == mat_X ) || ( NULL == mat_A ) || ( NULL == mat_B ) ) {
    return -1;
  }
#ifdef _DEVELOPPE_
  if ( mat_A->n_lignes != mat_A->n_colonnes ) {
    (void) fprintf(stderr, "%s,%s l.%u : La matrice A doit tre carre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -10;
  }
  if ( mat_X->n_lignes != mat_B->n_lignes ) {
    (void) fprintf(stderr, "%s,%s l.%u : Les vecteurs X et B doivent avoir le mme nombre de composantes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -11;
  }
  if ( mat_X->n_lignes != mat_A->n_lignes ) {
    (void) fprintf(stderr, "%s,%s l.%u : Matrices A et X incompatibles\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -12;
  }
  if ( mat_X->n_colonnes != mat_B->n_colonnes ) {
    /* Rq. Pas forcment 1, si systme rpt [genre inversion !] */
    (void) fprintf(stderr, "%s,%s l.%u : Les matrices X et B doivent avoir les mmes dimensions\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -13;
  }
#endif

  /* Nombre d'inconnues */
  taille_systeme = mat_A->n_colonnes;

  /* Cas spciaux, faciles */
  /* 1. Une inconnue : ax = b */
  if ( 1 == taille_systeme ) {
    double coefficient;

    coefficient = lit_element( mat_A, 0, 0 );
    if ( 0.0 == coefficient ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas de solution\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return MATRICE_LINEAIRE_PAS_DE_SOLUTION;
    }

    /* Rsolution directe */
    for ( i = 0 ; i < (int) mat_X->n_colonnes ; i++ ) {
      met_element( mat_X, 0, i, lit_element( mat_B, 0, i ) / coefficient );
    }
    return 0;
  }

  /* 2. Deux inconnues : mthode de Cramer [plus stable ?] */
  if ( 2 == taille_systeme ) {
    double determinant, Dx, Dy;

#ifdef _DEVELOPPE_
    (void) printf("%s,%s : Deux quations  deux inconnues\n",  __FILE__, __FUNCTION__);
#endif

    determinant = lit_element( mat_A, 0, 0 ) * lit_element( mat_A, 1, 1 ) - lit_element( mat_A, 0, 1 ) * lit_element( mat_A, 1, 0 );
    if ( 0.0 == determinant ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas de solution, dterminant nul\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return MATRICE_LINEAIRE_PAS_DE_SOLUTION;
    }

    for ( i = 0 ; i < (int) mat_X->n_colonnes ; i++ ) {
      Dx = lit_element( mat_B, 0, i ) * lit_element( mat_A, 1, 1 ) - lit_element( mat_B, 1, i ) * lit_element( mat_A, 0, 1 );
      Dy = lit_element( mat_A, 0, 0 ) * lit_element( mat_B, 1, i ) - lit_element( mat_A, 1, 0 ) * lit_element( mat_B, 0, i );

      met_element( mat_X, 0, i, Dx / determinant );
      met_element( mat_X, 1, i, Dy / determinant );
    }
    return 0;
  }

  /* On force une matrice quelconque */
  retour = MAT_ChangerType( mat_A, MATRICE_QUELCONQUE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de changer le type de la matrice !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
       
  /* Table des permutations */
  idx_sol = (int *) calloc( taille_systeme, sizeof(int));
  if ( idx_sol == NULL ) {
    ERR_Memoire( "Table de permutation des solutions", sizeof( int ) * taille_systeme );
    return -1;
  }

  /* Au dbut : inconnue 1 = coef 1, ..., inconnue n = coef n */
  for ( i = 0 ; i < (int) taille_systeme ; i++ )
    idx_sol[i] = i;

#ifdef DEBUG_GAUSS_
  printf("%s : matrice initiale =\n", __FUNCTION__ );
  MAT_AfficherMatrice( mat_A );
  MAT_AfficherMatrice( mat_B );
#endif

  for ( k = taille_systeme - 1 ; k >= 0 ; k-- ) { /* Nombre de passages pour rendre le systme diagonal */
    /* Recherche du pivot dans toute la matrice k * k */
    double pivot = 0.0;

    for (i = 0 ; i <= k ; i++)
      for (j = 0 ; j <= k ; j++) {
	double val;

	val = lit_element( mat_A, i, j ); /* Valeur de l'lment de matrice */
	if ( fabs(val) > fabs(pivot) ) { /* On cherche le plus grand pivot... */
	  l_pivot = j; /* Ligne du pivot   */
	  c_pivot = i; /* Colonne du pivot */
	  pivot = val;
	}
      }

    if ( pivot == 0.0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Matrice non-inversible (pivot nul)\n",
		     __FILE__, __FUNCTION__, __LINE__);
      free( idx_sol ); idx_sol = NULL;
      return MATRICE_LINEAIRE_PAS_DE_SOLUTION;
    }
#ifdef DEBUG_GAUSS_
    else
      printf("Pivot choisi, tour %d : %g\n", k, pivot);
#endif

    /* On s'arrange pour placer le pivot dans le coin infrieur droit de la matrice */
    if (l_pivot != k) { /* Il faut changer les lignes */
      MAT_EchangerLignes( mat_A, k, l_pivot ); /* Pour A */
      MAT_EchangerLignes( mat_B, k, l_pivot ); /* Pour B */

#ifdef DEBUG_GAUSS_
/*       printf("Aprs change des lignes :\n"); */
/*       MAT_AfficherMatrice( mat_A ); */
#endif
    }

    if (c_pivot != k) { /* Il faut changer les colonnes */
      int tmp = idx_sol[k];
      idx_sol[k] = idx_sol[c_pivot];
      idx_sol[c_pivot] = tmp; /* On change aussi les inconnues */

      MAT_EchangerColonnes( mat_A, k, c_pivot );

#ifdef DEBUG_GAUSS_
/*       printf("Aprs change des colonnes :\n"); */
/*       MAT_AfficherMatrice( mat_A ); */
#endif
    }

    /* On annule la colonne k en combinant les lignes grce au pivot */
    for (j = 0 ; j < (int) taille_systeme ; j++) {
      if ( j != k ) {
	double coefficient;

	coefficient = -lit_element( mat_A, j, k ) / pivot;
	MAT_CL_Lignes( mat_A, j, k, coefficient );
	MAT_CL_Lignes( mat_B, j, k, coefficient );

	/* On force  0 pour viter des erreurs d'arrondi aprs */
	met_element( mat_A, j, k, 0.0 );
      }
    }

#ifdef DEBUG_GAUSS_
    printf("Aprs combinaison linaire des lignes :\n");
    MAT_AfficherMatrice( mat_A );
    MAT_AfficherMatrice( mat_B );
#endif
  }

  /* Ici, on a une matrice diagonale + la colonne des y... */
  for ( i = 0 ; i < (int) taille_systeme ; i++ ) {
    double tmp;

    tmp = lit_element( mat_A, i,i ); /* lment sur la diagonale */
    if ( 0.0 == tmp ) {
      free( idx_sol ); idx_sol = NULL;
      (void) fprintf(stderr, "%s,%s l.%u : Valeur nulle sur la diagonale, matrice non-inversible\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return MATRICE_LINEAIRE_PAS_DE_SOLUTION;
    }

    mat_X->valeurs[ idx_sol[ i ] ] = mat_B->valeurs[ i ] / tmp;
  }

  free( idx_sol ); idx_sol = NULL;

  return 0; /* Pas de problme */
}
