/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Prparation des modles avant l'ajustement (partie gnrique)
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "couches.h" /* Pour NC_PreparerModele */
#include "decomposition.h"	/* Pour MODELE_Decomposition_PreparerAjustement */
#include "graphes.h"
#include "modeles.h"

#include "modeles_init.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Initialisations qui dpendent du modle -------- */

int MODELE_Initialiser(ModeleAjustement *modele)
{
  int retour = 0;

  if ( NULL == modele ) return -1;
  if ( NULL == modele->infos_modele.feff ) {
    (void) fprintf( stderr, "%s,%s l.%u : pas d'informations complmentaires !\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  switch ( modele->type )
  {
  case MODELE_CL_MODELES: /* Combinaison linaire de modles : on initialise chacun des termes */
    {
      unsigned int i;
      struct cl_modeles *infos;

      infos = modele->infos_modele.cl_modeles;
      for (i = 0 ; i < infos->nbr_termes ; i++) {
	retour |= MODELE_Initialiser( infos->modeles[i] );
      }
    }
    break;

  case MODELE_EXAFS_CHEMINS:
    retour = NC_PreparerModele( modele->infos_modele.feff );
    break;

  case MODELE_EXAFS_GNXAS:
  case MODELE_SEUIL_BASIQUE:
  case MODELE_FLUOX:
    (void) fprintf(stderr, "MODELE_Initialiser : type de modle non gr, %d\n", modele->type);
    retour = -5;
    break;

  case MODELE_DECOMPOSITION:
    retour = MODELE_Decomposition_PreparerAjustement( modele );
    break;

  case MODELE_GENERIQUE:
    retour = MODELE_Generique_PreparerAjustement( modele );
    break;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : type de modle inconnu, %d\n",
		    __FILE__, __FUNCTION__, __LINE__, modele->type );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    retour = -8;
  }

#ifdef _DEVELOPPE_
  printf( "Sortie de ==> %s <== [code retour : %d]\n",
	  __FUNCTION__, retour );
#endif
  return retour;
}

/* -------- Installer un graphe de rfrence -------- */

int MODELE_MettreReference(ModeleAjustement *modele, struct graphe *experience, boolean eff_anciens)
{
  if (NULL == modele) return -1;
  if (NULL == experience) return -2;

  if (modele->experience != NULL) {
    if (modele->experience == experience) return 0; /* Rien  faire... */

    /* On demande pour effacer les anciens... */
    if ( QUESTION == eff_anciens ) {
      if ( ( modele->modele != NULL ) || ( modele->residu != NULL ) ) {
	int reponse;

	reponse = BI_ERR_alerte( NULL, &Q_EffacerAnciens, VRAI );
	if ( 3 == reponse ) { /* On annule */
	  return +1;
	}
	if ( 1 == reponse )
	  eff_anciens = VRAI;
	else
	  eff_anciens = FAUX;
      }
    }

    /* On enlve l'ancienne rfrence */
    if ( FAUX != eff_anciens ) {
      if ( modele->modele == experience ) {
	(void) fprintf(stderr, "MODELE_MettreReference : impossible d'effacer l'ancien modle : on cherche  l'ajuster !\n");
	eff_anciens = FAUX;
      } else if ( modele->residu == experience ) {
	(void) fprintf(stderr, "MODELE_MettreReference : impossible d'effacer l'ancien rsidu : on cherche  l'ajuster !\n");
	eff_anciens = FAUX;
      }
    }
    MODELE_OterReference(modele, eff_anciens);
  }

  /* On installe la nouvelle rfrence */
  modele->experience = experience;
  return 0;
}

int MODELE_PreparerAjustement(ModeleAjustement *modele, int action_ancien_modele)
{
  char *titre;
  boolean effacer_anciens = FAUX;

  if ( NULL == modele ) return -1;
#ifdef _DEVELOPPE_
  printf( "On entre dans ==> %s <== \n", __FUNCTION__ );
#endif

  if ( VRAI == modele->drapeaux.en_cours ) {
#ifdef _DEVELOPPE_
    (void) fprintf( stderr, "%s,%s l.%u : ajustement en cours avec ce modle, modification impossible !\n",
		    __FILE__, __FUNCTION__, __LINE__);
#endif
    (void) BI_ERR_alerte( NULL, &ERR_AjustementEnCours, VRAI );
    return ERREUR_MODELE_AJUSTEMENT_EN_COURS;
  }
  if ( VRAI == modele->drapeaux.preparation ) {
    (void) fprintf(stderr, "MODELE_PreparerAjustement : ce modle est dj en cours de prparation !\n");
    return -9;
  }

  if ( NULL == modele->experience ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "MODELE_PreparerAjustement : pas de spectre exprimental !\n");
#endif
    (void) BI_ERR_alerte( NULL, &ERR_SansExperience, VRAI );
    return -2;
  }

  /* On perd l'information d'arrt forc, car elle n'a plus de sens */
  modele->drapeaux.arret_force = FAUX;

  if ( ( modele->x_min >= modele->experience->x_max ) || ( modele->x_max <= modele->experience->x_min ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "MODELE_PreparerAjustement : intervalle d'ajustement [%f, %f] hors du spectre  ajuster [%f, %f] !\n",
		   modele->x_min, modele->x_max, modele->experience->x_min, modele->experience->x_max);
#endif
    (void) BI_ERR_alerte( NULL, &ERR_IntervalleIncorrect, VRAI,
			  modele->x_min, modele->x_max, modele->experience->x_min, modele->experience->x_max);
    return -3;
  }

  if ( modele->reference != modele->experience ) {
    GRF_eff_isole( modele->reference ); /* Si la rfrence existe & diffre de l'exprience, c'est un graphe isol */
    modele->reference = NULL;
  }

  /* Le graphe qui est rellement ajust (sous-ensemble du graphe exprimental */
  if ( ( modele->x_min > modele->experience->x_min ) || ( modele->x_max < modele->experience->x_max ) ) {
    unsigned long nbr_pts, premier;

    /* On ne garde qu'un sous-ensemble du graphe exprimental [suppos tri] */
    nbr_pts = GRF_intervalle( modele->experience, modele->x_min, modele->x_max, &premier);
    if ( nbr_pts < 2 ) {
#ifdef _DEVELOPPE_
      (void) fprintf( stderr, "%s,%s l.%u : trop peu de points exprimentaux dans l'intervalle d'ajustement [%f, %f]\n",
		      __FILE__, __FUNCTION__, __LINE__, modele->x_min, modele->x_max);
#endif
      (void) BI_ERR_alerte( NULL, &ERR_IntervalleVide, VRAI, modele->x_min, modele->x_max, (int) nbr_pts);
      return -4;
    }
    if ( nbr_pts == modele->experience->nombre_point ) {
      modele->reference = modele->experience;
    } else { /* On extrait la partie intressante sous forme de graphe isol */
      modele->reference = GRF_ExtraitGraphe( modele->experience, premier, nbr_pts, VRAI );
      if ( NULL == modele->reference ) {
	ERR_Memoire( "Cration du graphe de rfrence", 0 );
	return -5;
      }
    }
  } else { /* On ajuste l'ensemble du spectre exprimental */
    modele->reference = modele->experience;
  }
  if ( VRAI == modele->drapeaux.pondere ) {
    /* A-t-on rellement les informations pour faire un ajustement pondr par les incertitudes ? */
    int reponse;

    /* Y a-t-il des incertitudes ? */
    if ( NULL == modele->reference->erreur_y ) {
      reponse = BI_ERR_alerte( NULL, &ERR_PasDIncertitude, VRAI );
      if ( 1 == reponse ) {
	modele->drapeaux.pondere = FAUX; /* Pondration par les barres d'erreur impossible */
      } else {
	return -20;
      }
    } else {
      unsigned long i;
      double *erreur;

      /* Oui... mais sont-elles toutes non-nulles ? */
      for ( i = 0, erreur = modele->reference->erreur_y->y ; i < modele->reference->nombre_point ; i++, erreur++ ) {
	if ( 0.0 == *erreur ) {
	  reponse = BI_ERR_alerte( NULL, &ERR_ErreurNulle, VRAI );
	  if ( 1 == reponse ) {
	    modele->drapeaux.pondere = FAUX; /* Pondration par les barres d'erreur impossible */
	    break;
	  } else {
	    return -21;
	  }
	}
      }
    }
  }

  /* Le rsidu & le modle */
  if ( modele->modele != NULL ) { /* Que faire de l'ancien modle ? */
    boolean automatique = VRAI;

    if ( Q_MODELE_ANCIENS_RESULTATS_DEMANDER == action_ancien_modele ) {
      automatique = FAUX;
      action_ancien_modele = BI_ERR_alerte( NULL, &Q_EffacerAnciens, VRAI );
    }
    if ( Q_MODELE_ANCIENS_RESULTATS_ANNULER == action_ancien_modele ) { /* On annule */
      if ( VRAI == automatique ) {
	(void) fprintf( stderr, "%s,%s l.%u : Ajustement annul car il y a encore les rsultats de l'ajustement prcdent\n",
			__FILE__, __FUNCTION__, __LINE__);
      }
      return -1;
    }
    modele->modele->drapeaux.protege = FAUX;
    if ( Q_MODELE_ANCIENS_RESULTATS_EFFACER == action_ancien_modele ) { /* On efface */
      GRF_efface( modele->modele, VRAI, FAUX );
      modele->modele = NULL;
      effacer_anciens = VRAI;
    }
  }
  titre = TXT_genere_chaine( "Mod"e_GRAVE"le de %s", modele->experience->nom );
  modele->modele = GRF_nouveau( modele->experience, modele->reference->nombre_point, modele->experience->type,
			        modele->experience->genre_donnees, titre );
  free( titre ); titre = NULL;
  if ( NULL == modele->modele ) {
    ERR_Memoire( "Graphe contenant le modle", 0 );
    return -6;
  }
  modele->modele->drapeaux.theorique = VRAI;

  if ( modele->residu != NULL ) { /* Que faire de l'ancien rsidu ? */
    modele->residu->drapeaux.protege = FAUX;
    if ( VRAI == effacer_anciens ) {
      GRF_efface( modele->residu, VRAI, FAUX );
      modele->residu = NULL;
    }
  }
  titre = TXT_genere_chaine( "R"e_AIGU"sidu du mod"e_GRAVE"le de %s", modele->experience->nom );
  modele->residu = GRF_nouveau(modele->modele, modele->reference->nombre_point, GR_ECART,
			       modele->experience->genre_donnees, titre);
  free( titre ); titre = NULL;
  if (NULL == modele->residu) {
    ERR_Memoire( "Graphe contenant les rsidus", 0 );
    GRF_efface(modele->modele, VRAI, VRAI);
    return -6;
  }

  /* Initialisation des X */
  memcpy( modele->modele->x, modele->reference->x, sizeof(double) * (modele->reference->nombre_point) );
  modele->modele->x_min = modele->reference->x_min;
  modele->modele->x_max = modele->reference->x_max;

  memcpy( modele->residu->x, modele->reference->x, sizeof(double) * (modele->reference->nombre_point) );
  modele->residu->x_min = modele->reference->x_min;
  modele->residu->x_max = modele->reference->x_max;

  /* Pour les tapes,... */
  modele->nbr_variables = 0;
  modele->iteration = 0;

  /* On protge les graphes... */
  GRF_ProtegerGraphe( modele->experience );
  GRF_ProtegerGraphe( modele->modele );
  GRF_ProtegerGraphe( modele->residu );

  /* Tout s'est bien pass */
  modele->drapeaux.preparation = VRAI; /* On va continuer  prparer l'ajustement */
#ifdef _DEVELOPPE_
  printf( "Sortie normale de ==> %s <== [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif
  return 0;
}

/* -------- Enlever le spectre exprimental -------- */

int MODELE_OterReference(ModeleAjustement *modele, boolean eff_modele)
{
  if (NULL == modele) return -1;

  if ( modele->reference != modele->experience ) {
    GRF_eff_isole(modele->reference); /* Si la rfrence existe, c'est un graphe isol */
    modele->reference = NULL;
  }

  modele->experience = NULL;

  if (VRAI == eff_modele) { /* On efface l'ancien modle */
    GRF_efface(modele->modele, VRAI, VRAI);
    GRF_efface(modele->residu, VRAI, VRAI);
  }

  modele->modele = NULL;
  modele->residu = NULL;

  return 0;
}
