/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Gestion des variables de l'ajustement
   Les variables sont organises en table, elles sont accessibles de l'extrieur par le numro dans cette table.
   Ordre des variables : dpend du modle ; pour une CL de modles : celles du modle 1, puis du modle 2, puis...
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "couches.h"		/* Pour NC_... */
#include "decomposition.h"	/* Modles de dcomposition linaire sur une base de spectres */
#include "feff.h"		/* Modles de type FEFF */
#include "modeles.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned int MODELE_CompterNC           (Modele_FEFF *modele);
static unsigned int MODELE_CompterDecomposition(ModeleAjustement *modele);

static int Modele_PGInitVariables(ModeleAjustement *modele);

static Variable *MODELE_QuelleVariable(ModeleAjustement *modele, unsigned int numero_variable);

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Compter le nombre de variables dans un ajustement -------- */

unsigned int MODELE_CompterVariables(ModeleAjustement *modele)
{
  unsigned int nbr_variables = 0, i;

  if ( NULL == modele ) return 0;

  switch ( modele->type )
  {
  case MODELE_MULTI_MODELES:
    if ( NULL == modele->infos_modele.modeles )
      return 0;

    for ( i = 0 ; i < modele->infos_modele.modeles->nbr_modeles ; i++ )
      nbr_variables += MODELE_CompterVariables( modele->infos_modele.modeles->modeles[i] );
    break;

  case MODELE_CL_MODELES:
    if ( NULL == modele->infos_modele.cl_modeles )
      return 0;

    for ( i = 0 ; i < modele->infos_modele.cl_modeles->nbr_termes ; i++ ) {
      nbr_variables += MODELE_CompterVariables( modele->infos_modele.cl_modeles->modeles[i] );
      if ( VAR_LIBRE == modele->infos_modele.cl_modeles->coefs[i].type )
	nbr_variables++; /* Le coefficient dans la CL est aussi variable */
    }
    break;

  case MODELE_EXAFS_CHEMINS:
    nbr_variables = MODELE_CompterNC( modele->infos_modele.feff );
    break;

  case MODELE_EXAFS_GNXAS:
  case MODELE_SEUIL_BASIQUE:
  case MODELE_FLUOX:
    (void) fprintf( stderr, "%s, %s l.%u : type de modle pas encore support [%d]\n",
		    __FILE__, __FUNCTION__, __LINE__, modele->type );
    break;

  case MODELE_DECOMPOSITION:
    nbr_variables = MODELE_CompterDecomposition( modele );
    break;

  default:
    (void) fprintf( stderr, "%s : type de modle inconnu [%d]\n",
		    __FUNCTION__, modele->type );
    return modele->nbr_variables; /* Suppose que c'est fourni par l'utilisateur... */
  }

  modele->nbr_variables = nbr_variables; /* Nombre de variables dans le modle considr */
  return nbr_variables;
}

/* -------- Dcompte pour un modle de type N couches --------
   Suppose que le modle a dj t contrl : toute variable ajustable
   est effectivement ajustable !
 */

static unsigned int MODELE_CompterNC(Modele_FEFF *modele)
{
  unsigned int nbr_variables = 0;
  Couche *couche;

  if ( NULL == modele ) return 0;

  /* 1. Variables gnrales */
  if ( VAR_LIBRE == modele->S0_2.type ) nbr_variables++;
  if ( ( VAR_LIBRE == modele->Delta_E.type ) && ( VRAI == modele->drapeaux.DE0_general ) ) nbr_variables++;
  if ( ( VAR_LIBRE == modele->Gamma.type   ) && ( VRAI == modele->drapeaux.GAM_general ) ) nbr_variables++;

  /* 2. Contrle couche par couche */
  for ( couche = modele->couches ; couche != NULL ; couche = couche->suivante ) {
    if ( FAUX == couche->utilisee ) continue;

    if ( VAR_LIBRE == couche->N.type ) nbr_variables++;
    if ( VAR_LIBRE == couche->R.type ) nbr_variables++;
    if ( VAR_LIBRE == couche->Debye_Waller.type ) nbr_variables++;
    if ( VAR_LIBRE == couche->Cumul_1.type ) nbr_variables++;
    if ( VAR_LIBRE == couche->Cumul_2.type ) nbr_variables++;
    if ( VAR_LIBRE == couche->Cumul_3.type ) nbr_variables++;
    if ( VAR_LIBRE == couche->Cumul_4.type ) nbr_variables++;

    if ( ( FAUX == modele->drapeaux.DE0_general ) && ( VAR_LIBRE == couche->Delta_E.type ) ) nbr_variables++;
    if ( ( FAUX == modele->drapeaux.GAM_general ) && ( VAR_LIBRE == couche->Gamma.type   ) ) nbr_variables++;
  }

  /* Dcompte termin */
  return nbr_variables;
}

/* ------ Dcompte pour un modle de dcomposition -------- */

static unsigned int MODELE_CompterDecomposition(ModeleAjustement *modele)
{
  struct decomposition *decomposition;
  struct element_base *base;
  unsigned long i, nbr_variables;

  decomposition = modele->infos_modele.decomposition;
  if ( ( NULL == decomposition->base ) || ( 0 == decomposition->nombre_bases ) ) {
    /* Aucune fonction de base indique ! */
    return 0;
  }

  /* A priori, chaque fonction de base conduit  une variable */
  nbr_variables = decomposition->nombre_bases;

  /* On contrle tout de mme que l'on peut l'utiliser */
  for ( i = 0, base = decomposition->base ; i < decomposition->nombre_bases ; i++, base++ ) {
    if ( NULL == base->graphe ) {
      /* Probablement un problme, en tout cas inutilisable ! */
      nbr_variables--;
      continue;
    }

    if ( VAR_CONSTANTE == base->coefficient.type ) {
      /* On a fix la contribution de cet lment */
      nbr_variables--;
      continue;
    }

    if ( ( base->graphe->x_min > modele->x_max ) ||
	 ( base->graphe->x_max < modele->x_min ) ) {
      /* Ce graphe est dfini sur une rgion en X hors modle : inutile de le considrer ! */
      nbr_variables--;
      continue;
    }
  }

  /* On renvoie le nombre de variables rellement utilises */
#ifdef _DEVELOPPE_
  (void) printf( "%s,%s l.%u : Nombre de variable dtect = %lu.\n",
		 __FILE__, __FUNCTION__, __LINE__, nbr_variables );

#endif
  return nbr_variables;
}

/* ~-~-~-~-~-~-~-~- Initialisations des variables d'ajustement ~-~-~-~-~-~-~-~- */

/* -------- Initialisation des variables pour l'ajustement -------- */

int MODELE_InitVariables(ModeleAjustement *modele)
{
  int retour;

  if ( NULL == modele ) return -1;

  if ( VRAI == modele->drapeaux.en_cours ) {
    (void) fprintf( stderr, "%s,%s l.%u : ajustement en cours sur ce modle !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    return 1;
  }

  /* On compte le nombre de variables dans le modle
     Remarque : a pu dj tre fait, mais n'est sans doute pas limitant en temps... et plus sr !
   */
  modele->nbr_variables = MODELE_CompterVariables( modele );
  if ( 0 == modele->nbr_variables ) {
    (void) fprintf( stderr, "%s : aucune variable dans le modle\n", __FUNCTION__);
    return -2;
  }

  if ( NULL != modele->infos_variables.pseudogradient ) {
    /* On efface les anciennes infos, car algorithme ou autre facteur a pu changer entretemps */
    free( modele->infos_variables.pseudogradient );
    modele->infos_variables.pseudogradient = NULL;
  }

  switch ( modele->algorithme )
  {
  case ALGO_PSEUDOGRADIENT:
    /* Infos pour chacune des variable ajuste */
    modele->infos_variables.pseudogradient = (PG_variable *) malloc( sizeof(PG_variable) * modele->nbr_variables );
    if ( NULL == modele->infos_variables.pseudogradient ) {
      ERR_Memoire( "Table des variables pour l'algorithme du pseudo-gradient", sizeof(PG_variable) * modele->nbr_variables );
      return -3;
    }
    /* On initialise ces variables */
    retour = Modele_PGInitVariables( modele );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant les variables / Pseudo-gradient [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return retour;
    }
    break;

  default:
    (void) fprintf( stderr, "%s,%s l.%u : algorithme inconnu (%d)\n",
		    __FILE__, __FUNCTION__, __LINE__, modele->algorithme );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -10;
  }

  modele->drapeaux.en_cours = VRAI; /* Une fois les variables prtes, l'ajustement est comme commenc ! */
  modele->drapeaux.preparation = FAUX; /* et la prparation est finie */
#ifdef _DEVELOPPE_
  printf( "Sortie normale de ==> %s <== [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif
  return 0;
}

/* -------- Initialisation des variables, pseudo-gradient -------- */

static int Modele_PGInitVariables(ModeleAjustement *modele)
{
  if ( NULL == modele ) return -1;

  switch ( modele->type )
  {
  case MODELE_CL_MODELES:
    break;

  case MODELE_EXAFS_CHEMINS:
    return NC_PGInitVariables( modele, 0 );

  case MODELE_EXAFS_GNXAS:
  case MODELE_SEUIL_BASIQUE:
  case MODELE_FLUOX:
    AVERTISSEMENT( "Type de modle pas encore gr (%d)", modele->type );
    break;

  case MODELE_DECOMPOSITION:
    /* Pas d'initialisation ncessaire : rsolution directe exacte puisque linaire... */
    break;

  case MODELE_GENERIQUE:
    return MODELE_Generique_PGInitVariables( modele, 0 );

  default:
    AVERTISSEMENT( "Type de modle inconnu (%d)", modele->type );
    return -2;
  }

  return 0;
}

/* -~-~-~-~-~-~-~-~ Manipulations des valeurs des variables -~-~-~-~-~-~-~-~ */

/* -------- Trouver la variable numero i d'un modele -------- */

static Variable *MODELE_QuelleVariable(ModeleAjustement *modele, unsigned int numero_variable)
{
  Variable *variable = NULL;

  /* Controles... */
  if ( NULL == modele ) return NULL;
  if ( NULL == modele->infos_variables.pseudogradient ) {
    (void) fprintf(stderr, "%s,%s l.%u : pas de variable dans ce modle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if ( numero_variable >= modele->nbr_variables ) {
    (void) fprintf(stderr, "%s : il n'y a que %u variables dans ce modle [%u demande] !\n",
		   __FUNCTION__, modele->nbr_variables, numero_variable);
    return NULL;
  }

  /* On la trouve, en fonction de l'algorithme */
  switch ( modele->algorithme )
  {
  case ALGO_PSEUDOGRADIENT:
    {
      struct pg_variable *var_pg;

      var_pg = modele->infos_variables.pseudogradient;
      variable = var_pg[ numero_variable ].var;
    }
    break;

  default :
    AVERTISSEMENT( "Algorithme inconnu ou pas encore support [%u]", modele->algorithme );
    break;
  }

  if ( NULL == variable ) {
    (void) fprintf( stderr, "%s,%s l.%u : variable NULL dans l'ajustement !\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  /* On renvoie cette variable [pointeur => contenu modifiable !] */
  return variable;
}

/* -------- Trouver les infos associes a la variable i du modele -------- */

char *MODELE_InfosVariable(ModeleAjustement *modele, unsigned int numero_variable)
{
  char *nom = NULL;
  int type = -1;

  if (NULL == modele) return NULL;
  if (NULL == modele->infos_variables.pseudogradient ) {
    (void) fprintf(stderr, "%s,%s l.%u : pas de variable dans ce modle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if ( numero_variable >= modele->nbr_variables ) {
    (void) fprintf(stderr, "%s,%s l.%u : variable %u demande, le modle n'en a que %u [donc 0  %u] !\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_variable, modele->nbr_variables, modele->nbr_variables - 1);
    return NULL;
  }

  switch ( modele->algorithme )
  {
  case ALGO_PSEUDOGRADIENT:
    {
      struct pg_variable *var_pg;

      var_pg = modele->infos_variables.pseudogradient;
      type = var_pg[ numero_variable ].type;
      switch ( modele->type )
      {
      case MODELE_EXAFS_CHEMINS:
	nom = NC_InfosVariable( type, var_pg[ numero_variable ].origine.couche );
	break;

      case MODELE_GENERIQUE:
	if ( NULL == var_pg[ numero_variable ].origine.parametre->symbole ) {
	  nom = strdup( "Sans symbole" );
	} else {
	  nom = strdup( var_pg[ numero_variable ].origine.parametre->symbole );
	}
	break;
      }
    }
    break;

  default :
    (void) fprintf(stderr, "%s,%s l.%u : algorithme inconnu ou pas encore support [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, modele->algorithme);
    break;
  }

  if ( 0 > type ) {
    (void) fprintf( stderr, "%s : variable de type inconnu dans l'ajustement, NE DEVRAIT JAMAIS ARRIVER !\n",
		    __FUNCTION__ );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return strdup("Type inconnu");
  }

  return nom;
}

/* -------- Donne l'unit de la variable indique ------------ */

char *MODELE_UniteVariable(ModeleAjustement *modele, unsigned int numero_variable)
{
  int type = -1;
  char *unite = NULL;

  if (NULL == modele) return NULL;
  if (NULL == modele->infos_variables.pseudogradient ) {
    (void) fprintf(stderr, "%s,%s l.%u : pas de variable dans ce modle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if ( numero_variable >= modele->nbr_variables ) {
    (void) fprintf(stderr, "%s,%s l.%u : variable %u demande, le modle n'en a que %u [donc 0  %u] !\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_variable, modele->nbr_variables, modele->nbr_variables - 1);
    return NULL;
  }

  switch ( modele->algorithme )
  {
  case ALGO_PSEUDOGRADIENT:
    {
      struct pg_variable *var_pg;

      var_pg = modele->infos_variables.pseudogradient;
      switch ( modele->type )
      {
      case MODELE_EXAFS_CHEMINS:
	type = var_pg[ numero_variable ].type;
	unite = NC_UniteVariable( type, var_pg[ numero_variable ].origine.couche );
	break;

      case MODELE_GENERIQUE:
	if ( NULL == var_pg[ numero_variable ].origine.parametre->symbole ) {
	  unite = NULL;		/* Pas d'unit dfinie */
	} else {
	  unite = strdup( var_pg[ numero_variable ].origine.parametre->unite );
	}
	break;
      }
    }
    break;

  default :
    (void) fprintf( stderr, "%s,%s l.%u : algorithme inconnu ou pas encore support [%u]\n",
		    __FILE__, __FUNCTION__, __LINE__, modele->algorithme );
    break;
  }

  /* On a bien trouv une unit ; on la renvoie [ LIBRER AVEC FREE()] */
  return unite;
}

/* -------- Donne la valeur dans la variable indique --------
 * modele          = le modele a utiliser
 * numero_variable = le numero de la variable
 * type            = le type de valeur a lire [VARIABLE_LIRE_VALEUR_xxx, cf. c_formel.h]
 */

double MODELE_ValeurVariable(ModeleAjustement *modele, unsigned int numero_variable, unsigned int type)
{
  Variable *variable = NULL;

  variable = MODELE_QuelleVariable( modele, numero_variable );
  if ( NULL == variable ) {
    (void) fprintf( stderr, "%s : impossible de trouver la variable %u dans ce modle !\n",
		    __FUNCTION__, numero_variable );
    return 0.0;
  }

  return VAR_LireValeur( variable, type );
}

/* -------- Fixe l'une des valeurs de la variable indiquee -------- */

int MODELE_FixerValeurVariable(ModeleAjustement *modele, unsigned int numero_variable, unsigned int type, double nouvelle_valeur)
{
  Variable *variable = NULL;

  /* On trouve la variable correspondante */
  variable = MODELE_QuelleVariable( modele, numero_variable );
  if ( NULL == variable ) {
    (void) fprintf( stderr, "%s : impossible de trouver la variable %u dans ce modle !\n",
		    __FUNCTION__, numero_variable );
    return -1;
  }

  return VAR_FixerValeur( variable, type, nouvelle_valeur );
}

/* -------- Remet la variable indique  sa valeur initiale -------- */

int MODELE_ReinitialiseVariable(ModeleAjustement *modele, unsigned int numero_variable)
{
  Variable *variable = NULL;

  /* On trouve la variable correspondante */
  variable = MODELE_QuelleVariable( modele, numero_variable );
  if ( NULL == variable ) {
    (void) fprintf( stderr, "%s : impossible de trouver la variable %u dans ce modle !\n",
		    __FUNCTION__, numero_variable );
    return -1;
  }

  if ( VAR_LIBRE == variable->type ) { /* Sinon, pas de sens ! */
    variable->contenu.valeur = variable->valeur_initiale;
  }

  /* Tout s'est bien pass */
  return 0;
}
