/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1997-1999
   Création : 27 novembre 1997)

   Troncature d'un graphe, en conservant le plus d'informations possible

  6, 12 janvier ; 16 & 17, 25, 31 mars ; 23 avril 1998
   * Adaptation à la nouvelle gestion des réticules
   * Quelques modifs pour stabiliser sous X
   * Affichage des bornes après modification dans GRF_tronque() et autres...
   * Inclus les messages d'erreur exafs (erreurs.h)

  22 & 23, 25, 27 novembre ; 2 & 3, 9, 15 décembre 1997
   * Troncature d'une TF : copie les infos de TF dans le graphe tronqué
   * Corrigé MSG_tronque() : mauvais arbre utilisé pour redessin du pointeur
   * Troncature des graphes : création des fonctions
            GRF_tronque() = ouverture de la fenêtre d'interface
	    RDS_tronque() = redessin de cette fenêtre
	    CLC_tronque() = clic sur cette fenêtre
	    MSG_tronque() = messages pour cette fenêtre
	    TRQ_calc()    = réalisation effective de la troncature
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "systeme.h"

#include "exafs.h"
#include "filtrage.h" /* Pour garder les infos de TF en cas de troncature */
#include "interface.h"
#include "operations.h"
#include "serie.h"
#include "tracer.h"

#include "op_tronquer.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static unsigned long TRQ_CompterPoints(Graphe *graphe, TRQ_infos *infos);

static int TRQ_ReduireX(Graphe *graphe, TRQ_infos *infos);
static int TRQ_Reduire (Graphe *graphe, TRQ_infos *infos, unsigned long nbr_points);
static int TRQ_Isoler  (Graphe *graphe, TRQ_infos *infos, unsigned long nbr_points);

static boolean TRQ_GarderPoint(TRQ_infos *infos, double x, double y);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Interface visible de l'extérieur ------------------\
 |  Renvoie 0 si tout s'est bien passé, !=0 en cas de problème |
 |                                                             |
 | graphe     : le graphe que l'on veut tronquer (en X ou en Y)|
 | infos      : les informations pour effectuer la troncature  |
 | en_place   : VRAI, on reste dans le même graphe             |
 | historique : VRAI, on sauve dans l'historique le protocole  |
 \------------------------------------------------------------ */

int TRQ_Tronquer(Graphe *graphe, TRQ_infos *infos, boolean en_place, boolean historique)
{
  unsigned long nbr_points = 0;

  if (NULL == graphe) return -1;
  if (NULL == infos) {
    (void) fprintf(stderr, "TRQ_Tronquer : pas d'information de troncature !\n");
    return -2;
  }

  if (VRAI == historique) {
    HIST_Titre(1, "Tronquer un graphe", "TRQ");
    HIST_Texte(1, 0, "Graphe original : %s\n", graphe->nom);

    if ( AXE_X & infos->axes ) {
      HIST_Texte(3, 0, "R"e_AIGU"gion conserv"e_AIGU"e en X : ");
      if ( TRQ_GARDER_INTERIEUR == infos->mode_x )
	HIST_Texte(3, 0, "[%g, %g]\n", infos->x_min, infos->x_max);
      else
	HIST_Texte(3, 0, "]-infini, %g] U [%g, +infini[\n\n", infos->x_min, infos->x_max);
    }
    if ( AXE_Y & infos->axes ) {
      HIST_Texte(3, 0, "R"e_AIGU"gion conserv"e_AIGU"e en Y : ");
      if ( TRQ_GARDER_INTERIEUR == infos->mode_y )
	HIST_Texte(3, 0, "[%g, %g]\n", infos->y_min, infos->y_max);
      else
	HIST_Texte(3, 0, "]-infini, %g] U [%g, +infini[\n\n", infos->y_min, infos->y_max);
    }

    if (VRAI == en_place)
      HIST_Texte(3, 0, "... troncature en place\n");
  }

  /* Un cas particulier fréquent : on ne travaille que sur les X */
  if ( AXE_X == infos->axes ) {
    if ( VRAI == en_place ) {
      return TRQ_ReduireX(graphe, infos);
    } else { /* On extrait un nouveau graphe : fonction déjà faite */
      Graphe *resultat;

      resultat = GRF_ExtraitGrapheI(graphe, infos->x_min, infos->x_max, infos->mode_x == TRQ_GARDER_EXTERIEUR, FAUX);
      if ( VRAI == historique ) {
	if (NULL == resultat)
	  HIST_Erreur(1, "Calcul annul"e_AIGU, NULL);
	else
	  HIST_Texte(1, 0, "Graphe tronqu"e_AIGU"  : %s\n", resultat->nom);
      }
    }

    /* Tout s'est bien passé ici */
    return 0;
  }

  /* Si on est là, c'est que l'on travaille aussi avec les Y */

  /* Vérification préliminaire */
  if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "TRQ_Tronquer : troncature en Y impossible pour un graphe à valeurs complexes [%s]\n", graphe->nom);
#endif
    (void) BI_ERR_alerte(NULL, &ERR_GrapheComplexe, VRAI);
    return -4;
  }

  /* On compte le nombre de points conservés finalement */
  nbr_points = TRQ_CompterPoints(graphe, infos);
  if (nbr_points < 2) {
    (void) fprintf(stderr, "TRQ_Tronquer : trop peu de points conservés avec cette troncature (2 min, %lu trouvés) !\n",
		   nbr_points);
    return -3;
  }
  if ( nbr_points == graphe->nombre_point ) {
    (void) fprintf(stderr, "TRQ_Tronquer : tous les points sont conservés, troncature inutile !\n");
    return +1;
  }

  /* On choisit quelle fonction on appelle */
  if ( VRAI == en_place ) {
    TRQ_Reduire( graphe, infos, nbr_points );
  } else {
    TRQ_Isoler( graphe, infos, nbr_points );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Décompte des points valables -------- */

static unsigned long TRQ_CompterPoints(Graphe *graphe, TRQ_infos *infos)
{
  unsigned long i, nbr_pts = 0;
  double *x, *y;

  /* Remarque : travail sur l'axe X au moyen des fonctions dans Graphes/extrait.c : non repris ici */
  if ( AXE_Y == infos->axes ) { /* Seul l'axe Y sert à définir les points corrects */
    if ( TRQ_GARDER_INTERIEUR == infos->mode_y ) {
      /* Vérifications sommaires */
      if ( infos->y_min > graphe->y_max ) return 0;
      if ( infos->y_max < graphe->y_min ) return 0;
      if ( (infos->y_min < graphe->y_min) && (infos->y_max > graphe->y_max) )
	return graphe->nombre_point;

      /* Contrôle précis, si nécessaire */
      for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
	if ( (*y >= infos->y_min) && (*y <= infos->y_max) )
	  nbr_pts++;
      }
    } else {
      /* Vérifications sommaires */
      if ( infos->y_min > graphe->y_max ) return graphe->nombre_point;
      if ( infos->y_max < graphe->y_min ) return graphe->nombre_point;
      if ( (infos->y_min < graphe->y_min) && (infos->y_max > graphe->y_max) )
	return 0;

      /* Contrôle précis, si nécessaire */
      for (i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++) {
	if ( (*y < infos->y_min) || (*y > infos->y_max) )
	  nbr_pts++;
      }
    }

    return nbr_pts;
  }

  for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
    if ( VRAI == TRQ_GarderPoint(infos, *x, *y)) {
      nbr_pts++;
    }
  }

  /* Décompte terminé */
  return nbr_pts;
}

/* -~-~-~-~-~-~-~-~ Troncature suivant l'axe des X uniquement -~-~-~-~-~-~-~-~ 
   Note : la versions avec création d'un nouveau graphe est dans Graphes/extrait.c */

static int TRQ_ReduireX(Graphe *graphe, TRQ_infos *infos)
{
  double *nv_x = NULL, *nv_y = NULL;
  unsigned long nombre_points;

  nombre_points = GRF_extrait_table(graphe, infos->x_min, infos->x_max, infos->mode_x == TRQ_GARDER_EXTERIEUR, &nv_x, &nv_y);
  if ( nombre_points < 2 ) {
    (void) fprintf(stderr, "TRQ_ReduireX : trop peu de points conservés avec cette troncature (2 min, %lu trouvés) !\n",
		   nombre_points);
    free(nv_x);
    free(nv_y);
    return -3;
  }
  /* On libère les anciennes tables du graphe et on met les nouvelles */
  free(graphe->x); graphe->x = nv_x;
  free(graphe->y); graphe->y = nv_y;

  if ( GR_DEVIATION == graphe->type ) {
    printf("TRQ_ReduireX : À FAIRE, adaptation des degrés de liberté\n");
  }

  graphe->nombre_point = nombre_points; /* Nouveau nombre de points */

  /* Les barres d'erreur */
  if ( graphe->erreur_x != NULL ) {
    printf("TRQ_ReduireX : À FAIRE, adaptation des erreurs en X\n");
  }
  if ( graphe->erreur_y != NULL ) {
    printf("TRQ_ReduireX : À FAIRE, adaptation des erreurs en Y\n");
  }

  /* La matrice de covariance */
  if ( graphe->covariance != NULL ) {
    printf("TRQ_ReduireX : À FAIRE, adaptation de la matrice de covariance\n");
  }

  GRF_bornes(graphe, DESORDONNE, DESORDONNE);

  /* Tout s'est bien passé */
  return 0;
}

/* -~-~-~-~-~-~-~-~ Troncature faisant intervenir au moins l'axe des Y -~-~-~-~-~-~-~-~
   On est donc sûr que le graphe contient des valeurs réelles en Y
 */

/* -------- Cas avec création d'un nouveau graphe -------- */

static int TRQ_Isoler(Graphe *graphe, TRQ_infos *infos, unsigned long nbr_points)
{
  Graphe *troncature;
  char *titre;
  double *x_grf, *y_grf;
  unsigned long i, j;

  /* On crée le graphe contenant le résultat */
  titre = TXT_genere_chaine(nom_graphe_tronque, graphe->nom);
  troncature = GRF_nouveau(graphe, nbr_points, graphe->type, graphe->genre_donnees, titre);
  free( titre ); titre = NULL;
  if ( NULL == troncature ) {
    (void) fprintf(stderr, "TRQ_Isoler : ERREUR dans GRF_nouveau()\n");
    return -2;
  }

  /* On recopie la partie intéressante */
  for ( i = 0, j = 0, x_grf = graphe->x, y_grf = graphe->y ;
	i < graphe->nombre_point ; i++, x_grf++, y_grf++ ) {
    if ( VRAI == TRQ_GarderPoint( infos, *x_grf, *y_grf) ) {
      troncature->x[ j ] = *x_grf;
      troncature->y[ j ] = *y_grf;

      if ( GR_DEVIATION == graphe->type ) {
	troncature->z.ddl[ j ] = graphe->z.ddl[ i ];
      }

      j++;
    } else {
      printf("%s : point rejeté, X=%g, Y=%g\n", __FUNCTION__, *x_grf, *y_grf );
    }
  }

  /* On adapte les bornes */
  GRF_bornes( troncature, DESORDONNE, DESORDONNE );

  /* S'il y a des erreurs, on les traite
     On ne peut pas faire un appel récursif, car les Y n'ont aucune raison d'être les mêmes pour les erreurs !
  */
  if ( graphe->erreur_x ) {
    (void) fprintf(stderr, "TRQ_Isoler : À FAIRE, traitement des erreurs en X\n");
  }
  if ( graphe->erreur_y ) {
    (void) fprintf(stderr, "TRQ_Isoler : À FAIRE, traitement des erreurs en Y\n");
  }

  /* Et la matrice de covariance... */
  if ( graphe->covariance != NULL ) {
    (void) fprintf(stderr, "TRQ_Isoler : À FAIRE, traitement de la matrice de covariance\n");
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Cas en place -------- */

static int TRQ_Reduire(Graphe *graphe, TRQ_infos *infos, unsigned long nbr_points)
{
  double *nv_x, *nv_y, *x_grf, *y_grf;
  unsigned long i, j;

  /* On réserve les tables de résultat */
  nv_x = (double *) malloc( sizeof( double ) * nbr_points );
  if ( NULL == nv_x ) {
    (void) fprintf(stderr, "TRQ_Reduire : mémoire insuffisante, table des X\n");
    (void) SYS_ERREUR( MANQUE_MEMOIRE, "tronquer.c", "TRQ_Reduire", "nv_x" );
    return -1;
  }
  nv_y = (double *) malloc( sizeof( double ) * nbr_points );
  if ( NULL == nv_y ) {
    (void) fprintf(stderr, "TRQ_Reduire : mémoire insuffisante, table des Y\n");
    (void) SYS_ERREUR( MANQUE_MEMOIRE, "tronquer.c", "TRQ_Reduire", "nv_y" );
    free( nv_x );
    return -1;
  }

  /* On les remplit avec les valeurs conservées */
  /* On recopie la partie intéressante */
  for ( i = 0, j = 0, x_grf = graphe->x, y_grf = graphe->y ;
	i < graphe->nombre_point ; i++, x_grf++, y_grf++ ) {
    if ( VRAI == TRQ_GarderPoint( infos, *x_grf, *y_grf) ) {
      nv_x[ j ] = *x_grf;
      nv_y[ j ] = *y_grf;

      j++;
    }
  }

  /* Faire la même chose avec les données complémentaires */
  if ( GR_DEVIATION == graphe->type ) {
    printf("TRQ_Reduire : À FAIRE, conservation des ddl\n");
  }

  /* On permute les tables */
  x_grf = graphe->x; graphe->x = nv_x; nv_x = x_grf;
  y_grf = graphe->y; graphe->y = nv_y; nv_y = y_grf;

  /* On traite les erreurs */
  if ( graphe->erreur_x ) {
    (void) fprintf(stderr, "TRQ_Reduire : À FAIRE, traitement des erreurs en X\n");
  }
  if ( graphe->erreur_y ) {
    (void) fprintf(stderr, "TRQ_Reduire : À FAIRE, traitement des erreurs en Y\n");
  }

  /* Et la matrice de covariance... */
  if ( graphe->covariance != NULL ) {
    (void) fprintf(stderr, "TRQ_Reduire : À FAIRE, traitement de la matrice de covariance\n");
  }

  /* On termine proprement */
  graphe->nombre_point = nbr_points;
  free( nv_x );
  free( nv_y );

  /* Tout s'est bien passé */
  GRF_bornes( graphe, DESORDONNE, DESORDONNE );
  return 0;
}

/* ------ Renvoie VRAI si on conserve le point indiqué, FAUX sinon ------ */

static boolean TRQ_GarderPoint(TRQ_infos *infos, double x, double y)
{
  if ( NULL == infos ) return FAUX;

  printf("%s : point à tester = (%g,%g) [Xmin=%g, Xmax=%g, mode=%d; Ymin=%g, Ymax=%g, mode=%d]",
	 __FUNCTION__, x, y, infos->x_min, infos->x_max, infos->mode_x, infos->y_min, infos->y_max, infos->mode_y );

  /* On contrôle d'abord, si nécessaire, les X */
  if ( infos->axes & AXE_X ) {
    if ( TRQ_GARDER_INTERIEUR == infos->mode_x ) { /* Acceptation : x_min <= X <= x_max */
      if ( x < infos->x_min ) {
	printf(" : X < X_min\n");
	return FAUX;
      }
      if ( x > infos->x_max ) {
	printf(" : X > X_max\n");
	return FAUX;
      }
    } else {		    /* Acceptation : X < x_min ou X > x_max */
      if ( (x >= infos->x_min) && (x <= infos->x_max) ) {
	printf(" : X_min <= X <= X_max\n");
	return FAUX;
      }
    }
  }

  /* On contrôle ensuite, si nécessaire, les Y */
  if ( infos->axes & AXE_Y ) {
    if ( TRQ_GARDER_INTERIEUR == infos->mode_y ) { /* Acceptation : y_min <= Y <= y_max */
      if ( y < infos->y_min ) {
	printf(" : Y < Y_min\n");
	return FAUX;
      }
      if ( y > infos->y_max ) {
	printf(" : Y > Y_max\n");
	return FAUX;
      }
    } else { /* Acceptation : Y < y_min ou Y > y_max */
      if ( (y >= infos->y_min) && (y <= infos->y_max) ) {
	printf(" : Y_min <= Y <= Y_max\n");
	return FAUX;
      }
    }
  }

  /* Si on atteint ça, c'est que le point est dans la bonne région */
  return VRAI;
}
