/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2006, Emmanuel CURIS
   Création : 3 aoûtt 2006

   Analyses statistiques

   Comparaisons multiples [de moyennes]
    (pour les comparaisons fonctionnelles de graphes : cf. comparer_graphes.c)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "graphes.h"
#include "proba.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Réalisation pratique ~~~~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Gestion de la structure de données ~-~-~-~-~-~-~-~-~ */

/* -------- Créer une table de comparaisons, vide, et l'initialiser -------- */

struct comparaison *STAT_CreerComparaisons(unsigned int nombre_comparaisons)
{
  struct comparaison *comparaisons;
  unsigned long i;

  if ( 0 == nombre_comparaisons ) return NULL;

  /* On réserve la mémoire [stockées sour forme de tableau] */
  comparaisons = (struct comparaison *) malloc( sizeof( struct comparaison ) * nombre_comparaisons );
  if ( NULL == comparaisons ) {
    ERR_Memoire( "Réservation des comparaisons", sizeof( struct comparaison ) * nombre_comparaisons );
    return NULL;
  }

  /* On initialise */
  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    comparaisons[ i ].methode  = STAT_IC_FRACTILE_EXACT; /* On préfère les calculs exacts en général... */
    comparaisons[ i ].drapeaux = 0; /* Rien de particulier en général */
    comparaisons[ i ].sens     = 0; /* Bilatéral en général */

    comparaisons[ i ].modalites[ 0 ] = NULL;  /* Aucune modalité associée pour l'instant */
    comparaisons[ i ].modalites[ 1 ] = NULL;

    comparaisons[ i ].difference = 0.0;
    comparaisons[ i ].variance_difference = 0.0;
    comparaisons[ i ].fractile = 0.0;
    comparaisons[ i ].ic_bas   = 0.0;
    comparaisons[ i ].ic_haut  = 0.0;
  }

  /* Terminé, sans souci */
  return comparaisons;
}

/* -------- Effacer une table de comparaisons multiples -------- */

void STAT_EffacerComparaisons(unsigned int nombre_comparaisons, struct comparaison *comparaisons)
{
  unsigned int i;

  if ( NULL == comparaisons ) {
    if ( nombre_comparaisons > 0 ) {
      AVERTISSEMENT( "Incohérence dans les paramètres [NULL pour %u comparaisons]", nombre_comparaisons );
    }
    return;
  }

  /* On libère pour chaque comparaison */
  for ( i = 0 ; i < nombre_comparaisons ; i++ ) {
    /* On supprime le lien aux modalités */
    comparaisons[ i ].modalites[ 0 ] = NULL;
    comparaisons[ i ].modalites[ 1 ] = NULL;

    /* Rien d'autre à faire ! */
  }

  /* On libère la table */
  free( comparaisons ); comparaisons = NULL;
}


/* -------- Initialiser des comparaisons multiples à un contrôle/témoin ------- */

struct comparaison *STAT_Init_ComparaisonsAUnTemoin(unsigned int nombre_modalites /* Total */, struct modalite *modalites, unsigned int modalite_temoin, int sens)
{
  struct comparaison *comparaisons;
  unsigned int i, j;

  if ( ( 0 == nombre_modalites ) || ( NULL == modalites ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune modalité pour les comparaisons !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  if ( modalite_temoin >= nombre_modalites ) {
    (void) fprintf(stderr, "%s, %s l.%u : Numéro de modalité témoin absurde : %u, devrait être entre 0 et %u\n",
		   __FILE__, __FUNCTION__, __LINE__, modalite_temoin, nombre_modalites - 1);
  }

  /* On prépare la table des comparaisons */
  comparaisons = STAT_CreerComparaisons( nombre_modalites - 1 );
  if ( NULL == comparaisons ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de créer les %u comparaisons\n",
		   __FILE__, __FUNCTION__, __LINE__, nombre_modalites - 1);
    return NULL;
  }

  /* On initialise en plaçant les modalités */
  for ( i = 0, j = 0 ; i < nombre_modalites ; i++ ) {
    if ( i == modalite_temoin ) continue;  /* Rien à faire pour celle-là ! */

    /* On travaille toujours sur les différences modalite[i] - témoin */
    comparaisons[ j ].modalites[ 0 ] = modalites + i;
    comparaisons[ j ].modalites[ 1 ] = modalites + modalite_temoin;

    /* On indique le sens des comparaisons */
    comparaisons[ j ].sens = sens;
    if ( sens != 0 ) {
      comparaisons[ j ].drapeaux |= STAT_CMP_UNILATERALE;
    }

    /* On fait les calculs, autant que possible */
    comparaisons[ j ].difference = modalites[ i ].moyenne - modalites[ modalite_temoin ].moyenne;

    /* La modalité suivante sera faite */
    j++;
  }

  /* Terminé sans souci */
  return comparaisons;
}

/* ~-~-~-~-~-~-~-~-~ Calculs génériques ~-~-~-~-~-~-~-~-~ */

int STAT_CMP_CalculerIC(struct comparaison *comparaison)
{
  if ( NULL == comparaison ) return -1;

  if ( 0 == ( comparaison->drapeaux & STAT_CMP_AVEC_IC ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Méthode de comparaison ne permettant pas de construire un IC de la différence !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }

  AVERTISSEMENT( "différence = %g, fractile = %g, variance = %g", comparaison->difference, comparaison->fractile, comparaison->variance_difference );

  if ( 0 == comparaison->sens ) {
    /* Test bilatéral */
    comparaison->ic_bas  = comparaison->difference - comparaison->fractile * comparaison->variance_difference;
    comparaison->ic_haut = comparaison->difference + comparaison->fractile * comparaison->variance_difference;
  } else if ( comparaison->sens > 0 ) {
    /* Unilatéral, différence positive attendue */
    comparaison->ic_bas  = comparaison->difference - comparaison->fractile * comparaison->variance_difference;
    comparaison->ic_haut = -5.0e100;		/* Arbitraire, de toute façon devrait être ignoré car IC [d_min, +infini[ ; si d_min > 0 alors diff. sig. */
  } else {
    /* Unilatéral, différence négative attendue */
    comparaison->ic_bas  = -5.0e100;		/* Arbitraire, de toute façon devrait être ignoré car IC ]-infini, d_max] ; si d_max < 0 alors diff. sig. */
    comparaison->ic_haut = comparaison->difference + comparaison->fractile * comparaison->variance_difference;
  }

  /* On a terminé */
  return 0;
}
