/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1996 - 2002
   Date de cration      : 13 mars 1998

   Toutes les fonctions de traitement statistique de l'E.X.A.F.S.

   Calcul des carts-types  partir de plusieurs spectres
                                       (  Somme_k ( f_moyen - f_k )^2    )
   FORMULE : cart_type = racine_carre( ------------------------------- )
                                       (       N_chantillons - 1        )
   Les graphe de dispersion ont la forme (X = valeur moyenne de A, Y = cart-type de A, ddl)

   ATTENTION : toutes les routines ici travaillent implicitement
               sur des graphes de type REL (soit GR_NORMAL)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "serie.h"

#include "interface.h"
#include "lissage.h"
#include "statistiques.h"
#include "tri.h"

#include "dispersion.txt"

/* ~~~~~~~~ Variables locales  ce module ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales  ce module ~~~~~~~~ */

static void init_mesures(Graphe *graphe, int nbr_ddl);

static void       DISP_ApresCalculs  (Graphe *moyen, boolean err_moyenne);
static Graphe    *DISP_Fusion        (Graphe **graphes, unsigned long nombre, Graphe *modele);
static boolean    DISP_Lissage       (Graphe *moyen, Graphe **graphes, unsigned long nombre);
static boolean    DISP_Moyenne       (Graphe *moyen, TRI_infos *infos, unsigned long nbr_graphes);
static Graphe    *DISP_MoyennerX     (Graphe **graphes, unsigned long nombre, Graphe *modele, int axes);
static TRI_infos *DISP_PrepareFusion (Graphe **graphes, unsigned long nombre);
static Graphe    *DISP_PrepareLissage(Graphe **graphes, unsigned long nombre, Graphe *modele, int axes, int type);
static boolean    DISP_PrepareModele (Graphe **graphes, unsigned long nombre, Graphe *modele);

static Graphe *disp_graphe_moyen(Graphe *modele, boolean erreurs_x, unsigned long nbr_graphes);

/* ~~~~~~~~~~~~~~~~ Calcul de l'cart-type de spectres E.X.A.F.S. ~~~~~~~~~~~~~~~~~ */

/* -------- Cration du graphe qui contiendra la moyenne -------- */

static Graphe *disp_graphe_moyen(Graphe *modele, boolean erreurs_x, unsigned long nbr_graphes __attribute__((__unused__)))
{
  int axe = AXE_Y; /* Erreurs en Y toujours cres */
  Graphe *moyenne, *erreurs;

  if (NULL == modele) return NULL;

  moyenne = GRF_nv_isole( modele->nombre_point, modele->type, modele->genre_donnees );
  if (NULL == moyenne) return NULL;

  /* Log dans l'historique */
  HIST_Texte(5, 0, "Intervalle utile en X : [%.10g ; %.10g]\n", modele->x_min, modele->x_max);
  HIST_Texte(5, 0, "Nombre de points restant : %lu\n", modele->nombre_point);
  HIST_Texte( 1, 0, txt_hist_nom_grf, moyenne->nom );

  /* Cration du graphe pour les erreurs aussi [le log sera fait aussi] */
  if (VRAI == erreurs_x) /* On veut aussi les erreurs en X */
    axe |= AXE_X;
  erreurs = GRF_creer_erreurs(moyenne, axe);
  if (NULL == erreurs) {
    GRF_efface(moyenne, VRAI, VRAI);
    return NULL;
  }

  return moyenne;
}

/* -------- Remplit z avec le nombre de mesures utilises pour la moyenne --------- */

static void init_mesures(Graphe *graphe, int nbr_ddl)
{
  int *z;
  unsigned long i;

  if (graphe == NULL) return;
  if (graphe->z.ddl == NULL) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de table pour les DDL !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }

#ifdef _DEVELOPPE_
/*   printf( "%s : remplir avec %u ddl\n", __FUNCTION__, nbr_ddl ); */
#endif

  for (i = 0, z = graphe->z.ddl ; i < graphe->nombre_point ; i++, z++) {
    *z = nbr_ddl;
  }
}

/* -------- Traitement aprs avoir fini les calculs -------------\
 | moyen       : le graphe  traiter (ses barres d'erreur)       |
 | err_moyenne : VRAI => on veut les erreurs sur la moyenne      |
 \-------------------------------------------------------------- */

static void DISP_ApresCalculs(Graphe *moyen, boolean err_moyenne)
{
  if (NULL == moyen) return;

  /* Si ncessaire, division par le nombre de mesures (= degrs de libert + 1)
     Donne alors la barre d'erreur sur la moyenne */
  if (VRAI == err_moyenne) {
    int *z;
    unsigned long i;

    /* Pour les X */
    if ( moyen->erreur_x != NULL ) {
      double *x;

      for (i = 0, x = moyen->erreur_x->y, z = moyen->erreur_x->z.ddl ;
	   i < moyen->nombre_point ; i++, x++, z++) {
	*x /= sqrt( (double) (*z + 1) );
	*z = -*z; /* On a les erreurs sur la moyenne ! */
      }
    }

    /* Pour les Y */
    if ( moyen->erreur_y != NULL ) {
      if ( GRF_FCT_COMPLEXE == moyen->genre_donnees ) {
	Complexe *y;

	for (i = 0, y = (Complexe *) moyen->erreur_y->y, z = moyen->erreur_y->z.ddl ;
	   i < moyen->nombre_point ; i++, y++, z++) {
	  y->re /= sqrt( (double) (*z + 1) );
	  y->im /= sqrt( (double) (*z + 1) );
	  *z = -*z; /* On a les erreurs sur la moyenne ! */
	}
      } else {
	double *y;

	for (i = 0, y = moyen->erreur_y->y, z = moyen->erreur_y->z.ddl ;
	   i < moyen->nombre_point ; i++, y++, z++) {
	  *y /= sqrt( (double) (*z + 1) );
	  *z = -*z; /* On a les erreurs sur la moyenne ! */
	}
      }
    }
  }

  GRF_bornes(moyen->erreur_x, ORDONNE, DESORDONNE); /* On recalcule les bornes */
  GRF_bornes(moyen->erreur_y, ORDONNE, DESORDONNE);
  GRF_bornes(moyen, ORDONNE, DESORDONNE); /* On calcule les bornes */
}

/* -~-~-~-~-~-~-~-~ Moyenne en considrant que les X sont en correspondance -~-~-~-~-~-~-~-~ */

/* -------- Interface gnrale -----------------------------------------------\
 |  Renvoie le graphe moyen avec ses barres d'erreur                          |
 |  SUPPOSE 1) que les graphes soient tris                                   |
 |          2) que les X soient cohrents (il n'y a pas de vrification)      |
 |  NOTE : les X tant arbitraires, ils sont pris gaux  la moyenne des X... |
 |                                                                            |
 | graphes   : les graphes  moyenner                                         |
 | erreurs_x : VRAI si on veut calculr les erreurs en X                       |
 \--------------------------------------------------------------------------- */

static Graphe *DISP_MoyennerX(Graphe **graphes, unsigned long nombre, Graphe *modele, int axes)
{
  Graphe *moyen; /* Le rsultat */
  TRI_infos *infos_tri;
  unsigned long nbr_moyennes;

  if ( (NULL == graphes) || (NULL == modele) ) return NULL;

  /* Contrle de la cohrence du nombre de points */
  if ( FAUX == GRF_ControleTaille( graphes, nombre, GRF_CONTROLE_N_MINIMAL, &(modele->nombre_point), VRAI ) ) return NULL;

  /* Construction des infos pour la moyenne */
  infos_tri = DISP_PrepareFusion(graphes, nombre);
  if ( NULL == infos_tri) return NULL;

  /* Prparation du spectre moyen */
  moyen = disp_graphe_moyen(modele, axes & AXE_X, nombre);
  if (moyen == NULL) {
    free(infos_tri);
    return NULL;
  }

  /* On calcule le spectre moyen, avant toute chose */
  nbr_moyennes = MOY_blocs(moyen->x, moyen->y, infos_tri, nombre, modele->nombre_point);
  if (nbr_moyennes != modele->nombre_point) {
    STAT_ERREUR(STAT_INCOHERENCE);
    GRF_efface(moyen, VRAI, VRAI);
    free(infos_tri);
    return NULL;
  }

  if (FAUX == DISP_Moyenne(moyen, infos_tri, nombre) ) { /* Erreur survenue */
    GRF_efface(moyen, VRAI, VRAI);
    free(infos_tri);
    return NULL;
  }
  free(infos_tri);

  /* Les degrs de libert */
  init_mesures( moyen->erreur_y, nombre - 1 );

  return moyen;
}

/* -------------- Routine principale -----------------------------------------------------------\
 |    Renvoie VRAI si tout s'est bien pass, FAUX en cas d'erreur                               |
 |    Utilise les graphes d'erreur prcrs pour le graphe moyen                                |
 |    SUPPOSE que les graphes traits sont tous tris par X croissants                          |
 |            que tous les graphes ont le mme nombre de point (= celui du graphe moyen)        |
 |            que les X soient cohrents (aucune vrification)                                  |
 |                                                                                              |
 | moyen       : le spectre moyen                                                               |
 | infos       : contient les infos sur tous les spectres ayant servi au calcul                 |
 |               (voir graphes.h pour la structure TRI_infos)                                   |
 | nbr_graphes : nombre de graphes ayant servi au calcul                                        |
 \--------------------------------------------------------------------------------------------- */

static boolean DISP_Moyenne(Graphe *moyen, TRI_infos *infos, unsigned long nbr_graphes)
{
  Graphe *x_disperse, *y_disperse;
  unsigned long i, j;

  if ( (NULL == moyen) || (infos == NULL) )
    return FAUX;

  if (nbr_graphes < 2) {
    EXF_ERREUR(DIV_ZERO, "Trop peu de graphes", txt_err_calc_annule);
    return FAUX;
  }

  x_disperse = moyen->erreur_x;
  y_disperse = moyen->erreur_y;

  if (NULL != x_disperse) { /* On calcule aussi les erreurs sur les X */
    register double *x_lu, *sigma_x;

    /* Calcul de l'cart-type sur les X */
    for (i = 0, x_lu = moyen->x, sigma_x = x_disperse->y ;
	 i < moyen->nombre_point ;
	 i++, x_lu++, sigma_x++) {
      double somme = 0;
      for (j = 0 ; j < nbr_graphes ; j++) {
	double carre = ( *x_lu - (infos[j].x)[i] );
	somme += carre * carre;
      }
      somme /= (double) (nbr_graphes - 1);
      
      *sigma_x = sqrt(somme);
    }

    GRF_bornes(x_disperse, ORDONNE, DESORDONNE);
  }

  if (y_disperse == NULL) return FAUX; /* Sert  rien du tout */

  /* Calcul de l'cart-type sur les Y */
  for (i = 0 ; i < moyen->nombre_point ; i++) { /* Pour chaque point */
    double somme = 0;
    for (j = 0 ; j < nbr_graphes ; j++) {
      double carre = ( moyen->y[i] - (infos[j].y)[i] );
      somme += carre * carre;
    }
    somme /= (nbr_graphes - 1);

    (y_disperse->y)[i] = sqrt(somme); /* cart type sur ce point = cart-type */
  }

  return VRAI;
}

/* -~-~-~-~-~-~-~-~ Moyenne par interpolation sur une grille en X impose -~-~-~-~-~-~-~-~ */

/* -------- Interface gnrale ----------------------------------\
 |  Renvoie le graphe moyen avec ses barres d'erreur             |
 |                                                               |
 | infos = les informations dfinies par l'interface             |
 | type  = comment calculer la grille des X                      |
 | axes  = sur quel axe (X, Y, X & Y) calculer les dispersions ? |
 \-------------------------------------------------------------- */

static Graphe *DISP_PrepareLissage(Graphe **graphes, unsigned long nombre, Graphe *modele, int axes __attribute__((__unused__)), int type)
{
  Graphe *moyen;
  unsigned long j, premier, j_ref = 0;

  if ( (NULL == graphes) || (NULL == modele) ) {
    return NULL;
  }

  /* Intervalle d'intersection des X de tous les spectres + construction de la table */
  if ( GRF_ControleX(graphes, nombre, GRF_CONTROLE_X_INTER, &(modele->x_min), &(modele->x_max) ) == FAUX ) {
    (void) fprintf(stderr, "DISP_PrepareLissage : erreur dans GRF_ControleX()\n");
    return NULL;
  }

  /* Nombre de points : ceux du graphe qui en a le plus dans l'intervalle retenu */
  modele->nombre_point = GRF_intervalle( graphes[0], modele->x_min, modele->x_max, &premier);
  for ( j = 1 ; j < nombre ; j++ ) {
    unsigned long n, p;

    n = GRF_intervalle( graphes[j], modele->x_min, modele->x_max, &p);
    if (n > modele->nombre_point) {
      modele->nombre_point = n;
      premier = p;
      j_ref = j;
    }
  }

  /* Prparation du spectre moyen */
  moyen = disp_graphe_moyen( modele, FAUX, nombre);
  if (NULL == moyen) {
    (void) fprintf(stderr, "DISP_PrepareLissage : ERREUR dans disp_graphe_moyen()\n");
    return NULL;
  }

  /* Les X pour lesquels on va calculer */
  if ( DISP_X_RECALER == type ) { /* Ceux du graphe qui en a le plus dans l'intervalle retenu */
    memcpy( moyen->x, graphes[j_ref]->x + premier, sizeof(double) * modele->nombre_point );
  } else { /* Rgulirement espacs entre x_min et x_max */
    double pas, *x;
    unsigned long i;

    pas = (modele->x_max - modele->x_min) / ( (double) (moyen->nombre_point - 1) );

    for ( i = 0, x = moyen->x ; i < moyen->nombre_point ; i++, x++ ) {
      *x = modele->x_min + pas * i; /* Plus long, mais erreurs ne s'accumulent pas /  *x = *(x - 1) + pas */
    }
  }

  if ( FAUX == DISP_Lissage(moyen, graphes, nombre) ) {
    GRF_efface(moyen, VRAI, VRAI);
    return NULL;
  }

  return moyen; /* Termin */
}

/* --------- Calcul effectif, variante pour l'utilisation des spectres lisss ---------\
 |    Renvoie le graphe contenant la dispersion sur les Y,                             |
 |                                                                                     |
 | moyen   : contient le spectre moyen                                                 |
 | graphes : les graphes dont on veut les carts-types (il y en a nombre)              |
 \------------------------------------------------------------------------------------ */

static boolean DISP_Lissage(Graphe *moyen, Graphe **graphes, unsigned long nombre)
{
  double **b, **c, **d;
  register double *x, *y;
  unsigned long i;

  if ( (graphes == NULL) || (moyen == NULL) )
    return FAUX;

  /* Prparation des tableaux des splines */
  b = (double **) calloc(nombre, sizeof(double *));
  c = (double **) calloc(nombre, sizeof(double *));
  d = (double **) calloc(nombre, sizeof(double *));
  if ( (b == NULL) || (c == NULL) || (d == NULL) ) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "dispersion.c", "disp_lisser", "b, c, d");
    free(d); free(c); free(b);
    return FAUX;
  }

  /* Calcul des coefficients des splines */
  for (i = 0 ; i < nombre ; i++) {
    int retour;

    retour = SPL_tt_prep(graphes[i], &(b[i]), &(c[i]), &(d[i]) );
    if ( retour != 0 ) { /* Erreur survenue */
      unsigned long j;

      (void) LIS_ERREUR( retour, graphes[i]->nom );
      for (j = 0 ; j < i ; j++) {
	free( b[j] ); free( c[j] ); free( d[j] );
      }
      free(b); free(c); free(d);
      return FAUX;
    }
  }

  /* On calcule la somme des Y_i / nombre_graphes
     Pour chaque point, on calcule la spline de chaque graphe et on fait la somme.
     NOTE : repose sur le fait qu' la cration, les Y des graphes sont initialiss  0 */

  for ( i = 0, x = moyen->x, y = moyen->y ; i < moyen->nombre_point ; i++, x++, y++ ) {
    unsigned long j;

    for (j = 0 ; j < nombre ; j++) {
      *y += SPL_tt_calc(graphes[j], *x, b[j], c[j], d[j]);
    }
    *y /= (double) (nombre);
  }

  /* Calcul de l'cart-type sur les Y */
  for (i = 0, x = moyen->x, y = moyen->y ; i < moyen->nombre_point ; i++, x++, y++) { /* Pour chaque point */
    double somme = 0.0;
    unsigned long j;

    for (j = 0 ; j < nombre ; j++) {
      double carre;

      carre = *y - SPL_tt_calc(graphes[j], *x, b[j], c[j], d[j]);
      somme += carre * carre;
    }
    somme /= (double) (nombre - 1);

    (moyen->erreur_y->y)[i] = sqrt(somme); /* cart type sur ce point */
  }

  /* Plus besoin des coefficients de la spline */  
  for (i = 0 ; i < nombre ; i++) {
    free(b[i]); free(c[i]); free(d[i]);    
  }
  free(b); free(c); free(d);

  /* Dispersions enregistres sous la forme (X, cart_type(f(X)) */
  init_mesures(moyen->erreur_y, nombre - 1);

  return VRAI; /* Termin sans problme */
}

/* ~~~~~~~~ Moyenne en fusionnant les graphes puis liminant les doublons ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* -------- Prparation des infos pour le tri/la fusion par blocs -------- */

static TRI_infos *DISP_PrepareFusion(Graphe **graphes, unsigned long nombre)
{
  TRI_infos *infos_fusion;
  unsigned long i;

  if ( NULL == graphes ) return NULL;

  /* On rserve */
  infos_fusion = (TRI_infos *) malloc( sizeof( TRI_infos ) * nombre );
  if (NULL == infos_fusion) return NULL;

  /* On remplit */
  for (i = 0 ; i < nombre ; i++) {
    infos_fusion[ i ].x = graphes[ i ]->x; /* les x */
    infos_fusion[ i ].y = graphes[ i ]->y; /* les y */
    infos_fusion[ i ].position = 0;
    infos_fusion[ i ].nbr_points = graphes[i]->nombre_point;
    infos_fusion[ i ].axe_erreurs = 0;
    if ( graphes[ i ]->erreur_x != NULL ) { /* Il y a dj des erreurs en X */
      infos_fusion[ i ].axe_erreurs |= AXE_X;
      infos_fusion[ i ].erreurs_x = graphes[ i ]->erreur_x->y;
    } else {
      infos_fusion[ i ].erreurs_x = NULL;
    }
    if ( graphes[ i ]->erreur_y != NULL ) { /* Il y a dj des erreurs en X */
      infos_fusion[ i ].axe_erreurs |= AXE_Y;
      infos_fusion[ i ].erreurs_y = graphes[ i ]->erreur_y->y;
    } else {
      infos_fusion[ i ].erreurs_y = NULL;
    }
  }

  return infos_fusion;
}

/* -------- Interface gnrale ----------------------------------\
 |  Renvoie le graphe moyen avec ses barres d'erreur             |
 |    Les barres d'erreur ne sont estimes que sur les doublons  |
 |  SUPPOSE que les spectres soient un minimum cohrent          |
 |                                                               |
 | graphes = les graphes  traiter (il y en a nombre)            |
 | modele  = quelques informations complmentaires               |
 \-------------------------------------------------------------- */

static Graphe *DISP_Fusion(Graphe **graphes, unsigned long nombre, Graphe *modele)
{
  TRI_infos *infos_fusion, resultat = { NULL, NULL, 0, 0, 0, NULL, NULL };
  Graphe *moyen;
  double *x_lit, *y_lit, *x_ecrit, *y_ecrit;
  double *y_err_ecrit;
  int *ddl_ecrit;
  unsigned long i, nbr_points;

  if ( (NULL == graphes) || (NULL == modele) ) return FAUX;

  /* Construction des infos pour la fusion */
  infos_fusion = DISP_PrepareFusion(graphes, nombre);
  if ( NULL == infos_fusion) return NULL;

  /* Nombre total de points */
  (void) GRF_ControleTaille(graphes, nombre, GRF_CONTROLE_N_SOMME, &(modele->nombre_point), FAUX);

  /* Prparation du spectre moyen */
  moyen = disp_graphe_moyen(modele, FAUX, nombre);
  if (moyen == NULL) {
    free(infos_fusion);
    return NULL;
  }

  /* Fusion : revient  faire un tri en blocs dont le rsultat est stock dans le graphe moyen */
  resultat.x = moyen->x;
  resultat.y = moyen->y;
  resultat.nbr_points = modele->nombre_point;
  TRI_Blocs( &resultat, infos_fusion, nombre );

  /* On parcourt maintenant le rsultat pour reprer les multiplons, les moyenner en calculant sigma */
  x_lit = moyen->x; y_lit = moyen->y; x_ecrit = x_lit; y_ecrit = y_lit;
  y_err_ecrit = moyen->erreur_y->y; ddl_ecrit = moyen->erreur_y->z.ddl;
  nbr_points = 0;
  i = 0;
  
  while ( i < moyen->nombre_point ) {
    unsigned long j, nbr_copies = 0; /* Sans compter l'original */
    double x_ref, *x;
    double somme_y;

    /* On cherche la premire valeur diffrente qui suit */
    x_ref = *x_lit; x = x_lit + 1;
    somme_y = *y_lit;
    if ( i >= moyen->nombre_point ) { /* On est arriv  la fin */
      *x_ecrit = x_ref; *y_ecrit = somme_y;
      *y_err_ecrit = 0.0; *ddl_ecrit = 0; /* Une seule copie */
      nbr_points++;
      break;
    }

    while ( *x == x_ref ) {
      nbr_copies++;
      somme_y += *(y_lit + nbr_copies);
      if ( i + nbr_copies >= moyen->nombre_point ) { /* On a atteint la fin */
	break;
      }
      x++;
    }
    *x_ecrit = x_ref; *y_ecrit = somme_y / (double) (nbr_copies + 1);
    *ddl_ecrit = nbr_copies;

    /* On calcule la variance */
    *y_err_ecrit = 0.0;
    if (nbr_copies > 0) { /* seulement s'il y a plusieurs valeurs... */
      for (j = 0 ; j <= nbr_copies ; j++) {
	double e;

	e = y_lit[j] - *y_ecrit;
	*y_err_ecrit += e * e;
      }
      *y_err_ecrit = sqrt(*y_err_ecrit / (double) (nbr_copies));
    }

    /* On passe au point suivant */
    nbr_points++;
    x_ecrit++; y_ecrit++;
    y_err_ecrit++; ddl_ecrit++;
    x_lit += 1 + nbr_copies;
    y_lit += 1 + nbr_copies;
    i += nbr_copies + 1;
  }

  /* Le graphe ne compte plus que nbr_points valeurs diffrentes */
  GRF_AdapteTaille(moyen, nbr_points);

  return moyen;
}

/* -------- Prparation du "modle" = paramtres du graphe moyen ---\
 |   Renvoie VRAI si tout va bien, FAUX en cas d'erreur             |
 |  en cas d'erreur, met GR_AUCUN comme type du modle              |
 |                                                                  |
 | graphes = les graphes  moyenner (il y en a nombre)              |
 | modele  = on y stockera ce qu'il faut...                         |
 \----------------------------------------------------------------- */

static boolean DISP_PrepareModele(Graphe **graphes, unsigned long nombre, Graphe *modele)
{
  unsigned long j;

  if ( (NULL == graphes) || (NULL == modele) ) return FAUX;
  modele->type = GR_AUCUN; /* Pour l'instant pas de type associ */

  /* Infos pour le modle */
  modele->nombre_point  = graphes[ 0 ]->nombre_point;
  modele->genre_donnees = graphes[ 0 ]->genre_donnees;
  modele->x_min = graphes[ 0 ]->x_min;
  modele->x_max = graphes[ 0 ]->x_max;

  modele->type     = graphes[ 0 ]->type;
  modele->drapeaux = graphes[ 0 ]->drapeaux;

  if ( GR_SSTYPE == modele->type ) { /* On voit s'il n'y a pas un type plus sympathique */
    for ( j = 1 ; j < nombre ; j++) {
      if ( GR_SSTYPE != graphes[ j ]->type ) {
	modele->type = graphes[ j ]->type;
	break;
      }
    }
  }

  return VRAI;
}

/* -------- Moyenne de plusieurs graphes avec calcul des barres d'erreur (carts-types) -------\
 |   Renvoie le graphe moyen avec ses barres d'erreur (NULL en cas d'erreur)                   |
 |                                                                                             |
 | graphes     : la table des graphes  moyenner (supposs plus ou moins compatibles)          |
 | nbr_graphes : le nombre de graphes dans cette table                                         |
 | type        : comment traiter les diffrences en X (cf statistiques.h pour les dfinitions) |
 | erreur_x    : VRAI : on veut aussi estimer les erreurs en X (si l'on peut)                  |
 | err_moyenne : VRAI : on veut les erreurs sur la moyenne (en gnral, peu utile)             |
 \-------------------------------------------------------------------------------------------- */

Graphe *MoyennerGraphes(Graphe **graphes, unsigned long nbr_graphes, int type, boolean erreur_x, boolean err_moyenne)
{
  int axes = AXE_Y;
  unsigned long j;
  Graphe *moyen = NULL, modele;

  if ( NULL == graphes ) return NULL;
  if ( nbr_graphes < 2 ) {
    if ( nbr_graphes > 0 ) { /* Appel avec 0 et graphes != NULL possible lors de l'extraction de fichiers exprimentaux ! */
      (void) BI_ERR_alerte(NULL, &ERR_PasAssezGraphes, VRAI);
    }
    return NULL;
  }

  if ( VRAI == erreur_x )
    axes |= AXE_X; /* On garde les erreurs en X */

  /* Premier graphe : donne le type et tout a */
  if ( FAUX == DISP_PrepareModele(graphes, nbr_graphes, &modele) ) /* Erreur survenue */
    return NULL;

  /* L'historique (dbut) */
  HIST_Titre(1, "Calcul des barres d'erreur", "DISP");
  HIST_Texte(2, FAUX, "Nombre de graphes choisis : %lu\n", nbr_graphes);
  HIST_Texte(5, FAUX, "Graphes choisis :\n");
  for (j = 0 ; j < nbr_graphes ; j++) {
    HIST_Texte(5, FAUX, " - %s\n", graphes[j]->nom);
  }
  HIST_Texte(2, FAUX, "Traitement des X : ");

  switch ( type )
  {
  case DISP_X_MOYENNER: /* X postuls compatibles entre eux */
    HIST_Texte(2, FAUX, "suppos"e_AIGU"s compatibles\n");
    moyen = DISP_MoyennerX(graphes, nbr_graphes, &modele, axes);
    break;

  case DISP_X_INTERPOLER: /* On fait des interpolations */
    HIST_Texte(2, FAUX, "interpolation sur la r"e_AIGU"gion commune\n");
    moyen = DISP_PrepareLissage(graphes, nbr_graphes, &modele, axes, type);
    break;

  case DISP_X_RECALER:    /* On recale sur le premier spectre */
    HIST_Texte(2, FAUX, "calage sur les X du premier\n");
    moyen = DISP_PrepareLissage(graphes, nbr_graphes, &modele, axes, type);
    break;

  case DISP_X_FUSION: /* On fusionne puis limine les doublons */
    HIST_Texte(2, FAUX, "fusion & traitement des doublons\n");
    moyen = DISP_Fusion(graphes, nbr_graphes, &modele);
    break;

  default:
    HIST_Texte(2, FAUX, "INCONNUE\n");
    fprintf(stderr, "MoyennerGraphes : Mthode de traitement des X inconnue (%d)\n", type);
  }

  if ( NULL == moyen ) {
    HIST_Titre(-1, "Calcul du spectre moyen & des erreurs annul"e_AIGU, "DISP");
    return NULL;
  }

  DISP_ApresCalculs( moyen, err_moyenne );

  return moyen;
}

