/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999, Emmanuel CURIS
   Cration de ce fichier : 26 avril 1999

   Module d'analyse statistique
   Routines de bruitage d'un spectre et de gnration d'un processus alatoire
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "systeme.h"

#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int PA_GenererBruitBlanc(Graphe *graphe, struct processus_aleatoire *processus);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Routines gnriques : appel depuis l'extrieur ~-~-~-~-~-~-~-~-~ */

/* -------- Gnrer un graphe avec les valeurs d'un processus alatoire
    Renvoie le graphe cr, c'est un graphe ISOL
            le processus est suppos centr, rduit

  processus   : dfinition du processus alatoire  utiliser
  nbr_valeurs : nombre de valeurs  gnrer
*/

Graphe *PA_GenererStationnaire(struct processus_aleatoire *processus, unsigned long nbr_valeurs)
{
  Graphe *resultat;
  int retour = -1;

  /* Contrle prliminaire */
  if ( NULL == processus ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune information indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  /* On cre le graphe rsultat */
  resultat = GRF_nv_isole( nbr_valeurs, GR_EPREUVE, GRF_SERIE );
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de gnrer le graphe pour les valeurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On gnre les valeurs, avec le modle adapt */
  switch ( processus->type )
  {
  case PA_BRUIT_BLANC:
    retour = PA_GenererBruitBlanc( resultat, processus );    
    break;

  case PA_AR:
    OPTION_A_FAIRE( "Processus autorgressif" );
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type de processus inconnu [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, processus->type);
  }

  /* On contrle le calcul */
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le calcul, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    GRF_eff_isole( resultat );
    return NULL;
  }

  /* On contrle les bornes */
  GRF_bornes( resultat, NE_PAS_FAIRE, DESORDONNE );

  /* On renvoie le rsultat */
  return resultat;
}

/* -------- Gnrer une ralisation d'un processus alatoire dans le graphe indiqu --------
    Renvoie  0 si tout s'est bien pass
            <0 en cas d'erreur

      Le processus alatoire est distribu avec la mme famille de loi en chaque point
                              = valeur en Y dans le graphe indiqu
			     s = valeur de l'erreur sur Y dans le graphe indiqu
      Les points sont indpendants
 */

int PA_GenererSurGraphe(struct processus_aleatoire *processus, Graphe *graphe)
{
  unsigned long i;
  double *y, sigma;

  if ( ( NULL == graphe ) || ( NULL == processus ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Paramtres absurdes [graphe=%p, processus=%p]\n",
		   __FILE__, __FUNCTION__, __LINE__, (void *) graphe, (void *) processus);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }
  if ( NULL == graphe->erreur_y ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune erreur indique : le bruit sera donc nul !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    /* On rcupre l'cart-type dans les erreurs sur Y originelles */
    sigma = graphe->erreur_y->y[ i ];
    if ( sigma <= 0.0 ) {
      /* Le cas nul peut arriver : valeur parfaite... */
      if ( sigma < 0.0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreur ngative [%g], impossible\n",
		       __FILE__, __FUNCTION__, __LINE__, sigma);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      }
      continue;
    }

    /* On en dduit les paramtres de la loi */
    switch ( processus->loi.famille )
    {
    case DP_GAUSSIENNE:
      processus->loi.parametres.Gauss.mu    = graphe->y[ i ];
      processus->loi.parametres.Gauss.sigma = sigma;
      break;

    default:
      AVERTISSEMENT( "Loi inconnue ou  faire = %d", processus->loi.famille );
    }

    /* On tire au hasard la valeur */
    *y = STAT_NombreAleatoire( &(processus->loi) );
  }

  /* Termin : on a maintenant, dans le graphe, une collection de valeurs altoires, indpendantes */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Briques lmentaires pour les processus alatoires ~-~-~-~-~-~-~-~-~ */

/* -------- Bruit blanc : points indpendants -------- */

static int PA_GenererBruitBlanc(Graphe *graphe, struct processus_aleatoire *processus)
{
  unsigned long i;
  double *y;

  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    *y = STAT_NombreAleatoire( &(processus->loi) );
  }

  return 0;
}
