/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1999, Emmanuel CURIS
   Cration : 20 mai 1999

   Module de calculs statistiques

   Tri des donnes (utilis essentiellement en stats, d'o sa place, mais aussi ailleurs...)
   Permet soit le tri d'un graphe en X, avec tri des Y coupl
          soit le tri d'un tableau simplement
   Utilise la mthode de tri par fusion (par blocs), sauf prcision contraire
   Le tri se fait toujours pas ordre croissant.
   Toutes les donnes tries sont supposes tre des valeurs relles
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* Pour HUGE_VAL */

#include "systeme.h"

#include "exafs.h"

#include "statistiques.h"
#include "tri.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void fusion(double *y1, double *y2, 
		   unsigned long i1, unsigned long i2, unsigned long i3, unsigned long max);


static int TRI_ControlerBloc( TRI_infos *bloc );

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~- Tri d'un seul tableau par ordre croissant ~-~-~-~- */

/* ------ Fonction d'appel du tri : organise les diffrents tris par rcurrence -------\
 |  Renvoi l'adresse du tableau tri (NULL en cas d'erreur)                            |
 |  En cas d'erreur, le tableau initial n'a pas t modifi                            |
 |  Attention, le tableau tri n'est pas forcment le mme que le tableau d'origine :  |
 |  tout dpend du nombre de passes effectues                                         |
 |  (consquence : le tableau d'origine a trs bien pu tre libr, donc tre perdu !) |
 |                                                                                     |
 | x            = le tableau  trier                                                   |
 | nbr_elements = le nombre d'lments dans ce tableau                                 |
 \------------------------------------------------------------------------------------ */

double *TRI_Tableau(double *x, unsigned long nbr_elements)
{
  double *x_trie = NULL;
  unsigned long i, j = 1;

  /* On rserve le second tableau : tri par change d'un tableau  l'autre... */
  x_trie = (double *) calloc(nbr_elements, sizeof(double));
  if (x_trie == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Stats/tri.c", "tri_tableau", "x_trie");
    return NULL;
  }

  /* On ralise effectivement le tri. Principe du tri par fusion :
       1) on tri les lments des couples { (0,1), (2,3), ..., (i, i+1) }
       2) on tri les lments des quadruplets { (0,1,2,3), (4,5,6,7), ... }
       ...
       3) on tri les deux demi-tableaux finaux -> le tableau global final est tri.
  */

  while (j < nbr_elements) {
    double *temp;

    for (i = 0 ; i < nbr_elements ; i += 2 * j) {
      fusion(x, x_trie, i, i + j, i + 2 * j, nbr_elements);
    }
    j <<= 1;

    temp = x; x = x_trie; x_trie = temp;
  }

  free(x_trie); /* x contient les X tris par ordre croissant */
  return x;
}

/* ------ Ralisation du tri par fusion - fusion effective de 2 zones conscutives ------\
 |                                                                                       |
 | x     = le tableau originel                                                           |
 | x_tri = le tableau contenant les valeurs tries                                       |
 |    la routine fusionne, par ordre croissant, les rgions [i1, i2[ et [i2, i3[         |
 | max   = le nombre d'lments dans le tableau                                          |
 \-------------------------------------------------------------------------------------- */

static void fusion(double *x, double *x_tri,
		   unsigned long i1, unsigned long i2, unsigned long i3, unsigned long max)
{
  unsigned long i, j, k;
  double x_1, x_2;

  /* On vrifie qu'on ne dpasse pas le tableau */
  if (i2 > max) i2 = max;
  if (i3 > max) i3 = max;
 
  i = i1; j = i2;
  if (i < i2)
    x_1 = x[i];
  else
    x_1 = HUGE_VAL;
  if (j < i3)
    x_2 = x[j];
  else
    x_2 = HUGE_VAL;

  for (k = i1 ; k < i3 ; k++) {
    if (x_1 < x_2) {
      x_tri[k] = x_1; i++;
      if (i < i2)
	x_1 = x[i];
      else
	x_1 = HUGE_VAL;
    } else {
      x_tri[k] = x_2; j++;
      if (j < i3)
	x_2 = x[j];
      else
	x_2 = HUGE_VAL;
    }
  }
}

/* ~-~-~-~-~-~-~-~-~ Tri par fusion de 2 tableaux coupls, plusieurs blocs, avec incertitudes , hors structure de graphe ~-~-~-~-~-~-~-~-~ */

/* -------- Controle qu'un bloc est utilisable --------
     Renvoie  0 si tout va bien
             <0 en cas de problme
 */

static int TRI_ControlerBloc(TRI_infos *bloc)
{
  if ( NULL == bloc ) {
    (void) fprintf(stderr, "%s,%s l.%u : Bloc NULL\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  if ( NULL == bloc->x ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de X indiqu\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  if ( NULL == bloc->y ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de Y indiqu\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  if ( ( bloc->axe_erreurs & AXE_X ) && ( NULL == bloc->erreurs_x ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'erreurs en X, pourtant demandes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }

  if ( ( bloc->axe_erreurs & AXE_Y ) && ( NULL == bloc->erreurs_y ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'erreurs en Y, pourtant demandes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  /* Tout va bien */
  return 0;
}

/* -------- Fusion & Tri de plusieurs blocs de donnes simultanment --------
    Renvoie 0   si tout s'est bien pass
            <=0 en cas d'erreur
    Tri par X croissants avec tri simultan des Y, chaque bloc est suppos prtri
 */

int TRI_Blocs(TRI_infos *trie, TRI_infos *blocs, unsigned long nbr_blocs)
{
  double x_min;
  int retour;
  unsigned long i, num_bloc, num_mini, total;

  /* On controle que les donnes passes sont correctes */
  retour = TRI_ControlerBloc( trie );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Zone destination incorrecte [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  for ( i = 0, total = 0 ; i < nbr_blocs ; i++ ) {
    retour = TRI_ControlerBloc( blocs + i );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Zone source %lu sur %lu incorrecte [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, i + 1, nbr_blocs, retour);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return -2;
    }

    total += blocs[ i ].nbr_points;
  }

  if ( total > trie->nbr_points ) { /* Si la zone destination est trop grande, ce n'est pas grave */
    (void) fprintf(stderr, "%s,%s l.%u : La zone destination est trop petite [%lu points, il en faut %lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, trie->nbr_points, total);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -3;
  }

  /* On remplit tous les points de la zone destination, en cherchant le plus petit 1er point de chacun des blocs */
  for ( i = 0 ; i < total ; i++ ) {
    /* On cherche le premier bloc non termin */
    num_mini = 0;
    while ( blocs[ num_mini ].position >= blocs[ num_mini ].nbr_points ) {
      num_mini++;
      if ( num_mini >= nbr_blocs ) {
	(void) fprintf(stderr, "%s,%s l.%u : Tous les blocs sont finis mais il reste des points  faire !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	return -20;
      }
    }

    /* Point courant du premier bloc non-termin */
    x_min = *( blocs[ num_mini ].x + blocs[ num_mini ].position );

    /* On recherche le plus petit point des blocs restants */
    for ( num_bloc = num_mini ; num_bloc < nbr_blocs ; num_bloc++ ) {
      if ( blocs[ num_bloc ].position < blocs[ num_bloc ].nbr_points ) { /* Il reste des points  lire */
	if ( *(blocs[ num_bloc ].x  + blocs[ num_bloc ].position) < x_min ) { /* Premier point du bloc n */
	  num_mini = num_bloc;
	  x_min = *( blocs[ num_bloc ].x + blocs[ num_bloc ].position);
	}
      }
    }
    
    /* Maintenant, x_min contient la valeur minimale et num_mini le numro du bloc associ */
    trie->x[ i ] = x_min;                          /* On enregistre le premier point */
    trie->y[ i ] = *( blocs[ num_mini ].y + blocs[ num_mini ].position );
    if ( trie->axe_erreurs & AXE_X ) { /* On veut une erreur en X */
      if ( blocs[ num_mini ].axe_erreurs & AXE_X ) { /* Il y en avait une  l'origine */
	trie->erreurs_x[ i ] = blocs[ num_mini ].erreurs_x[ blocs[ num_mini ].position ];
      } else {
	trie->erreurs_x[ i ] = 0.0;
      }
    }

    /* On a lu un point de ce bloc */
    blocs[ num_mini ].position++;
  }

  /* Tout s'est bien pass */
  return 0;
}
