/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2004, Emmanuel CURIS

   Calcul théorique de l'absorption des éléments
   Utilisation de la méthode de Ebel & coll. XRS 32:6 (2003) 442-451   
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

#include "systeme.h"
#include "ascii_utils.h"
#include "chimie_utils.h" /* Données de base sur les atomes */
#include "txt_utils.h"

#include "absorption.h"
#include "science.h" /* Pour la conversion E <-> longueur d'onde */
#include "theorie.h" /* Pour la gestion de tables de valeurs */

/* ~~~~~~~~ Variables locales : table des coefficients de Ebel ~~~~~~~~ */

/* Coefficients de Ebel par sous-couche
   Les valeurs sont en cm^2/g ; en simple precision dans le fichier :
   float suffit et economise de la memoire

   Formule : ln( mu/rho ) = f( ln(E) ) polynome d'ordre 5
   Valable entre 1 et 300 keV
*/

#ifdef _DEVELOPPE_
# ifndef PREFIXE_REP
#  define PREFIXE_REP "/home/curis/Lase/Donnees/"
# endif
#endif

#define EBEL_CHEMIN_TABLE PREFIXE_REP"/Ebel/"

#define EBEL_Z_MAX PLUTONIUM

static float *Ebel_Krange = NULL; /* Données pour       E > K , section efficace totale */
static float *Ebel_Lrange = NULL; /* Données pour K  >= E > L3, section efficace totale */
static float *Ebel_Mrange = NULL; /* Données pour L3 >= E > M5, section efficace totale */
static float *Ebel_Nrange = NULL; /* Données pour M5 >= E > N7, section efficace totale */

static float *Ebel_Sauts[ NOMBRE_SEUILS ] = { /* Hauteurs de seuil */
  NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL
#ifdef AVEC_TRANSURANIENS_
  , NULL, NULL, NULL, NULL,
#endif
};

static float *Ebel_ParSeuil[ NOMBRE_SEUILS ] = { /* Sections efficaces par seuil */
  NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL
#ifdef AVEC_TRANSURANIENS_
  , NULL, NULL, NULL, NULL,
#endif
};

static float *Ebel_elastique   = NULL; /* Données de diffusion élastique */
static float *Ebel_inelastique = NULL; /* Données de diffusion inélastique */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int Ebel_ChargerTables (void);
static int Ebel_ReserverTables(void);
static int Ebel_LitFichier    (const char *nom_fichier);

/* ~~~~~~~~ Réalisation pratique ~~~~~~~~ */

/* -------- Lecture des tables de coefficients --------
     Renvoie 0    si tout s'est bien passe
             != 0 en cas d'erreur
*/

static int Ebel_ChargerTables(void)
{
  DIR *dossier;
  int retour;
  struct dirent *fichier;

  if ( Ebel_elastique != NULL ) {
    (void) fprintf(stderr, "%s,%s l.%u : Tables déjà chargées !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1; /* Erreur mineure */
  }

  /* On réserve les tables */
  retour = Ebel_ReserverTables( );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver les tables [erreur %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    Ebel_EffacerTables( ); /* Pour effacer en cas de chargement partiel */
    return -1;
  }

  /* On ouvre le répertoire */
  dossier = opendir( EBEL_CHEMIN_TABLE );
  if ( NULL == dossier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Le dossier de données n'existe pas\n",
		   __FILE__, __FUNCTION__, __LINE__);
    Ebel_EffacerTables();
    return -50;
  }

  /* On parcourt le répertoire pour charger fichier par fichier */
  fichier = readdir( dossier );
  while ( fichier != NULL ) {
    unsigned long lg_nom;
    int retour;

    /* On vérifie que le fichier convient */
    if ( '.' == fichier->d_name[ 0 ] ) { /* Fichier caché, ignoré */
      fichier = readdir( dossier );
      continue;
    }
    lg_nom = strlen( fichier->d_name );
    if ( ( lg_nom != 4 ) && ( lg_nom != 5 ) ) {
      AVERTISSEMENT( "Fichier ignoré : %s [nom trop court ou trop long]", fichier->d_name );
      fichier = readdir( dossier );
      continue;
    }
    if ( ( fichier->d_name[ lg_nom - 1 ] != 'U' ) || ( fichier->d_name[ lg_nom - 2 ] != 'M' ) ||
	 ( fichier->d_name[ lg_nom - 3 ] != '.' ) ) {
      AVERTISSEMENT( "Fichier ignoré : %s [extension incorrecte]", fichier->d_name );
      fichier = readdir( dossier );
      continue;
    }

    /* On lit le fichier correspondant */
    retour = Ebel_LitFichier( fichier->d_name );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR en lisant le fichier %s\n",
		     __FILE__, __FUNCTION__, __LINE__, fichier->d_name);
    }

    /* On passe au fichier suivant */
    fichier = readdir( dossier );
  }

  /* Parcours du dossier terminé */
  (void) closedir( dossier );

  /* Chargement des tables effectué, tout va bien */
  return 0;
}

/* -------- Lire un fichier de données de la compilation par Ebel -------------------- */

static int Ebel_LitFichier(const char *nom_fichier)
{
  char *nom_complet;
  double coefs[ 6 ];
  int c __attribute__((__unused__)), Z = 0, seuil = SEUIL_K, i, Z_min = HYDROGENE;
  FILE *fichier;

  /* Créer le nom complet du fichier à lire*/
  nom_complet = TXT_genere_chaine( "%s%s", EBEL_CHEMIN_TABLE, nom_fichier );
  if ( NULL == nom_complet ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de former le nom de fichier complet\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On ouvre le fichier */
  fichier = fopen( nom_complet, "r" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de lire le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom_complet);
    free( nom_complet ); nom_complet = NULL;
    return -2;
  }
  free( nom_complet ); nom_complet = NULL;

  /* On ignore la première ligne */
  c = ignore_ligne( fichier );

  /* On lit le numéro atomique -- la fin de la ligne est inutile */
  fscanf( fichier, "%u", &Z );
  if ( ( Z < HYDROGENE ) || ( Z > EBEL_Z_MAX ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Élément non prévu dans les tables d'Ebel originales, Z = %d\n",
		   __FILE__, __FUNCTION__, __LINE__, Z);
    (void) fclose( fichier );
    return -5;
  }
  c = ignore_ligne( fichier );

  /* On ignore la ligne suivante : commentaires */
  c = ignore_ligne( fichier );

  /* Deux lignes suivantes : valeurs pour la diffusion élastique et inélastique */
  lit_valeurs( fichier, 6, 0, coefs, NULL );
  for ( i = 0 ; i < 6 ; i++ ) {
    Ebel_elastique[ (Z - HYDROGENE) * 6 + i ] = (float) coefs[ i ];
  }
  lit_valeurs( fichier, 6, 0, coefs, NULL );
  for ( i = 0 ; i < 6 ; i++ ) {
    Ebel_inelastique[ (Z - HYDROGENE) * 6 + i ] = (float) coefs[ i ];
  }

  /* On ignore la ligne suivante : commentaires */
  c = ignore_ligne( fichier );

  /* On lit les seuils, un par un */
  do {
    double h_seuil;
    char nom_seuil[ 3 ];

    /* On lit le nom du seuil */
    (void) lit_cars( fichier, nom_seuil, 2 );
/*     printf( "Fichier %s, seuil : %s [élément %s]\n", nom_fichier, nom_seuil, ATM_Z_vers_symbole( Z ) ); */

    /* De quel seuil s'agit-il ? */
    if ( 'K' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_K;
    } else if ( 'L' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_LI + ( nom_seuil[ 1 ] - '1' );
    } else if ( 'M' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_MI + ( nom_seuil[ 1 ] - '1' );
    } else if ( 'N' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_NI + ( nom_seuil[ 1 ] - '1' );
    } else if ( 'O' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_OI + ( nom_seuil[ 1 ] - '1' );
    } else if ( 'P' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_PI + ( nom_seuil[ 1 ] - '1' );
    } else if ( 'Q' == nom_seuil[ 0 ] ) {
      seuil = SEUIL_QI + ( nom_seuil[ 1 ] - '1' );
    } else {
      seuil = -58;
      /* Lecture des seuils terminée, normalement */
      break;
    }

    /* À partir de quel élément ce seuil existe-t-il ? */
    Z_min = SEUIL_PremierElement( seuil );    

    /* Lecture de l'énergie de seuil et de sa hauteur */
    fscanf( fichier, "%*f %lf", &h_seuil );
    c = ignore_ligne( fichier ); /* La fin de la ligne est inutile */

    /* On lit les coefficients pour l'absorption par couche */
    if ( Z >= Z_min ) {
      lit_valeurs( fichier, 6, 0, coefs, NULL );
      for ( i = 0 ; i < 6 ; i++ ) {
	Ebel_ParSeuil[ seuil ][ (Z - Z_min) * 6 + i ] = (float) coefs[ i ];
      }
      /* ... et on mémorise la hauteur de seuil */
      TH_Table_Remplir( Ebel_Sauts[ seuil ], Z, Z_min, EBEL_Z_MAX, h_seuil );
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : Seuil %s pour l'élément %s, qui ne devrait pas en avoir. Ignoré\n",
		     __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( seuil ), ATM_Z_vers_symbole( Z ) );
      c = ignore_ligne( fichier );
    }
    /* On se prépare pour le seuil suivant */
  } while ( seuil >= 0 ); /* On a bien trouvé un seuil */


  /* On ignore la ligne suivante : commentaires */
  c = ignore_ligne( fichier );

  /* On lit l'absorbance totale, seuil K */
  for( i = 0 ; i < 6 ; i++ ) {
    fscanf( fichier, "%f", Ebel_Krange + (Z - HYDROGENE) * 6 + i );
  }

  /* On lit l'absorbance totale, seuils L */
  Z_min = SEUIL_PremierElement( SEUIL_LI );
  if ( Z > Z_min ) {
    for( i = 0 ; i < 6 ; i++ ) {
      fscanf( fichier, "%f", Ebel_Lrange + (Z - Z_min) * 6 + i );
    }
  } else {
    c = ignore_ligne( fichier );
  }

  /* On lit l'absorbance totale, seuils M */
  Z_min = SEUIL_PremierElement( SEUIL_MI );
  if ( Z > Z_min ) {
    for( i = 0 ; i < 6 ; i++ ) {
      fscanf( fichier, "%f", Ebel_Mrange + (Z - Z_min) * 6 + i );
    }
  } else {
    c = ignore_ligne( fichier );
  }

  /* On lit l'absorbance totale, seuils N */
  Z_min = SEUIL_PremierElement( SEUIL_NI );
  if ( Z > Z_min ) {
    for( i = 0 ; i < 6 ; i++ ) {
      fscanf( fichier, "%f", Ebel_Nrange + (Z - Z_min) * 6 + i );
    }
  } else {
    c = ignore_ligne( fichier );
  }

  /* Plus besoin de tout cela */
  (void) fclose( fichier );

  /* Chargement réussi, tout s'est bien passé */
  return 0;
}

/* -------- Effacer les tables de la mémoire quand on n'a plus besoin d'elles -------- */

void Ebel_EffacerTables(void)
{
  unsigned int i;

  free( Ebel_elastique ); Ebel_elastique = NULL;
  free( Ebel_inelastique ); Ebel_inelastique = NULL;

  for ( i = 0 ; i < NOMBRE_SEUILS ; i++ ) {
    free( Ebel_ParSeuil[ i ] ); Ebel_ParSeuil[ i ] = NULL;
    free( Ebel_Sauts[ i ] ); Ebel_Sauts[ i ] = NULL;
  }

  free( Ebel_Krange ); Ebel_Krange = NULL;
  free( Ebel_Lrange ); Ebel_Lrange = NULL;
  free( Ebel_Mrange ); Ebel_Mrange = NULL;
  free( Ebel_Nrange ); Ebel_Nrange = NULL;
}

/* -------- Réserver les tables en mémoire -------- */

static int Ebel_ReserverTables(void)
{
  unsigned int i;

  /* Tables de diffusion : tous les éléments */
  Ebel_elastique = (float *) malloc( sizeof(float) * 6 * EBEL_Z_MAX ); /* 6 coefficients par élément */
  if ( NULL == Ebel_elastique ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table de diffusion élastique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  Ebel_inelastique = (float *) malloc( sizeof(float) * 6 * EBEL_Z_MAX ); /* 6 coefficients par élément */
  if ( NULL == Ebel_inelastique ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table de diffusion inélastique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Tables de section efficace par seuil */
  for ( i = 0 ; i < NOMBRE_SEUILS ; i++ ) {
    int Zmin;

    Zmin = SEUIL_PremierElement( SEUIL_K + i );
    if ( Zmin > EBEL_Z_MAX ) {
      AVERTISSEMENT( "Données pour le seuil %s non-disponibles", TH_TexteSeuil( SEUIL_K + i ) );
      Ebel_ParSeuil[ i ] = NULL;
      Ebel_Sauts[ i ] = NULL;
      continue;
    }

    /* 6 coefficients par élément */
    Ebel_ParSeuil[ i ] = (float *) malloc( sizeof(float) * 6 * ( EBEL_Z_MAX - Zmin + 1 ) );
    if ( NULL == Ebel_ParSeuil[ i ] ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table du seuil %s\n",
		     __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( SEUIL_K + i ) );
      return -3;
    }

    Ebel_Sauts[ i ] = TH_Creer_Table( Zmin, EBEL_Z_MAX );
    if ( NULL == Ebel_Sauts[ i ] ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table de hauteur des seuils %s\n",
		     __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( SEUIL_K + i ) );
      return -4;
    }
  }

  /* Tables des sections efficaces totales */
  Ebel_Krange = (float *) malloc( sizeof(float) * 6 * EBEL_Z_MAX );
  if ( NULL == Ebel_Krange ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table pour E > K\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -8;
  }

  Ebel_Lrange = (float *) malloc( sizeof(float) * 6 * ( EBEL_Z_MAX - SEUIL_PremierElement( SEUIL_LI ) + 1 ) );
  if ( NULL == Ebel_Lrange ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table pour K >= E > L1\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -9;
  }

  Ebel_Mrange = (float *) malloc( sizeof(float) * 6 * ( EBEL_Z_MAX - SEUIL_PremierElement( SEUIL_MI ) + 1 ) );
  if ( NULL == Ebel_Mrange ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table pour L1 >= E > M1\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -10;
  }

  Ebel_Nrange = (float *) malloc( sizeof(float) * 6 * ( EBEL_Z_MAX - SEUIL_PremierElement( SEUIL_NI ) + 1 ) );
  if ( NULL == Ebel_Nrange ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la table pour M1 >= E > N1\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -11;
  }

  /* Création réussie, on renvoie 0 */
  return 0;
}

/* -------- Lire les coefficients dans la table -------- 
    Renvoie 0   si tout s'est bien passé
            !=0 sinon

  Z       : élément pour lequel on veut les informations
  seuil   : le seuil qui nous intéresse
  total   : VRAI = on veut la section efficace totale => on suppose en fait pour (seuil précédent) <= E <= seuil
            FAUX = on veut la section efficace pour le seuil indiqué uniquement
  c       : les 6 coefficients
  h_seuil : quelle hauteur de seuil correspond aux valeurs de la table ? [uniquement pour les sections efficaces par seuil !]
*/

int TH_coef_Ebel(int Z, int seuil, boolean total, double c[ 6 ], double *h_seuil)
{
  unsigned int i;

  if ( NULL == c ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de table pour stocker les coefficients !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* On initialise la table à 0.0 */
  for ( i = 0 ; i < 6 ; i++ )
    c[ i ] = 0.0;

  if ( ( Z < HYDROGENE ) || ( Z > EBEL_Z_MAX ) ) { /* Élément non tabulé */
    (void) fprintf(stderr, "%s,%s l.%u : Élément non-tabulé [Z = %d / %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, Z, ATM_Z_vers_symbole( Z ));
    return -2;
  }

  /* Les tables sont-elles chargées ? */
  if ( NULL == Ebel_inelastique ) { /* Non */
    int retour;

    retour = Ebel_ChargerTables();
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de charger les tables de coefficients\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -3;
    }
  }

  /* Les seuils "spéciaux" */
  if ( DIFFUSION_INCOHERENTE == seuil ) {
    for ( i = 0 ; i < 6 ; i++ ) {
      c[ i ] = (double) Ebel_inelastique[ 6 * ( Z - HYDROGENE ) + i ];
    }
  } else if ( DIFFUSION_COHERENTE == seuil ) {
    for ( i = 0 ; i < 6 ; i++ ) {
      c[ i ] = (double) Ebel_elastique[ 6 * ( Z - HYDROGENE ) + i ];
    }
  } else if ( APRES_SEUIL_K == seuil ) { /* Section efficace totale pour E > seuil K */
    for ( i = 0 ; i < 6 ; i++ ) {
      c[ i ] = (double) Ebel_Krange[ 6 * ( Z - HYDROGENE ) + i ];
    }
  } else {
    int Z_min;

    /* Quel est le premier élément à présenter le seuil considéré ? */
    Z_min = SEUIL_PremierElement( seuil );
    if ( VRAI == total ) {
      if ( Z < Z_min ) {
	(void) fprintf(stderr, "%s,%s l.%u : Pas de seuil %s pour l'élément %s, la région n'a pas de sens\n",
		       __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( seuil ), ATM_Z_vers_symbole( Z ));
	for ( i = 0 ; i < 6 ; i++ ) {
	  c[ i ] = -2.0;
	}
	return +5;
      }

      Z_min = SEUIL_PremierElement( seuil + 1 );
      if ( Z < Z_min ) {
	/* Section efficace totale nulle : on est avant le premier seuil de l'élément */
      } else {
	/* En fonction du seuil, on utilise la bonne table */
	switch (seuil)
	{
	case SEUIL_K:   /* On est entre K et LI */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Lrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_LI ) ) + i ];
	  }
	  break;

	case SEUIL_LI:  /* On est entre LI et LII */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Lrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_LI ) ) + i ];
	  }
	  /* On introduit la hauteur de seuil : division de l'exponentielle par cette valeur ->
	     - ln(valeur) du coefficient constant dans l'exponentielle */
	  c[ 0 ] -= log( TH_Table_Lire( Ebel_Sauts[ SEUIL_LI ], Z, SEUIL_PremierElement( SEUIL_LI ), EBEL_Z_MAX ) );
	  break;

	case SEUIL_LII: /* entre LII et LIII */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Lrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_LI ) ) + i ];
	  }
	  /* On introduit la hauteur de seuil : division de l'exponentielle par cette valeur ->
	     - ln(valeur) du coefficient constant dans l'exponentielle */
	  c[ 0 ] -= log( TH_Table_Lire( Ebel_Sauts[ SEUIL_LI  ], Z, SEUIL_PremierElement( SEUIL_LI  ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_LII ], Z, SEUIL_PremierElement( SEUIL_LII ), EBEL_Z_MAX ) );
	  break;

	case SEUIL_LIII: /* On est entre LIII et MI */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Mrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_MI ) ) + i ];
	  }
	  break;

	case SEUIL_MI: /* On est entre MI et MII */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Mrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_MI ) ) + i ];
	  }
	  /* On introduit la hauteur de seuil : division de l'exponentielle par cette valeur ->
	     - ln(valeur) du coefficient constant dans l'exponentielle */
	  c[ 0 ] -= log( TH_Table_Lire( Ebel_Sauts[ SEUIL_MI ], Z, SEUIL_PremierElement( SEUIL_MI ), EBEL_Z_MAX ) );
	  break;

	case SEUIL_MII: /* On est entre MII et MIII */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Mrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_MI ) ) + i ];
	  }
	  /* On introduit la hauteur de seuil : division de l'exponentielle par cette valeur ->
	     - ln(valeur) du coefficient constant dans l'exponentielle */
	  c[ 0 ] -= log( TH_Table_Lire( Ebel_Sauts[ SEUIL_MI  ], Z, SEUIL_PremierElement( SEUIL_MI  ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_MII ], Z, SEUIL_PremierElement( SEUIL_MII ), EBEL_Z_MAX ) );
	  break;

	case SEUIL_MIII: /* On est entre MIII et MIV */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Mrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_MI ) ) + i ];
	  }
	  /* On introduit la hauteur de seuil : division de l'exponentielle par cette valeur ->
	     - ln(valeur) du coefficient constant dans l'exponentielle */
	  c[ 0 ] -= log( TH_Table_Lire( Ebel_Sauts[ SEUIL_MI   ], Z, SEUIL_PremierElement( SEUIL_MI   ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_MII  ], Z, SEUIL_PremierElement( SEUIL_MII  ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_MIII ], Z, SEUIL_PremierElement( SEUIL_MIII ), EBEL_Z_MAX ) );
	  break;

	case SEUIL_MIV: /* On est entre MIV et MV */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Mrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_MI ) ) + i ];
	  }
	  /* On introduit la hauteur de seuil : division de l'exponentielle par cette valeur ->
	     - ln(valeur) du coefficient constant dans l'exponentielle */
	  c[ 0 ] -= log( TH_Table_Lire( Ebel_Sauts[ SEUIL_MI   ], Z, SEUIL_PremierElement( SEUIL_MI   ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_MII  ], Z, SEUIL_PremierElement( SEUIL_MII  ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_MIII ], Z, SEUIL_PremierElement( SEUIL_MIII ), EBEL_Z_MAX ) *
			 TH_Table_Lire( Ebel_Sauts[ SEUIL_MIV  ], Z, SEUIL_PremierElement( SEUIL_MIV  ), EBEL_Z_MAX ) );
	  break;

	case SEUIL_NI: /* On est entre MV et NI */
	  for ( i = 0 ; i < 6 ; i++ ) {
	    c[ i ] = (double) Ebel_Nrange[ 6 * ( Z - SEUIL_PremierElement( SEUIL_NI ) ) + i ];
	  }
	  break;
	  
	default: /* Seuil pas encore géré */
	  AVERTISSEMENT( "Seuil inconnu ou non-géré [%s]", TH_TexteSeuil( seuil ) );
	  return -5;
	}
      }
    } else {
      if ( Z < Z_min ) {
	(void) fprintf(stderr, "%s,%s l.%u : Pas de seuil %s pour l'élément %s, section efficace nulle\n",
		       __FILE__, __FUNCTION__, __LINE__, TH_TexteSeuil( seuil ), ATM_Z_vers_symbole( Z ));
      } else {
	for ( i = 0 ; i < 6 ; i++ ) {
	  c[ i ] = (double) Ebel_ParSeuil[ seuil ][ 6 * (Z - Z_min) + i ];
	}
	if ( h_seuil != NULL ) {
	  *h_seuil = TH_Table_Lire( Ebel_Sauts[ seuil ], Z, Z_min, EBEL_Z_MAX );
	}
      }
    }
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Calculer le coefficient d'absorption selon la méthode d'Ebel --------
    Renvoie la valeur du coefficient d'absorption pour Z, énergie indiqués, en cm^2/g
            < 0 en cas d'erreur

    energie : l'énergie pour laquelle on veut faire le calcul [en eV]
    Z       : numéro atomique de l'élément considéré
    c       : les 6 coefficients à utiliser
 */

double TH_Ebel(double energie, int Z __attribute__((__unused__)), double c[ 6 ])
{
  double mu = 0.0, ln_E;
  int i;

  if ( energie <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Énergie négative [%g], calcul impossible\n",
		   __FILE__, __FUNCTION__, __LINE__, energie);
    return -1.0;
  }

  energie /= 1000.0; /* Coefficients pour une énergie en keV */
  ln_E = log( energie );

  /* Calcul du polynome */
  for ( i = 5 ; i > 0 ; i-- ) {
    mu += c[ i ];
    mu *= ln_E;
  }
  mu += c[ 0 ];

  /* Calcul du coefficient final */
  mu = exp( mu );

  /* Calcul terminé */
  return mu;
}
