/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'un spectre XAS par transmission
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "extrait.h" /* Pour MU_Transmission */
#include "experience.h"
#include "graphes.h"

#ifdef _DEVELOPPE_
# include "images.h" /* Pour suivre les coupes de faisceau */
#endif

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int XAS_Trans_I0Ct_PrjCte(Experience *experience, Graphe *grf_i0, Graphe *grf_i1);
static int XAS_Trans_PrjCte     (Experience *experience, Graphe *grf_i0, Graphe *grf_i1);
static int XAS_TransQuelconque (Experience *experience, Graphe *grf_i0, Graphe *grf_i1);

static double TRANS_CalculerSignal(Experience *experience, Faisceau *faiceau, double energie, Faisceau *f_transmis);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Calcul d'un spectre transmis complet [appelable extrieurement] -------- */

Graphe *EXP_SpectreXAS_Transmission( Experience *experience )
{
  int retour = -1;
  boolean i0_constant = VRAI; /* Idalement, il n'y a pas besoin de recalculer I0  chaque pas */
  boolean proj_constante = VRAI; /* Mme chose, mais pour la projection du faisceau sur l'chantillon */
  Graphe *spectre, *grf_i0, *grf_i1;

  ENTREE_FONCTION("EXP_SpectreXAS_Transmission", "experience=%p", experience);

  /* Vrifications : juste pour viter un plantage, sinon suppos contrl avant avec EXP_ControlerOptions() */
  if ( NULL == experience ) {
    SORTIE_FONCTION("EXP_SpectreXAS_Transmission", "%p [ERREUR : experience==NULL]", NULL);
    return NULL;
  }
  if ( NULL == experience->i0 ) {
    SORTIE_FONCTION("EXP_SpectreXAS_Transmission", "%p [ERREUR : pas de dtecteur I0]", NULL);
    return NULL;
  }
  if ( NULL == experience->i1 ) {
    SORTIE_FONCTION("EXP_SpectreXAS_Transmission", "%p [ERREUR : pas de dtecteur I1]", NULL);
    return NULL;
  }

  /* Premire tape : construire les spectres */
  grf_i0 = GRF_nv_GrilleXP(NULL, GR_I0, GR_NORMAL, "Signal I0", FAUX,
			   experience->E_debut, experience->E_fin, experience->pas_E );
  if ( NULL == grf_i0 ) {
    (void) fprintf(stderr, "EXP_SpectreXAS_Transmission : ERREUR dans GRF_nv_GrilleXP() / graphe I0\n");
    SORTIE_FONCTION("EXP_SpectreXAS_Transmission", "%p [ERREUR : grf_i0]", NULL);
    return NULL;
  }
  grf_i1 = GRF_nv_GrilleXP(grf_i0, GR_I1, GR_NORMAL, "Signal I1", FAUX,
			   experience->E_debut, experience->E_fin, experience->pas_E );
  if ( NULL == grf_i1 ) {
    (void) fprintf(stderr, "EXP_SpectreXAS_Transmission : ERREUR dans GRF_nv_GrilleXP() / graphe I1\n");
    GRF_efface( grf_i0, VRAI, VRAI );
    SORTIE_FONCTION("EXP_SpectreXAS_Transmission", "%p [ERREUR : grf_i1]", NULL);
    return NULL;
  }

  /* Faut-il recalculer le i0 en chaque point ? */
  if ( experience->faisceau->proprietes & FAISCEAU_DECROIT ) {
    i0_constant = FAUX; /* Il faut, car l'intensit du faisceau change avec l'nergie */
    proj_constante = FAUX; /* La projection du faisceau sur l'chantillon est donc modifie */
  }
  if ( experience->i0->proprietes & DETECTEUR_ABSORBANT ) {
    i0_constant = FAUX; /* Il faut, car le dtecteur ne rpond alors pas pareil pour chaque nergie */
    proj_constante = FAUX; /* La projection du faisceau sur l'chantillon est donc modifie */
  }
  if ( experience->i0->proprietes & DETECTEUR_NRJ_VARIABLE ) {
    i0_constant = FAUX; /* Il faut, car le dtecteur ne rpond pas pareil pour chaque nergie */
  }

  if ( ( VRAI == i0_constant ) && ( VRAI == proj_constante ) ) {
    retour = XAS_Trans_I0Ct_PrjCte( experience, grf_i0, grf_i1 );
    /* On calcule le signal sur I0, et son effet sur le faisceau */
  } else if ( VRAI == proj_constante ) {
    retour = XAS_Trans_PrjCte( experience, grf_i0, grf_i1 );
  } else {
    retour = XAS_TransQuelconque( experience, grf_i0, grf_i1 );
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le calcul du signal\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Maintenant que l'on connat I0 et I1, on peut en dduire le mu observ */
  spectre = MU_Transmission( grf_i0, grf_i1 ); /* S'occupe de tout, y compris les erreurs si ncessaire ! */
  if ( NULL == spectre ) {
    (void) fprintf(stderr, "EXP_SpectreXAS_Transmission : ERREUR dans MU_Transmission \n");
    spectre = grf_i1; /* On renvoie alors I1 tout simplement... */
  }

  /* On a fini, tout s'est bien pass */
  SORTIE_FONCTION("EXP_SpectreXAS_Transmission", "%p", spectre);
  return spectre;
}

/* -------- Cas le plus simple : calcul en transmission
 * Le courant I0 ne varie pas avec l'nergie
 * La projection du faisceau sur l'chantillon ne varie pas avec l'nergie
 * ------------------- */

static int XAS_Trans_I0Ct_PrjCte(Experience *experience, Graphe *grf_i0, Graphe *grf_i1)
{
  Faisceau *f_echantillon, *f_apres_i0, *f_transmis;
  double i0, *energie, *i1;
  unsigned long i;

#ifdef _DEVELOPPE_
  double i0p, i0pp;
#endif

  ENTREE_FONCTION( __FUNCTION__ , "experience=%p grf_i0=%p grf_i1=%p", experience, grf_i0, grf_i1);

  /* On calcule I0 */
  f_apres_i0 = FAISCEAU_Copier( experience->faisceau );
  if ( NULL == f_apres_i0 ) {
    (void) fprintf(stderr, "XAS_Trans_I0Ct_PrjCte : impossible de crer le faisceau aprs I0\n");
    SORTIE_FONCTION(__FUNCTION__, "%d [ERREUR faisceau aprs I0]", -1);
    return -1;
  }
  DANS_FONCTION("Calcul du I0 commun", 0);
  i0 = DETECTEUR_FaisceauTraverse_I(experience->i0, experience->faisceau, grf_i0->x[ 0 ], f_apres_i0 );
#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "%s,%s l.%u : i0=%g, aprs I=%g [g_E=%p p=%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, i0, f_apres_i0->intensite, (void *) f_apres_i0->g_E, f_apres_i0->proprietes);
  GRF_titre( f_apres_i0->g_E, "g(E) aprs I0");
  IMG_Titre( f_apres_i0->g_espace, "Coupe aprs I0");
#endif

  /* Valable  toutes les nergies */
  for ( i = 0 ; i < grf_i0->nombre_point ; i++ ) {
    grf_i0->y[ i ] = i0;
  }
  grf_i0->y_min = i0; grf_i0->y_max = i0;

  /* Et on projette cela sur l'chantillon */
  DANS_FONCTION("Projection sur l'chantillon", 0);
  f_echantillon = FAISCEAU_ProjeterSurEchantillon( f_apres_i0, experience->echantillon );
  if ( f_echantillon != f_apres_i0 ) { /* On n'a plus besoin de lui */
    FAISCEAU_Effacer( f_apres_i0 );
  }
  if ( NULL == f_echantillon ) {
    (void) fprintf(stderr, "XAS_Trans_I0Cte_PrjCte : ERREUR dans FAISCEAU_ProjeterSurEchantillon()\n");
    SORTIE_FONCTION(__FUNCTION__, "%d [ERREUR faisceau avant chantillon]", -2);
    return -2;
  }

#ifdef _DEVELOPPE_
  if ( f_echantillon != experience->faisceau ) {
    GRF_titre( f_echantillon->g_E, "g(E) avant chantillon");
    IMG_Titre( f_echantillon->g_espace, "Coupe avant chantillon");
/*     IMG_bornes( f_echantillon->g_espace ); */
  }

  i0p = DETECTEUR_FaisceauTraverse_I(experience->i0, f_echantillon, grf_i0->x[ 0 ], NULL );
#endif

  /* On en dduit, ensuite, le "squelette" du faisceau aprs l'chantillon */
  DANS_FONCTION("Projection aprs l'chantillon", 0);
  f_transmis = FAISCEAU_GeometrieApresEchantillon( f_echantillon, experience->echantillon );
  if ( NULL == f_transmis ) {
    (void) fprintf(stderr, "XAS_Trans_I0Ct_PrjCte : ERREUR dans FAISCEAU_GeometrieApresEchantillon\n");
    if ( f_echantillon != experience->faisceau ) { /* Il devient donc inutile ! */
      FAISCEAU_Effacer( f_echantillon );
    }
    SORTIE_FONCTION(__FUNCTION__, "%d [ERREUR faisceau aprs chantillon]", -3);
    return -3;
  }
#ifdef _DEVELOPPE_
  if ( f_echantillon != experience->faisceau ) {
    IMG_Titre( f_transmis->g_espace, "Coupe aprs chantillon");
    GRF_titre( f_transmis->g_E, "g(E) aprs chantillon");
  }
  i0pp = DETECTEUR_FaisceauTraverse_I(experience->i0, f_transmis, grf_i0->x[ 0 ], NULL );
  if ( (i0p != 0.0 ) && ( fabs( ( i0pp - i0p ) / i0p ) > 1.0e-8 ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Intensits des faisceaux diffrentes ! [%g - %g = %g / %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, i0pp, i0p, i0pp - i0p, ( i0pp - i0p ) / i0p );
    *(char *)NULL = 0;
  }
#endif

  /* On calcule le signal mesur par I1 pour chaque nergie */
  DANS_FONCTION("Calcul pour chaque nergie", 0);
  for ( i = 0, energie = grf_i1->x, i1 = grf_i1->y ; i < grf_i1->nombre_point ; i++, energie++, i1++ ) {
    *i1 = TRANS_CalculerSignal( experience, f_echantillon, *energie, f_transmis );
#ifdef _DEVELOPPE_
    printf("%s : E=%g I0=%g [%g : %g] I1=%g\n", __FUNCTION__, *energie, i0, i0p, i0pp, *i1);
#endif
  }

  /* Bornes... */
  GRF_bornes( grf_i1, NE_PAS_FAIRE, DESORDONNE );

  /* On a fini, tout s'est bien pass */
  SORTIE_FONCTION(__FUNCTION__, "%d", 0);
  return 0;
}

/* -------- Cas intermdiaire : I0 change, mais pas le faisceau sur l'chantillon -------- */

static int XAS_Trans_PrjCte(Experience *experience, Graphe *grf_i0, Graphe *grf_i1)
{
  Faisceau *f_echantillon, *f_apres_i0, *f_transmis;
  double *energie, *i1, *i0;
  unsigned long i;

  /* Prparation des faisceaux intermdiaires */
  /* 1. En sortie de I0 */
  f_apres_i0 = FAISCEAU_Copier( experience->faisceau );
  if ( NULL == f_apres_i0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le faisceau aprs I0\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* 2. Sur l'chantillon */
  f_echantillon = FAISCEAU_ProjeterSurEchantillon( f_apres_i0, experience->echantillon );
  if ( f_echantillon != f_apres_i0 ) {
    FAISCEAU_Effacer( f_apres_i0 ); /* Plus besoin de lui */
  }
  if ( NULL == f_echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_ProjeterSurEchantillon()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* 3. Aprs l'chantillon */
  f_transmis = FAISCEAU_GeometrieApresEchantillon( f_echantillon, experience->echantillon );
  if ( NULL == f_transmis ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_GeometrieApresEchantillon()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( f_echantillon );
    return -3;
  }

  /* En chaque point du spectre... */
  for ( i = 0, energie = grf_i1->x, i0 = grf_i0->y, i1 = grf_i1->y ; i < grf_i1->nombre_point ; i++, energie++, i0++, i1++ ) {
    /* Signal avant l'chantillon */
    *i0 = DETECTEUR_FaisceauTraverse_I(experience->i0, experience->faisceau, *energie, NULL );

    /* Signal aprs l'chantillon */
    *i1 = TRANS_CalculerSignal( experience, f_echantillon, *energie, f_transmis );

#ifdef _DEVELOPPE_
    printf("%s : E=%g I0=%g I1=%g\n", __FUNCTION__, *energie, *i0, *i1);
#endif
  }

  /* Plus besoin d'eux */
  if ( f_transmis != f_echantillon )
    FAISCEAU_Effacer( f_transmis );
  FAISCEAU_Effacer( f_echantillon );

  /* Tout s'est bien pass */
  GRF_bornes( grf_i0, NE_PAS_FAIRE, DESORDONNE );
  GRF_bornes( grf_i1, NE_PAS_FAIRE, DESORDONNE );
  return 0;
}

/* -------- Cas le plus gnral : il faut tout refaire  chaque fois -------- */

static int XAS_TransQuelconque(Experience *experience, Graphe *grf_i0, Graphe *grf_i1)
{
  Faisceau *f_echantillon, *f_apres_i0, *f_transmis = NULL;
  double *energie, *i1, *i0;
  unsigned long i;

  /* Prparation des faisceaux intermdiaires */
  /* 1. En sortie de I0 */
  f_apres_i0 = FAISCEAU_Copier( experience->faisceau );
  if ( NULL == f_apres_i0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le faisceau aprs I0\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* En chaque point du spectre... */
  for ( i = 0, energie = grf_i1->x, i0 = grf_i0->y, i1 = grf_i1->y ; i < grf_i1->nombre_point ; i++, energie++, i0++, i1++ ) {
    /* Signal avant l'chantillon */
    *i0 = DETECTEUR_FaisceauTraverse_I(experience->i0, experience->faisceau, *energie, f_apres_i0 );

    /* Et on projette cela sur l'chantillon */
    f_echantillon = FAISCEAU_ProjeterSurEchantillon( f_apres_i0, experience->echantillon );
    if ( NULL == f_echantillon ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_ProjeterSurEchantillon()\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -2;
    }
    if ( NULL == f_transmis ) {
      f_transmis = FAISCEAU_GeometrieApresEchantillon( f_echantillon, experience->echantillon );
      if ( NULL == f_transmis ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_GeometrieApresEchantillon()\n",
		       __FILE__, __FUNCTION__, __LINE__);
	if ( f_echantillon != f_apres_i0)
	  FAISCEAU_Effacer( f_echantillon );
	FAISCEAU_Effacer( f_apres_i0 );
	return -3;
      }
    }

    /* Signal aprs l'chantillon */
    *i1 = TRANS_CalculerSignal( experience, f_echantillon, *energie, f_transmis );

    /* On efface pour le recrer correctement  l'nergie suivante... */
    if ( f_echantillon != f_apres_i0)
      FAISCEAU_Effacer( f_echantillon );

#ifdef _DEVELOPPE_
    printf("%s : E=%g I0=%g I1=%g\n", __FUNCTION__, *energie, *i0, *i1);
#endif
  }

  /* Plus besoin d'eux */
  FAISCEAU_Effacer( f_apres_i0 );

  /* Tout s'est bien pass */
  GRF_bornes( grf_i0, NE_PAS_FAIRE, DESORDONNE );
  GRF_bornes( grf_i1, NE_PAS_FAIRE, DESORDONNE );
  return 0;
}

/* -------- Calculer le signal de transmission (faisceau ponctuel) -------- */

static double TRANS_CalculerSignal(Experience *experience, Faisceau *faisceau, double energie, Faisceau *f_transmis)
{
  int retour;
  double i1;

  ENTREE_FONCTION("TRANS_CalculerSignal", "experience=%p faisceau=%p energie=%g f_transmis=%p", experience, faisceau, energie, f_transmis);

  /* On calcule le faisceau en sortie d'chantillon */
  retour = FAISCEAU_TraverserEchantillon( faisceau, experience->echantillon, energie, f_transmis );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "TRANS_CalculerSignal : ERREUR dans FAISCEAU_TraverserEchantillon\n");
    SORTIE_FONCTION("TRANS_CalculerSignal", "%g [ERREUR]", -1.0);
    return -1.0;
  }

  /* Ici, f_transmis contient donc le faisceau en sortie immdiate de l'chantillon */

  /* ---- A FAIRE : Exprience Absorbante --- */
  if ( experience->proprietes & EXPERIENCE_ABSORBANTE ) {
    printf("TRANS_CalculerSignal :  FAIRE, CAS D'UNE EXPRIENCE ABSORBANTE\n");
  }

  /* Quel signal obtient-t-on dans le dtecteur i1 ? */
  i1 = DETECTEUR_FaisceauTraverse_I( experience->i1, f_transmis, energie, NULL );
#ifdef _DEVELOPPE_
  if ( is_inf( i1 ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : INTENSIT INFINIE [E=%g eV]\n",
		   __FILE__, __FUNCTION__, __LINE__, energie);
    abort();
  }
#endif

  if ( experience->ordre > 0 ) {
    (void) fprintf(stderr, "TRANS_CalculerSignal :  FAIRE, corrections de fluorescence vers I0 et I1\n");
  }

  /* On a fini, tout s'est bien pass */
  SORTIE_FONCTION("TRANS_CalculerSignal", "%g", i1);
  return i1;
}
