/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 1998 - 2003

   Reconstruction du coefficient d'absorption atomique global, mu(0, total)
   Principe : on s'intresse au seuil S de l'atome A, dans une matrice autre. Alors

                                                       mu(0, total)
                                                   /-------------------\
   mu = mu(autres atomes) + mu(A, autres seuils) + mu(A,S) * ( 1 + khi )
        \______________________________________/   \_____/         \_/
                 mu(matrice), continu               mu(0)          XAS
                   non considr ici             intrinsque

   De plus, on suppose que le mu(0) intrinsque peut s'crire :
                                            N
   mu(A,S) = mu(seuil) + mu(apres_seuil) +  S mu(raie_i)
                                           i=1
   avec
 - mu(seuil)       = marche d'escalier [idale], Arc tg,... [0 avant seuil, Cte aprs, avec transition plus ou moins douce]
 - mu(apres_seuil) = correction pour avoir la dcroissance aprs seuil
 - mu(raie_i)      = absorption due aux transitions vers les tats lis
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "chimie_utils.h"
#include "absorption.h"
#include "extrait.h" /* Constantes : types de mu(0) */
#include "graphes.h"
#include "interface.h"
#include "lissage.h"
#include "science.h" /* Pour la conversion E <-> longueur d'onde */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Calculs ~-~-~-~-~-~-~-~-~ */

/* -------- Routine gnrale : calcule mu(apres_seuil) dans un graphe -------- */

int TH_CalculerMu0Intrinseque(Graphe *graphe, int modele, double e0, int Z, int seuil)
{
  double *x, *y;
  unsigned long i;

  if ( NULL == graphe ) return -1;

  if ( MODELE_MU0_CONSTANT == modele ) {
    /* Correspond  la hauteur de seuil => sera pris en compte plus tard */
    return 0;
  }
  if ( e0 <= 0.0 ) {
    e0 = TH_DonneesSeuil( Z, seuil, NULL, NULL );
  }

  /* Calcul en chaque point du spectre */
  for ( i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++ ) {
    if ( *x < e0 ) {
      *y = 0.0; /* Le seuil ne contribue pas avant le seuil ! */
    } else {
      *y = TH_CalculerMu0( Z, *x, modele, seuil, e0 );
    }
  }

  GRF_bornes( graphe, NE_PAS_FAIRE, DESORDONNE );

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Calcul de mu(apres_seuil)  une nergie donne pour un seuil unitaire ----------
   Renvoie la valeur de mu0 pour corriger la marche d'escalier horizontale.
   Pour avoir le mu0 total, il faut donc rajouter 1 (puisque le seuil est postul unitaire)
 */

double TH_CalculerMu0(int Z, double energie, int methode, int seuil, double e_seuil)
{
  double mu0_ap_seuil = 0.0, h_seuil = 1.0;

  if ( e_seuil <= 0.0 ) {
    e_seuil = TH_DonneesSeuil( Z, seuil, NULL, NULL );
    if ( e_seuil <= 0.0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : nergie de seuil inconnue [Z=%d, seuil=%d]\n",
		     __FILE__, __FUNCTION__, __LINE__, Z, seuil);
      return 0.0;
    }
  }

  if ( energie < e_seuil ) return 0.0; /* Pas d'absorption avant seuil, de toute faon */

  switch ( methode )
  {
  case MODELE_MU0_CONSTANT:
    mu0_ap_seuil = 1.0;
    break;

  case MODELE_MU0_LENGELER:
    mu0_ap_seuil = 11.0 / 3.0 - ( 8.0 / 3.0 ) * ( energie / e_seuil );
    (void) fprintf(stderr, "%s,%s l.%u : h(seuil=%g)=%g -- mu(%g)=%g\n",
		   __FILE__, __FUNCTION__, __LINE__, e_seuil, h_seuil, energie, mu0_ap_seuil);
    break;

  case MODELE_MU0_VICTOREEN:
    {
      double C_apres[ 2 ], C_avant[ 2 ];

      TH_coef_Victoreen( Z, seuil, C_avant, C_avant + 1 );
      TH_coef_Victoreen( Z, seuil - 1, C_apres, C_apres + 1 );

      h_seuil  = TH_Victoreen( e_seuil, Z, C_apres[ 0 ], C_apres[ 1 ] );
      h_seuil -= TH_Victoreen( e_seuil, Z, C_avant[ 0 ], C_avant[ 1 ] );

      mu0_ap_seuil  = TH_Victoreen( energie, Z, C_apres[ 0 ], C_apres[ 1 ] );
      mu0_ap_seuil -= TH_Victoreen( energie, Z, C_avant[ 0 ], C_avant[ 1 ] );
    }
    break;

  case MODELE_MU0_MCMASTER:
    {
      double A_apres[ 4 ], A_avant[ 4 ];

      TH_coef_McMaster( Z, seuil, A_avant, A_avant + 1, A_avant + 2, A_avant + 3 );
      TH_coef_McMaster( Z, seuil - 1, A_apres, A_apres + 1, A_apres + 2, A_apres + 3 );

      h_seuil  = TH_McMaster( e_seuil, Z, A_apres[ 0 ], A_apres[ 1 ], A_apres[ 2 ], A_apres[ 3 ] );
      h_seuil -= TH_McMaster( e_seuil, Z, A_avant[ 0 ], A_avant[ 1 ], A_avant[ 2 ], A_avant[ 3 ] );

      mu0_ap_seuil  = TH_McMaster( energie, Z, A_apres[ 0 ], A_apres[ 1 ], A_apres[ 2 ], A_apres[ 3 ] );
      mu0_ap_seuil -= TH_McMaster( energie, Z, A_avant[ 0 ], A_avant[ 1 ], A_avant[ 2 ], A_avant[ 3 ] );
    }
    break;

  case MODELE_MU0_EBEL:
    {
      double c[ 6 ];

      TH_coef_Ebel( Z, seuil, FAUX, c, NULL );

      h_seuil = TH_Ebel( e_seuil, Z, c );

      mu0_ap_seuil = TH_Ebel( energie, Z, c );
    }
    break;

  case MODELE_MU0_HEITLER:
    {
      double ln_mu0, racine;

      energie = ( energie - e_seuil ) / e_seuil;
      racine = sqrt( energie );

      ln_mu0  = 1.0 - atan( racine ) - log( 1.0 + energie );
      ln_mu0 *= 4.0;
      ln_mu0 -= log( 1.0 - exp( -2.0 * M_PI / racine ) );

      mu0_ap_seuil = exp( ln_mu0 );
    }
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Mthode inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, methode);
    mu0_ap_seuil = 0.0;
  }

  if ( h_seuil != 0.0 ) {
    mu0_ap_seuil /= h_seuil; /* On se ramne  un seuil de 1 */
  }
  mu0_ap_seuil -= 1.0; /* Et on ne garde que le terme correctif par rapport  la marche d'escalier */

  return mu0_ap_seuil;
}

/* -------- Combiner l'EXAFS et mu(A,S) pour obtenir mu(0, total) --------
   S'il y a de l'exafs, il est dj dans graphe
   On utilise les X de graphes, le rsultat est dans graphe
   Les deux graphes DOIVENT avoir les MMES X
 */

int TH_CalculerMu0Total(struct graphe *graphe, struct graphe *mu0_intrinseque)
{
  unsigned long i;
  double *y;
  Graphe *retour;

  if ( ( NULL == graphe ) || ( NULL == mu0_intrinseque ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : L'un des graphes est NULL [graphe=%p, mu0_i=%p]\n",
		   __FILE__, __FUNCTION__, __LINE__, (void *) graphe, (void *) mu0_intrinseque);
    return -1;
  }

  /* On commence par ajouter 1  l'EXAFS : on a (1 + khi(E)) dans graphe
     Ne modifie pas les barres d'erreur */
  for ( i = 0, y = graphe->y ; i < graphe->nombre_point ; i++, y++ ) {
    *y += 1.0;
  }
  graphe->y_min += 1.0;
  graphe->y_max += 1.0;

  /* On fait ensuite le produit [s'occupe des erreurs] */
  retour = GRF_ProduitGraphes( graphe, mu0_intrinseque, VRAI, VRAI, NULL );
  if ( NULL == retour ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_ProduitGraphes()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Interface utilisateur ~-~-~-~-~-~-~-~-~ */
/* -------- Construire un mu0 complet, avec les etapes successives --------
 * Renvoie un graphe ISOLE cree pour l'occasion qui contient le resultat
 *
 *  Z : numero atomique de l'element dont on calcule le seuil
 * e0 : energie de seuil [en eV]
 */

Graphe *TH_ConstruireMu0(double e_min, double e_max, double pas, struct infos_construction_mu0 *parametres)
{
  Graphe *mu0_total = NULL, *mu0 = NULL;
  int retour;

  if ( NULL == parametres ) {
    (void) fprintf( stderr, "%s : pas de parametres indiques ! Calcul impossible...\n",
		    __FUNCTION__ );
    return NULL;
  }

  /* Prparation du graphe rsultat */
  mu0_total = GRF_nv_GrilleXP( NULL, GR_MOINSMAT, GR_NORMAL, "[TH_ConstruireMu0] Graphe resultat", VRAI, e_min, e_max, pas );
  if ( NULL == mu0_total ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe rsultat\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Calcul effectif
     1. Prparation des oscillations EXAFS */
  if ( parametres->options & TH_CALCUL_MU0_AVEC_EXAFS ) {
    Graphe *exafs = NULL;

    exafs = GRF_Copier( parametres->exafs.signal_exafs, NULL, VRAI );
    if ( NULL == exafs ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de prparer les oscillations EXAFS\n",
		     __FILE__, __FUNCTION__, __LINE__);
      GRF_eff_isole( mu0_total );
      return NULL;
    }

    /* On supprime la pondration */
    retour = EXF_k_p_n( exafs, -(parametres->exafs.ponderation), FAUX );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Problme dans EXF_k_p_n() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      if ( retour < 0 ) {
	GRF_eff_isole( exafs );
	GRF_eff_isole( mu0_total );
	return NULL; /* Erreur grave */
      }
    }

    /* On convertit dans l'espace des E */
    retour = EXF_k_v_e( exafs, parametres->seuil.energie, FAUX );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Problme dans EXF_k_v_e() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      if ( retour < 0 ) {
	GRF_eff_isole( exafs );
	GRF_eff_isole( mu0_total );
	return NULL; /* Erreur grave */
      }
    }

    /* Et on le recale dans le graphe rsultat */
    retour = LIS_RecalerGraphe( exafs, mu0_total, INTERPOLE_LINEAIRE, INTERPOLE_BORDS_0_Y );

    /* Plus besoin de lui, donc ! */
    GRF_eff_isole( exafs );
  }

  /* 2. Prparation de mu(apres_seuil) */
  if ( parametres->options & TH_CALCUL_MU0_AVEC_MU0 ) {
    if ( VRAI == OBJ_choisit( SIMULER_MU, SM_MU0_THEORIQUE, QUESTION ) ) {
      mu0 = GRF_nv_GrilleXP( NULL, GR_MU0, GR_NORMAL, "[TH_ConstruireMu0] : le mu0 simule", VRAI, e_min, e_max, pas );
      if ( NULL == mu0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le coefficient intrinsque aprs seuil\n",
		       __FILE__, __FUNCTION__, __LINE__);
	GRF_bornes( mu0_total, NE_PAS_FAIRE, DESORDONNE );
	return mu0_total;
      }

      /* On calcule pour le modle prcis */
      retour = TH_CalculerMu0Intrinseque( mu0, parametres->mu0.modele_theorique, parametres->seuil.energie, parametres->Z, parametres->seuil.type );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TH_CalculerMu0() [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	GRF_eff_isole( mu0 );
	GRF_eff_isole( mu0_total );
	return NULL;
      }

      /* Et on adapte  la hauteur de seuil indique */
      GRF_FacteurMultiplicatif( mu0, parametres->seuil.hauteur, AXE_Y, NULL );
    } else {
      Graphe *mu0_exp;
      double h_seuil, *x, *y;
      int code_erreur;
      unsigned long i;

      mu0_exp = GRF_Copier( parametres->mu0.mu0_exp , NULL, VRAI );
      if ( NULL == mu0_exp ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de prparer le coefficient intrinsque aprs seuil\n",
		       __FILE__, __FUNCTION__, __LINE__);
	GRF_bornes( mu0_total, NE_PAS_FAIRE, DESORDONNE );
	return mu0_total;
      }

      /* On supprime la pondration eventuelle ici aussi */
      retour = EXF_k_p_n( mu0_exp, -( parametres->mu0.ponderation ), FAUX );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Problme dans EXF_k_p_n() [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	if ( retour < 0 ) {
	  GRF_eff_isole( mu0_exp );
	  GRF_eff_isole( mu0_total );
	  return NULL; /* Erreur grave */
	}
      }

      /* On suppose que ce mu0 est dans l'espace des k, on convertit donc dans l'espace des E. */
      retour = EXF_k_v_e( mu0_exp, parametres->seuil.energie, FAUX );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Problme dans EXF_k_v_e() [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	if ( retour < 0 ) {
	  GRF_eff_isole( mu0_exp );
	  GRF_eff_isole( mu0_total );
	  return NULL; /* Erreur grave */
	}
      }

      /* Valeur du saut pour le seuil, dans le graphe indiqu */
      h_seuil = GRF_LireValeur( mu0_exp, parametres->seuil.energie, INTERPOLE_LINEAIRE );
      if ( h_seuil <= 0.0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Hauteur de seuil absurde [%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, h_seuil);
      }

      /* On ne garde que la partie  correction  */
      for ( i = 0, x = mu0_exp->x, y = mu0_exp->y ; i < mu0_exp->nombre_point ; i++, x++, y++ ) {
	if ( *x < parametres->seuil.energie ) {
	  *y = 0.0;
	  /* Il faudrait aussi mettre  zro la barre d'erreur, si elle existe ! */
	} else {
	  *y -= h_seuil;
	  /* Il faudrait tenir compte de l'erreur ! */
	}
      }
      GRF_bornes( mu0_exp, NE_PAS_FAIRE, DESORDONNE );

      /* On passe pour la hauteur de seuil indique */
      if ( h_seuil != 0.0 ) {
	GRF_FacteurMultiplicatif( mu0_exp, parametres->seuil.hauteur / h_seuil, AXE_Y, NULL );
      } else {
	GRF_FacteurMultiplicatif( mu0_exp, parametres->seuil.hauteur, AXE_Y, NULL );
      }

      /* On case sur la mme grille que les X proposs */
      mu0 = GRF_CalerSurGraphe( mu0_exp, INTERPOLE_LINEAIRE, INTERPOLE_BORDS_0_Y, mu0_total, &code_erreur );
      GRF_eff_isole( mu0_exp );
      if ( NULL == mu0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_CalerSurGraphe() [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, code_erreur);
	GRF_eff_isole( mu0_total );
	return NULL;
      }
    }
  }

  /* 3. Calcul de mu(A,S) proprement dit */
  if ( parametres->options & TH_CALCUL_MU0_AVEC_SEUIL ) {
    if ( NULL == mu0 ) { /* Il n'y a pas de correction aprs seuil */
      mu0 = GRF_nv_GrilleXP( NULL, GR_MU0, GR_NORMAL, NULL, VRAI, e_min, e_max, pas );
      if ( NULL == mu0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le coefficient intrinsque aprs seuil\n",
		       __FILE__, __FUNCTION__, __LINE__);
	GRF_eff_isole( mu0_total );
	return NULL;
      }
    }

    retour = TH_SimulerSeuil( mu0, parametres->seuil.energie, parametres->seuil.hauteur, parametres->seuil.largeur, parametres->seuil.methode );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TH_SimulerSeuil() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      GRF_eff_isole( mu0_total );
      GRF_eff_isole( mu0 );
      return NULL;
    }
  }

  /* 4. On combine tout a dans le rsultat */
  retour = TH_CalculerMu0Total( mu0_total, mu0 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TH_CalculerMu0Total() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* 5. On ajoute les raies avant seuil */
  if ( parametres->options & TH_CALCUL_MU0_AVEC_RAIES ) {
    OPTION_A_FAIRE( "Raies du prseuil  ajouter" );
  }

  GRF_eff_isole( mu0 );
  GRF_bornes( mu0_total, NE_PAS_FAIRE, DESORDONNE );
  return mu0_total;
}

const char *TH_TexteModeleMu0(int modele)
{
  switch ( modele )
  {
  case MODELE_MU0_CONSTANT:
    return "Constant";

  case MODELE_MU0_LENGELER:
    return "Lengeler-Eisenberger";

  case MODELE_MU0_VICTOREEN:
    return "Victorine";

  case MODELE_MU0_MCMASTER:
    return "McMaster";

  case MODELE_MU0_HEITLER:
    return "Heitler";

  case MODELE_MU0_EBEL:
    return "Ebel";

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Modle inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, modele);
    return "Inconnu";
  }
}
