/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'un modle tridimensionnel d'chantillon
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

/* #include "graphe.h" */
#include "echantillon.h"
#include "images.h"
#include "lissage.h" /* Pour GRF_Lirevaleur & les interpolations 3D */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static Echantillon *ECH_ReserverEchantillon( int type );
static int          ECH_InitialiserValeurs ( Echantillon *echantillon );

static int ECH_CopierContenu  (Echantillon *original, Echantillon *copie);
static int ECH_EchangerContenu(Echantillon *ech_1, Echantillon *ech_2);

inline static double ECH_MuTotal(struct graphe *mu, struct graphe *te, struct graphe *ti, double energie, double *mu_photoelectrique);

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* -------- Crer un chantillon,  partir du pas de la grille --------
   Renvoie le pointeur sur l'chantillon cr, si tout va bien
           NULL en cas d'erreur

   larg, haut, epaisseur : les dimensions de l'chantillon (rectangulaire) en MICRONS
   pas_l, pas_h, pas_e   : taille de subdivision de l'chantillon en MICRONS aussi
   proprietes            : caractristiques de l'chantillon [cf. echantillon.h]
 */

Echantillon *ECH_CreerEchantillonP(double larg, double haut, double epaisseur, double pas_l, double pas_h, double pas_e,
				   int type, unsigned int proprietes)
{
  unsigned int n_h, n_l, n_e;

  /* Vrifications prliminaires */
  if ( ( larg <= 0.0 ) || ( haut <= 0.0 ) || ( epaisseur <= 0.0 ) ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonP : Dimensions absurdes [l=%g, h=%g, e=%g]\n", larg, haut, epaisseur);
    return NULL;
  }
  if ( ( pas_l <= 0.0 ) || ( pas_h <= 0.0 ) || ( pas_e <= 0.0 ) ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonP : Pas absurdes [dl=%g, dh=%g, de=%g]\n", pas_l, pas_h, pas_e);
    return NULL;
  }
  if ( ( pas_l >= larg ) || ( pas_h >= haut ) || ( pas_e >= epaisseur ) ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonP : Pas trop grand [dl=%g, dh=%g, de=%g]\n", pas_l, pas_h, pas_e);
    (void) fprintf(stderr, "                                       [ l=%g,  h=%g,  e=%g]\n", larg, haut, epaisseur);
    return NULL;
  }

  /* On en dduit le nombre de cases dans chaque direction */
  n_l = 1 + (unsigned int) ( larg / pas_l );
  n_h = 1 + (unsigned int) ( haut / pas_h );
  n_e = 1 + (unsigned int) ( epaisseur / pas_e );

  /* On cre effectivement l'chantillon */
  return ECH_CreerEchantillonN( larg, haut, epaisseur, n_l, n_h, n_e, type, proprietes );
}

/* -------- Crer un chantillon,  partir du nombre de cases dans chaque direction --------
   Renvoie le pointeur sur l'chantillon cr, si tout va bien
           NULL en cas d'erreur

   larg, haut, epaisseur : les dimensions de l'chantillon (rectangulaire) en MICRONS
   n_l, n_h, n_e         : nombre de cases  raliser dans chaque direction
   proprietes            : caractristiques de l'chantillon [cf. echantillon.h]
 */

Echantillon *ECH_CreerEchantillonN(double larg, double haut, double epaisseur,
				   unsigned int n_l, unsigned int n_h, unsigned int n_e,
				   int type, unsigned int proprietes)
{
  int retour;
  Echantillon *echantillon;

  if ( ( larg <= 0.0 ) || ( haut <= 0.0 ) || ( epaisseur <= 0.0 ) ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonN : Dimensions absurdes [l=%g, h=%g, e=%g]\n", larg, haut, epaisseur);
    return NULL;
  }
  if ( ( 1 >= n_l ) || ( 1 >= n_h ) || ( 1 >= n_e ) ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonN : grille incorrecte [nl=%u, nh=%u, ne=%u]\n", n_l, n_h, n_e);
    return NULL;
  }

  /* On rserve l'chantillon */
  echantillon = ECH_ReserverEchantillon( type );
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonN : ERREUR dans ECH_ReserverEchantillon()\n");
    return NULL;
  }

  /* On remplit sa gomtrie et ses proprits */
  echantillon->larg = larg;
  echantillon->haut = haut;
  echantillon->epaisseur = epaisseur;

  echantillon->nbr_x = n_l; /* Largeur   : axe des X */
  echantillon->nbr_y = n_e; /* paisseur : axe des Y */
  echantillon->nbr_z = n_h; /* Hauteur   : axe des Z */

  echantillon->pas_x = larg / (double) ( n_l - 1 );
  echantillon->pas_y = epaisseur / (double) ( n_e - 1 );
  echantillon->pas_z = haut / (double) ( n_h - 1 );

  echantillon->proprietes = proprietes;

  /* On initialise les valeurs des cases */
  retour = ECH_InitialiserValeurs( echantillon );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "ECH_CreerEchantillonN : ERREUR dans ECH_InitialiserValeurs()\n");
    free( echantillon );
    return NULL;
  }

  /* Tout s'est bien pass */
  return echantillon;
}

/* -------- Rservation proprement dite -------- */

static Echantillon *ECH_ReserverEchantillon( int type )
{
  Echantillon *echantillon;

  /* Rservation */
  echantillon = ( Echantillon * ) calloc( 1, sizeof( Echantillon ) );
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "ECH_ReserverEchantillon : Mmoire insuffisante\n");
    return NULL;
  }

  /* Prinitialisation : aucune information dans l'chantillon
     calloc() met tout  0, donc rien de spcial  faire... */
  echantillon->type = type;
  switch ( type )
  {
  case ECHANTILLON_MATERIEL:
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      echantillon->valeurs.homogene.composition = NULL;
      echantillon->valeurs.homogene.concentration = NULL;
    } else {
      echantillon->valeurs.inhomogene = NULL;
    }
    break;

  case ECHANTILLON_GRADIENT:
    echantillon->valeurs.homogene.composition = NULL;
    echantillon->valeurs.homogene.concentration = NULL;
    break;

  case ECHANTILLON_SPECTRE:
    echantillon->valeurs.spectre = NULL;
    break;

  default:
    (void) fprintf(stderr, "ECH_ReserverEchantillon : type d'chantillon inconnu [%d]\n", type);
    free (echantillon);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return NULL;
  }

  /* Fini, tout va bien */
  return echantillon;
}

/* -------- Initialiser la table des valeurs, en fonction du type d'chantillon -------- */
 
static int ECH_InitialiserValeurs(Echantillon *echantillon)
{
  unsigned long nbr_cases;

  /* Combien de cases au total ? */
  nbr_cases = echantillon->nbr_x * echantillon->nbr_y * (unsigned long) echantillon->nbr_z;
#ifdef _DEVELOPPE_
  if ( nbr_cases > 100000 ) {
    printf("ECH_InitialiserValeurs : chantillon avec beaucoup de cases (%lu), calculs longs...\n", nbr_cases);
  }
#endif

  switch ( echantillon->type )
  {
  case ECHANTILLON_MATERIEL:
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) { /* Une seule valeur de composition suffit ! */
      echantillon->valeurs.homogene.composition = NULL;
      echantillon->valeurs.homogene.concentration = NULL;
    } else {
      echantillon->valeurs.inhomogene = (struct matiere **) calloc( nbr_cases, sizeof(struct matiere *) );
      if ( NULL == echantillon->valeurs.inhomogene ) {
	(void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante / chantillon matriel inhomogne\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -5;
      }
    }
    break;

  case ECHANTILLON_GRADIENT:
    echantillon->valeurs.homogene.composition = NULL;
    echantillon->valeurs.homogene.concentration = (double *) calloc( nbr_cases, sizeof( double ) );
    if ( NULL == echantillon->valeurs.homogene.concentration ) {
      (void) fprintf(stderr, "ECH_InitialiserValeurs : Impossible de rserver la table des pondrations\n");
      return -3;
    } else {
      unsigned long i;

      /* Par dfaut, on fait comme si on tait homogne... */
      for ( i = 0 ; i < nbr_cases ; i++ ) {
	echantillon->valeurs.homogene.concentration[ i ] = 1.0;
      }
    }
    break;

  case ECHANTILLON_SPECTRE:
    /* On a forcment un spectre par case, mme pour un chantillon homogne... */
    echantillon->valeurs.spectre = (struct spectre_local **) calloc( nbr_cases, sizeof(struct spectre_local *) );
    if ( NULL == echantillon->valeurs.spectre ) {
      (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante / chantillon spectral\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -5;
    }
    break;

  default:
    (void) fprintf(stderr, "ECH_InitialiserValeurs : type d'chantillon inconnu [%d]\n", echantillon->type);
    return -1;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Redimensionner un chantillon, en changeant le pas
   ==> le contenu en commun peut tre perdu
   ------------------------------------------------------------ */

int ECH_ChangerGrille(Echantillon *echantillon, double nv_largeur, double nv_hauteur, double nv_epaisseur,
		       double nvpas_l, double nvpas_h, double nvpas_e)
{
  Echantillon *nv_echantillon;
  int retour;

  if ( NULL == echantillon ) return -1;

  /* On cre un nouvel chantillon avec le mme pas */
  nv_echantillon = ECH_CreerEchantillonP( nv_largeur, nv_hauteur, nv_epaisseur, nvpas_l, nvpas_h, nvpas_e,
					  echantillon->type, echantillon->proprietes );
  if ( NULL == nv_echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le nouvel chantillon\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On copie le contenu, autant que possible */
  if ( ECHANTILLON_MATERIEL == echantillon->type ) {
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      nv_echantillon->valeurs.homogene.composition = echantillon->valeurs.homogene.composition;
    } else {
      OPTION_A_FAIRE( "Adaptation de la composition au nouveau pas / chantillon inhomogne" );
    }
  } else {
    /* La composition est la mme */
    nv_echantillon->valeurs.homogene.composition = echantillon->valeurs.homogene.composition;

    OPTION_A_FAIRE( "Adaptation de la composition au nouveau pas / chantillon gradient" );
  }

  /* On installe les nouvelles valeurs  la place des anciennes */
  retour = ECH_EchangerContenu( echantillon, nv_echantillon );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'change des contenus [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    ECH_EffacerEchantillon( nv_echantillon, VRAI );
    return -3;
  }

  /* On efface l'chantillon temporaire, qui contient maintenant les valeurs de l'ancien */
  ECH_EffacerEchantillon( nv_echantillon, VRAI );

  /* Termin avec succs */
  return 0;
}

/* -------- Redimensionner un chantillon, sans changer le pas
   ==> le contenu en commun n'est pas perdu
   ------------------------------------------------------------ */

int ECH_Redimensionner(Echantillon *echantillon, double nv_largeur, double nv_hauteur, double nv_epaisseur)
{
  Echantillon *nv_echantillon;
  int retour;

  if ( NULL == echantillon ) return -1;

  /* On cre un nouvel chantillon avec le mme pas */
  nv_echantillon = ECH_CreerEchantillonP( nv_largeur, nv_hauteur, nv_epaisseur,
					  echantillon->pas_x, echantillon->pas_z, echantillon->pas_y,
					  echantillon->type, echantillon->proprietes );
  if ( NULL == nv_echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le nouvel chantillon\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On copie le contenu, autant que possible */
  retour = ECH_CopierContenu( echantillon, nv_echantillon );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la copie du contenu [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    ECH_EffacerEchantillon( nv_echantillon, VRAI );
    return -3;
  }

  /* On installe les nouvelles valeurs  la place des anciennes */
  retour = ECH_EchangerContenu( echantillon, nv_echantillon );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'change des contenus [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    ECH_EffacerEchantillon( nv_echantillon, VRAI );
    return -3;
  }

  /* On efface l'chantillon temporaire, qui contient maintenant les valeurs de l'ancien */
  ECH_EffacerEchantillon( nv_echantillon, VRAI );

  /* Termin avec succs */
  return 0;
}

/* -------- Copie du contenu d'un chantillon -------- */

static int ECH_CopierContenu(Echantillon *original, Echantillon *copie)
{
  if ( ECHANTILLON_MATERIEL == original->type ) {
    if ( original->proprietes & ECHANTILLON_HOMOGENE ) {
      copie->valeurs.homogene.composition = original->valeurs.homogene.composition;
      return 0;
    }

    /* La composition est la mme */
    copie->valeurs.homogene.composition = original->valeurs.homogene.composition;

    OPTION_A_FAIRE( "chantillon inhomogne / copie du contenu [table des densits]" );
    return -1;
  }

  OPTION_A_FAIRE( "chantillon gradient / copie du contenu" );
  return -2;
}

/* -------- changer le contenu de deux chantillons -------- */

static int ECH_EchangerContenu(Echantillon *ech_1, Echantillon *ech_2)
{
  double tmp_d;
  int tmp_i;
  unsigned int tmp_ui;

  /* Les types */
  tmp_i = ech_1->type; ech_1->type = ech_2->type; ech_2->type = tmp_i;

  /* Les dimensions */
  tmp_d = ech_1->larg; ech_1->larg = ech_2->larg; ech_2->larg = tmp_d;
  tmp_d = ech_1->haut; ech_1->haut = ech_2->haut; ech_2->haut = tmp_d;
  tmp_d = ech_1->epaisseur; ech_1->epaisseur = ech_2->epaisseur; ech_2->epaisseur = tmp_d;

  /* Les pas */
  tmp_d = ech_1->pas_x; ech_1->pas_x = ech_2->pas_x; ech_2->pas_x = tmp_d;
  tmp_d = ech_1->pas_y; ech_1->pas_y = ech_2->pas_y; ech_2->pas_y = tmp_d;
  tmp_d = ech_1->pas_z; ech_1->pas_z = ech_2->pas_z; ech_2->pas_z = tmp_d;

  /* Le nombre de cases */
  tmp_ui = ech_1->nbr_x; ech_1->nbr_x = ech_2->nbr_x; ech_2->nbr_x = tmp_ui;
  tmp_ui = ech_1->nbr_y; ech_1->nbr_y = ech_2->nbr_y; ech_2->nbr_y = tmp_ui;
  tmp_ui = ech_1->nbr_z; ech_1->nbr_z = ech_2->nbr_z; ech_2->nbr_z = tmp_ui;

  /* Les proprits */
  tmp_ui = ech_1->proprietes; ech_1->proprietes = ech_2->proprietes; ech_2->proprietes = tmp_ui;

  /* Le contenu */
  {
    Echantillon tmp_ech;

    tmp_ech.valeurs = ech_1->valeurs; ech_1->valeurs = ech_2->valeurs; ech_2->valeurs = tmp_ech.valeurs;
  }

  /* Le reste est pour usage interne, ne devrait pas exister lors de l'appel */
#ifdef _DEVELOPPE_
  if ( ( ech_1->nombre != NULL ) || ( ech_2->nombre != NULL ) )
    AVERTISSEMENT( "Contenu interne dfini IGNOR [%s/%s]", ech_1->nom, ech_2->nom );
#endif

  return 0;
}

/* Effacement d'un chantillon */

int ECH_EffacerEchantillon(Echantillon *echantillon, boolean forcer)
{
  if ( NULL == echantillon ) return +1;

  if ( FAUX == forcer ) {
    int reponse = 0;

    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- poser la question !\n",
		   __FILE__, __FUNCTION__, __LINE__);

    if ( 2 == reponse ) return +1; /* On annule */
  }

  /* Libration du contenu */
  free( echantillon->nom );
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    if ( echantillon->type == ECHANTILLON_SPECTRE ) {
      free( echantillon->valeurs.spectre );
    }
  } else {
    if ( echantillon->type == ECHANTILLON_MATERIEL ) {
      free( echantillon->valeurs.inhomogene );
    } else if ( ECHANTILLON_SPECTRE == echantillon->type ) {
      free( echantillon->valeurs.spectre );
    } else {
      free( echantillon->valeurs.homogene.concentration );
    }
  }

  free( echantillon->nombre );

  /* Effacement des spectres inutiles */

  /* Libration de l'chantillon */
  free( echantillon );

  /* Tout s'est bien pass */
  return 0;
}

int ECHANTILLON_Nommer(Echantillon *echantillon, const char *nom)
{
  if ( NULL == echantillon ) return -1;
  if ( NULL == nom ) return +1;

  free( echantillon->nom );

  echantillon->nom = strdup( nom );
  if ( NULL == echantillon->nom ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mmoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +2;
  }

  return 0;
}

/* -------- Gestion des index de tables -------- */

inline int ECH_Case_v_IJK(Echantillon *echantillon, unsigned long n_case, unsigned int *i, unsigned int *j, unsigned int *k)
{
  unsigned long N_plan;

  if ( NULL == echantillon ) return -1;

  N_plan = echantillon->nbr_x * echantillon->nbr_y; /* Nombre de cases par tranche */
  if ( k != NULL ) *k = n_case / N_plan; /* numro de la tranche */

  /* Reste : position dans la tranche */
  n_case = n_case % N_plan;
  if ( j != NULL ) *j = n_case / echantillon->nbr_x; /* numro de la ligne */

  /* Reste : position dans la ligne = numro de colonne */
  if ( i != NULL ) *i = n_case % echantillon->nbr_x;

  /* Tout s'est bien pass */
  return 0;
}

inline int ECH_Case_v_XYZ(Echantillon *echantillon, unsigned long n_case, double *x, double *y, double *z)
{
  int retour;
  unsigned int i, j, k;

  retour = ECH_Case_v_IJK( echantillon, n_case, &i, &j, &k);
  if ( retour != 0 ) {
    (void) fprintf(stderr, "ECH_Case_v_XYZ : ERREUR dans ECH_Case_v_IJK() [code %d]\n", retour);
    return -1;
  }

/*   printf( "[i=%u j=%u k=%u]", i, j, k ); */
  *x = i * echantillon->pas_x - echantillon->larg / 2.0;
  *y = j * echantillon->pas_y;
  *z = k * echantillon->pas_z - echantillon->haut / 2.0;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Trouver le numro de case  partir des index sur les trois axes --------
   Renvoie 0 en cas d'erreur, le numro de la case sinon

 echantillon : celui sur lequel on travaille
 i = index sur l'axe des X [axe de plus bas niveau]
 j = index sur l'axe des Y [axe intermdiaire]
 k = index sur l'axe des Z [axe de plus haut niveau]
 ----------------------------------------------------------------------------------- */

inline unsigned long ECH_IJK_v_Case(Echantillon *echantillon, unsigned int i, unsigned int j, unsigned int k)
{
  unsigned long n_case;

  ENTREE_FONCTION("ECH_IJK_v_Case", "echantillon=%p i=%u j=%u k=%u", echantillon, i, j, k);

  if ( ( i >= echantillon->nbr_x ) || ( j >= echantillon->nbr_y ) || ( k >= echantillon->nbr_z ) ) {
    (void) fprintf(stderr, "ECH_IJK_v_Case : Paramtre incorrect [%u/%u %u/%u %u/%u]\n",
		   i, echantillon->nbr_x - 1, j, echantillon->nbr_y - 1, k, echantillon->nbr_z - 1);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    SORTIE_FONCTION("ECH_IJK_v_Case", "%lu [Index trop grand, max I=%u J=%u K=%u]", 0, echantillon->nbr_x, echantillon->nbr_y, echantillon->nbr_z);
    return 0;
  }

  n_case  = k * (unsigned long) echantillon->nbr_y;
  n_case += j;
  n_case *= (unsigned long) echantillon->nbr_x;
  n_case += i;

#ifdef _DEVELOPPE_
  if ( n_case >= echantillon->nbr_x * echantillon->nbr_y * (unsigned long) echantillon->nbr_z ) {
    (void) fprintf(stderr, "%s : Numro de case ABSURDE [%lu, max %lu] ; x = %u [%u] y=%u [%u] z=%u [%u]\n", __FUNCTION__,
		   n_case, echantillon->nbr_x * echantillon->nbr_y * (unsigned long) echantillon->nbr_z,
		   i, echantillon->nbr_x, j, echantillon->nbr_y, k, echantillon->nbr_z);
/*     FIN_ANTICIPEE( FIN_ERREUR_GRAVE ); */
  }
#endif

  SORTIE_FONCTION("ECH_IJK_v_Case", "%lu", n_case);
  return n_case;
}

inline unsigned long ECH_XYZ_v_Case(Echantillon *echantillon, double x, double y, double z)
{
  unsigned long n_case;
  unsigned int i, j, k;
  int retour;

  retour = ECH_XYZ_v_IJK( echantillon, x, y, z, &i, &j, &k );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : On est hors de l'chantillon\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  n_case = ECH_IJK_v_Case(echantillon, i, j, k);
  return n_case;
}

inline int ECH_IJK_v_XYZ(Echantillon *echantillon, unsigned int i, unsigned int j, unsigned int k, double *x, double *y, double *z)
{
  *x = i * echantillon->pas_x - echantillon->larg / 2.0;
  *y = j * echantillon->pas_y;
  *z = k * echantillon->pas_z - echantillon->haut / 2.0;

/*   printf("%s : (%u, %u, %u) => (%g,%g,%g)\n", __FUNCTION__, i, j, k, *x, *y, *z); */

  return 0;
}

/* -------- Conversion (x, y, z) ==> index de case
     La case (i, j, k) contient une valeur en (x0, y0, z0)
     Par principe, (x0, y0, z0) est au centre de la case, c'est--dire en (i+1/2, j+1/2, k+1/2)
     ==> la case occupe donc l'espace de ( x_min
*/

inline int ECH_XYZ_v_IJK(Echantillon *echantillon, double x, double y, double z, unsigned int *i, unsigned int *j, unsigned int *k)
{
/*   printf("%s : (%g,%g,%g) => ", __FUNCTION__, x, y, z ); */

  /* On centre par rapport  l'chantillon */
  x += echantillon->larg / 2.0;
  z += echantillon->haut / 2.0;

  /* Contrle : on est bien dans l'chantillon */
  if ( x < 0.0 ) return -1; if ( x > echantillon->larg ) return +1;
  if ( y < 0.0 ) return -2; if ( y > echantillon->epaisseur ) return +2;
  if ( z < 0.0 ) return -3; if ( z > echantillon->haut ) return +3;

  /* Conversion vers le numro de case : toujours de (x, y, z) plus petits
     => (x, y, z) est compris entre (i, j, k) et (i+1, j+1, k+1) */
  if ( i != NULL ) *i = x / echantillon->pas_x;
  if ( j != NULL ) *j = y / echantillon->pas_y;
  if ( k != NULL ) *k = z / echantillon->pas_z;

/*   printf("(%u, %u, %u)\n", *i, *j, *k); */

  return 0;
}

/* -------- Contrle des paramtres des options -------- */

int ECH_ControlerOptions(Echantillon *echantillon)
{
  unsigned long nbr_cases;
  int retour;

  if ( NULL == echantillon ) return -1;

  /* Dimensions [normalement, contrl  la cration, mais...] */
  if ( ( echantillon->larg <= 0.0 ) || ( echantillon->haut <= 0.0 ) || ( echantillon->epaisseur <= 0.0 ) ) {
    (void) fprintf(stderr, "ECH_ControlerOptions : Dimensions absurdes [L=%g H=%g e=%g]\n", echantillon->larg, echantillon->haut, echantillon->epaisseur);
    return -2;
  }

  /* Nombre de cases */
  nbr_cases = echantillon->nbr_x * echantillon->nbr_y * echantillon->nbr_z;
  if ( 0 == nbr_cases ) {
    (void) fprintf(stderr, "ECH_ControlerOptions : nombre de case absurde [%lu]\n", nbr_cases);
    return -2;
  }

  /* Composition --- suivant le type d'chantillon */
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    retour = Matiere_ControlerComposition( echantillon->valeurs.homogene.composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "ECH_ControlerOptions : ERREUR dans Matiere_ControlerComposition() [Homogne]\n");
      return -3;
    }
  } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    unsigned long i;
    double *concentrations;

    /* La composition est-elle correcte ? */
    retour = Matiere_ControlerComposition( echantillon->valeurs.homogene.composition );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "ECH_ControlerOptions : ERREUR dans Matiere_ControlerComposition() [Gradient]\n");
      return -3;
    }

    /* Les concentrations locales sont-elles correctes ? [long, mais moins qu'un calcul absurde !] */
    for ( i = 0, concentrations = echantillon->valeurs.homogene.concentration ; i < nbr_cases ; i++, concentrations++ ) {
      if ( *concentrations < 0.0 ) {
	(void) fprintf(stderr, "ECH_ControlerOptions : concentration locale absurde [%g]\n", *concentrations);
	return -3;
      }
    }
  } else if ( ECHANTILLON_MATERIEL == echantillon->type ) {
    (void) fprintf(stderr, "ECH_ControlerOptions :  FAIRE --- chantillon inhomogne\n");
  }

  /* Aucun problme trouv */
  return 0;
}

/* -------- Connatre la densit en un point de l'chantillon -------- */

double ECH_LireDensite(Echantillon *echantillon, Vecteur3D *point)
{
  double densite = 0.0;

  if ( NULL == echantillon ) return 0.0; /* Pas d'chantillon, pas de densit ! */

  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) { /* Le point n'a pas d'importance */
    struct matiere *composition;

    composition = echantillon->valeurs.homogene.composition;
    if ( NULL == composition ) {
      (void) fprintf(stderr, "%s,%s l.%u : Composition NULL\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return 0.0;
    }

    return composition->densite;
  }

  if ( NULL == point ) {
    (void) fprintf(stderr, "%s,%s l.%u : Point non prcis\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return 0.0;
  }

  if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    unsigned int i, j, k;
    unsigned long n_case;
    int position;
    double coins[ 8 ] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 }, x_coin, y_coin, z_coin;
    double u_x, u_y, u_z;

    position = ECH_XYZ_v_IJK( echantillon, point->x, point->y, point->z, &i, &j, &k );
    if ( position != 0 ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Point (%g, %g, %g) hors-chantillon [Lx=%g - %g, Hz=%g - %g, ey=%g - %g]\n",
		     __FILE__, __FUNCTION__, __LINE__, point->x, point->y, point->z, -echantillon->larg/2.0, echantillon->larg/2.0,
		     -echantillon->haut/2.0, echantillon->haut/2.0, 0.0, echantillon->epaisseur);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
#endif
      return 0.0; /* Hors-chantillon */
    }

    /* L, le point est entre (i, j, k) et (i+1, j+1, k+1)
       On fait donc une interpolation 3D de ces valeurs pour avoir le rsultat */
    /* 1. Valeurs des coordonnes fractionnaires */
    ECH_IJK_v_XYZ( echantillon, i, j, k, &x_coin, &y_coin, &z_coin );
    u_x = point->x - x_coin; u_y = point->y - y_coin; u_z = point->z - z_coin;
    u_x /= echantillon->pas_x; u_y /= echantillon->pas_y; u_z /= echantillon->pas_z;

    /* 2. Valeurs aux sommets [avec traitement des bords] */ 
    n_case = ECH_IJK_v_Case( echantillon, i, j, k );
    coins[ 0 ] = echantillon->valeurs.homogene.concentration[ n_case ];

    if ( i + 1 < echantillon->nbr_x ) {
      n_case = ECH_IJK_v_Case( echantillon, i + 1, j, k );
      coins[ 1 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    } else {
      u_x = 0.0;
    }

    if ( ( i + 1 < echantillon->nbr_x ) && ( j + 1 < echantillon->nbr_y ) ) {
      n_case = ECH_IJK_v_Case( echantillon, i + 1, j + 1, k );
      coins[ 2 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    }

    if ( j + 1 < echantillon->nbr_y ) {
      n_case = ECH_IJK_v_Case( echantillon, i, j + 1, k );
      coins[ 3 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    } else {
      u_y = 0.0;
    }

    if ( k + 1 < echantillon->nbr_z ) {
      n_case = ECH_IJK_v_Case( echantillon, i, j, k + 1 );
      coins[ 4 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    } else {
      u_z = 0.0;
    }

    if ( ( i + 1 < echantillon->nbr_x ) && ( k + 1 < echantillon->nbr_z ) ) {
      n_case = ECH_IJK_v_Case( echantillon, i + 1, j, k + 1 );
      coins[ 5 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    }

    if ( ( i + 1 < echantillon->nbr_x ) && ( j + 1 < echantillon->nbr_y ) && ( k + 1 < echantillon->nbr_z ) ) {
      n_case = ECH_IJK_v_Case( echantillon, i + 1, j + 1, k + 1);
      coins[ 6 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    }

    if ( ( j + 1 < echantillon->nbr_y ) && ( k + 1 < echantillon->nbr_z ) ) {
      n_case = ECH_IJK_v_Case( echantillon, i, j + 1, k + 1 );
      coins[ 7 ] = echantillon->valeurs.homogene.concentration[ n_case ];
    }

    /* 3. Calcul proprement dit */
    densite  = (1.0 - u_x) * (1.0 - u_y) * (1.0 - u_z) * coins[ 0 ];
    densite +=        u_x  * (1.0 - u_y) * (1.0 - u_z) * coins[ 1 ];
    densite +=        u_x  *        u_y  * (1.0 - u_z) * coins[ 2 ];
    densite += (1.0 - u_x) *        u_y  * (1.0 - u_z) * coins[ 3 ];
    densite += (1.0 - u_x) * (1.0 - u_y) *        u_z  * coins[ 4 ];
    densite +=        u_x  * (1.0 - u_y) *        u_z  * coins[ 5 ];
    densite +=        u_x  *        u_y  *        u_z  * coins[ 6 ];
    densite += (1.0 - u_x) *        u_y  *        u_z  * coins[ 7 ];

#ifdef _DEVELOPPE_
    if ( densite < 0.0 ) {
      (void) printf("\n%s : Densit <0 [%g] u_x=%g u_y=%g u_z=%g\n", __FUNCTION__, densite, u_x, u_y, u_z);
      for ( i = 0 ; i < 8 ; i++ ) {
	printf("d[ %u ] = %g ", i, coins[ i ] );
      }
      printf("\n");
      abort();
    }
#endif

  } else {
    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- chantillon inhomogne\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  return densite;
}

/* -------- Connatre le coefficient d'absorption total en un point de l'chantillon -------- */

double ECH_CalculerMuTotal(Echantillon *echantillon, Vecteur3D *point, double energie, double *mu_photoelectrique)
{
  double mu_total = 0.0;
  struct graphe *grf_mu = NULL, *grf_te = NULL, *grf_ti = NULL;
  struct matiere *composition;

  if ( NULL == echantillon ) return 0.0; /* Pas d'chantillon, pas d'absorption ! */

  if ( ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) ||
       ( ECHANTILLON_GRADIENT == echantillon->type ) ) { /* Le point d'impact n'a pas d'importance ! */
    composition = echantillon->valeurs.homogene.composition;
    if ( NULL == composition ) {
      (void) fprintf(stderr, "ECH_CalculerMuTotal : Composition NULL\n");
      return 0.0;
    }

    grf_mu = composition->mu_total;
    grf_te = composition->te_total;
    grf_ti = composition->ti_total;

    mu_total  = ECH_MuTotal( grf_mu, grf_te, grf_ti, energie, mu_photoelectrique );
    mu_total *= composition->densite;
    if ( mu_photoelectrique != NULL ) *mu_photoelectrique *= composition->densite;

    /* On veut aussi le terme correctif */
    if ( ( point != NULL ) && ( ECHANTILLON_GRADIENT == echantillon->type ) ) {
      double densite;

      densite = ECH_LireDensite( echantillon, point );
#ifdef _DEVELOPPE_
      if ( 0.0 == densite ) {
	(void) fprintf(stderr, "%s,%s l.%u : Densit nulle [point (%g, %g, %g)]\n",
		       __FILE__, __FUNCTION__, __LINE__, point->x, point->y, point->z);
      }
#endif
      mu_total *= densite;
      if ( mu_photoelectrique != NULL ) *mu_photoelectrique *= densite;
    }
  } else {
    unsigned long n_case;

    if ( NULL == point ) {
#ifdef _DEVELOPPE_ /* Car appel sans point normal pour init. fluorescence... */
      (void) fprintf(stderr, "ECH_CalculerMuTotal : Pas de point prcis !\n");
#endif
      return 0.0;
    }

    /* De quelle case parle-t-on ? */
    n_case = ECH_XYZ_v_Case( echantillon, point->x, point->y, point->z );

    /* Composition en cette case */
    composition = echantillon->valeurs.inhomogene[ n_case ];
    if ( NULL == composition ) {
      return 0.0; /* Trou : aucune absorption ! */
    }

    grf_mu = composition->mu_total;
    grf_te = composition->te_total;
    grf_ti = composition->ti_total;

    mu_total  = ECH_MuTotal( grf_mu, grf_te, grf_ti, energie, mu_photoelectrique );
    mu_total *= composition->densite;
    if ( mu_photoelectrique != NULL ) *mu_photoelectrique *= composition->densite;
  }

  /* On a fini le calcul */
  return mu_total;
}

inline static double ECH_MuTotal(struct graphe *mu, struct graphe *te, struct graphe *ti, double energie, double *mu_photoelectrique)
{
  double mu_total = 0.0;

  /* Terme d'absorption photolectrique */
  mu_total = GRF_LireValeur( mu, energie, INTERPOLE_LINEAIRE );
  if ( NULL != mu_photoelectrique ) *mu_photoelectrique = mu_total;

  /* Terme de diffusion lastique */
  mu_total += GRF_LireValeur( te, energie, INTERPOLE_LINEAIRE );

  /* Terme de diffusion inlastique */
  mu_total += GRF_LireValeur( ti, energie, INTERPOLE_LINEAIRE );

  /* Calcul termin */
  return mu_total;
}

/* ~-~-~-~-~-~-~-~- Ralisation de coupes de l'chantillon ~-~-~-~-~-~-~-~- */

/* -------- Construction d'une tranche en Y -------- */

struct image *ECH_TrancheVersImage_Y(Echantillon *echantillon, unsigned int tranche_y, struct image *image)
{
  unsigned i, j;

  if ( NULL == echantillon ) return NULL;
  if ( tranche_y >= echantillon->nbr_y ) {
    (void) fprintf(stderr, "%s,%s l.%u : Numro de tranche invalide [%u / max %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, tranche_y, echantillon->nbr_y);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  ENTREE_FONCTION( __FUNCTION__, "echantillon=%p [%s] tranche_y=%u image=%p", echantillon, echantillon->nom, tranche_y, image);

  /* Prparation de l'image */
  if ( NULL == image ) {
    image = IMG_CreerImage( NULL, NULL, IMG_GENRE_REELLE, IMG_TYPE_CARTOGRAPHIE, echantillon->nbr_x, echantillon->nbr_y );
    if ( NULL == image ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans creer_image()\n",
		     __FILE__, __FUNCTION__, __LINE__);
      SORTIE_FONCTION( __FUNCTION__, "%p [ERREUR dans IMG_CreerImage()]", NULL);
      return NULL;
    }
  } else { /* Contrler que les dimensions sont cohrentes */
    if ( ( image->larg != echantillon->nbr_x ) || ( image->haut != echantillon->nbr_z ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Dimensions de l'image [%u  %u] incompatibles avec celles de l'chantillon [%u  %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, image->larg, image->haut, echantillon->nbr_x, echantillon->nbr_z);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      SORTIE_FONCTION( __FUNCTION__, "%p [Dimensions incompatibles]", NULL);
      return NULL;
    }
  }

  IMG_ChangerEchelle( image, AXE_X | AXE_Y, echantillon->pas_x, echantillon->pas_z );
  IMG_ChangerRepere ( image, AXE_X | AXE_Y, -echantillon->larg / 2.0, -echantillon->haut / 2.0 );

  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    double densite;
    double *valeurs;

    if ( NULL == echantillon->valeurs.homogene.composition ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Attention, pas de composition !\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      densite = 0.0;
    } else {
      densite = echantillon->valeurs.homogene.composition->densite;
    }

    for ( i = 0, valeurs = image->valeurs.R ; i < echantillon->nbr_z ; i++ ) { /* Indice de ligne */
      for ( j = 0 ; j < echantillon->nbr_x ; j++, valeurs++ ) { /* Indice de colonne */
 	*valeurs = densite;
      }
    }
  } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    double densite;
    double *valeurs;

    if ( NULL == echantillon->valeurs.homogene.composition ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Attention, pas de composition !\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      densite = 0.0;
    } else {
      densite = echantillon->valeurs.homogene.composition->densite;
    }

    for ( i = 0, valeurs = image->valeurs.R ; i < echantillon->nbr_z ; i++ ) { /* Indice de ligne = sur les Z */
      for ( j = 0 ; j < echantillon->nbr_x ; j++, valeurs++ ) { /* Indice de colonne = sur les X */
	*valeurs = densite * echantillon->valeurs.homogene.concentration[ ECH_IJK_v_Case( echantillon, j, tranche_y, i ) ];
      }
    }
  } else {
    struct matiere **valeurs;

    /* Premire version ultra-simple : copie bte et mchante des pointeurs */
    for ( i = 0, valeurs = image->valeurs.M ; i < echantillon->nbr_z ; i++ ) { /* Indice de ligne = sur les Z */
      for ( j = 0 ; j < echantillon->nbr_x ; j++, valeurs++ ) { /* Indice de colonne = sur les X */
	*valeurs = echantillon->valeurs.inhomogene[ ECH_IJK_v_Case( echantillon, j, tranche_y, i ) ];
      }
    }
  }

  /* On met les bornes pour l'image */
  IMG_bornes( image );

  /* Calcul fini correctement */
  SORTIE_FONCTION( __FUNCTION__, "%p [%s]", image, image->nom);
  return image;
}

/* ~-~-~-~-~-~-~-~- Reconstruction de l'chantillon  partir de coupes ~-~-~-~-~-~-~-~- */

int ECH_ImageVersTranche_Y(Echantillon *echantillon, unsigned int tranche_y, struct image *coupe)
{
  unsigned int i, j;
  unsigned long n_case;
  double x_img, y_img;

  /* Contrles pralables */
  if ( NULL == echantillon ) return -1;
  if ( NULL == coupe ) return -2;
  if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
    (void) fprintf(stderr, "%s,%s l.%u : chantillon homogne, impossible de modifier la rpartition spatiale.\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +1;
  }
  if ( tranche_y >= echantillon->nbr_y ) {
    (void) fprintf(stderr, "%s,%s l.%u : Numro de tranche invalide [%u / max %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, tranche_y, echantillon->nbr_y);
    return -3;
  }

#ifdef _DEVELOPPE_
  printf("%s : chantillon, X = %g -> %g par pas de %g [%u points => %g]\n", __FUNCTION__, -echantillon->larg / 2.0, echantillon->larg / 2.0,
	 echantillon->pas_x, echantillon->nbr_x, echantillon->larg / echantillon->nbr_x );
  printf("         Z = %g -> %g par pas de %g [%u points => %g]\n", -echantillon->haut / 2.0, echantillon->haut / 2.0,
	 echantillon->pas_y, echantillon->nbr_y, echantillon->haut / echantillon->nbr_y );
  printf("  Image, X = %g -> %g par pas de %g [%u points]\n", coupe->x_min, coupe->x_min + coupe->l_reel * (coupe->larg - 1), coupe->l_reel, coupe->larg);
  printf("         Y = %g -> %g par pas de %g [%u points]\n", coupe->y_min, coupe->y_min + coupe->h_reel * (coupe->haut - 1), coupe->h_reel, coupe->haut);
#endif

  /* chantillon de type gradient ? */
  if ( ECHANTILLON_GRADIENT == echantillon->type ) {
    /* Dans ce cas, l'image contient les CORRECTIONS  la densit => 1 : comme la composition indique */
    for ( i = 0 ; i < echantillon->nbr_x ; i++ ) {
      x_img = echantillon->pas_x * i - echantillon->larg / 2.0; /* X utilis */

      for ( j = 0 ; j < echantillon->nbr_z ; j++ ) {
	y_img = echantillon->pas_z * j - echantillon->haut / 2.0;

	n_case = ECH_IJK_v_Case( echantillon, i, tranche_y, j );
	echantillon->valeurs.homogene.concentration[ n_case ] = IMG_lit_valeur( coupe, x_img, y_img, IMG_INTERPOLE_BILINEAIRE, NULL );
      }
    }
  } else {
    int retour;
    unsigned long n_case_image;

    /* Dans ce cas, l'image contient les compositions */
    for ( i = 0 ; i < echantillon->nbr_x ; i++ ) {
      x_img = echantillon->pas_x * i - echantillon->larg / 2.0; /* i : varie sur les X => indice de colonne, donne le X utilis */

      for ( j = 0 ; j < echantillon->nbr_z ; j++ ) { /* j : varie sur les Z => indice de ligne, donne le Y utilis */
	y_img = echantillon->pas_z * j - echantillon->haut / 2.0;

	n_case = ECH_IJK_v_Case( echantillon, i, tranche_y, j );
	retour = IMG_XY_v_N( coupe, x_img, y_img, &n_case_image );
	printf("%s : on va lire en X=%g, Y=%g : case n%lu [retour : %d]\n", __FUNCTION__, x_img, y_img, n_case_image, retour );
	if ( retour != 0 ) {
	  echantillon->valeurs.inhomogene[ n_case ] = NULL; /* On met un trou ! */
	} else {
	  echantillon->valeurs.inhomogene[ n_case ] = coupe->valeurs.M[ n_case_image ];
	}
      }
    }
  }

  /* Tout s'est bien pass */
  return 0;
}
