/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'une description du faisceau excitant
   Projection d'un faisceau dans un plan donné (coupe)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "echantillon.h"
#include "faisceau.h"
#include "graphes.h" /* Pour les contrôles avant effacement ! */
#include "images.h"
#include "lissage.h"
#include "science.h"		/* Pour M_PI, au cas où... */

#include "th_faisceau.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define PRECISION_BORDS_FAISCEAU (1.0e-2) /* Précision sur les dimensions du faisceau, en microns [1/100 µ] */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static Faisceau *FSC_Projection_Etendu  (Faisceau *faisceau, double pas_x, double pas_z, double energie, Graphe *absorbance);
static Faisceau *FSC_Projection_Ponctuel(Faisceau *faisceau, double energie, Graphe *absorbance);
static Faisceau *FSC_ProjectionOrthogonale(Faisceau *faisceau, double pas_x, double pas_z, double energie, Graphe *absorbance);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Création ~-~-~-~-~-~-~-~- */

/* -------- Création d'une structure décrivant un faisceau d'excitation --------\
 |  Renvoie la structure créée si tout va bien
 |          NULL en cas d'erreur
 |
 | larg, haut  : dimensions du faisceau, en MICRONS (nulles : faisceau ponctuel)
 | proprietes  : propriétés du faisceau [cf. echantillons.h]
 | harmoniques : nombre d'harmoniques présentes (0 : faisceau pur)
 | I           : intensité de référence (u. arb.)
 \----------------------------------------------------------------------------- */

Faisceau *FAISCEAU_Creer(double larg, double haut, unsigned int proprietes, unsigned int harmoniques, double I)
{
  Faisceau *faisceau;

  /* Contrôle des dimensions */
  if ( ( larg < 0.0 ) || ( haut < 0.0 ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : dimensions absurdes (larg=%g, haut=%g)\n",
		    __FILE__, __FUNCTION__, __LINE__, larg, haut );
    return NULL;
  }

  /* On réserve la mémoire */
  faisceau = ( Faisceau * ) calloc( 1, sizeof( Faisceau ) );
  if ( NULL == faisceau ) {
    (void) fprintf(stderr, "FAISCEAU_Creer : mémoire insuffisante\n");
    return NULL;
  }

  /* On initialise : partie commune
     1. organisation spatiale : dimensions, position,... */
  faisceau->largeur = larg;
  faisceau->hauteur = haut;
  faisceau->orientation.position.x = 0.0; /* Par défaut, le faisceau passe par le centre du repère */
  faisceau->orientation.position.y = -1.0e6; /* Mais il démarre bien avant I0 [1 m] pour être tranquille */
  faisceau->orientation.position.z = 0.0;
  faisceau->orientation.alpha = 0.0; /* Il est orthogonal à la face avant de l'échantillon */
  faisceau->orientation.theta = 0.0; /* donc a le même repère local que le repère absolu */
  faisceau->orientation.phi   = 0.0;

  faisceau->point_focal.x = 0.0; /* Par défaut, le faisceau est parallèle */
  faisceau->point_focal.y = 0.0; /* ou focalisé sur la face avant de l'échantillon, au centre */
  faisceau->point_focal.z = 0.0;

  /* 2. propriétés */
  faisceau->intensite = I;
  faisceau->proprietes = proprietes;

  /* 3. infos complémentaires */
  faisceau->g_E = NULL; /* Pas de graphe de chromatisme pour l'instant */
  faisceau->nbr_harmoniques = 0;
  faisceau->i_harmoniques = NULL;
  faisceau->g_harmoniques = NULL;
    
  if ( harmoniques > 0 ) {
    int retour;

    retour = FAISCEAU_CreerHarmoniques( faisceau, harmoniques );
    if ( retour != 0 ) {
      (void) fprintf( stderr, "%s,%s l.%u :  ERREUR dans FAISCEAU_CreerHarmoniques [code %d]\n",
		      __FILE__, __FUNCTION__, __LINE__, retour);
    }
  }
  faisceau->g_energie = NULL;
  faisceau->g_espace  = NULL; /* Faisceau ponctuel ou homogène spatialement */

  /* Tout s'est bien passé, initialisation terminée */
  return faisceau;
}

/* -------- Réserver pour les harmoniques -------- */

int FAISCEAU_CreerHarmoniques(Faisceau *faisceau, unsigned int nbr_harmoniques)
{
  unsigned int i;
  float *table;
  struct graphe **table_g;

  ENTREE_FONCTION("FAISCEAU_CreerHarmoniques", "faisceau=%p nbr_harmoniques=%u", faisceau, nbr_harmoniques);

  if ( NULL == faisceau ) {
    SORTIE_FONCTION("FAISCEAU_CreerHarmoniques", "%d [ERREUR : faisceau=NULL]", -1);
    return -1;
  }

  /* Pas de changement -> rien à faire ! */
  if ( nbr_harmoniques == faisceau->nbr_harmoniques ) {
    SORTIE_FONCTION("FAISCEAU_CreerHarmoniques", "%d [rien à faire]", 0);
    return 0;
  }

  if ( 0 == nbr_harmoniques ) {
    /* Plus d'harmoniques : on libère tout */
    free( faisceau->i_harmoniques ); faisceau->i_harmoniques = NULL;
    free( faisceau->g_harmoniques ); faisceau->g_harmoniques = NULL;

    /* Tout s'est bien passé */
    faisceau->nbr_harmoniques = 0;
    SORTIE_FONCTION("FAISCEAU_CreerHarmoniques", "%d [suppression]", 0);
    return 0;
  }

  /* Ici, il faut donc réserver la table des intensités relatives */
  table = (float *) realloc( faisceau->i_harmoniques, nbr_harmoniques * sizeof( float ) );
  if ( NULL == table ) {
    /* Original : non modifié -> on renvoie tel quel */
    (void) fprintf(stderr, "FAISCEAU_CreerHarmoniques : Impossible d'adapter la table des intensités des harmoniques.\n");
    SORTIE_FONCTION("FAISCEAU_CreerHarmoniques", "%d [ERREUR : table=NULL]", -1);
    return -1;
  }
  faisceau->i_harmoniques = table;

  /* La table des résolutions en énergie */
  table_g = ( struct graphe ** ) realloc( faisceau->g_harmoniques, nbr_harmoniques * sizeof( struct graphe * ) );
  if ( NULL == table_g ) {
    (void) fprintf(stderr, "FAISCEAU_Creer : impossible d'adapter la table d'harmoniques.\n");
    SORTIE_FONCTION("FAISCEAU_CreerHarmoniques", "%d [ERREUR : table_g=NULL]", -2);
    return -2;
  }
  faisceau->g_harmoniques = table_g;

  /* On initialise les nouvelles harmoniques */
  for ( i = faisceau->nbr_harmoniques ; i < nbr_harmoniques ; i++ ) {
    table[ i ] = 0.0;
    table_g[ i ] = NULL;
  }

  /* On installe le nouveau nombre d'harmoniques */
  faisceau->nbr_harmoniques = nbr_harmoniques;

  /* Tout s'est bien passé */
  SORTIE_FONCTION("FAISCEAU_CreerHarmoniques", "%d", 0);
  return 0;
}

/* -------- Réserver une coupe -------- */

Image *FAISCEAU_ReserverCoupe(double larg, double haut, double pas_x, double pas_z)
{
  Image *coupe;
  unsigned int n_x, n_z;

  ENTREE_FONCTION("FAISCEAU_ReserverCoupe", "larg=%g haut=%g pas_x=%g pas_z=%g", larg, haut, pas_x, pas_z);

  /* Vérifications préliminaires */
  if ( ( larg <= 0.0 ) || ( haut <= 0.0 ) ) {
    (void) fprintf(stderr, "FAISCEAU_ReserverCoupe : dimensions absurdes [%g, %g]\n", larg, haut);
    SORTIE_FONCTION("FAISCEAU_ReserverCoupe", "%p [dimensions absurdes]", NULL);
    return NULL;
  }
  if ( ( pas_x <= 0.0 ) || ( pas_z <= 0.0 ) ) {
    (void) fprintf(stderr, "FAISCEAU_ReserverCoupe : pas absurdes [%g, %g]\n", pas_x, pas_z);
    SORTIE_FONCTION("FAISCEAU_ReserverCoupe", "%p [pas absurdes]", NULL);
    return NULL;
  }

  /* Nombre de points dans chaque direction */
  n_x = 1 + (unsigned int) ( larg / pas_x );
  n_z = 1 + (unsigned int) ( haut / pas_z );
  DANS_FONCTION("Grille, n_x = %u, n_z = %u", n_x, n_z);

  /* Réservation */
  coupe = IMG_CreerImage( NULL, NULL, IMG_GENRE_REELLE, IMG_TYPE_INTENSITES, n_x, n_z );
  if ( NULL == coupe ) {
    (void) fprintf(stderr, "FAISCEAU_ReserverCoupe : ERREUR dans creer_image()\n");
    SORTIE_FONCTION("FAISCEAU_ReserverCoupe", "%p [coupe=NULL]", NULL);
    return NULL;
  }

  /* Mise à jour du repère de l'image & de sa grille */
  /* On travaille avec un faisceau centré sur son repère local */
  IMG_ChangerEchelle( coupe, AXE_X | AXE_Y, pas_x, pas_z );
  IMG_ChangerRepere ( coupe, AXE_X | AXE_Y, -larg / 2.0, -haut / 2.0 );

  /* On a fini */
  SORTIE_FONCTION("FAISCEAU_ReserverCoupe", "%p", coupe);
  return coupe;
}

/* -------- Associer une coupe au faisceau -------- */

int FAISCEAU_AssocierCoupe(Faisceau *faisceau, Image *coupe, boolean copie, unsigned int nbr_x, unsigned int nbr_y)
{
  int retour;

  if ( NULL == faisceau ) return -1;
  if ( NULL == coupe ) return -2;

  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau ponctuel, pas de structure spatiale !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -9;
  }

  if ( ( 0.0 == faisceau->largeur ) || ( 0.0 == faisceau->hauteur ) ) {
    (void) fprintf( stderr, "%s,%s l.%u : Faisceau rectangulaire %g×%g, pas de structure spatiale !\n",
		    __FILE__, __FUNCTION__, __LINE__, faisceau->largeur, faisceau->hauteur );
    return -10;
  }

  /* On ôte le lien avec l'ancien image */
  if ( faisceau->g_espace != NULL ) {
    if ( VRAI == faisceau->g_espace->drapeaux.isolee ) {
      IMG_EffacerImage( faisceau->g_espace, VRAI );
    }
    faisceau->g_espace = NULL;
  }

  /* On fait une association directe */
  if ( FAUX == copie ) {
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Association directe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -4;
  }

  /* Les pas sont-ils réalistes */
  if ( ( nbr_x < 2 ) || ( nbr_y < 2 ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Valeur du pas absurde [X  %u - Y : %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_x, nbr_y);
    return -15;
  }

  faisceau->g_espace = IMG_CreerImage( NULL, coupe, IMG_GENRE_REELLE, IMG_TYPE_INTENSITES, nbr_x, nbr_y );
  if ( NULL == faisceau->g_espace ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer l'image de coupe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  /* Pas de l'image : celui indiqué */
  faisceau->g_espace->l_reel = faisceau->largeur / (double) ( nbr_x - 1 );
  faisceau->g_espace->h_reel = faisceau->hauteur / (double) ( nbr_y - 1 );

  /* Image centrée sur le faisceau */
  faisceau->g_espace->x_min = -faisceau->largeur / 2.0;
  faisceau->g_espace->y_min = -faisceau->hauteur / 2.0;

  /* Et on recale l'image fournie sur celle créée */
  retour = IMG_CalerSurImage( faisceau->g_espace, coupe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_CalerSurImage [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    IMG_EffacerImage( faisceau->g_espace, VRAI );
    faisceau->g_espace = NULL;
    return -20;
  }

  /* Terminé, tout s'est bien passé */
  return 0;
}

/* -------- Associer une courbe g(E) au faisceau --------
   Principe : g(E) doit être une densité de probabilité.
   Comme on travaille en considérant le polychromatique comme la somme des monochromatiques pour chaque énergie,
   il suffit en fait d'en faire une loi de probabilité

   => la somme des valeurs doit valoir 1.
*/

int FAISCEAU_AssocierSpectre(Faisceau *faisceau, struct graphe *g_E, boolean copie)
{
  double somme_valeurs = 0.0, *y;
  unsigned long i;

  if ( NULL == faisceau ) return -1;
  if ( NULL == g_E ) return -2;

  if ( g_E == faisceau->g_E ) {
    return 0; /* Rien à faire ! */
  }

  /* Décompte et vérification */
  for ( y = g_E->y, i = 0 ; i < g_E->nombre_point ; i++, y++ ) {
    if ( *y < 0.0 ) {
      (void) BI_ERR_alerte( NULL, &ERR_gE_ValeurNegative, VRAI );
      return +1;
    }

    somme_valeurs += *y;
  }

  if ( 0.0 == somme_valeurs ) {
      (void) BI_ERR_alerte( NULL, &ERR_gE_GrapheNul, VRAI );
    return +2;
  }

  if ( VRAI == copie ) {
    g_E = GRF_Copier( g_E, NULL, VRAI );
    if ( NULL == g_E ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Copier()\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -5;
    }
  }

  /* On peut donc remplacer l'ancien faisceau */
  if ( faisceau->g_E != NULL ) {
    if ( VRAI == faisceau->g_E->drapeaux.isole ) {
      GRF_eff_isole( faisceau->g_E );
      faisceau->g_E = NULL;
    }
  }

  /* On renormalise le graphe */
  GRF_FacteurMultiplicatif( g_E, 1.0 / somme_valeurs, AXE_Y, NULL );

  /* Attribution réussie */
  faisceau->g_E = g_E;
  return 0;
}

/* -------- Changer le nom -------- */

int FAISCEAU_Nommer(Faisceau *faisceau, const char *nom)
{
  if ( NULL == faisceau ) return -1;
  if ( NULL == nom ) return -2;

  free( faisceau->nom );
  faisceau->nom = strdup( nom );
  if ( NULL == faisceau->nom ) {
    (void) fprintf(stderr, "%s,%s l.%u : mémoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  return 0;
}

/* ~-~-~-~-~-~-~-~- Effacement ~-~-~-~-~-~-~-~- */

int FAISCEAU_Effacer(Faisceau *faisceau)
{
  if ( NULL == faisceau ) return VRAI;

  /* On libère les informations internes */
  if ( faisceau->g_energie != NULL ) {
    if ( VRAI == faisceau->g_energie->drapeaux.isole ) {
      GRF_eff_isole( faisceau->g_energie );
    }
    faisceau->g_energie = NULL;
  }

  if ( faisceau->g_E != NULL ) {
    if ( VRAI == faisceau->g_E->drapeaux.isole ) {
      GRF_eff_isole( faisceau->g_E );
    }
    faisceau->g_E = NULL;
  }

  if ( faisceau->nbr_harmoniques > 0 ) {
    if ( faisceau->g_harmoniques != NULL ) {
      unsigned int i;

      for ( i = 0 ; i < faisceau->nbr_harmoniques ; i++ ) {
	if ( faisceau->g_harmoniques[ i ] != NULL ) {
	  if ( VRAI == faisceau->g_harmoniques[ i ]->drapeaux.isole ) {
	    GRF_eff_isole( faisceau->g_harmoniques[ i ] );
	  }
	  faisceau->g_harmoniques[ i ] = NULL;
	}
      }
    }
  }
  free( faisceau->g_harmoniques );
  free( faisceau->i_harmoniques );

  /* On libère la coupe, mais il faudrait un contrôle d'isolement comme pour les graphes */
/*   effacer_image( faisceau->g_espace, VRAI ); */

  /* On n'a plus besoin de la structure */
  free( faisceau );

  /* Effacement réalisé sans difficulté */
  return VRAI;
}

/* ~-~-~-~-~-~-~-~- Copie ~-~-~-~-~-~-~-~- */

Faisceau *FAISCEAU_Copier(Faisceau *faisceau)
{
  Faisceau *copie;

  if ( NULL == faisceau ) return NULL;

  /* On crée un faisceau de même squelette */
  copie = FAISCEAU_Creer( faisceau->largeur, faisceau->hauteur, faisceau->proprietes, faisceau->nbr_harmoniques, faisceau->intensite);
  if ( NULL == copie ) {
    (void) fprintf(stderr, "FAISCEAU_Copier : Impossible de créer la copie du faisceau\n");
    return NULL;
  }

  /* On copie les informations complémentaires. A priori, modifiables => copie "physique"
     Mais sans intérêt particulier -> copie sous forme de graphe isolé */

  /* Dimensions, intensité : copiés à la création */
  copie->orientation = faisceau->orientation;
  copie->point_focal = faisceau->point_focal;

#ifdef _DEVELOPPE_
  copie->g_energie = GRF_Copier( faisceau->g_energie, NULL, FAUX );
  copie->g_E = GRF_Copier( faisceau->g_E, NULL, FAUX );
#else
  copie->g_energie = GRF_Copier( faisceau->g_energie, NULL, VRAI );
  copie->g_E = GRF_Copier( faisceau->g_E, NULL, VRAI );
#endif

  if ( copie->nbr_harmoniques > 0 ) {
    unsigned int i;

    for ( i = 0 ; i < copie->nbr_harmoniques ; i++ ) {
      copie->i_harmoniques[ i ] = faisceau->i_harmoniques[ i ];

      if ( copie->g_harmoniques != NULL ) {
	copie->g_harmoniques[ i ] = GRF_Copier( faisceau->g_harmoniques[ i ], NULL, VRAI );
      }
    }
  }

  /* Cette copie est isolée */
  copie->g_espace = IMG_CopierImage( faisceau->g_espace, NULL, VRAI );

  return copie;
}

/* ~-~-~-~-~-~-~-~- Utilisation ~-~-~-~-~-~-~-~- */

double FAISCEAU_LireIntensite(Faisceau *faisceau, double x, double y, double energie)
{
  double intensite = 0.0;

  if ( NULL == faisceau ) return 0.0;

  ENTREE_FONCTION("FAISCEAU_LireIntensite", "faisceau=%p [%g] x=%g y=%g energie=%g", faisceau, faisceau->intensite, x, y, energie);

  /* Intensité nominale */
  intensite = faisceau->intensite;
#ifdef _DEVELOPPE_
  if ( is_inf( intensite ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : INTENSITÉ INFINIE [E=%g eV, x=%g, y=%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, energie, x, y);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
  }
#endif

  /* Effet de la structure spatiale */
  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
#ifdef _DEVELOPPE_
    if ( ( x != 0.0 ) || ( y != 0.0 ) ) {
      (void) fprintf(stderr, "FAISCEAU_LireIntensite : Appel en (%g, %g) pour un faisceau ponctuel !\n", x, y);
    }
#endif
  } else {
    double x_max, y_max;

    /* Limites du faisceau "image" : X [-x_max, x_max], Y [-y_max, y_max] */
    x_max = faisceau->largeur / 2.0; y_max = faisceau->hauteur / 2.0;

    /* Est-on dans le rectangle définissant le faisceau ?
       On considère qu'une précision plus grande que [...] est inutile ;
       dans cette marge, on force à être dans le faisceau ! */
    if ( x < -x_max ) {
      if ( ( -x_max - x ) > PRECISION_BORDS_FAISCEAU ) {
#ifdef _DEVELOPPE_
	SORTIE_FONCTION("FAISCEAU_LireIntensite", "%g [X < %g]", 0.0, -x_max );
	printf("%s : x=%g, y=%g hors faisceau [X, %g->%g, x-Xmin = %g, %s]\n",
	       __FUNCTION__, x, y, -x_max, x_max, -x - x_max, faisceau->nom );
#endif
	return 0.0;
      }
      x = -x_max; /* On se cale sur le vrai bord */
    } else if ( x > x_max ) {
      if ( ( x - x_max ) > PRECISION_BORDS_FAISCEAU ) {
#ifdef _DEVELOPPE_
	SORTIE_FONCTION("FAISCEAU_LireIntensite", "%g [X > %g]", 0.0, x_max );
	printf("%s : x=%g, y=%g hors faisceau [X, %g->%g, Xmax-x = %g, %s]\n",
	       __FUNCTION__, x, y, -x_max, x_max, x - x_max, faisceau->nom );
#endif
	return 0.0;
      }
      x = x_max; /* On se cale sur le vrai bord */
    }

    if ( y < -y_max ) {
      if ( ( -y_max - y ) > PRECISION_BORDS_FAISCEAU ) {
#ifdef _DEVELOPPE_
	SORTIE_FONCTION("FAISCEAU_LireIntensite", "%g [Y < %g]", 0.0, -y_max );
	printf("%s : y=%g, y=%g : ON EST HORS FAISCEAU [Y : %g -> %g ; y-Ymin = %g]\n",
	       __FUNCTION__, x, y, -y_max, y_max, -y - y_max );
#endif
	return 0.0;
      }
      y = -y_max; /* On se cale sur le vrai bord */
    } else if ( y > y_max ) {
      if ( ( y - y_max ) > PRECISION_BORDS_FAISCEAU ) {
#ifdef _DEVELOPPE_
	SORTIE_FONCTION("FAISCEAU_LireIntensite", "%g [Y > %g]", 0.0, y_max );
	printf("%s : x=%g, y=%g : ON EST HORS FAISCEAU [Y : %g -> %g ; Ymax-y = %g]\n",
	       __FUNCTION__, x, y, -y_max, y_max, y - y_max );
#endif
	return 0.0;
      }
      y = y_max; /* On se cale sur le vrai bord */
    }

    /* Faisceau inhomogène : correction à l'intensité dans l'image */
    if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) {
      boolean dehors;

      intensite *= IMG_lit_valeur( faisceau->g_espace, x, y, IMG_INTERPOLE_BILINEAIRE, &dehors );
#ifdef _DEVELOPPE_
      if ( is_inf( intensite ) ) {
	(void) fprintf(stderr, "%s,%s l.%u : INTENSITÉ INFINIE [E=%g eV, x=%g, y=%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, energie, x, y);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      }
      if ( VRAI == dehors ) {
	/* Pas d'affichage car arrive fréquemment en fluorescence : init pour un faisceau étendu
	   C'est alors normal => pas de message d'avertissement ! */
/* 	printf("%s : x=%g, y=%g : ON EST HORS FAISCEAU [X : %g -> %g, Y : %g -> %g]\n", __FUNCTION__, x, y, */
/* 	       - faisceau->largeur/2.0, faisceau->largeur / 2.0, -faisceau->hauteur / 2.0, faisceau->hauteur / 2.0 ); */
      }
#endif
    }
    DANS_FONCTION("Effet de surface : intensite = %g", intensite);
#ifdef _DEVELOPPE_
/*     printf("%s : x=%g, y=%g, I=%g, E=%g\n", __FUNCTION__, x, y, intensite, energie ); */
#endif
  }

  /* Effet de l'énergie */
  if ( energie > 0.0 ) {
    if ( faisceau->proprietes & FAISCEAU_DECROIT ) {
      intensite *= GRF_LireValeur( faisceau->g_energie, energie, INTERPOLE_LINEAIRE );
    }
  }

  /* Calcul terminé */
  SORTIE_FONCTION("FAISCEAU_LireIntensite", "%g", intensite);
#ifdef _DEVELOPPE_
  /* Sans affichage car normal dès que l'on a un faisceau projeté, à cause des erreurs d'arrondi */
/*   if ( 0.0 == intensite ) { */
/*     AVERTISSEMENT( "Intensité nulle [%s, X=%g, Y=%g]", faisceau->nom, x, y ); */
/*   } */
#endif
  return intensite;
}

/* -------- Contrôler la cohérence des divers paramètres -------- */

int FAISCEAU_ControlerOptions(Faisceau *faisceau, double energie)
{
  double y_faisceau;
  int reponse;

  if ( NULL == faisceau ) return -1;

  /* Dimensions */
  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Un faisceau ponctuel ne peut pas être inhomogène [Corrigé]\n");
#endif
      reponse = BI_ERR_alerte( NULL, &ERR_PonctuelInhomogene, VRAI );
      if ( 2 == reponse ) return +8;

      faisceau->proprietes &= ~ FAISCEAU_INHOMOGENE;
    }
    if ( faisceau->proprietes & FAISCEAU_FOCALISE ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Un faisceau ponctuel ne peut pas être focalisé [Corrigé]\n");
#endif
      reponse = BI_ERR_alerte( NULL, &ERR_PonctuelFocalise, VRAI );
      if ( 2 == reponse ) return +8;

      faisceau->proprietes &= ~ FAISCEAU_FOCALISE;
    }
  } else {
    if ( ( faisceau->largeur < 0.0 ) || ( faisceau->hauteur < 0.0 ) ) {
      (void) BI_ERR_alerte( NULL, &ERR_DimensionsAbsurdes, FAUX, faisceau->largeur, faisceau->hauteur);
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Dimensions de faisceau absurdes [%g × %g]\n", faisceau->largeur, faisceau->hauteur);
#endif
      return -2;
    }

    if ( 0.0 == faisceau->largeur ) {
      if ( 0.0 == faisceau->hauteur ) { /* Faisceau ponctuel */
	(void) fprintf(stderr, "%s,%s l.%u : Faisceau ponctuel, non-marqué comme tel\n",
		       __FILE__, __FUNCTION__, __LINE__);
	faisceau->proprietes |= FAISCEAU_PONCTUEL;
      } else {
	(void) fprintf(stderr, "FAISCEAU_ControlerOptions : Faisceau infiniment fin [L=0], non géré\n");
	return +1;
      }
    } else if ( 0.0 == faisceau->hauteur ) {
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Faisceau infiniment fin [H=0], non géré\n");
      return +2;
    }
  }

  /* Intensité */
  if ( faisceau->intensite <= 0.0 ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Intensité absurde [%g]\n", faisceau->intensite );
#endif
    (void) BI_ERR_alerte( NULL, &ERR_IntensiteAbsurde, FAUX, faisceau->intensite);
    return -3;
  }

  /* Direction de propagation */
  y_faisceau = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
  if ( y_faisceau <= 0.0 ) {
    Vecteur3D v_directeur;

#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Direction de propagation incorrecte\n");
#endif
    (void) REPERE_VecteurNormal( &(faisceau->orientation), &v_directeur );
    (void) BI_ERR_alerte( NULL, &ERR_DirectionAbsurde, FAUX,
			  faisceau->orientation.theta * 180.0 / M_PI, faisceau->orientation.phi * 180.0 / M_PI, /* Angles */
			  v_directeur.x, v_directeur.y, v_directeur.z); /* Coordonnées */
    return -4;
  }

  /* Focalisation */
  if ( faisceau->proprietes & FAISCEAU_FOCALISE ) {
    (void) fprintf(stderr, "FAISCEAU_ControlerOptions : À FAIRE --- Contrôle des ouvertures de focalisation\n");
  }

  /* Décroissance avec l'énergie nominale */
  if ( faisceau->proprietes & FAISCEAU_DECROIT ) {
    if ( NULL == faisceau->g_energie ) {
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : I(E) variable, mais non-indiquée\n");
      return -6;
    }
  }

  /* Chromatisme */
  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    double e_minimale;

    if ( NULL == faisceau->g_E ) {
      int reponse;

      reponse = BI_ERR_alerte( NULL, &ERR_gE_GrapheAbsent, VRAI );
      if ( 2 == reponse ) {
#ifdef _DEVELOPPE_
	(void) fprintf(stderr, "FAISCEAU_ControlerOptions : faisceau polychromatique, pas de graphe g(E) pour le fondamental\n");
#endif
	return -7;
      }
      faisceau->proprietes &= ~FAISCEAU_POLYCHROMATIQUE;
    }

    e_minimale = energie * ( 1.0 + faisceau->g_E->x_min );
    if ( e_minimale <= 0.0 ) {
      (void) BI_ERR_alerte( NULL, &ERR_gE_GrapheAbsurde, VRAI, energie );
      return -15;
    }
  }

  /* Harmoniques */
  if ( faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    unsigned int i;
    float H_totale = 0.0;

    if ( NULL == faisceau->i_harmoniques ) {
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Pas de table d'harmoniques\n");
      return -8;
    }

    for ( i = 0 ; i < faisceau->nbr_harmoniques ; i++ ) {
      if ( faisceau->i_harmoniques[ i ] < 0.0 ) {
	(void) fprintf(stderr, "FAISCEAU_ControlerOptions : Intensité d'harmonique absurde [%g, %ue harmonique]\n", faisceau->i_harmoniques[ i ], i + 1);
	return -8;
      }

      H_totale += faisceau->i_harmoniques[ i ];
    }
    if ( 0.0 == H_totale ) {
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Toutes les harmoniques sont éteintes, calcul des harmoniques inutile\n");
      faisceau->proprietes &= ~FAISCEAU_HARMONIQUES;
    }

    if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
      if ( NULL == faisceau->g_harmoniques ) {
	(void) fprintf(stderr, "FAISCEAU_ControlerOptions : Pas de table d'harmoniques / g(E)\n");
	return -8;
      }

      for ( i = 0 ; i < faisceau->nbr_harmoniques ; i++ ) {
	if ( NULL == faisceau->g_harmoniques[ i ] ) {
	  (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Harmonique %u dans graphe g(E)\n", i + 1);
	  /* Pas forcément faux (harmonique éteinte, p. ex.) => on se contente de le signaler */
	}
      }
    }
  }

  /* Homogénéité */
  if ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) {
    if ( NULL == faisceau->g_espace ) {
      (void) fprintf(stderr, "FAISCEAU_ControlerOptions : Pas de carte d'intensité\n");
      return -9;
    }
  }

  /* Pas de problème détecté */
  return 0;
}

/* -------- Projeter le faisceau sur l'échantillon --------
    Renvoie le faisceau tel qu'il est sur la surface de l'échantillon,
            c'est-à-dire avec une coupe exprimée dans le repère absolu
            avec l'origine du repère dans le plan d'entrée du faisceau
    ATTENTION, dans un tel faisceau l'axe de propagation n'est pas forcément orthogonal à la coupe !
 */

Faisceau *FAISCEAU_ProjeterSurEchantillon(Faisceau *faisceau, struct echantillon *echantillon)
{
  boolean dehors = FAUX; /* VRAI : le faisceau est entièrement en-dehors de l'échantillon */
  Faisceau *f_projete;

  if ( NULL == faisceau ) return NULL;
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'échantillon --- pas de contrôle de débordement\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Choix important : le faisceau est-il ponctuel ou non ? */
  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    f_projete = FSC_Projection_Ponctuel( faisceau, 0.0, NULL );
    if ( NULL == f_projete ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FSC_Projection_Ponctuel\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* On vérifie que l'on est bien dans l'échantillon
       Facile puisque le centre du repère local est sur le plan de l'échantillon
    */
    if ( f_projete->orientation.position.x < -(echantillon->larg / 2.0) )
      dehors = VRAI;
    else if ( f_projete->orientation.position.x > (echantillon->larg / 2.0) )
      dehors = VRAI;
    else if ( f_projete->orientation.position.z < -(echantillon->haut / 2.0) )
      dehors = VRAI;
    else if ( f_projete->orientation.position.z > (echantillon->haut / 2.0) )
      dehors = VRAI;
  } else {
    double pas_larg = -1.0, pas_haut = -1.0;

    /* On se cale sur la grille de l'échantillon */
    pas_larg = echantillon->pas_x;
    pas_haut = echantillon->pas_z;

    /* Mais si le faisceau est de taille inférieure à la grille, on prend plus petit */
    if ( faisceau->largeur <= 10.0 * pas_larg )
      pas_larg = faisceau->largeur / 10.0;
    if ( faisceau->hauteur <= 10.0 * pas_haut )
      pas_haut = faisceau->hauteur / 10.0;

    if ( ( 0.0 == faisceau->orientation.theta ) && ( 0.0 == faisceau->orientation.phi ) ) {
      f_projete = FSC_ProjectionOrthogonale( faisceau, pas_larg, pas_haut, 0.0, NULL);
    } else {
      f_projete = FSC_Projection_Etendu( faisceau, pas_larg, pas_haut, 0.0, NULL);
      if ( NULL == f_projete ) {
	(void) fprintf(stderr, "FAISCEAU_ProjeterSurEchantillon : ERREUR dans FSC_Projection_Etendu\n");
	return NULL;
      }
    }

    /* On vérifie que l'on rencontre l'échantillon */
    (void) fprintf(stderr, "FAISCEAU_ProjeterSurEchantillon : À FAIRE --- Contrôler que l'on rencontre l'échantillon\n");
  }

  /* Contrôle : est-on bien sur l'échantillon ? */
  if ( VRAI == dehors ) {
    (void) fprintf(stderr, "%s,%s l.%u : Le faisceau ne rencontre pas l'échantillon --- il taperait en (%g, %g, %g) ; échantillon de (%g, %g, %g) à (%g, %g, %g)\n",
		   __FILE__, __FUNCTION__, __LINE__, f_projete->orientation.position.x, f_projete->orientation.position.y, f_projete->orientation.position.z,
		   -echantillon->larg/2.0, 0.0, -echantillon->haut/2.0, echantillon->larg/2.0, echantillon->epaisseur, echantillon->haut/2.0 );
    printf( "Les différences : en X, %g ; en Y : %g\n", f_projete->orientation.position.x - (-echantillon->larg/2.0), f_projete->orientation.position.z - (-echantillon->haut/2.0) );
    if ( f_projete != faisceau ) {
      FAISCEAU_Effacer( f_projete );
    }
    return NULL;
  }

  /* On a fini la projection, tout va bien */
  return f_projete;
}

/* -------- Projection d'un faisceau ponctuel -------- */

static Faisceau *FSC_Projection_Ponctuel(Faisceau *faisceau, double energie, struct graphe *absorbance)
{
  Vecteur3D point_impact;
  Faisceau *f_projete;
  double distance;
  int retour;

  /* Coordonnées du point d'impact */
  retour = FAISCEAU_ProjeterCentre( faisceau, 0.0, &(point_impact.x), &(point_impact.z) );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_ProjeterCentre()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  point_impact.y = 0.0;

  /* Distance entre ce point et la source du faisceau */
  distance = U3D_Distance( &point_impact, &(faisceau->orientation.position) );

  /* Le faisceau est identique, SAUF S'IL Y A ABSORPTION PAR L'EXPÉRIENCE */
  if ( NULL == absorbance ) { /* Manip. dans le vide */
    f_projete = faisceau;
  } else { /* L'expérience absorbe : tenir compte de l'atténuation résultante */
    f_projete = FAISCEAU_Copier( faisceau );
    if ( NULL == f_projete ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FAISCEAU_Copier\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
  }

  /* Cela ne pose pas de problème de modifier la position du faisceau, puisque l'on reste sur la même droite ! */
  f_projete->orientation.position = point_impact;

  /* Atténuation due à l'environnement expérimental (air par exemple) */
  if ( absorbance != NULL ) {
    double mu;

    /* Le graphe doit contenir le coefficient d'absorption linéique de l'environnement expérimental */
    mu = GRF_LireValeur( absorbance, energie, INTERPOLE_LINEAIRE );
    f_projete->intensite *= exp( -mu * distance );
  }

  /* On a fini, tout s'est bien passé */
  return f_projete;
}

/* -------- Projection orthogonale d'un faisceau étendu ---------- */

static Faisceau *FSC_ProjectionOrthogonale(Faisceau *faisceau, double pas_x, double pas_z, double energie, struct graphe *absorbance)
{
  double x_p[ 5 ], y_p[ 5 ];
  double x_min = 0.0, y_min = 0.0, x_max = 0.0, y_max = 0.0; /* Région couverte par la projection */
  double f_transmise = 1.0; /* Fraction absorbée par l'expérience */
  int retour;
  unsigned int i;
  Faisceau *f_projete = NULL;
  Image *coupe = NULL;

  ENTREE_FONCTION(__FUNCTION__, "faisceau=%p pas_x=%g pas_z=%g energie=%g absorbance=%p", faisceau, pas_x, pas_z, energie, absorbance);

  /* Dimensions de la projection */
  retour = FAISCEAU_ProjeterLimites( faisceau, 0.0, x_p, y_p );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FSC_ProjectionOrthogonale : ERREUR dans FAISCEAU_ProjeterLimites()\n");
    SORTIE_FONCTION(__FUNCTION__, "%p [ERREUR]", NULL);
    return NULL;
  }

  /* Rectangle encadrant le faisceau */
  x_min = x_p[ 0 ]; y_min = y_p[ 0 ];
  x_max = x_p[ 0 ]; y_max = y_p[ 0 ];
  for ( i = 1 ; i < 4 ; i++ ) {
    if ( x_p[ i ] < x_min ) x_min = x_p[ i ];
    if ( y_p[ i ] < y_min ) y_min = y_p[ i ];
    if ( x_p[ i ] > x_max ) x_max = x_p[ i ];
    if ( y_p[ i ] > y_max ) y_max = y_p[ i ];
  }
#ifdef _DEVELOPPE_
  printf("%s : Dimensions du faisceau projeté : %g, %g -> %g, %g\n", __FUNCTION__, x_min, y_min, x_max, y_max);
#endif

  /* Fraction transmise par l'expérience */
  if ( absorbance != NULL ) {
    double distance; /* Parcourue par le faisceau avant de toucher l'échantillon --- identique en tout point */
    double mu;
    Vecteur3D point_impact;

    /* Distance parcourue */
    point_impact.x = x_p[ 4 ];
    point_impact.y = 0.0;
    point_impact.z = y_p[ 4 ];
    distance  = U3D_Distance( &point_impact, &(faisceau->orientation.position) );
    distance *= 1.0e-4; /* Distance en microns, on la veut en cm pour pouvoir utiliser mu/rho */

    /* Absorption par l'expérience */
    mu = GRF_LireValeur( absorbance, energie, INTERPOLE_LINEAIRE );

    /* Fraction de rayonnement transmis */
    f_transmise = exp( - mu * distance );
  }

  /* Cas simples : rotations de 0, 90, 180, 270 degrés */
  if ( 0.0 == faisceau->orientation.alpha ) {
    if ( NULL == absorbance ) {
      SORTIE_FONCTION(__FUNCTION__, "%p [0 sans absorbance]", faisceau);
#ifdef _DEVELOPPE_
/*       printf("%s : faisceau sortant = faisceau entrant\n", __FUNCTION__ ); */
#endif
      return faisceau; /* Le faisceau est absolument identique ! */
    }

    f_projete = FAISCEAU_Copier( faisceau );
    f_projete->intensite *= f_transmise;
    SORTIE_FONCTION(__FUNCTION__, "%p [0]", f_projete);
    return f_projete;
  }

  if ( M_PI_2 == faisceau->orientation.alpha ) {
    (void) fprintf(stderr, "FSC_ProjectionOrthogonale : À FAIRE --- rotation de 90 degrés\n");
    SORTIE_FONCTION(__FUNCTION__, "%p [90]", f_projete);
    return f_projete;
  }

  if ( M_PI == faisceau->orientation.alpha ) {
    (void) fprintf(stderr, "FSC_ProjectionOrthogonale : À FAIRE --- rotation de 180 degrés\n");
    SORTIE_FONCTION(__FUNCTION__, "%p [180]", f_projete);
    return f_projete;
  }

  if ( 3.0 * M_PI_2 == faisceau->orientation.alpha ) {
    (void) fprintf(stderr, "FSC_ProjectionOrthogonale : À FAIRE --- rotation de -90 degrés\n");
    SORTIE_FONCTION(__FUNCTION__, "%p [270]", f_projete);
    return f_projete;
  }

  /* -- Rotation quelconque -- */

  /* Rotation de l'image associée au faisceau */
  if ( NULL == faisceau->g_espace ) { /* Faisceau initialement homogène */
    Image *img_tmp;
    unsigned int j;
    double *valeurs;

    /* Grille arbitraire, puisque de toute façon elle est homogène ! */
    img_tmp = FAISCEAU_ReserverCoupe( faisceau->largeur, faisceau->hauteur, faisceau->largeur / 10.0, faisceau->hauteur / 10.0 );
    if ( NULL == img_tmp ) {
      (void) fprintf(stderr, "FSC_ProjectionOrthogonale : ERREUR dans FAISCEAU_ReserverCoupe()\n");
      SORTIE_FONCTION(__FUNCTION__, "%p [ERREUR]", NULL);
      return NULL;
    }
    for ( i = 0, valeurs = img_tmp->valeurs.R ; i < 11 ; i++ ) {
      for ( j = 0 ; j < 11 ; j++ ) {
	*valeurs = 1.0;
      }
    }
    coupe = IMG_rotation_quelconque( img_tmp, faisceau->orientation.alpha, pas_x, pas_z );
    IMG_EffacerImage( img_tmp, VRAI );
  } else {
    coupe = IMG_rotation_quelconque( faisceau->g_espace, faisceau->orientation.alpha, pas_x, pas_z );
  }
  if ( NULL == coupe ) {
    (void) fprintf(stderr, "FSC_ProjectionOrthogonale : ERREUR dans IMG_rotation_quelconque()\n");
    SORTIE_FONCTION(__FUNCTION__, "%p [coupe = NULL]", NULL);
    return NULL;
  }

  /* Génération du faisceau */
  f_projete = FAISCEAU_Creer( coupe->larg * coupe->l_reel, coupe->haut * coupe->h_reel,
			      faisceau->proprietes | FAISCEAU_INHOMOGENE, faisceau->nbr_harmoniques, faisceau->intensite );
  if ( NULL == f_projete ) {
    (void) fprintf(stderr, "FSC_ProjectionOrthogonale : ERREUR dans FAISCEAU_Creer()\n");
    SORTIE_FONCTION(__FUNCTION__, "%p [f_projete = NULL]", NULL);
    return NULL;
  }
  f_projete->orientation = faisceau->orientation; /* Même orientation, mais LA NORMALE N'EST PLUS NORMALE à la coupe */
  f_projete->point_focal = faisceau->point_focal; /* Même point focal */
  f_projete->g_energie   = faisceau->g_energie;   /* Même propriétés énergétiques */
  f_projete->g_E         = faisceau->g_E;
  for ( i = 0 ; i < f_projete->nbr_harmoniques ; i++ ) {
    f_projete->i_harmoniques[ i ] = faisceau->i_harmoniques[ i ];
    if ( f_projete->g_harmoniques != NULL ) {
      f_projete->g_harmoniques[ i ] = faisceau->g_harmoniques[ i ];
    }
  }
  f_projete->g_espace    = coupe; /* Structure spatiale : vient d'être calculée */

  SORTIE_FONCTION(__FUNCTION__, "%p", f_projete);
  return f_projete;
}

/* -------- Projection d'un faisceau étendu ---------- */

static Faisceau *FSC_Projection_Etendu(Faisceau *faisceau, double pas_x, double pas_z, double energie, struct graphe *absorbance)
{
  double x_p[ 5 ], y_p[ 5 ];
  double x_min = 0.0, y_min = 0.0, x_max = 0.0, y_max = 0.0; /* Région couverte par la projection */
  double M_projection[ 4 ];
  int retour;
  unsigned int i;
  Faisceau *f_projete;
  Image *coupe;

  /* Dimensions de la projection */
  retour = FAISCEAU_ProjeterLimites( faisceau, 0.0, x_p, y_p );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FSC_Projection_Etendu : ERREUR dans FAISCEAU_ProjeterLimites()\n");
    return NULL;
  }

  /* Rectangle encadrant le faisceau */
  x_min = x_p[ 0 ]; y_min = y_p[ 0 ];
  x_max = x_p[ 0 ]; y_max = y_p[ 0 ];
  for ( i = 1 ; i < 4 ; i++ ) {
    if ( x_p[ i ] < x_min ) x_min = x_p[ i ];
    if ( y_p[ i ] < y_min ) y_min = y_p[ i ];
    if ( x_p[ i ] > x_max ) x_max = x_p[ i ];
    if ( y_p[ i ] > y_max ) y_max = y_p[ i ];
  }
#ifdef _DEVELOPPE_
  printf("%s : Dimensions du faisceau projeté : %g, %g -> %g, %g\n", __FUNCTION__, x_min, y_min, x_max, y_max);
#endif

  /* Matrice de projection */
  {
    double c_alpha, c_theta, c_phi, s_alpha, s_theta, s_phi;
    double tg_phi, tg_theta;

    /* Précalcul des sinus & cosinus, puis des tangentes */
    c_alpha = cos( faisceau->orientation.alpha ); s_alpha = sin( faisceau->orientation.alpha );
    c_theta = cos( faisceau->orientation.theta ); s_theta = sin( faisceau->orientation.theta );
    c_phi   = cos( faisceau->orientation.phi )  ; s_phi   = sin( faisceau->orientation.phi );

    tg_theta = s_theta / c_theta;
    tg_phi = s_phi / c_phi;

    /* Calcul de la matrice de projection [en lignes] */
    M_projection[ 0 ] = c_alpha / c_theta + s_alpha * tg_theta * tg_phi;
    M_projection[ 1 ] = c_theta * s_alpha + s_alpha * s_theta * tg_theta - c_alpha * tg_theta * tg_phi;

    M_projection[ 2 ]= -s_alpha / c_phi;
    M_projection[ 3 ] = c_alpha / c_phi;
  }

  /* Quel pas utiliser pour construire la carte ? */
  if ( pas_x <= 0.0 ) { /* en X : pas à choisir */
    if ( faisceau->g_espace != NULL ) /* On prend le même */
      pas_x = faisceau->g_espace->l_reel;
    else {
      (void) fprintf(stderr, "FSC_Projection_Etendu : À FAIRE, choisir le pas en X\n");
    }
  }
  if ( pas_z <= 0.0 ) { /* en Y : pas à choisir */
    if ( faisceau->g_espace != NULL ) /* On prend le même */
      pas_z = faisceau->g_espace->h_reel;
    else {
      (void) fprintf(stderr, "FSC_Projection_Etendu : À FAIRE, choisir le pas en Y\n");
    }
  }

  /* Projection de l'image : simple transformation tant que la distance ne joue pas ! */
  if ( NULL == faisceau->g_espace ) { /* Faisceau initialement homogène */
    Image *img_tmp;
    unsigned int j;
    double *valeurs;

    /* Grille arbitraire, puisque de toute façon elle est homogène ! */
    img_tmp = FAISCEAU_ReserverCoupe( faisceau->largeur, faisceau->hauteur, faisceau->largeur / 10.0, faisceau->hauteur / 10.0 );
    if ( NULL == img_tmp ) {
      (void) fprintf(stderr, "FSC_Projection_Etendu : ERREUR dans FAISCEAU_ReserverCoupe()\n");
      return NULL;
    }
    /* Initialisée à 1 : aucune correction */
    for ( i = 0, valeurs = img_tmp->valeurs.R ; i < 11 ; i++ ) {
      for ( j = 0 ; j < 11 ; j++, valeurs++ ) {
	*valeurs = 1.0;
      }
    }
    coupe = IMG_TransformationQuelconque( img_tmp, M_projection, pas_x, pas_z );
    IMG_EffacerImage( img_tmp, VRAI );
  } else {
    coupe = IMG_TransformationQuelconque( faisceau->g_espace, M_projection, pas_x, pas_z );
  }
  if ( NULL == coupe ) {
    (void) fprintf(stderr, "FSC_Projection_Etendu : ERREUR dans IMG_TransformationQuelconque()\n");
    return NULL;
  }

  /* Génération du faisceau */
  f_projete = FAISCEAU_Creer( ( x_max - x_min ), ( y_max - y_min ),
			      faisceau->proprietes | FAISCEAU_INHOMOGENE, faisceau->nbr_harmoniques, faisceau->intensite );
  if ( NULL == f_projete ) {
    (void) fprintf(stderr, "FSC_Projection_Etendu : ERREUR dans FAISCEAU_Creer()\n");
    return NULL;
  }
#ifdef _DEVELOPPE_
  FAISCEAU_Nommer( f_projete, "Faisceau projeté" );
#endif
  f_projete->orientation = faisceau->orientation; /* Même orientation, mais LA NORMALE N'EST PLUS NORMALE à la coupe */
  f_projete->point_focal = faisceau->point_focal; /* Même point focal */
  f_projete->g_energie   = faisceau->g_energie;   /* Même propriétés énergétiques */
  f_projete->g_E         = faisceau->g_E;
  for ( i = 0 ; i < f_projete->nbr_harmoniques ; i++ ) {
    f_projete->i_harmoniques[ i ] = faisceau->i_harmoniques[ i ];
    if ( f_projete->g_harmoniques != NULL ) {
      f_projete->g_harmoniques[ i ] = faisceau->g_harmoniques[ i ];
    }
  }

  /* On associe la structure spatiale que l'on vient de calculer */
  f_projete->g_espace = FAISCEAU_ReserverCoupe( f_projete->largeur, f_projete->hauteur, pas_x, pas_z );
  if ( NULL == f_projete->g_espace ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver la coupe pour le projeté\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FAISCEAU_Effacer( f_projete ); f_projete = NULL;
    return NULL;
  }
  IMG_CalerSurImage( f_projete->g_espace, coupe );

  /* Plus besoin de la coupe originale */
  IMG_EffacerImage( coupe, VRAI ); coupe = NULL;

/*   f_projete->g_espace    = coupe; */ /* Structure spatiale : vient d'être calculée */
  /* Ce modèle de projection suppose que g(E) reste le même en tout point du faisceau.
     Ce n'est pas rigoureusement vrai dans le cas d'une expérience absorbante avec faisceau incliné,
     mais on peut supposer la différence négligeable devant le reste... */
  if ( absorbance != NULL ) {
    double mu __attribute__((__unused__));

    /* Absorption par l'expérience */
    mu = GRF_LireValeur( absorbance, energie, INTERPOLE_LINEAIRE );

    (void) fprintf(stderr, "À FAIRE : FSC_Projection_Etendu --- expérience absorbante\n");
  }

  return f_projete;
}
