/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Construction d'une description de configuration exprimentale
   Interface graphique pour le paramtrage de la composition d'un chantillon
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "chimie_utils.h" /* Analyse des formules chimiques... */
#include "echantillon.h"
#include "graphes.h"
#include "science.h" /* Pour les constantes physiques */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_CONSTITUANT    21
#define LGMAX_NOM_ELEMENT        15
#define LGMAX_NOM_GRAPHE_TOTAL   22
#define LGMAX_NOM_GRAPHE_PARTIEL 16

struct infos_affichage {
  unsigned int prem_composant;   /* Numro du premier constituant affich dans la liste */
  unsigned int composant_choisi; /* Numro du constituant affich */
  unsigned int prem_element;     /* Numro du premier lment affich dans la liste */
  unsigned int element_choisi;   /* Numro de l'lment affich */
  struct matiere *composition;
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GER_arbre_composition(unsigned int objet, int x, int y);

static void enregistrer_composant( struct infos_affichage *infos );
static void installer_formulation( struct infos_affichage *infos, boolean sens, struct composant *composant );
static void installer_composant( struct infos_affichage *infos, struct composant *composant, float f_massique );
static void prepare_arbre_composition(void);

static int IndexVZ( struct composant *composant, unsigned long element );

static const char *ObjetVAtome    ( unsigned int arbre, unsigned long element );
static const char *ObjetVComposant( unsigned int arbre, unsigned long composant );

/* ~-~-~-~-~-~-~-~- Gestion de l'interface utilisateurs ~-~-~-~-~-~-~-~- */

int Matiere_ModifierComposition( struct matiere *composition )
{
  struct infos_affichage infos;
  int retour;

  if ( NULL == composition ) return -1;

  if ( FAUX == arbres[ PARAM_COMPO ].drapeaux.prepare ) {
    prepare_arbre_composition();
  }

  /* On prpare l'affichage */
  infos.prem_composant   = 0;
  infos.composant_choisi = 0;
  infos.prem_element     = 0;
  infos.element_choisi   = 0;
  infos.composition = composition;

  /* Densit, nom,... */
  met_chaine( PARAM_COMPO, PC_NOM_COMPOSIT, 0, composition->nom, TXT_COPIER );
  OBJ_MettreNombre( PARAM_COMPO, PC_DENSITE, (double) composition->densite, 0.0 );

  /* On affiche la liste des composants */
  if ( NULL == composition->composants ) {
    ARB_PreparerDefileur_O( PARAM_COMPO, PC_CONSTITUANT_1, PC_DERNIER_CONST, LGMAX_NOM_CONSTITUANT, PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES,
			    0, NULL, 0, 0, ObjetVComposant );
  } else {
    ARB_PreparerDefileur_O( PARAM_COMPO, PC_CONSTITUANT_1, PC_DERNIER_CONST, LGMAX_NOM_CONSTITUANT, PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES,
			    composition->nbr_composants, NULL, 0, 0, ObjetVComposant );
  }

  /* On affiche le premier composant */
  if ( NULL == composition->composants ) {
    installer_composant( &infos, NULL, 0.0 );
  } else {
    installer_composant( &infos, composition->composants[ 0 ], composition->fraction_massique[ 0 ] );
  }

  ARB_MettreComplement( PARAM_COMPO, &infos );
  retour = ARB_Bloquant( PARAM_COMPO, "Composition locale", principale, 0 );
  ARB_MettreComplement( PARAM_COMPO, NULL );
  if ( PC_ANNULER == retour ) return 0;

  /* On mmorise les dernires modifications pour le constituant affich */
  enregistrer_composant( &infos );

  /* On mmorise la densit, le nom */
  composition->densite = (float) OBJ_LireNombre( PARAM_COMPO, PC_DENSITE, NULL );
  Matiere_Nommer( composition, lit_chaine( PARAM_COMPO, PC_NOM_COMPOSIT, 0, NULL, TXT_COPIER ) );

  /* S'il y avait des graphes associs, ils ne sont plus valables puisque la composition a chang */
  GRF_eff_isole( composition->fluo ); composition->fluo = NULL;
  GRF_eff_isole( composition->mu_total ); composition->mu_total = NULL;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Enregistrement des donnes dans l'arbre -------- */

static void enregistrer_composant( struct infos_affichage *infos )
{
  struct matiere *composition;

  composition = infos->composition;
  if ( NULL == composition ) return;
  if ( 0 == composition->nbr_composants ) return; /* Rien  enregistrer s'il n'y a encore aucun composant ! */

  /* On affiche le % en masse, on stocke la fraction massique */
  composition->fraction_massique[ infos->composant_choisi ] = OBJ_LireNombre( PARAM_COMPO, PC_PROPORTION, NULL ) / 100.0;

  composition->composants[ infos->composant_choisi ]->densite = (float) OBJ_LireNombre( PARAM_COMPO, PC_DST_COMPOSANT, NULL );

  /* Mmorisation du nom du composant */
  if ( VRAI == OBJ_choisit( PARAM_COMPO, PC_AV_TOTAL, QUESTION ) ) {
    Composant_ChangerNom( composition->composants[ infos->composant_choisi ], lit_chaine( PARAM_COMPO, PC_NOM, 0, NULL, TXT_COPIER ) );
  } else {
    char *nv_nom;

    nv_nom = TXT_genere_chaine( "%s [%s]", lit_chaine( PARAM_COMPO, PC_NOM, 0, NULL, TXT_COPIER ), lit_chaine( PARAM_COMPO, PC_FORMULE, 0, NULL, TXT_COPIER ) );
    Composant_ChangerNom( composition->composants[ infos->composant_choisi ], nv_nom );
    free( nv_nom );
  }
}

/* -------- Affichages dans l'arbre -------- */

static void installer_composant( struct infos_affichage *infos, struct composant *composant, float f_massique )
{
  if ( NULL == composant ) { /* Pas de composant -> les options suivantes n'ont pas de sens */
    OBJ_active( PARAM_COMPO, PC_NOM, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_FORMULE, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_EFFACER, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_DENSITE, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_PROP_DENSITE, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_AUTO, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_CHOIX_SPECTRE, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_ACCEPTER, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_MISEAJOUR, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_SPECTRE_TOTAL, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_AV_TOTAL, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_DST_COMPOSANT, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_PROPORTION, FORCE_NON );

    ARB_PreparerDefileur_O( PARAM_COMPO, PC_PREMIER_ATOME, PC_DERNIER_ATOME, LGMAX_NOM_ELEMENT, PCO_ATM_AVANT, PCO_FND_ATM, PCO_BOUTON_ATM, PCO_ATM_APRES,
			    0, NULL, 0, 0, ObjetVAtome );

    return;
  }

  OBJ_active( PARAM_COMPO, PC_DENSITE, FORCE_OUI );
  OBJ_active( PARAM_COMPO, PC_PROP_DENSITE, FORCE_OUI );
  OBJ_active( PARAM_COMPO, PC_EFFACER, FORCE_OUI );

  OBJ_active( PARAM_COMPO, PC_ACCEPTER, FORCE_OUI );
  OBJ_active( PARAM_COMPO, PC_AV_TOTAL, FORCE_OUI );

  OBJ_active( PARAM_COMPO, PC_NOM, FORCE_OUI );
  if ( composant->nom[ strlen( composant->nom ) - 1] == ']' ) {
    char *nom, *formule;

    nom = strdup( composant->nom );
    if ( NULL == nom ) {
      nom = composant->nom; /* Pas de problme particulier, juste que la fin sera perdue dans le dfileur... */
    }

    /* Dbut de la formule brute */
    formule = strrchr( nom, '[' );
    if ( formule != nom )
      *(formule - 1) = 0;
    else
      *formule = 0;
    formule++;

    formule[ strlen( formule ) - 1 ] = 0;

    met_chaine( PARAM_COMPO, PC_NOM, 0, nom, TXT_COPIER );
    met_chaine( PARAM_COMPO, PC_FORMULE, 0, formule, TXT_COPIER );

    if ( nom != composant->nom ) free( nom );
  } else {
    met_chaine( PARAM_COMPO, PC_NOM, 0, composant->nom, TXT_COPIER );
  }

  OBJ_active( PARAM_COMPO, PC_DST_COMPOSANT, FORCE_OUI );
  OBJ_MettreNombre( PARAM_COMPO, PC_DST_COMPOSANT, (double) composant->densite, 0.0 );

  OBJ_active( PARAM_COMPO, PC_PROPORTION, FORCE_OUI );
  OBJ_MettreNombre( PARAM_COMPO, PC_PROPORTION, 100.0 * (double) f_massique, 0.0 );

  if ( NULL == composant->mu_total ) {
    installer_formulation( infos, VRAI, composant );
  } else {
    infos->prem_element = 0; /* On recommence au dbut */
    infos->element_choisi = 0;

    installer_formulation( infos, FAUX, composant );
  }
}

/* -------- Installer la composition du constituant indiqu -------- */

static void installer_formulation( struct infos_affichage *infos, boolean sens, struct composant *composant )
{
  int nbr_elements;
  int Z;

  if ( ( FAUX == sens) && ( NULL == composant->mu_total ) ) { /* Pas de spectre d'absorption total : on force l'affichage de la composition */
    infos->prem_element = 0; /* On recommence au dbut, pour tre tranquille... */
    infos->element_choisi = 0;

    sens = VRAI;
  }

  if ( FAUX == sens ) { /* Inutile, car un spectre total est fourni */
    ARB_PreparerDefileur_O( PARAM_COMPO, PC_PREMIER_ATOME, PC_DERNIER_ATOME, LGMAX_NOM_ELEMENT, PCO_ATM_AVANT, PCO_FND_ATM, PCO_BOUTON_ATM, PCO_ATM_APRES,
			    0, NULL, 0, 0, ObjetVAtome );

    OBJ_active( PARAM_COMPO, PC_FORMULE, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_AUTO, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_CHOIX_SPECTRE, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_PROP_DENSITE, FORCE_NON );

    OBJ_choisit( PARAM_COMPO, PC_AV_TOTAL, FORCE_OUI );
    OBJ_active( PARAM_COMPO, PC_SPECTRE_TOTAL, FORCE_OUI );

    met_chaine( PARAM_COMPO, PC_SPECTRE_TOTAL, LGMAX_NOM_GRAPHE_TOTAL, composant->mu_total->nom, TXT_COPIER );

    return;
  }

  /* Si l'on est l, c'est que l'on veut l'affichage de la composition */
  OBJ_choisit( PARAM_COMPO, PC_AV_TOTAL, FORCE_NON );
  OBJ_active( PARAM_COMPO, PC_SPECTRE_TOTAL, FORCE_NON );

  OBJ_active( PARAM_COMPO, PC_FORMULE, FORCE_OUI );
  OBJ_active( PARAM_COMPO, PC_MISEAJOUR, FORCE_OUI );
  OBJ_active( PARAM_COMPO, PC_PROP_DENSITE, FORCE_OUI );

  nbr_elements = Composant_NombreElements( composant );
  if ( nbr_elements < 0 ) {
    (void) fprintf(stderr, "installer_formulation : ERREUR dans Composant_NombreElements() [code %d]\n", nbr_elements);
    return;
  }

  /* On affiche l'arbre de la composition */
  ARB_PreparerDefileur_O( PARAM_COMPO, PC_PREMIER_ATOME, PC_DERNIER_ATOME, LGMAX_NOM_ELEMENT, PCO_ATM_AVANT, PCO_FND_ATM, PCO_BOUTON_ATM, PCO_ATM_APRES,
			  (unsigned long) nbr_elements, NULL, infos->element_choisi, infos->prem_element, ObjetVAtome );

  if ( 0 == nbr_elements ) {
    OBJ_active( PARAM_COMPO, PC_AUTO, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_CHOIX_SPECTRE, FORCE_NON );

    return;
  }

  OBJ_active( PARAM_COMPO, PC_AUTO, FORCE_OUI );

  /* Quel est l'lment actuellement affich ? */
  Z = IndexVZ( composant, infos->element_choisi );

  /* On affiche ventuellement le spectre complmentaire */
  if ( NULL == composant->mu[ Z - 1 ] ) {
    OBJ_choisit( PARAM_COMPO, PC_AUTO, FORCE_OUI );
    OBJ_active( PARAM_COMPO, PC_CHOIX_SPECTRE, FORCE_NON );

    met_chaine( PARAM_COMPO, PC_CHOIX_SPECTRE, LGMAX_NOM_GRAPHE_PARTIEL, "Pas de spectre", TXT_COPIER );
  } else {
    OBJ_choisit( PARAM_COMPO, PC_AUTO, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_CHOIX_SPECTRE, FORCE_OUI );

    met_chaine( PARAM_COMPO, PC_CHOIX_SPECTRE, LGMAX_NOM_GRAPHE_PARTIEL, composant->mu[ Z - 1 ]->nom, TXT_COPIER );
  }
}

/* -------- Connaissant le numro du constituant slectionn, trouver le Z correspondant */

static int IndexVZ( struct composant *composant, unsigned long element )
{
  unsigned int i, i_element;

  i = 0;
  i_element = 0;
  do {
    while ( composant->atomes[ i ] <= 0.0 ) {
      i++;
      if ( i >= PLUTONIUM ) {
	(void) fprintf(stderr, "IndexVZ : Il n'y a pas autant d'lments diffrents... [element=%lu, max=%u]\n", element, i_element);
	return 0;
      }
    }

    /* Ici, l'lment i est utilis */
    i_element++;
    i++;
  } while ( i_element <= element );

  return i;
}

/* -------- Nom du i-me lment du constituant actuel -------- */

static const char *ObjetVAtome( unsigned int arbre, unsigned long element )
{
  struct composant *composant;
  struct infos_affichage *infos;
  int Z;

  ENTREE_FONCTION("ObjetVAtome", "arbre=%u element=%lu", arbre, element);

  infos = ( struct infos_affichage * ) ARB_LireComplement( arbre );
  if ( NULL == infos ) {
    SORTIE_FONCTION("ObjetVAtome", "%p [ERREUR : Pas d'infos d'affichage]", NULL);
    (void) fprintf(stderr, "ObjetVAtome : Pas d'infos d'affichage\n");
    return NULL;
  }

  if ( NULL == infos->composition->composants ) {
    SORTIE_FONCTION("ObjetVAtome", "%p [ERREUR : Pas de table des composants]", NULL);
    (void) fprintf(stderr, "ObjetVAtome : Pas de table des composants\n");
    return NULL;
  }
  composant = infos->composition->composants[ infos->composant_choisi ];
  if ( NULL == composant ) {
    SORTIE_FONCTION("ObjetVAtome", "%p [ERREUR : Composant choisi NULL]", NULL);
    (void) fprintf(stderr, "ObjetVAtome : Composant choisi NULL\n");
    return NULL;
  }

  /* On cherche le premier lment utilis */
  Z = IndexVZ( composant, element );

  /* Ici, i est donc le numro atomique de l'lment cherch */
  SORTIE_FONCTION("ObjetVAtome", "%p", ATM_Z_vers_nom( Z ));
  return ATM_Z_vers_nom( Z );
}

/* -------- Nom du i-me composant de la composition -------- */

static const char *ObjetVComposant( unsigned int arbre, unsigned long composant )
{
  struct infos_affichage *infos;

  infos = ( struct infos_affichage * ) ARB_LireComplement( arbre );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "ObjetVComposant : Pas d'infos d'affichage\n");
    return NULL;
  }

  if ( NULL == infos->composition->composants ) {
    (void) fprintf(stderr, "ObjetVComposant : Pas de table des composants\n");
    return NULL;
  }

  return infos->composition->composants[ composant ]->nom;
}

/* -------- Prparation de l'arbre -------- */

static void prepare_arbre_composition(void)
{
  if ( FAUX == arbres[ PARAM_COMPO ].drapeaux.init ) {
    int retour;

    OPTION_A_FAIRE( "Documentation : modification d'une composition" );
    retour = ARB_Initialiser( PARAM_COMPO, NULL, GER_arbre_composition, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans l'initialisation de l'arbre [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return;
    }
  }

  /* Liste des constituants impliqus */
  ARB_EnjoliverDefileur( PARAM_COMPO, PC_FOND_LISTE, PC_CONSTITUANT_1, PC_DERNIER_CONST,
			 PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES, ARB_ASCENSEUR_DROITE );

  /* Liste des atomes impliqus dans un constituant */
  ARB_EnjoliverDefileur( PARAM_COMPO, PC_FND_ATOMES, PC_PREMIER_ATOME, PC_DERNIER_ATOME,
			 PCO_ATM_AVANT, PCO_FND_ATM, PCO_BOUTON_ATM, PCO_ATM_APRES, ARB_ASCENSEUR_DROITE );

  /* Position d'autres objets */
  OBJ_deplace( PARAM_COMPO, PC_NOM_CONST, 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( PARAM_COMPO, PC_NOM      , 0, -1, DEPLACE_PIXELS );

  OBJ_deplace( PARAM_COMPO, PC_TXT_FORMULE, 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( PARAM_COMPO, PC_FORMULE    , 0, -1, DEPLACE_PIXELS );

  OBJ_deplace( PARAM_COMPO, PC_MISEAJOUR    , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( PARAM_COMPO, PC_TXT_SPC_TOTAL, -2,  0, DEPLACE_PIXELS );
  OBJ_deplace( PARAM_COMPO, PC_TXT_SPC_TOTAL, 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( PARAM_COMPO, PC_SPECTRE_TOTAL, 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( PARAM_COMPO, PC_AV_TOTAL     , 0, 500, DEPLACE_RELATIF );

  OBJ_deplace( PARAM_COMPO, PC_CHOIX_SPECTRE, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( PARAM_COMPO, PC_TXT_DENSITE , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAM_COMPO, PC_DENSITE     , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAM_COMPO, PC_PROP_DENSITE, 0, +1, DEPLACE_PIXELS );

  /* Types spciaux plus pratiques */
  OBJ_ChangerType( PARAM_COMPO, PC_FORMULE, G_TXT_INFINI );

  /* L'arbre est pret */
  arbres[ PARAM_COMPO ].drapeaux.prepare = VRAI;
}

/* -------- Gestion de l'arbre -------- */

static int GER_arbre_composition(unsigned int objet, int x, int y)
{
  struct infos_affichage *infos;
  int etat;

  infos = ( struct infos_affichage * ) ARB_LireComplement( PARAM_COMPO );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "GER_arbre_composition : Pas d'infos d'affichage\n");
    return BI_EVENEMENT_TRAITE;
  }

  if ( ( objet >= PC_CONSTITUANT_1 ) && ( objet <= PC_DERNIER_CONST ) ) {
    unsigned long nv_constituant = infos->composant_choisi,
      prem_affiche = infos->prem_composant;

    (void) ARB_GererDefileur_O( PARAM_COMPO, PC_CONSTITUANT_1, PC_DERNIER_CONST, LGMAX_NOM_CONSTITUANT, PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES,
				infos->composition->nbr_composants, NULL, &nv_constituant, &prem_affiche, ObjetVComposant,
				objet, y );
    infos->prem_composant = prem_affiche;
    if ( nv_constituant != infos->composant_choisi ) {
      enregistrer_composant( infos );
      infos->composant_choisi = nv_constituant;
      /* On met  jour le dfileur, car le nom a pu changer ! */
      ARB_PreparerDefileur_O( PARAM_COMPO, PC_CONSTITUANT_1, PC_DERNIER_CONST, LGMAX_NOM_CONSTITUANT, PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES,
			      infos->composition->nbr_composants, NULL, infos->composant_choisi, infos->prem_composant, ObjetVComposant );
      installer_composant( infos, infos->composition->composants[ nv_constituant ], infos->composition->fraction_massique[ nv_constituant ] );
    }

    return BI_EVENEMENT_TRAITE;
  }

  if ( ( ( objet >= PC_PREMIER_ATOME ) && ( objet <= PC_DERNIER_ATOME ) ) ||
       ( PCO_ATM_AVANT == objet ) || ( PCO_FND_ATM == objet ) || ( PCO_BOUTON_ATM == objet ) || ( PCO_ATM_APRES == objet ) ) {
    unsigned long nv_atome, prem_affiche;
    struct composant *composant;

    composant = infos->composition->composants[ infos->composant_choisi ];
    nv_atome = infos->element_choisi;
    prem_affiche = infos->prem_element;

    (void) ARB_GererDefileur_O( PARAM_COMPO, PC_PREMIER_ATOME, PC_DERNIER_ATOME, LGMAX_NOM_ELEMENT, PCO_ATM_AVANT, PCO_FND_ATM, PCO_BOUTON_ATM, PCO_ATM_APRES,
				Composant_NombreElements( composant ), NULL, &nv_atome, &prem_affiche, ObjetVAtome, objet, y );
    infos->prem_element = (unsigned int) prem_affiche;
    if ( nv_atome != infos->element_choisi ) {
/*       enregistrer_composant( infos ); */ /* Pour ne pas perdre le nom, la formule, la f. massique en rinstallant */
      infos->element_choisi = nv_atome;
      installer_formulation( infos, VRAI, infos->composition->composants[ infos->composant_choisi ] );
    }

    return BI_EVENEMENT_TRAITE;
  }

  switch ( objet )
  {
  case PC_SOLIDE:
    OBJ_choisit( PARAM_COMPO, objet, FORCE_OUI );
    OBJ_active( PARAM_COMPO, PCESOLVANT, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_TMP, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_PRESSION, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_MASSE_MOLAIRE, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case PC_SOLUTION:
    OBJ_choisit( PARAM_COMPO, objet, FORCE_OUI );
#ifdef _DEVELOPPE_
    OBJ_active( PARAM_COMPO, PCESOLVANT, FORCE_OUI );
#endif
    OBJ_active( PARAM_COMPO, PC_TMP, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_PRESSION, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_MASSE_MOLAIRE, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case PC_GAZ:
    OBJ_choisit( PARAM_COMPO, objet, FORCE_OUI );
    OBJ_active( PARAM_COMPO, PCESOLVANT, FORCE_NON );
    OBJ_active( PARAM_COMPO, PC_TMP, FORCE_OUI );
    OBJ_active( PARAM_COMPO, PC_PRESSION, FORCE_OUI );
    OBJ_active( PARAM_COMPO, PC_MASSE_MOLAIRE, FORCE_OUI );
    return BI_EVENEMENT_TRAITE;

  case PC_NOUVEAU: /* Crer un nouveau constituant */
    {
      struct composant *nouveau;

      OBJ_choisit( PARAM_COMPO, objet, FORCE_OUI );
      enregistrer_composant( infos ); /* Mme si on ne cre pas, peut toujours servir... */
      nouveau = Matiere_AjouterComposant( infos->composition, NULL );
      if ( nouveau != NULL ) {
	infos->composant_choisi = infos->composition->nbr_composants - 1;
	ARB_PreparerDefileur_O( PARAM_COMPO, PC_CONSTITUANT_1, PC_DERNIER_CONST, LGMAX_NOM_CONSTITUANT, PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES,
				infos->composition->nbr_composants, NULL, infos->composant_choisi, infos->prem_composant, ObjetVComposant );
	installer_composant( infos, nouveau, 1.0 );
      }
      OBJ_choisit( PARAM_COMPO, objet, FORCE_NON );
    }
    return BI_EVENEMENT_TRAITE;

  case PC_EFFACER:
    {
      (void) fprintf(stderr, "GER_arbre_composition :  FAIRE --- ter un constituant de la liste\n");
    }
    return BI_EVENEMENT_TRAITE;

  case PC_CHOISIR:
    {
      struct composant *nouveau;

      OBJ_choisit( PARAM_COMPO, objet, FORCE_OUI );
      enregistrer_composant( infos ); /* Mme si on ne choisit pas, peut toujours servir... */
      nouveau = ChoisirComposant( ARB_QuelleFenetre( arbres + PARAM_COMPO, FAUX ) );
      if ( nouveau != NULL ) {
	nouveau = Matiere_AjouterComposant( infos->composition, nouveau );
	if ( nouveau != NULL ) {
	  infos->composant_choisi = infos->composition->nbr_composants - 1;
	  ARB_PreparerDefileur_O( PARAM_COMPO, PC_CONSTITUANT_1, PC_DERNIER_CONST, LGMAX_NOM_CONSTITUANT, PC_LST_AVANT, PC_LST_FOND, PC_LST_BOUTON, PC_LST_APRES,
				  infos->composition->nbr_composants, NULL, infos->composant_choisi, infos->prem_composant, ObjetVComposant );
	  installer_composant( infos, nouveau, 1.0 );
	}
	OBJ_choisit( PARAM_COMPO, objet, FORCE_NON );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case PC_PROP_DENSITE: /* Estimation de la densit globale : moyenne pondre des densits partielles */
    (void) Matiere_EstimerDensite( infos->composition );
    OBJ_MettreNombre( PARAM_COMPO, PC_DENSITE, (double) infos->composition->densite, 0.0 );
    return BI_EVENEMENT_TRAITE;

  case PC_AUTO:
    etat = OBJ_choisit( PARAM_COMPO, PC_AUTO, QUESTION );
    if ( VRAI == etat ) { /* On enlve le calcul automatique -> on cherche un seuil */
      Graphe *nv_graphe;

      nv_graphe = GRF_choisir( GR_MOINSMAT, FAUX, VRAI, NULL );
      if ( NULL != nv_graphe ) {
	int Z;

	Z = IndexVZ( infos->composition->composants[ infos->composant_choisi ], infos->element_choisi );
	infos->composition->composants[ infos->composant_choisi ]->mu[ Z - 1 ] = nv_graphe;
      }
    } else { /* On remet le calcul automatique */
      int Z;

      Z = IndexVZ( infos->composition->composants[ infos->composant_choisi ], infos->element_choisi );
      infos->composition->composants[ infos->composant_choisi ]->mu[ Z - 1 ] = NULL;
    }
    installer_composant( infos, infos->composition->composants[ infos->composant_choisi ], infos->composition->fraction_massique[ infos->composant_choisi ] );
    return BI_EVENEMENT_TRAITE;

  case PC_CHOIX_SPECTRE:
    {
      Graphe *nv_graphe;

      nv_graphe = GRF_choisir( GR_MOINSMAT, FAUX, VRAI, NULL );
      if ( NULL != nv_graphe ) {
	int Z;

	Z = IndexVZ( infos->composition->composants[ infos->composant_choisi ], infos->element_choisi );
	infos->composition->composants[ infos->composant_choisi ]->mu[ Z - 1 ] = nv_graphe;
	installer_composant( infos, infos->composition->composants[ infos->composant_choisi ], infos->composition->fraction_massique[ infos->composant_choisi ] );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case PC_AV_TOTAL:
    etat = OBJ_choisit( PARAM_COMPO, objet, ECHANGE );
    if ( ( VRAI == etat ) && ( NULL == infos->composition->composants[ infos->composant_choisi ]->mu_total ) ) {
      infos->composition->composants[ infos->composant_choisi ]->mu_total = GRF_choisir( GR_MUMATRICE, FAUX, VRAI, NULL );
    }
    installer_formulation( infos, !etat, infos->composition->composants[ infos->composant_choisi ] );
    return BI_EVENEMENT_TRAITE;

  case PC_SPECTRE_TOTAL:
    {
      Graphe *nv_graphe;

      nv_graphe = GRF_choisir( GR_MUMATRICE, FAUX, VRAI, NULL );
      if ( NULL != nv_graphe ) {
	infos->composition->composants[ infos->composant_choisi ]->mu_total = nv_graphe;
	installer_formulation( infos, FAUX, infos->composition->composants[ infos->composant_choisi ] );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case PC_MISEAJOUR:
    {
      char *formule;

      formule = lit_chaine( PARAM_COMPO, PC_FORMULE, 0, NULL, TXT_REMPLACE );
      etat = Composant_AnalyserFormule( infos->composition->composants[ infos->composant_choisi ], formule );
      if ( etat != 0 ) {
	(void) fprintf(stderr, "GER_arbre_composition : ERREUR dans Composant_AnalyserFormule\n");
      }
      installer_formulation( infos, VRAI, infos->composition->composants[ infos->composant_choisi ] );
    }
    return BI_EVENEMENT_TRAITE;

  case PC_PROP_DS_CMP:
    if ( VRAI == OBJ_choisit( PARAM_COMPO, PC_GAZ, QUESTION ) ) {
      double temperature, pression, masse_molaire;
      double densite;

      temperature = OBJ_LireNombre( PARAM_COMPO, PC_TMP, NULL ) + 273.15; /* Affiche en C, travail en K */
      pression = OBJ_LireNombre( PARAM_COMPO, PC_PRESSION, NULL ) * 1.0e5; /* Affiche en bar, travail en Pa */
      masse_molaire = infos->composition->composants[ infos->composant_choisi ]->masse_molaire;

      densite = masse_molaire * pression / ( GAZ_PARFAIT * temperature ); /* Donne la masse volumique en g/m */
      printf("Densit : %g g/m\n", densite);
      densite /= 1.0e6; /* et on la veut en g/cm */

      infos->composition->composants[ infos->composant_choisi ]->densite = densite;
    } else { /* Solide, liquide : moyenne pondre des densits atomiques... */
      (void) Composant_EstimerDensite( infos->composition->composants[ infos->composant_choisi ] );
    }
    OBJ_MettreNombre( PARAM_COMPO, PC_DST_COMPOSANT, infos->composition->composants[ infos->composant_choisi ]->densite, 0.0 );
    return BI_EVENEMENT_TRAITE;
  }

#ifdef _DEVELOPPE_
  (void) fprintf(stderr, "GER_arbre_composition : objet non gr [%u, clic : x=%d, y=%d]\n", objet, x, y);
#endif
  return BI_EVENEMENT_NON_TRAITE;
}
