/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modles structurels d'une station exprimentale
   Paramtrer l'chantillon [interface utilisateur]
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"
#include "exafs.h"

#include "absorption.h" /* Pour calcul de la diffusion */
#include "echantillon.h"
#include "images.h"
#include "interface.h"
#include "science.h" /* pour M_PI, au cas-o */
#include "serie.h"
#include "statistiques.h" /* pour gaussienne() */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_COMPOSITION_UNIQUE    26
#define LGMAX_NOM_COMPOSITION_SELECTEUR 24

struct echantillon *premier_echantillon = NULL;

#define OPT_COMPO_APPLIQUER_IMAGE 0 /* Appliquer une image extrieure */
#define OPT_COMPO_ALEATOIRE       1 /* Gnrer une image alatoire */
#define OPT_COMPO_ALEATOIRE_TOUT  3 /* Gnrer alatoirement un chantillon complet */
#define OPT_COMPO_MU_IDEAL        5 /* Coefficient d'absorption idal en un point */
#define OPT_COMPO_FLUO_IDEALE     6 /* Spectre de fluorescence idal en un point */

#define NOMBRE_OPTIONS_COMPOSITION 7

#include "param_echantillon.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void preparer_arbre(struct echantillon *echantillon);
static int GER_arbre(unsigned int objet, int x, int y);

static double *ECH_DemanderRepartition(unsigned long *nbr_valeurs);

static const char *ECH_Composition_v_Texte(unsigned int arbre, unsigned long numero_objet);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

Echantillon *ECH_CreerEchantillon(void)
{
  double larg, haut, epaisseur;
  double pas_l, pas_h, pas_e;
  int retour, type, proprietes = 0;
  Echantillon *echantillon;

  /* On dsactive la partie composition */
  preparer_arbre( NULL );

  /* On appelle l'arbre */
  ARB_MettreComplement( PARAMS_ECHAN, NULL );
  retour = ARB_Bloquant( PARAMS_ECHAN, titre_param_echantillon, principale, 0 );
  if ( PE_ANNULER == retour )
    return NULL;

  /* Dimensions demandes */
  larg = OBJ_LireNombre( PARAMS_ECHAN, PE_LARGEUR, NULL ) * 1e4; /* Affichage en cm, interne en microns */
  haut = OBJ_LireNombre( PARAMS_ECHAN, PE_HAUTEUR, NULL ) * 1e4;
  epaisseur = OBJ_LireNombre( PARAMS_ECHAN, PE_EPAISSEUR, NULL ) * 1e4;

  /* Pas utiliss */
  pas_l = OBJ_LireNombre( PARAMS_ECHAN, PE_PAS_LARGEUR, NULL ) * 1e3; /* Affichage en mm, interne en microns */
  pas_h = OBJ_LireNombre( PARAMS_ECHAN, PE_PAS_HAUTEUR, NULL ) * 1e3;
  pas_e = OBJ_LireNombre( PARAMS_ECHAN, PE_PAS_EPAISSEUR, NULL ) * 1e3;
#ifdef _DEVELOPPE_
  printf( "%s : L=%g/%g H=%g/%g e=%g/%g\n", __FUNCTION__, larg, pas_l, haut, pas_h, epaisseur, pas_e );
#endif

  /* Organisation interne */
  if ( VRAI == OBJ_choisit( PARAMS_ECHAN, PE_HOMOGENE, QUESTION ) ) {
    type = ECHANTILLON_MATERIEL;
    proprietes = ECHANTILLON_HOMOGENE;
  } else if ( VRAI == OBJ_choisit( PARAMS_ECHAN, PE_GRADIENT, QUESTION ) ) {
    type = ECHANTILLON_GRADIENT;
  } else {
    type = ECHANTILLON_MATERIEL;
  }

  /* On cre l'chantillon */
  echantillon = ECH_CreerEchantillonP( larg, haut, epaisseur, pas_l, pas_h, pas_e, type, proprietes );
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "ECH_CreerEchantillon : ERREUR dans ECH_CreerEchantillonP()\n");
    return NULL;
  }
  /* Mmorisation du nom */
  ECHANTILLON_Nommer( echantillon, lit_chaine( PARAMS_ECHAN, PE_NOM, 0, NULL, TXT_COPIER ) );

  /* On a fini, tout s'est bien pass */
  return echantillon;
}

int ECHANTILLON_Parametrer(Echantillon *echantillon)
{
  int retour;
  double nv_larg, nv_haut, nv_epais, nvpas_l, nvpas_h, nvpas_e;

  if ( NULL == echantillon ) return -1;

  /* On installe les paramtres */
  met_chaine( PARAMS_ECHAN, PE_NOM, 0, echantillon->nom, TXT_COPIER );

  OBJ_MettreNombre( PARAMS_ECHAN, PE_LARGEUR  , echantillon->larg / 1.0e4, 0.0 ); /* Affichage en cm, interne en microns */
  OBJ_MettreNombre( PARAMS_ECHAN, PE_HAUTEUR  , echantillon->haut / 1.0e4, 0.0 );
  OBJ_MettreNombre( PARAMS_ECHAN, PE_EPAISSEUR, echantillon->epaisseur / 1.0e4, 0.0 );

  OBJ_MettreNombre( PARAMS_ECHAN, PE_PAS_LARGEUR  , echantillon->pas_x / 1.0e3, 0.0 ); /* Affichage en mm, interne en microns */
  OBJ_MettreNombre( PARAMS_ECHAN, PE_PAS_HAUTEUR  , echantillon->pas_z / 1.0e3, 0.0 );
  OBJ_MettreNombre( PARAMS_ECHAN, PE_PAS_EPAISSEUR, echantillon->pas_y / 1.0e3, 0.0 );

  /* On active la partie composition */
  preparer_arbre( echantillon );

  /* On appelle l'arbre */
  ARB_MettreComplement( PARAMS_ECHAN, echantillon );
  retour = ARB_Bloquant( PARAMS_ECHAN, titre_param_echantillon, principale, 0 );
  ARB_MettreComplement( PARAMS_ECHAN, NULL );
  if ( PE_ANNULER == retour )
    return +1;

  /* Mmorisation des paramtres, si ncessaire */
  ECHANTILLON_Nommer( echantillon, lit_chaine( PARAMS_ECHAN, PE_NOM, 0, NULL, TXT_COPIER ) );

  /* Nouvelles dimensions */
  nv_larg  = OBJ_LireNombre( PARAMS_ECHAN, PE_LARGEUR, NULL ) * 1.0e4; /* cm ->  */
  nv_haut  = OBJ_LireNombre( PARAMS_ECHAN, PE_HAUTEUR, NULL ) * 1.0e4; /* cm ->  */
  nv_epais = OBJ_LireNombre( PARAMS_ECHAN, PE_EPAISSEUR, NULL ) * 1.0e4; /* cm ->  */

  /* Nouveau pas, ventuellement */
  nvpas_l = OBJ_LireNombre( PARAMS_ECHAN, PE_PAS_LARGEUR, NULL ) * 1.0e3; /* mm ->  */
  nvpas_h = OBJ_LireNombre( PARAMS_ECHAN, PE_PAS_HAUTEUR, NULL ) * 1.0e3; /* mm ->  */
  nvpas_e = OBJ_LireNombre( PARAMS_ECHAN, PE_PAS_EPAISSEUR, NULL ) * 1.0e3; /* mm ->  */

  /* Ces dimensions ont-elle chang ? */
  if ( ( fabs( nv_larg - echantillon->larg ) > 1.e-3 ) || /* Prcision de plus du 1/1000 de  inutile ! */
       ( fabs( nv_haut - echantillon->haut ) > 1.e-3 ) || ( fabs( nv_epais - echantillon->epaisseur ) > 1.e-3 ) ) {
    if ( ( fabs( nvpas_l - echantillon->pas_x ) > 1.e-3 ) || /* Prcision de plus du 1/1000 de  inutile ! */
	 ( fabs( nvpas_h - echantillon->pas_z ) > 1.e-3 ) || ( fabs( nvpas_e - echantillon->pas_y ) > 1.e-3 ) ) {
      retour = ECH_ChangerGrille( echantillon,  nv_larg, nv_haut, nv_epais, nvpas_l, nvpas_h, nvpas_e );
    } else {
      /* Pas inchang */
      retour = ECH_Redimensionner( echantillon, nv_larg, nv_haut, nv_epais );
    }
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de redimensionner l'chantillon [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      return +1;
    }
  }

  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Mmorisation des paramtres\n",
		 __FILE__, __FUNCTION__, __LINE__);

  return 0;
}

/* ~-~-~-~-~-~-~-~- Gestion de l'arbre ~-~-~-~-~-~-~-~- */

static int GER_arbre(unsigned int objet, int x, int y)
{
  Echantillon *echantillon;

  echantillon = (Echantillon *) ARB_LireComplement( PARAMS_ECHAN );

  switch ( objet )
  {
  case PE_HOMOGENE:
  case PE_GRADIENT:
  case PE_INHOMOGENE:
    OBJ_choisit( PARAMS_ECHAN, objet, FORCE_OUI );
    preparer_arbre( echantillon );
    if ( echantillon != NULL ) {
      (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Modification du type d'chantillon\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }
    return BI_EVENEMENT_TRAITE;

  case PE_CMP_UNIQUE:
    echantillon->valeurs.homogene.composition = Matiere_ChoisirComposition( PARAMS_ECHAN, PE_CMP_UNIQUE, LGMAX_NOM_COMPOSITION_UNIQUE, NULL );
    return BI_EVENEMENT_TRAITE;
  }

  /* Bouton non-gr */
  (void) fprintf(stderr, "%s,%s l.%u : Bouton %u non-gr [x=%d, y=%d]\n",
		 __FILE__, __FUNCTION__, __LINE__, objet, x, y);
  return BI_EVENEMENT_NON_TRAITE;
}

static void preparer_arbre( Echantillon *echantillon )
{
  /* Quels boutons activer ? */
  if ( NULL == echantillon ) {
    OBJ_active( PARAMS_ECHAN, PE_CMP_UNIQUE, FORCE_NON );
    met_chaine( PARAMS_ECHAN, PE_CMP_UNIQUE, LGMAX_NOM_COMPOSITION_UNIQUE, texte_sans_composition, TXT_COPIER );

    OBJ_active( PARAMS_ECHAN, PE_AJOUTER, FORCE_NON );
    OBJ_active( PARAMS_ECHAN, PE_ENLEVER, FORCE_NON );

    ARB_PreparerDefileur_O( PARAMS_ECHAN, PE_COMPO_1, PE_COMPO_N, LGMAX_NOM_COMPOSITION_SELECTEUR, PE_CMP_AVANT, PE_FND_ASC, PE_BOUTON_ASC, PE_CMP_APRES,
			    0, NULL, 0, 0, ECH_Composition_v_Texte );

    met_chaine( PARAMS_ECHAN, PE_OK, 0, texte_bouton_OK_creer, TXT_REMPLACE );
  } else {
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      OBJ_choisit( PARAMS_ECHAN, PE_HOMOGENE, FORCE_OUI );

      OBJ_active( PARAMS_ECHAN, PE_CMP_UNIQUE, FORCE_OUI );

      OBJ_active( PARAMS_ECHAN, PE_AJOUTER, FORCE_NON );
      OBJ_active( PARAMS_ECHAN, PE_ENLEVER, FORCE_NON );

      ARB_PreparerDefileur_O( PARAMS_ECHAN, PE_COMPO_1, PE_COMPO_N, LGMAX_NOM_COMPOSITION_SELECTEUR, PE_CMP_AVANT, PE_FND_ASC, PE_BOUTON_ASC, PE_CMP_APRES,
			      0, NULL, 0, 0, ECH_Composition_v_Texte );

      if ( NULL == echantillon->valeurs.homogene.composition ) {
	met_chaine( PARAMS_ECHAN, PE_CMP_UNIQUE, LGMAX_NOM_COMPOSITION_UNIQUE, texte_sans_composition, TXT_COPIER );
      } else {
	met_chaine( PARAMS_ECHAN, PE_CMP_UNIQUE, LGMAX_NOM_COMPOSITION_UNIQUE, echantillon->valeurs.homogene.composition->nom, TXT_COPIER );
      }
    } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      OBJ_choisit( PARAMS_ECHAN, PE_GRADIENT, FORCE_OUI );

      OBJ_active( PARAMS_ECHAN, PE_CMP_UNIQUE, FORCE_OUI );

      OBJ_active( PARAMS_ECHAN, PE_AJOUTER, FORCE_NON );
      OBJ_active( PARAMS_ECHAN, PE_ENLEVER, FORCE_NON );

      ARB_PreparerDefileur_O( PARAMS_ECHAN, PE_COMPO_1, PE_COMPO_N, LGMAX_NOM_COMPOSITION_SELECTEUR, PE_CMP_AVANT, PE_FND_ASC, PE_BOUTON_ASC, PE_CMP_APRES,
			      0, NULL, 0, 0, ECH_Composition_v_Texte );

      if ( NULL == echantillon->valeurs.homogene.composition ) {
	met_chaine( PARAMS_ECHAN, PE_CMP_UNIQUE, LGMAX_NOM_COMPOSITION_UNIQUE, texte_sans_composition, TXT_COPIER );
      } else {
	met_chaine( PARAMS_ECHAN, PE_CMP_UNIQUE, LGMAX_NOM_COMPOSITION_UNIQUE, echantillon->valeurs.homogene.composition->nom, TXT_COPIER );
      }
    } else {
      OBJ_choisit( PARAMS_ECHAN, PE_INHOMOGENE, FORCE_OUI );

      OBJ_active( PARAMS_ECHAN, PE_CMP_UNIQUE, FORCE_NON );

      OBJ_active( PARAMS_ECHAN, PE_AJOUTER, FORCE_OUI );
      OBJ_active( PARAMS_ECHAN, PE_ENLEVER, FORCE_OUI );

      (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Compositions multiples\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }

    met_chaine( PARAMS_ECHAN, PE_OK, 0, texte_bouton_OK_modifier, TXT_REMPLACE );
  }

  if ( VRAI == arbres[ PARAMS_ECHAN ].drapeaux.prepare ) return;

  /* Disposition des objets */
  OBJ_deplace( PARAMS_ECHAN, PE_TXT_H    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_ECHAN, PE_HAUTEUR  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_ECHAN, PE_TXT_E    , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_ECHAN, PE_EPAISSEUR, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_ECHAN, PE_PAS_HAUTEUR  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_ECHAN, PE_TXT_MM_H     , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_ECHAN, PE_PAS_EPAISSEUR, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( PARAMS_ECHAN, PE_TXT_MM_E     , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( PARAMS_ECHAN, PE_CMP_UNIQUE, 0, +1, DEPLACE_PIXELS );

  ARB_EnjoliverDefileur( PARAMS_ECHAN, PE_FND_LISTE, PE_COMPO_1, PE_COMPO_N,
			 PE_CMP_AVANT, PE_FND_ASC, PE_BOUTON_ASC, PE_CMP_APRES, ARB_ASCENSEUR_DROITE );

  arbres[ PARAMS_ECHAN ].drapeaux.prepare = VRAI;
  arbres[ PARAMS_ECHAN ].fonction.arbre = GER_arbre;
}

static const char *ECH_Composition_v_Texte(unsigned int arbre, unsigned long objet)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- arbre=%u, objet=%lu\n",
		 __FILE__, __FUNCTION__, __LINE__, arbre, objet);

  return NULL;
}

/* ~-~-~-~-~-~-~-~- Routines pour le slecteur universel ~-~-~-~-~-~-~-~- */

/* -------- Le chanage -------- */

int ECH_ChainerEchantillon(Echantillon *avant, Echantillon *echantillon)
{
  if ( NULL == echantillon ) return -1;

  if ( NULL == premier_echantillon ) {
    premier_echantillon = echantillon;
    return 0;
  }

  /* Ajout en dbut de liste */
  if ( NULL == avant ) {
    echantillon->suivant = premier_echantillon;
    premier_echantillon = echantillon;

    return 0;
#ifdef _DEVELOPPE_
  } else {
    Echantillon *d;

    for ( d = premier_echantillon ; d != avant ; d = d->suivant ) {
      if ( NULL == d ) {
	(void) fprintf(stderr, "%s,%s l.%u : le dtecteur indiqu comme avant n'est pas dans la liste !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	abort();
      }
    }
#endif
  }


  /* Ajout aprs celui indiqu */
  echantillon->suivant = avant->suivant;
  avant->suivant = echantillon;

  return 0;
}

int ECH_DechainerEchantillon(Echantillon *echantillon)
{
  Echantillon *avant;

  if ( NULL == echantillon ) return -1;

  if ( echantillon == premier_echantillon ) {
    premier_echantillon = echantillon->suivant;
    return 0;
  }

  for ( avant = premier_echantillon ; avant->suivant != echantillon ; avant = avant->suivant ) {
    if ( NULL == avant ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : dtecteur hors-chanage\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      return +1;
    }
  }

  avant->suivant = echantillon->suivant;
  return 0;
}

/* -------- Trouver l'chantillon demand -------- */

unsigned long ECHANTILLON_E_Vers_N(Echantillon *premier, Echantillon *echantillon)
{
  Echantillon *idx_e;
  unsigned long i;

  if ( NULL == premier ) premier = premier_echantillon;
  if ( NULL == echantillon ) return 0;

  for ( idx_e = premier, i = 0 ; idx_e != echantillon ; idx_e = idx_e->suivant, i++ ) {
    if ( NULL == idx_e ) {
      (void) fprintf(stderr, "%s,%s l.%u : chantillon non-trouv [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, echantillon->nom);
      return 0;
    }
  }

  return i;
}

Echantillon *ECHANTILLON_N_Vers_E(Echantillon *premier, unsigned long n)
{
  Echantillon *echantillon;
  unsigned long i;

  if ( NULL == premier ) premier = premier_echantillon;
  if ( 0 == n ) return premier;

  for ( i = 0, echantillon = premier ; i < n ; i++, echantillon = echantillon->suivant ) {
    if ( NULL == echantillon ) {
      (void) fprintf(stderr, "%s,%s l.%u : n=%lu, seulement %lu dtecteurs\n",
		     __FILE__, __FUNCTION__, __LINE__, n + 1, ECHANTILLON_Compter() );
      return NULL;
    }
  }

  /* On a trouv */
  return echantillon;
}

int ECHANTILLON_Charger_I(unsigned long n_echantillon)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_echantillon);

  /* Tout s'est bien pass */
  return 0;
}

int ECHANTILLON_Copier_I(unsigned long n_original)
{
  Echantillon *original;

  original = ECHANTILLON_N_Vers_E( premier_echantillon, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECHANTILLON_N_Vers_E\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE\n",
		 __FILE__, __FUNCTION__, __LINE__);

  return 0;
}

int ECHANTILLON_Effacer_I(unsigned long n_original)
{
  Echantillon *original;

  original = ECHANTILLON_N_Vers_E( premier_echantillon, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECHANTILLON_N_Vers_E\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  ECH_EffacerEchantillon( original, FAUX );

  return 0;
}

int ECHANTILLON_Modifier_I(unsigned long n_original)
{
  Echantillon *original;

  original = ECHANTILLON_N_Vers_E( premier_echantillon, n_original );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECHANTILLON_N_Vers_D\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  ECHANTILLON_Parametrer( original );

  return 0;
}

int ECHANTILLON_Nouveau_I (unsigned long n_avant)
{
  Echantillon *echantillon, *avant;
  int retour;

  /* On le cre */
  echantillon = ECH_CreerEchantillon();
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECH_CreerEchantillon()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On cherche o le placer dans la liste */
  avant = premier_echantillon;
  if ( NULL == avant ) {
    retour = ECH_ChainerEchantillon( NULL, echantillon );
  } else {
    avant = ECHANTILLON_N_Vers_E( premier_echantillon, n_avant );
    retour = ECH_ChainerEchantillon( avant, echantillon );
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECH_ChainerEchantillon()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    ECH_EffacerEchantillon( echantillon, VRAI );
    return -1;
  }

  /* On le modifie */
  ECHANTILLON_Parametrer( echantillon );

  return 0;
}

int ECHANTILLON_Sauver_I(unsigned long n_echantillon)
{
  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE [%lu]\n",
		 __FILE__, __FUNCTION__, __LINE__, n_echantillon);

  /* Tout s'est bien pass */
  return 0;
}

unsigned long ECHANTILLON_Compter(void)
{
  unsigned long i;
  Echantillon *d;

  for ( d = premier_echantillon, i = 0 ; d != NULL ; d = d->suivant, i++);
  return i;
}

const char *ECHANTILLON_TrouverNom(unsigned int arbre __attribute__((__unused__)), unsigned long objet)
{
  Echantillon *d;

  d = ECHANTILLON_N_Vers_E( premier_echantillon, objet );
  if ( NULL == d ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans ECHANTILLON_N_Vers_D\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  return d->nom;
}

/* ~-~-~-~-~-~-~-~- Modification de la composition : fentre ~-~-~-~-~-~-~-~- */

int ECH_SousMenuComposition(Echantillon *echantillon, struct image *coupe, unsigned int n_tranche)
{
  int option_choisie;

  if ( NULL == echantillon ) return -1;
  if ( NULL == coupe ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR --- Pas de coupe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  option_choisie = SS_menu_texte( options_composition, NOMBRE_OPTIONS_COMPOSITION, titre_ssmenu_composition, -1, ROOT );
  if ( option_choisie < 0 ) return +1; /* Rien  faire */

  switch ( option_choisie )
  {
  case OPT_COMPO_APPLIQUER_IMAGE:
    {
      struct image modele;

      if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
	return 0; /* Pas de sens pour un chantillon homogne */
      }

      /* On prpare le modle d'image : carte de concentration ou carte de composition, suivant l'chantillon considr */
      if ( echantillon->type == ECHANTILLON_GRADIENT ) {
	modele.genre = IMG_GENRE_REELLE;
      } else {
	modele.genre = IMG_GENRE_COMPO;
      }
      modele.type = IMG_TYPE_CARTOGRAPHIE;

      /* On demande quelle image doit tre utilise */
      coupe = I_IMG_ChoisirUnique( &modele, I_FILTREIMAGE_GENRE, FAUX, VRAI, NULL );
      if ( coupe != NULL ) {
	int reponse;

	(void) fprintf(stderr, "%s,%s l.%u : Coupe retenue = \"%s\"\n",
		       __FILE__, __FUNCTION__, __LINE__, coupe->nom);

	/* On demande si l'on veut l'appliquer  toutes les tranches ou juste  la courante */
	reponse = BI_ERR_alerte( NULL, &Q_ToutEchantillon, VRAI );
	if ( 1 == reponse ) {
	  /* On assigne la nouvelle coupe  la tranche */
	  ECH_ImageVersTranche_Y( echantillon, n_tranche, coupe );
	} else {
	  /* On assigne  tout l'chantillon */
	  unsigned int i;

	  for ( i = 0 ; i < echantillon->nbr_y ; i++ ) {
	    ECH_ImageVersTranche_Y( echantillon, i, coupe );
	  }
	}
      }
    }
    break;

  case OPT_COMPO_ALEATOIRE:
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      return 0; /* Pas de sens pour un chantillon homogne */
    } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      double variance;
      unsigned long i, n_cases;
      double *valeurs;
      int reponse;
      struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, BI_EDITABLE_MIN, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };

      /* On demande la variance */
      nv_valeur.texte = "s"DEUX_EXPOSANT" =";
      nv_valeur.valeur_initiale.reel = 0.01;

      reponse = BI_DemandeValeur( NULL, &Q_ValeurVariance, &nv_valeur );
      if ( reponse == (int) Q_ValeurVariance.bouton_annuler ) return +1;

      variance = nv_valeur.valeur_donnee.reel;
      OPTION_A_FAIRE( "Proposer le choix de la loi aussi !" );

      /* On gnre l'image */
      n_cases = coupe->haut * (unsigned long) coupe->larg;
      for ( i = 0, valeurs = coupe->valeurs.R ; i < n_cases ; i++, valeurs++ ) {
	double valeur;

	valeur = gaussienne() * variance + 1.0;
	if ( valeur < 0.0 ) valeur = 0.0;
	*valeurs = valeur;
      }

      /* On refait les bornes pour tre tranquille */
      IMG_bornes( coupe );
    } else {
      double *probas;
      struct matiere **valeurs;
      unsigned int num_compo;
      unsigned long i, nbr_valeurs, nbr_cases;

      probas = ECH_DemanderRepartition( &nbr_valeurs );
      if ( NULL == probas ) return +1;

      /* On gnre l'image */
      nbr_cases = coupe->haut * (unsigned long) coupe->larg;
      for ( i = 0, valeurs = coupe->valeurs.M ; i < nbr_cases ; i++, valeurs++ ) {
	num_compo = STAT_TirerValeurDiscrete( nbr_valeurs, probas );
/* 	printf("%s : composition tire = %u [%s]\n", __FUNCTION__, num_compo, (COMPOSITION_N_Vers_C( NULL, num_compo ))->nom ); */

	if ( num_compo == ( nbr_valeurs - 1 ) ) { /* Trou */
	  *valeurs = NULL;
	} else {
	  *valeurs = COMPOSITION_N_Vers_C( NULL, num_compo );
/* 	  printf("%s : composition mise = %u [%s]\n", __FUNCTION__, num_compo, (*valeurs)->nom ); */
	}
      }

      /* Plus besoin de la table des probas */
      free( probas ); probas = NULL;

      /* On refait les bornes pour tre tranquille */
      IMG_bornes( coupe );
    }

    /* On assigne la nouvelle coupe  la tranche */
    ECH_ImageVersTranche_Y( echantillon, n_tranche, coupe );
    break;

  case OPT_COMPO_ALEATOIRE_TOUT:
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      return 0; /* Pas de sens pour un chantillon homogne */
    } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
      double variance;
      unsigned long i, n_cases;
      int reponse;
      struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, BI_EDITABLE_MIN, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };

      /* On demande la variance */
      nv_valeur.texte = "s"DEUX_EXPOSANT" =";
      nv_valeur.valeur_initiale.reel = 0.01;

      reponse = BI_DemandeValeur( NULL, &Q_ValeurVariance, &nv_valeur );
      if ( reponse == (int) Q_ValeurVariance.bouton_annuler ) return +1;

      variance = nv_valeur.valeur_donnee.reel;
      OPTION_A_FAIRE( "Proposer le choix de la loi aussi !" );

      /* Combien de cases  gnrer ? */
      n_cases = 1 + ECH_IJK_v_Case( echantillon, echantillon->nbr_x - 1, echantillon->nbr_y - 1, echantillon->nbr_z - 1 );

      /* On gnre l'chantillon */
      for ( i = 0 ; i < n_cases ; i++ ) {
	double valeur;

	valeur = gaussienne() * variance + 1.0;
	if ( valeur < 0.0 ) valeur = 0.0;
	echantillon->valeurs.homogene.concentration[ i ] = valeur;
      }
    } else {
      double *probas;
      unsigned int num_compo;
      unsigned long i, nbr_valeurs, nbr_cases;

      probas = ECH_DemanderRepartition( &nbr_valeurs );
      if ( NULL == probas ) return +1;

      /* Combien de cases  gnrer ? */
      nbr_cases = 1 + ECH_IJK_v_Case( echantillon, echantillon->nbr_x - 1, echantillon->nbr_y - 1, echantillon->nbr_z - 1 );

      /* On gnre l'chantillon */
      for ( i = 0 ; i < nbr_cases ; i++ ) {
	num_compo = STAT_TirerValeurDiscrete( nbr_valeurs, probas );
/* 	printf("%s : composition tire = %u [%s]\n", __FUNCTION__, num_compo, (COMPOSITION_N_Vers_C( NULL, num_compo ))->nom ); */

	if ( num_compo == ( nbr_valeurs - 1 ) ) { /* Trou */
	  echantillon->valeurs.inhomogene[ i ] = NULL;
	} else {
	  echantillon->valeurs.inhomogene[ i ] = COMPOSITION_N_Vers_C( NULL, num_compo );
/* 	  printf("%s : composition mise = %u [%s]\n", __FUNCTION__, num_compo, (*valeurs)->nom ); */
	}
      }

      /* Plus besoin de la table des probas */
      free( probas ); probas = NULL;
    }

    /* On refait la coupe pour tre  jour */
    ECH_TrancheVersImage_Y( echantillon, n_tranche, coupe );
    break;

  case OPT_COMPO_MU_IDEAL: /* On cherche  connatre le coefficient d'absorption idal */
    {
      struct matiere *composition = NULL;
      double densite = 1.0;
      struct graphe *mu_ideal;
      char *nom = NULL;
      int reponse;

      /* Doit-on prendre en compte la diffusion ? */
      reponse = BI_ERR_alerte( NULL, &Q_InclureDiffusion, VRAI );
      if ( 3 == reponse ) return 0; /* On annule */

      /* Quelle sont la composition, la densit ? */
      if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
	composition = echantillon->valeurs.homogene.composition;
	nom = TXT_genere_chaine( MU" id"e_AIGU"al, %s [%s homog"e_GRAVE"ne, d=%f]", echantillon->nom, composition->nom, composition->densite );
      } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	composition = echantillon->valeurs.homogene.composition;
	nom = TXT_genere_chaine( MU" id"e_AIGU"al, %s [%s gradient, d=%f]", echantillon->nom, composition->nom, composition->densite * densite );
      } else {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE ---  idal pour chantillon inhomogne\n",
		       __FILE__, __FUNCTION__, __LINE__);
      }
      if ( NULL == composition ) {
	(void) fprintf(stderr, "%s,%s l.%u : Pas de composition => pas d'absorption\n",
		       __FILE__, __FUNCTION__, __LINE__);
	free( nom );
	return -5;
      }
      densite *= composition->densite;

      if ( NULL == serie_courante ) {
	SERIE_GRF_CreerSerie( NULL, "Absorption par l'"e_AIGU"chantillon" );
      }

      /* On rcupre /rho, calcul pour l'affichage normalement ! */
      mu_ideal = GRF_Copier(composition->mu_total, serie_courante, FAUX );
      if ( NULL == mu_ideal ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Copier()\n",
		       __FILE__, __FUNCTION__, __LINE__);
	free( nom );
	return -6;
      }
      /* On s'assure qu'il s'agit bien d'un graphe isol avec ses propres X
	 (sinon, quand on changera les conditions de l'exprience, on risque d'avoir des surprises dsagrables) */
      GRF_ChangerGenre( mu_ideal, GRF_FCT_REELLE );

      /* On veut le coefficient total : on ajoute les autres termes [pas forcment prts, car pas fait pour l'afficaghe] */
      if ( 1 == reponse ) {
	/* Diffusion lastique */
	Matiere_CalculerTau( composition, mu_ideal, VRAI, DIFFUSION_COEF_McMaster, VRAI );

	/* Diffusion inlastique */
	Matiere_CalculerTau( composition, mu_ideal, FAUX, DIFFUSION_COEF_McMaster, VRAI );

	/* On a un vrai analogue de spectre exprimental */
	GRF_ChangerType( mu_ideal, GR_BRUT );
      }

      /* On veut  => on multiplie par la densit [en g/cm^3] */
      GRF_FacteurMultiplicatif( mu_ideal, densite, AXE_Y, "g*cm^-3" );

      /* On lui donne un nom plus clair... */
      GRF_titre( mu_ideal, nom );
      free( nom );
    }
    break;

  case OPT_COMPO_FLUO_IDEALE:
    {
      struct matiere *composition = NULL;
      struct graphe *fluo_ideale;
      char *nom = NULL;

      /* Quelle sont la composition, la densit ? */
      if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
	composition = echantillon->valeurs.homogene.composition;
	nom = TXT_genere_chaine( "Fluorescence id"e_AIGU"al, %s [%s homog"e_GRAVE"ne]", echantillon->nom, composition->nom );
      } else if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	composition = echantillon->valeurs.homogene.composition;
	nom = TXT_genere_chaine( "Fluorescence id"e_AIGU"al, %s [%s gradient]", echantillon->nom, composition->nom);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- fluo idale pour chantillon inhomogne\n",
		       __FILE__, __FUNCTION__, __LINE__);
      }
      if ( NULL == composition ) {
	(void) fprintf(stderr, "%s,%s l.%u : Pas de composition => pas de fluorescence\n",
		       __FILE__, __FUNCTION__, __LINE__);
	free( nom );
	return -5;
      }

      if ( NULL == serie_courante ) {
	SERIE_GRF_CreerSerie( NULL, "Fluorescence par l'"e_AIGU"chantillon" );
      }

      /* On rcupre /rho, calcul pour l'affichage normalement ! */
      fluo_ideale = GRF_Copier( composition->fluo, serie_courante, FAUX );
      if ( NULL == fluo_ideale ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Copier()\n",
		       __FILE__, __FUNCTION__, __LINE__);
	free( nom );
	return -6;
      }
      /* On s'assure qu'il s'agit bien d'un graphe isol avec ses propres X
	 (sinon, quand on changera les conditions de l'exprience, on risque d'avoir des surprises dsagrables) */
      GRF_ChangerGenre( fluo_ideale, GRF_FCT_REELLE );

      /* On lui donne un nom plus clair... */
      GRF_titre( fluo_ideale, nom );
      free( nom );
    }
    break;
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- On demande la composition de l'chantillon -------- */

static double *ECH_DemanderRepartition(unsigned long *nbr_valeurs)
{
  struct infos_demande *valeurs;
  struct matiere *composition;
  unsigned long i, nbr_compos;
  int retour;
  double *probas = NULL;

  /* Combien de compositisions sont disponibles ? */
  nbr_compos = COMPOSITION_Compter();
  if ( 0 == nbr_compos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune composition dfinie !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On rserve pour demander pour chaque composition (y compris les trous) */
  valeurs = (struct infos_demande *) malloc( sizeof( struct infos_demande ) * (nbr_compos + 1) );
  if ( NULL == valeurs ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de prparer la table des valeurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  /* On initialise chacune des valeurs */
  for ( i = 0, composition = premiere_composition ; i < nbr_compos ; i++, composition = composition->suivante ) {
    valeurs[ i ].type = BI_CHAMP_DEMANDE_REEL;
    valeurs[ i ].texte = composition->nom;
    valeurs[ i ].unite = "%";

    valeurs[ i ].proprietes = BI_EDITABLE_MIN | BI_EDITABLE_MAX | BI_EDITABLE_FRACTION | BI_EDITABLE_UNITE;
    valeurs[ i ].min = 0.0; valeurs[ i ].max = 100.0; /* Fraction massique, entre 0 et 100% */
    valeurs[ i ].valeur_initiale.reel = 100.0 / (double) (nbr_compos); /* Mlange homogne quimassique */
    valeurs[ i ].valeur_defaut = valeurs[ i ].valeur_initiale;
    valeurs[ i ].erreur = -1.0; /* Pas d'erreur associe */
  }

  /* et le trou... */
  valeurs[ nbr_compos ].type = BI_CHAMP_DEMANDE_REEL;
  valeurs[ nbr_compos ].proprietes = BI_EDITABLE_MIN | BI_EDITABLE_MAX | BI_EDITABLE_FRACTION | BI_EDITABLE_UNITE;
  valeurs[ nbr_compos ].texte = texte_nom_trous;
  valeurs[ nbr_compos ].unite = "%";
  valeurs[ nbr_compos ].min = 0.0; valeurs[ nbr_compos ].max = 100.0; /* Fraction massique, entre 0 et 100% */
  valeurs[ nbr_compos ].valeur_initiale.reel = 0.0; /* Mlange homogne quimassique sans trou */
  valeurs[ nbr_compos ].valeur_defaut.reel = 0.0;
  valeurs[ nbr_compos ].erreur = -1.0; /* Pas d'erreur associe */

  /* On appelle la bote de dialogue */
  retour = BI_DemandeMultiple( NULL, &Q_DemanderComposition, nbr_compos + 1, valeurs );
  if ( retour != (int) Q_DemanderComposition.bouton_ok ) {
    free( valeurs );
    return NULL; /* On annule ou une erreur est survenue */
  }

  /* On rserve la table des probabilits, associe aux compositions dans le mme ordre */
  probas = (double *) malloc( sizeof(double) * (nbr_compos + 1) );
  if ( NULL == probas ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mmoire insuffisante pour crer la table des probabilits\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On remplit cette table avec les rponses donnes */
  for ( i = 0 ; i <= nbr_compos ; i++ ) {
    probas[ i ] = valeurs[ i ].valeur_donnee.reel / 100.0; /* Les valeurs sont donnes en pourcentages */
  }

  /* Plus besoin des infos pour la demande */
  free( valeurs ); valeurs = NULL;

  /* Cration russie */
  *nbr_valeurs = nbr_compos + 1;
  return probas;
}
