/* LASE - Logiciel d'Analyse des Spectres EXAFS
   LASE - (C) 1998-2003, Emmanuel CURIS

   Calcul du signal obtenu en transmission
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "echantillon.h"
#include "faisceau.h"
#include "graphes.h"
#include "images.h"
#include "lissage.h" /* pour GRF_LireValeur */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int FTE_Monochromatique(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie);
static int FTE_Polychromatique(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie);

static int FTE_Gradient_Ponctuel  (Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie);
static int FTE_Homogene_Ponctuel  (Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie);
static int FTE_Inhomogene_Ponctuel(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie);

static int FTE_Homogene_Etendu(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* -------- Calculer le faisceau transmis par un échantillon ----------------------\
 |   Renvoie NULL en cas d'erreur, le faisceau transmis sinon                      |
 |                                                                                 |
 | echantillon : l'échantillon à traverser                                         |
 | faisceau    : le faisceau arrivant dessus [DÉJÀ CORRIGÉ DE TOUS LES EFFETS,     |
 |                                         y compris l'atténuation avec l'énergie] |
 | energie     : l'énergie de travail                                              |
 | f_transmis  : le faisceau après avoir traversé l'échantillon                    |
 \-------------------------------------------------------------------------------- */

int FAISCEAU_TraverserEchantillon(Faisceau *faisceau, Echantillon *echantillon, double energie, Faisceau *f_transmis)
{
  int retour = -10;

  ENTREE_FONCTION("FAISCEAU_TraverserEchantillon", "faisceau=%p echantillon=%p energie=%g f_transmis=%p", faisceau, echantillon, energie, f_transmis);

  if ( ( NULL == echantillon ) || ( NULL == faisceau ) || ( NULL == f_transmis ) ) {
    SORTIE_FONCTION("FAISCEAU_TraverserEchantillon", "%d [ERREUR paramètre absurde]", -1);
    return -1;
  }

  /* Traitement du fondamental */
  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    retour = FTE_Polychromatique( echantillon, faisceau, f_transmis, energie );
  } else {
    retour = FTE_Monochromatique( echantillon, faisceau, f_transmis, energie );
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FAISCEAU_TraverserEchantillon : ERREUR dans le calcul du faisceau transmis / fondamental\n");
    return -20;
  }

  /* On s'occupe des harmoniques éventuelles */
  if ( faisceau->proprietes & FAISCEAU_HARMONIQUES ) {
    Faisceau f_harmonique, f_trs_harmonique;
    unsigned int i_h;

    f_harmonique = *faisceau;
    f_harmonique.proprietes &= ~ FAISCEAU_HARMONIQUES;

    f_trs_harmonique = *f_transmis;
    f_trs_harmonique.proprietes &= ~ FAISCEAU_HARMONIQUES;

    for ( i_h = 0 ; i_h < faisceau->nbr_harmoniques ; i_h++ ) {
      f_harmonique.intensite = faisceau->intensite * faisceau->i_harmoniques[ i_h ];
      if ( 0.0 == f_harmonique.intensite ) continue; /* Harmonique éteinte */
      f_harmonique.g_E = faisceau->g_harmoniques[ i_h ];

      if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
	retour = FTE_Polychromatique( echantillon, &f_harmonique, &f_trs_harmonique, energie * ( i_h + 2 ) );
      } else {
	retour = FTE_Monochromatique( echantillon, &f_harmonique, &f_trs_harmonique, energie * ( i_h + 2 ) );
      }
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR / harmonique %u\n",
		       __FILE__, __FUNCTION__, __LINE__, i_h);
      }

      if ( f_transmis->intensite > 0.0 )
	f_transmis->i_harmoniques[ i_h ] = f_trs_harmonique.intensite / f_transmis->intensite;
      else {
	f_transmis->i_harmoniques[ i_h ] = f_trs_harmonique.intensite;
      }
    }
  }

  /* On a fini, sans problème */
  SORTIE_FONCTION("FAISCEAU_TraverserEchantillon", "%d, f_transmis->i=%g", 0, f_transmis->intensite);
  return 0;
}

/* -------- Cas d'un faisceau monochromatique : facile... -------- */

static int FTE_Monochromatique(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie)
{
  int retour = -1;

  ENTREE_FONCTION( __FUNCTION__, "echantillon=%p faisceau=%p f_transmis=%p energie=%g", echantillon, faisceau, f_transmis, energie );

  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    DANS_FONCTION("Faisceau ponctuel", 0);
    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      retour = FTE_Homogene_Ponctuel( echantillon, faisceau, f_transmis, energie );
    } else {
      if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	retour = FTE_Gradient_Ponctuel( echantillon, faisceau, f_transmis, energie );
      } else {
	retour = FTE_Inhomogene_Ponctuel( echantillon, faisceau, f_transmis, energie );
      }
    }

    /* L'intensité du faisceau transmis est celle initiale * l'atténuation (I = I0 * exp( - mu * l ) ) */
  } else {
    unsigned int i, j;
    double *valeurs;

    DANS_FONCTION("Faisceau surfacique", 0);
    /* On construit la coupe après l'échantillon, si elle n'existe pas */
    if ( NULL == f_transmis->g_espace ) {
      double pas_x, pas_y;

      /* On choisit le plus précis entre faisceau/10 et échantillon ! */
      pas_x = f_transmis->largeur / 10.0;
      if ( echantillon->pas_x < pas_x ) pas_x = echantillon->pas_x;

      pas_y = f_transmis->hauteur / 10.0;
      if ( echantillon->pas_y < pas_y ) pas_y = echantillon->pas_y;

      f_transmis->g_espace = FAISCEAU_ReserverCoupe( f_transmis->largeur, f_transmis->hauteur, pas_x, pas_y );
      if ( NULL == f_transmis->g_espace ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de préparer la coupe\n",
		       __FILE__, __FUNCTION__, __LINE__);
	SORTIE_FONCTION( __FUNCTION__, "%g [Impossible de créer la coupe]", -2.0 );
	return -2.0;
      }
      IMG_Titre( f_transmis->g_espace, "Coupe après l'échantillon" );
    }

    /* On l'initialise à zéro, de toute façon */
    for ( i = 0, valeurs = f_transmis->g_espace->valeurs.R ; i < f_transmis->g_espace->larg ; i++ ) {
      for ( j = 0 ; j < f_transmis->g_espace->haut ; j++, valeurs++ ) {
	*valeurs = 0.0;
      }
    }

    if ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) {
      DANS_FONCTION("Échantillon homogène", 0);
      retour = FTE_Homogene_Etendu( echantillon, faisceau, f_transmis, energie );
    } else {
      Faisceau pseudo_faisceau, pseudo_transmis;
      double x, y; /* Coordonnées du point d'impact */
      double *valeurs;
      int (*fonction_calcul)(Echantillon *, Faisceau *, Faisceau *, double);

      DANS_FONCTION("Échantillon inhomogène", 0);
      if ( ECHANTILLON_GRADIENT == echantillon->type ) {
	fonction_calcul = FTE_Gradient_Ponctuel;
      } else {
	fonction_calcul = FTE_Inhomogene_Ponctuel;
      }

      /* Certainement pas le plus efficace, mais le plus simple pour l'instant :
	 On utilise un pseudo-faisceau ponctuel... */
      pseudo_faisceau = *faisceau;
      pseudo_faisceau.proprietes |= FAISCEAU_PONCTUEL;

      for ( i = 0, valeurs = f_transmis->g_espace->valeurs.R ; i < f_transmis->g_espace->haut ; i++ ) {

	y = f_transmis->g_espace->y_min + i * f_transmis->g_espace->h_reel;
	for ( j = 0 ; j < f_transmis->g_espace->larg ; j++, valeurs++ ) {
	  x = f_transmis->g_espace->x_min + j * f_transmis->g_espace->l_reel;

	  /* (x, y) : coordonnée d'entrée du faisceau */
	  pseudo_faisceau.orientation.position.x = x;
	  pseudo_faisceau.orientation.position.y = 0.0;
	  pseudo_faisceau.orientation.position.z = y;

	  /* Intensité en entrée, pour ce point */
	  pseudo_faisceau.intensite = FAISCEAU_LireIntensite( faisceau, x, y, 0.0 ); /* L'effet de l'énergie sera traité après */
	  if ( 0.0 == pseudo_faisceau.intensite ) { 
#ifdef _DEVELOPPE_
/* 	    printf("%s E=%g (x,y,z)=(%g,%g,%g) I=%g\n", __FUNCTION__, energie, x, pseudo_faisceau.orientation.position.y, y, pseudo_faisceau.intensite); */
#endif
	    continue;
	  }
	  
	  retour = fonction_calcul( echantillon, &pseudo_faisceau, &pseudo_transmis, energie );
#ifdef _DEVELOPPE_
	  if ( is_inf( pseudo_transmis.intensite ) ) {
	    (void) fprintf(stderr, "%s,%s l.%u : INTENSITÉ INFINIE [E=%g eV (%g [%u], %g [%u])]\n",
			   __FILE__, __FUNCTION__, __LINE__, energie, x, i, y, j);
	    abort();
	  }
/*  	  printf("%s E=%g (x,y,z)=(%g,%g,%g) I=%g -> %g\n", __FUNCTION__, energie, x, pseudo_faisceau.orientation.position.y, y, */
/* 		 pseudo_faisceau.intensite, pseudo_transmis.intensite ); */
#endif

	  /* Valeur de l'intensité transmise */
	  *valeurs = pseudo_transmis.intensite;
	}
      }
    }

    f_transmis->intensite = 1.0; /* Les vraies valeurs sont dans la coupe ! */
    f_transmis->proprietes |= FAISCEAU_INHOMOGENE; /* Il faut donc l'utiliser ! */

/*     IMG_bornes( f_transmis->g_espace ); */ /* Normalement inutile, mais si affichage... */
  }

  /* On a fini */
  SORTIE_FONCTION( __FUNCTION__, "%g", retour );
  return retour;
}

/* -------- Cas d'un faisceau Polychromatique : plus délicat... -------- */

static int FTE_Polychromatique(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie)
{
  double e_vraie, *de, I_vraie, *gE;
  int retour;
  unsigned long i;
  Faisceau pseudo_faisceau, ps_sortie;

  pseudo_faisceau = *faisceau;
  pseudo_faisceau.proprietes &= ~FAISCEAU_POLYCHROMATIQUE;
  pseudo_faisceau.g_E = NULL;

  if ( NULL == f_transmis->g_E ) {
#ifdef _DEVELOPPE_
    char *nv_titre;

    (void) fprintf(stderr, "%s,%s l.%u : Attention, pas de courbe g(E) : on la fait\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    f_transmis->g_E = GRF_Copier( faisceau->g_E, NULL, FAUX /* VRAI */ );
    if ( NULL == f_transmis->g_E ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Copier\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
#ifdef _DEVELOPPE_
    nv_titre = TXT_genere_chaine("g(E) transmis, E=%g", energie );
    GRF_titre( f_transmis->g_E, nv_titre );
    free( nv_titre );
#endif
  }

  ps_sortie = *f_transmis; /* Pour avoir la bonne position => des distances correctes ! */

  /* Faisceau ponctuel : rien de particulier, g(E) est unique et tout va bien */
  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
   /* Pour chacune des énergies enregistrées */
    for ( i = 0, de = faisceau->g_E->x, gE = faisceau->g_E->y ; i < faisceau->g_E->nombre_point ; i++, de++, gE++ ) {
      /* Énergie prise en compte */
      e_vraie = energie * ( 1.0 + *de );

      /* Intensité pour cette énergie [si FAISCEAU_DECROIT, pris en compte dans Monochromatique [par LireIntensite]] */
      I_vraie = faisceau->intensite * *gE;
      pseudo_faisceau.intensite = I_vraie;

      /* Contribution de cette énergie */
      retour = FTE_Monochromatique( echantillon, &pseudo_faisceau, &ps_sortie, e_vraie );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FTE_Monochromatique / E = %g *( 1.0 + %g ) = %g\n",
		       __FILE__, __FUNCTION__, __LINE__, energie, *de, e_vraie);
      }
      f_transmis->g_E->y[ i ] = ps_sortie.intensite / faisceau->intensite;
    }

    f_transmis->intensite = faisceau->intensite;
  } else { /* Faisceau surfacique... */
    if ( NULL == f_transmis->g_espace ) { /* On aura de toute façon besoin de la coupe ! */
      (void) fprintf(stderr, "%s,%s l.%u : Il manque la coupe du faisceau\n",
		     __FILE__, __FUNCTION__, __LINE__);
#ifdef _DEVELOPPE_
      abort();
#endif
      return -581;
    }

    if ( IMG_GENRE_REELLE == f_transmis->g_espace->genre ) {
      /* Échantillon & faisceau homogènes => g(E) identique en tout point */

      /* Pour chacune des énergies enregistrées */
      for ( i = 0, de = faisceau->g_E->x, gE = faisceau->g_E->y ; i < faisceau->g_E->nombre_point ; i++, de++, gE++ ) {
	/* Énergie prise en compte */
	e_vraie = energie * ( 1.0 + *de );

	/* Intensité pour cette énergie [si FAISCEAU_DECROIT, pris en compte dans Monochromatique [par LireIntensite]] */
	I_vraie = faisceau->intensite * *gE;
	pseudo_faisceau.intensite = I_vraie;

	/* Contribution de cette énergie */
	retour = FTE_Monochromatique( echantillon, &pseudo_faisceau, &ps_sortie, e_vraie );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FTE_Monochromatique / E = %g *( 1.0 + %g ) = %g\n",
			 __FILE__, __FUNCTION__, __LINE__, energie, *de, e_vraie);
	}
	/* Attention, ps_sortie.intensite = 1.0 et les valeurs sont dans la coupe.
	   Mais avec échantillon et faisceau ponctuels, la coupe est homogène (normalement) */
#ifdef _DEVELOPPE_
/* 	IMG_bornes( f_transmis->g_espace ); */
/* 	if ( f_transmis->g_espace->val_min != f_transmis->g_espace->val_max ) { */
/* 	  (void) fprintf(stderr, "%s,%s l.%u : Coupe non-homogène, absurde compte-tenu des conditions [min=%g, max=%g, D=%g]\n", */
/* 			 __FILE__, __FUNCTION__, __LINE__, f_transmis->g_espace->val_min, f_transmis->g_espace->val_max, */
/* 			 f_transmis->g_espace->val_max - f_transmis->g_espace->val_min); */
/* 	  abort(); */
/* 	} */
#endif
	f_transmis->g_E->y[ i ] = f_transmis->g_espace->valeurs.R[ 0 ] / faisceau->intensite;
#ifdef _DEVELOPPE_ /* uniquement, car échantillon fortement absorbant peut conduire à cela... */
	if ( 0.0 >= f_transmis->g_E->y[ i ] ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Courbe g(E) absurde... [E = %g /%g ; I = %g ; It=%g]\n",
			 __FILE__, __FUNCTION__, __LINE__, energie, e_vraie, faisceau->intensite, ps_sortie.intensite );
	}
#endif
      }

      f_transmis->intensite = faisceau->intensite;
    } else {
      /* Échantillon ou faisceau inhomogène => g(E) différent en chaque point */
      Image *coupe_mono, *coupe_sortie;
      struct graphe **grf_gE;
      unsigned long i_case, nbr_valeurs;

      /* On prépare une pseudo-carte pour le faisceau monochromatique */
      coupe_mono = IMG_CreerImageIsolee( IMG_GENRE_REELLE, IMG_TYPE_INTENSITES, f_transmis->g_espace->larg, f_transmis->g_espace->haut );
      if ( NULL == coupe_mono ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la pseudo-coupe\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -487;
      }
      coupe_mono->l_reel = f_transmis->g_espace->l_reel; coupe_mono->h_reel = f_transmis->g_espace->h_reel;
      coupe_mono->x_min = f_transmis->g_espace->x_min; coupe_mono->y_min = f_transmis->g_espace->y_min;

      /* Idem pour la sortie */
      coupe_sortie = IMG_CopierImage( coupe_mono, NULL, VRAI );
      if ( NULL == coupe_sortie ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la pseudo-coupe\n",
		       __FILE__, __FUNCTION__, __LINE__);
	IMG_EffacerImage( coupe_mono, FAUX );
	return -248;
      }

      nbr_valeurs = coupe_mono->larg * (unsigned long) coupe_mono->haut;

      /* Ce sont donc elles que l'on utilisera ! */
      pseudo_faisceau.proprietes |= FAISCEAU_INHOMOGENE; /* Pour être sûr de l'utiliser, même si le faisceau incident est homogène  */
      pseudo_faisceau.g_espace = coupe_mono;
      ps_sortie.g_espace = coupe_sortie;

      /* Pour chacune des énergies enregistrées */
      for ( i = 0, de = faisceau->g_E->x ; i < faisceau->g_E->nombre_point ; i++, de++ ) {
	/* Énergie prise en compte */
	e_vraie = energie * ( 1.0 + *de );

	/* On initialise la pseudo-coupe. Suppose deux choses :
	   (1) que faisceau->g_E contienne bien g(E), ie que le faisceau avant échantillon soit de g(E) unique
	   (2) que les cartes avant et après soient bien exactement de même dimension
	   C'est normalement FAISCEAU_GeometrieApresEchantillon() qui s'arrange pour que (2) soit vraie
	                     FAISCEAU_ProjeterSurEchantillon() qui s'arrange pour que (1) soit vraie
	*/
	if ( faisceau->g_espace != NULL ) {
	  for ( i_case = 0 ; i_case < nbr_valeurs ; i_case++ ) {
	    coupe_mono->valeurs.R[ i_case ] = faisceau->g_E->y[ i ] * faisceau->g_espace->valeurs.R[ i_case ] ;
	  }
	} else { /* Faisceau homogène en entrée, mais pas en sortie (échantillon inhomogène) */
	  for ( i_case = 0 ; i_case < nbr_valeurs ; i_case++ ) {
	    coupe_mono->valeurs.R[ i_case ] = faisceau->g_E->y[ i ] ;
	  }
	}

	/* Intensité pour cette énergie [si FAISCEAU_DECROIT, pris en compte dans Monochromatique [par LireIntensite]] */
	pseudo_faisceau.intensite = faisceau->intensite;

	/* Contribution de cette énergie */
	retour = FTE_Monochromatique( echantillon, &pseudo_faisceau, &ps_sortie, e_vraie );
	if ( retour != 0 ) {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FTE_Monochromatique / E = %g *( 1.0 + %g ) = %g\n",
			 __FILE__, __FUNCTION__, __LINE__, energie, *de, e_vraie);
	}
	/* On mémorise la coupe dans les g(E) de l'image résultat.
	   Suppose que les g(E) en chaque point existent : c'est FAISCEAU_GeometrieApresEchantillon() qui s'en occupe */
	for ( i_case = 0, grf_gE = f_transmis->g_espace->valeurs.G ; i_case < nbr_valeurs ; i_case++, grf_gE++ ) {
	  (*grf_gE)->y[ i ] = coupe_sortie->valeurs.R[ i_case ];
	}
      }

      f_transmis->intensite = 1.0;

      /* Plus besoin des pseudo-coupes */
      IMG_EffacerImage( coupe_mono, FAUX );
      IMG_EffacerImage( coupe_sortie, FAUX );
    }
  }

#ifdef _DEVELOPPE_
/*   GRF_bornes( f_transmis->g_E, NE_PAS_FAIRE, DESORDONNE ); */
/*   printf("%s : E=%g => bornes de g(E) = %g, %g [DE = %g]\n", __FUNCTION__, energie, f_transmis->g_E->y_min, f_transmis->g_E->y_max, f_transmis->g_E->y_max - f_transmis->g_E->y_min ); */
#endif

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Cas d'un échantillon homogène et d'un faisceau rectangulaire ---------- */

int FTE_Homogene_Etendu(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie)
{
  Vecteur3D v_directeur, point_entree, point_sortie;
  double x, y;
  double mu_total, distance = 0.0, I0, I_transmise;
  double *valeurs;
  int retour;
  unsigned int i, j;

  /* On calcule l'atténuation pour l'énergie nominale */
  mu_total = ECH_CalculerMuTotal( echantillon, NULL, energie, NULL );

  /* Vecteur de propagation du faisceau */
  REPERE_VecteurNormal( &(faisceau->orientation), &v_directeur );

  /* Calcul : contribution de chaque pseudo-faisceau ponctuel */
  for ( i = 0, valeurs = f_transmis->g_espace->valeurs.R ; i < f_transmis->g_espace->haut ; i++ ) {

    y = f_transmis->g_espace->y_min + i * f_transmis->g_espace->h_reel; /* (x, y) : coordonnée d'entrée du faisceau */
    for ( j = 0 ; j < f_transmis->g_espace->larg ; j++, valeurs++ ) {
      x = f_transmis->g_espace->x_min + j * f_transmis->g_espace->l_reel;

      point_entree.x = x;
      point_entree.y = 0.0;
      point_entree.z = y;

      /* Intensité du faisceau entrant en ce point */
      I0 = FAISCEAU_LireIntensite( faisceau, x, y, 0.0 );
      if ( 0.0 == I0 ) continue;

      /* Coordonnées de sortie */
      retour = ECH_TrouverSortieFaisceau( echantillon, &point_entree, &v_directeur, &point_sortie, FAUX );
      if ( 0.0 == retour ) { /* On traverse */
	distance = U3D_Distance( &point_entree, &point_sortie );
	distance /= 1.0e4; /* Coordonnées en microns, mais (mu/rho) en cm^-1 => conversion en cm */

	I_transmise = I0 * exp( - mu_total * distance );
      } else {
	I_transmise = I0;
	(void) fprintf(stderr, "FTE_Homogene_Etendu : le faisceau sort de l'échantillon !\n");
      }

      *valeurs = I_transmise;
#ifdef _DEVELOPPE_
/*       printf("%s : E=%g X=%g Y=%g I0=%g I_transmise=%g distance=%g mu_total=%g\n", __FUNCTION__, energie, x, y, I0, I_transmise, distance, mu_total); */
#endif
    }
  }
  f_transmis->intensite = 1.0; /* Les vraies valeurs sont dans la coupe ! */
  f_transmis->proprietes |= FAISCEAU_INHOMOGENE; /* Il faut donc l'utiliser ! */

  /* Calcul terminé, tout s'est bien passé */
  return 0;
}

/* -------- Cas d'un échantillon homogène & d'un faisceau ponctuel [idéal] -------- */

static int FTE_Homogene_Ponctuel(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie)
{
  double mu_total, distance;

  /* On calcule l'atténuation pour l'énergie nominale */
  mu_total = ECH_CalculerMuTotal( echantillon, NULL, energie, NULL );

  /* Distance parcourue par le faisceau dans l'échantillon */
  distance = U3D_Distance( &(faisceau->orientation.position), &(f_transmis->orientation.position) );
  distance /= 1.0e4; /* Coordonnées en microns, mais (mu/rho) en cm^-1 => conversion en cm */

  /* Ne joue que sur l'intensité */
  f_transmis->intensite = faisceau->intensite * exp( -mu_total * distance );
#ifdef _DEVELOPPE_
  if ( f_transmis->intensite <= 0.0 ) {
    (void) fprintf(stderr, "FTE_Homogene_Ponctuel : Faisceau entièrement absorbé [distance=%g, mu=%g, I0=%g, exp=%g, E=%g]\n",
		   distance, mu_total, faisceau->intensite, exp( -mu_total * distance ), energie );
    
  } else {
/*     printf("%s : E=%g I0=%g I_transmise=%g distance=%g mu_total=%g\n", __FUNCTION__, energie, faisceau->intensite, f_transmis->intensite, distance, mu_total); */
  }
#endif
  
  /* On a fini, tout s'est bien passé */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Échantillon inhomogène ~-~-~-~-~-~-~-~- */

/* -------- Gradient de concentration --------
 * On suppose que la densité varie linéairement entre deux points pour lesquels elle est connue
 * Dans ce cas, cela revient à utiliser, pour la tranche, le coefficient d'absorption moyen entre les deux
 */

static int FTE_Gradient_Ponctuel(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie)
{
  double d_entree, d_sortie;
  double mu_total;
  double dd;
  double I0;
  unsigned int i;
  Vecteur3D u_faisceau;
  Vecteur3D p_entree __attribute__((__unused__)), p_sortie;

  /* Vecteur directeur de la droite */
  u_faisceau.x = sin( faisceau->orientation.theta );
  u_faisceau.y = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
  if ( 0.0 == u_faisceau.y ) {
    (void) fprintf(stderr, "%s,%s l.%u : Faisceau parallèle à l'échantillon\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
  u_faisceau.z = -cos( faisceau->orientation.theta ) * sin( faisceau->orientation.phi );

  /* Distance parcourue en passant d'une tranche à l'autre */
  dd  = echantillon->pas_y / u_faisceau.y;
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

  /* On calcule l'atténuation pour l'énergie nominale */
  mu_total = ECH_CalculerMuTotal( echantillon, NULL, energie, NULL );
#ifdef _DEVELOPPE_
/*   printf("%s(E=%g) : µ=%gm dd=%g\n", __FUNCTION__, energie, mu_total, dd); */
#endif

  /* Intensité en entrée */
  I0 = FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, energie ); /* Effets de I(E) pris en comptes... */

  /* Initialisation, tranche d'entrée
     (on met dans _sortie car au premier tour de boucle, deviendra entree...) */
  p_sortie = faisceau->orientation.position; /* Point d'impact sur la face d'entrée de l'échantillon */
  d_sortie = ECH_LireDensite( echantillon, &p_sortie );

  /* Pour chaque tranche de l'échantillon */
  for ( i = 0 ; i < echantillon->nbr_y - 1 ; i++ ) {
    double mu_moyen, t;

    p_entree = p_sortie; /* On repart de là où l'on était parti */
    d_entree = d_sortie;

    p_sortie = faisceau->orientation.position; /* Point d'impact sur la face de sortie de la tranche */
    t = ( i + 1 ) * echantillon->pas_y / u_faisceau.y;
    p_sortie.x += t * u_faisceau.x;
    p_sortie.y += t * u_faisceau.y;
    p_sortie.z += t * u_faisceau.z;
    d_sortie = ECH_LireDensite( echantillon, &p_sortie );

    mu_moyen = mu_total * ( d_sortie + d_entree ) / 2.0;
#ifdef _DEVELOPPE_
/*     printf("%g [%g %g] -", mu_moyen, d_entree, d_sortie); */
#endif

    I0 *= exp( -mu_moyen * dd ); /* Attenuation due à cette tranche */
  }

  /* Ne joue que sur l'intensité */
  f_transmis->intensite = I0;
#ifdef _DEVELOPPE_
/*   printf("\n%s(E=%g) : Épaisseur traversée %g ; I0=%g, I=%g\n", __FUNCTION__, energie, U3D_Distance( &p_sortie, &(faisceau->orientation.position) ), */
/* 	  FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, energie ), I0 ); */
  if ( f_transmis->intensite <= 0.0 ) {
    (void) fprintf(stderr, "%s : Faisceau entièrement absorbé [distance=??, mu=%g, I0=%g, exp=??, E=%g]\n", __FUNCTION__,
		   mu_total, faisceau->intensite, energie );
    
  }
#endif
  
  /* On a fini, tout s'est bien passé */
  return 0;
}

/* -------- Inhomogène : composition variable -------- */

static int FTE_Inhomogene_Ponctuel(Echantillon *echantillon, Faisceau *faisceau, Faisceau *f_transmis, double energie)
{
  double dd;
  double mu_entree, mu_sortie;
  double I0;
  unsigned int i;
  Vecteur3D u_faisceau;
  Vecteur3D p_entree __attribute__((__unused__)), p_sortie;

  /* Vecteur directeur de la droite */
  u_faisceau.x = sin( faisceau->orientation.theta );
  u_faisceau.y = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
  if ( 0.0 == u_faisceau.y ) {
    (void) fprintf(stderr, "FLUO_Init_Ponctuel : Faisceau parallèle à l'échantillon\n");
    return -2;
  }
  u_faisceau.z = -cos( faisceau->orientation.theta ) * sin( faisceau->orientation.phi );

  /* Distance parcourue en passant d'une tranche à l'autre */
  dd  = echantillon->pas_y / u_faisceau.y;
  dd /= 10000.0; /* mu/rho est en cm^-1, les distances dans l'échantillon en microns... */

  /* Intensité en entrée */
  I0 = FAISCEAU_LireIntensite( faisceau, 0.0, 0.0, energie ); /* Effets de I(E) pris en comptes... */

  /* Initialisation, tranche d'entrée */
  p_sortie  = faisceau->orientation.position; /* Point d'impact sur la face d'entrée de l'échantillon */
  mu_sortie = ECH_CalculerMuTotal( echantillon, &p_sortie, energie, NULL );

  /* Pour chaque tranche de l'échantillon */
  for ( i = 0 ; i < echantillon->nbr_y - 1 ; i++ ) {
    double mu_moyen, t;

    p_entree  = p_sortie; /* On repart de là où l'on était parti */
    mu_entree = mu_sortie;

    p_sortie = faisceau->orientation.position; /* Point d'impact sur la face de sortie de la tranche */
    t = ( i + 1 ) * echantillon->pas_y / u_faisceau.y;
    p_sortie.x += t * u_faisceau.x;
    p_sortie.y += t * u_faisceau.y;
    p_sortie.z += t * u_faisceau.z;
    mu_sortie = ECH_CalculerMuTotal( echantillon, &p_sortie, energie, NULL );

    mu_moyen = ( mu_sortie + mu_entree ) / 2.0;

    I0 *= exp( -mu_moyen * dd ); /* Attenuation due à cette tranche */
  }

  /* Ne joue que sur l'intensité */
  f_transmis->intensite = I0;
#ifdef _DEVELOPPE_
  if ( f_transmis->intensite <= 0.0 ) {
    (void) fprintf(stderr, "%s : Faisceau entièrement absorbé [distance=??, mu=??, I0=%g, exp=??, E=%g]\n", __FUNCTION__,
		   faisceau->intensite, energie );
    
  }
#endif
  
  /* On a fini, tout s'est bien passé */
  return 0;
}

/* -------- Générer la géométrie du faisceau transmis en sortie de l'échantillon --------
     Renvoie le faisceau transmis, NULL en cas d'erreur

   faisceau : le faisceau en entrée (supposé plaqué sur la face avant)
   echantillon : l'échantillon à traverser

   On ne s'occupe ici QUE de la géométrie. Le reste est à la charge d'autres routines.
 */

Faisceau *FAISCEAU_GeometrieApresEchantillon(Faisceau *faisceau, Echantillon *echantillon)
{
  int retour;
  Faisceau *f_transmis = NULL;
  Vecteur3D direction, point_sortie;

  if ( NULL == faisceau ) return NULL;
  if ( NULL == echantillon ) {
    (void) fprintf(stderr, "FAISCEAU_GeometrieApresEchantillon : Pas d'échantillon --- faisceau inchangé\n");
    return faisceau;
  }

  if ( faisceau->proprietes & FAISCEAU_FOCALISE ) {
    printf("FAISCEAU_GeometrieApresEchantillon : À FAIRE --- Faisceau focalisé rectangulaire\n");
  } else {
    /* Faisceau rectangulaire parallèle ou ponctuel : la géométrie est inchangée, seul le repère local bouge ! */
    f_transmis = FAISCEAU_Copier( faisceau );
  }

  if ( NULL == f_transmis ) {
    (void) fprintf(stderr, "FAISCEAU_GeometrieApresEchantillon : Impossible de créer le faisceau transmis\n");
    return NULL;
  }

  /* On cherche la position du nouveau point central du faisceau == positon du repère local */
  /* 1. Direction du faisceau : axe des Y du repère local */
  direction.x = sin( faisceau->orientation.theta );
  direction.y = cos( faisceau->orientation.theta ) * cos( faisceau->orientation.phi );
  direction.z = - cos( faisceau->orientation.theta ) * sin( faisceau->orientation.phi );

  /* 2. On cherche le point de sortie */
  ECH_TrouverSortieFaisceau( echantillon, &(faisceau->orientation.position), &direction, &point_sortie, FAUX );

  /* 3. Ce point de sortie est donc le centre du faisceau transmis */
  f_transmis->orientation.position = point_sortie;

  /* Faisceau ponctuel : on a terminé ! */
  if ( faisceau->proprietes & FAISCEAU_PONCTUEL ) {
    return f_transmis;
  }

  /* 4. Si le faisceau n'est pas ponctuel : la face de sortie est-elle unique ? */
  printf("FAISCEAU_GeometrieApresEchantillon : À FAIRE --- Contrôle de l'unicité de la face de sortie\n");

  /* 5. Faisceau non-ponctuel : on crée une coupe qui contiendra les résultats */
  f_transmis->g_espace = IMG_CopierImage( faisceau->g_espace, NULL, VRAI );
  if ( NULL == f_transmis->g_espace ) {
    double pas_x, pas_y;

    /* On choisit le plus précis entre faisceau/10 et échantillon ! */
    pas_x = f_transmis->largeur / 10.0;
    if ( echantillon->pas_x < pas_x ) pas_x = echantillon->pas_x;

    pas_y = f_transmis->hauteur / 10.0;
    if ( echantillon->pas_y < pas_y ) pas_y = echantillon->pas_y;

    f_transmis->g_espace = FAISCEAU_ReserverCoupe( f_transmis->largeur, f_transmis->hauteur, pas_x, pas_y );
    if ( NULL == f_transmis->g_espace ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de préparer la coupe\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FAISCEAU_Effacer( f_transmis );
      /*     SORTIE_FONCTION( __FUNCTION__, "%g [Impossible de créer la coupe]", -2.0 ); */
      return NULL;
    }
    IMG_Titre( f_transmis->g_espace, "Coupe après l'échantillon" );
  }

  if ( faisceau->proprietes & FAISCEAU_POLYCHROMATIQUE ) {
    unsigned long i, nbr_valeurs;

    /* 6. Faisceau polychromatique : doit-on créer une image avec g(E) pour chaque graphe ? */
    if ( ( echantillon->proprietes & ECHANTILLON_HOMOGENE ) && ( FAUX == ( faisceau->proprietes & FAISCEAU_INHOMOGENE ) ) ) {
      /* Échantillon et faisceau homogènes -> pas besoin ! */
      return f_transmis;
    }

#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : echantillon inhomogène [%d], faisceau inhomogène [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, echantillon->proprietes & ECHANTILLON_HOMOGENE, faisceau->proprietes & FAISCEAU_INHOMOGENE );
#endif
    f_transmis->proprietes |= FAISCEAU_INHOMOGENE; /* Pour être sûr d'utiliser la coupe, même si l'original était homogène */

    /* 7. On transforme donc la coupe résultat en coupe contenant un graphe par g(E) */
    retour = IMG_ChangerGenre( f_transmis->g_espace, IMG_GENRE_GRAPHE, NULL, NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_ChangerGenre() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
      FAISCEAU_Effacer( f_transmis );
      return NULL;
    }

    /* 8. On crée les graphes g(E) [isolés] en chaque point pour cette image */
    nbr_valeurs = f_transmis->g_espace->larg * (unsigned long) f_transmis->g_espace->haut;
    for ( i = 0 ; i < nbr_valeurs ; i++ ) {
      Graphe *nv_graphe;

      nv_graphe = GRF_Copier( faisceau->g_E, NULL, VRAI );
      if ( NULL == nv_graphe ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de créer les g(E)\n",
		       __FILE__, __FUNCTION__, __LINE__);
	FAISCEAU_Effacer( f_transmis );
	return NULL;
      }
      /* On doit garder la structure spatiale du faisceau malgré tout ! */
      if ( faisceau->g_espace != NULL ) {
	GRF_FacteurMultiplicatif( nv_graphe, faisceau->g_espace->valeurs.R[ i ], AXE_Y, NULL );
      }
      f_transmis->g_espace->valeurs.G[ i ] = nv_graphe;
    }

    /* On n'utilise donc pas g_E unique ! */
    GRF_eff_isole( f_transmis->g_E ); f_transmis->g_E = NULL;
  }

  /* On a fini, tout s'est bien passé */
  return f_transmis;
}
