/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998, Emmanuel CURIS

   Module d'affichage graphique des rsultats

   Cration d'une figure
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#ifdef _DEVELOPPE_
# include "graphes.h"		/* Normalement inutile ! */
#endif
#include "figures.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int BLOC_Initialiser(Bloc_Figure *bloc, unsigned int type);
static int BLOC_Liberer    (Bloc_Figure *bloc);

/* ~~~~~~~~ Realisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~ Gestion de la figure dans son ensemble ~-~-~-~-~-~-~-~ */

/* -------- Crer une nouvelle figure -------- */

Figure *FIG_CreerFigure(void)
{
  Figure *figure;

  /* On rserve la mmoire */
  figure = (Figure *) malloc( sizeof( Figure ) );
  if ( NULL == figure ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mmoire insuffisante pour crer une nouvelle figure\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On l'initialise : aucun bloc au dbut, taille A4 en paysage, pas de nom */
  figure->largeur = 297;
  figure->hauteur = 210;

  figure->marge_gauche = 10; /* 1 cm de marge partout */
  figure->marge_droite = 10;
  figure->marge_haute = 10;
  figure->marge_basse = 10;

  figure->titre = NULL;

  figure->premier_bloc = NULL;
  figure->bloc_actif = NULL;

  /* Termin */
  return figure;
}

/* -------- Effacer une figure -------- */

int FIG_EffacerFigure(Figure *figure)
{
  int retour;

  if ( NULL == figure ) return +1;  /* Erreur mineure ! */

  /* Il n'y a plus de bloc actif, forcment... */
  figure->bloc_actif = NULL;

  /* On efface tous les blocs de la figure */
  while ( figure->premier_bloc != NULL ) {
    retour = FIG_EffacerBloc( figure, figure->premier_bloc );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : chec en voulant effacer les blocs de la figure !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      /* chec en effaant le bloc ! */
      return -1;
    }
  }

  /* On efface le titre de la figure */
  free( figure->titre ); figure->titre = NULL;

  /* On peut maintenant effacer la figure */
  free( figure ); figure = NULL;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Changer le nom d'une figure -------- */

int FIG_RenommerFigure(Figure *figure, const char *nouveau_titre)
{
  if ( NULL == figure ) return +1;  /* Erreur mineure */
  if ( NULL == nouveau_titre ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de titre indiqu !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;  /* Pas de titre indiqu ! */
  }

  /* On efface l'ancien nom */
  free( figure->titre ); figure->titre = NULL;

  /* On mmorise une copie du nouveau nom : sera libre au prochain changement de nom ou  l'effacement de la figure */
  figure->titre = strdup( nouveau_titre );

  /* On a termin sans souci */
  return 0;
}

/* ~-~-~-~-~-~-~-~ Gestion des blocs composants une figure ~-~-~-~-~-~-~-~ */

/* -------- Changer le bloc actif d'une figure -------- */

void FIG_ChangerBlocActif(Figure *figure, Bloc_Figure *nv_actif)
{
  if ( NULL == figure ) return;
  if ( NULL == nv_actif ) {
    if ( figure->bloc_actif != NULL ) {
      figure->proprietes |= FIGURE_A_REDESSINER;
    }

    figure->bloc_actif = NULL;
  } else {
    Bloc_Figure *bloc;

    /* On contrle que le bloc est bien dans la figure indique */
    bloc = figure->premier_bloc;
    while ( bloc != nv_actif ) {
      if ( NULL == bloc ) {
	(void) fprintf(stderr, "%s, %s l.%u : Bloc actif indiqu pas dans la figure !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	return;
      }
      bloc = bloc->suivant;
    }

    /* On change */
    figure->bloc_actif = nv_actif;

    /* On doit donc redessiner */
    figure->proprietes |= FIGURE_A_REDESSINER;
  }
}

/* -------- Crer un nouveau bloc -------- */

Bloc_Figure *FIG_CreerBloc(Figure *figure, Bloc_Figure *avant, unsigned int type, double xBG_mm, double yBG_mm, double larg_mm, double haut_mm)
{
  Bloc_Figure *nv_bloc;
  int retour;

  /* Contrles */
  if ( NULL == figure ) {
    AVERTISSEMENT( "Aucune figure [%p]", (void *) figure );
    return NULL;
  }

  if ( avant != NULL ) {
    Bloc_Figure *bloc;

    /* Le bloc prcdent indiqu appartient-il bien  la figure ? */
    bloc = figure->premier_bloc;
    while ( bloc != avant ) {
      bloc = bloc->suivant;
      if ( NULL == bloc ) {
	(void) fprintf(stderr, "%s, %s l.%u : bloc %p indiqu comme prcdent, mais pas dans la figure %p\n",
		       __FILE__, __FUNCTION__, __LINE__, (void *) avant, (void *) figure);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	return NULL;
      }
    }
  }

  /* Rservation du bloc */
  nv_bloc = ( Bloc_Figure * ) malloc( sizeof( Bloc_Figure ) );
  if ( NULL == nv_bloc ) {
    (void) fprintf(stderr, "%s, %s l.%u : Mmoire insuffisante pour rserver un nouveau bloc\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* Initialisation du bloc */
  retour = BLOC_Initialiser( nv_bloc, type );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans l'initialisation [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    if ( retour < 0 ) {		/* Erreur grave, cration annule */
      free( nv_bloc ); nv_bloc = 0;
      return NULL;
    }
  }

  /* Initialisation des coordonnes */
  nv_bloc->dessin.x_BG_pc = (  xBG_mm / (double) figure->largeur ) * 100.0;
  nv_bloc->dessin.y_BG_pc = (  yBG_mm / (double) figure->hauteur ) * 100.0;
  nv_bloc->dessin.larg_pc = ( larg_mm / (double) figure->largeur ) * 100.0;
  nv_bloc->dessin.haut_pc = ( haut_mm / (double) figure->hauteur ) * 100.0;
  figure->proprietes |= FIGURE_A_REECHELONNER | FIGURE_A_REDESSINER; /* Au dessin [demand], il faudra convertir en coordonnes cran ! */

  /* Chanage */
  if ( NULL == avant ) {
    nv_bloc->suivant = figure->premier_bloc;
    figure->premier_bloc = nv_bloc;
  } else {
    nv_bloc->suivant = avant->suivant;
    avant->suivant = nv_bloc;
  }

  /* Cration termine avec succs */
  return nv_bloc;
}

/* -------- Rserver les infos complmentaires du bloc -------- */

static int BLOC_Initialiser(Bloc_Figure *bloc, unsigned int type)
{
  int retour = 0;

  /* Indpendamment du type : les axes */
  bloc->dessin.proprietes = CADRE2D_AVEC_CADRE | CADRE2D_ECRAN;
  bloc->dessin.X_secondaire = NULL;
  bloc->dessin.Y_secondaire = NULL;
  bloc->dessin.couleur_bord = NOIR;
  bloc->dessin.couleur_fond = BLANC;

  AXE_Initialiser( &(bloc->dessin.axe_X), FAUX );
  bloc->dessin.axe_X.type_graduation = GRADUATION_AUCUNE;

  AXE_Initialiser( &(bloc->dessin.axe_Y), VRAI );
  bloc->dessin.axe_Y.type_graduation = GRADUATION_AUCUNE;

  /* En fonction du type */
  bloc->type = type;
  switch ( type )
  {
  case BLOCFIGURE_TYPE_GRAPHES:
    bloc->donnees.graphes.nbr_graphes = 0;
    bloc->donnees.graphes.infos_graphe = NULL;

    /* Complment pour les axes */
    bloc->dessin.axe_X.type_graduation = GRADUATION_DECIMALE;
    bloc->dessin.axe_Y.type_graduation = GRADUATION_DECIMALE;
    break;

  case BLOCFIGURE_TYPE_TEXTE:
    bloc->donnees.texte = NULL;
    retour = -1;
    break;

  case BLOCFIGURE_TYPE_LEGENDE:
    bloc->donnees.a_legender = NULL;
    retour = -1;
    break;

  case BLOCFIGURE_TYPE_IMAGE:
    bloc->donnees.image = NULL;
    break;

  case BLOCFIGURE_TYPE_SURFACE:
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type de bloc inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, type);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    retour = +1;
    break;
  }

  /* Initialisation russie */
  return retour;
}

/* -------- Effacer un bloc et son contenu -------- */

int FIG_EffacerBloc(Figure *figure, Bloc_Figure *bloc)
{
  Bloc_Figure *bloc_l;

  if ( ( NULL == figure ) || ( NULL == bloc ) ) return 0;

  /* On commence par changer, si ncessaire, le bloc actif */
  if ( bloc == figure->bloc_actif ) {
    /* On passe au suivant ; si dernier, on passe au premier */
    if ( bloc->suivant != NULL ) {
      figure->bloc_actif = bloc->suivant;
    } else if ( figure->bloc_actif != figure->premier_bloc ) {
      figure->bloc_actif = figure->premier_bloc;
    } else {
      figure->bloc_actif = NULL;
    }
  }

  /* On sort le bloc du chanage */
  if ( bloc == figure->premier_bloc ) {
    figure->premier_bloc = figure->premier_bloc->suivant;
  } else {
    Bloc_Figure *avant;

    avant = figure->premier_bloc;
    while ( avant->suivant != bloc ) {
      avant = avant->suivant;
      if ( NULL == avant ) {
	(void) fprintf(stderr, "%s, %s l.%u : Le bloc indiqu [%p] n'est pas dans la figure donne [%p]\n",
		       __FILE__, __FUNCTION__, __LINE__, (void *) bloc, (void *) figure);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	return -1;
      }
    }

    avant->suivant = bloc->suivant;
    bloc->suivant = NULL;
  }

  /* On contrle qu'il n'est pas lgend par un autre bloc */
  bloc_l = figure->premier_bloc;
  while ( bloc_l != NULL ) {
    if ( BLOCFIGURE_TYPE_LEGENDE == bloc_l->type ) {
      if ( bloc == bloc_l->donnees.a_legender ) { /* On lgende le bloc que l'on est en train d'effacer ! */
	Bloc_Figure *suivant;

	/* Quel sera le suivant  regarder ? */
	suivant = bloc_l->suivant;

	/* On efface donc ce bloc lgende aussi... */
	FIG_EffacerBloc( figure, bloc_l );

	/* Et on passe au suivant [ATTENTION, bloc_l a t libr -> son contenu n'est plus utilisable] */
	bloc_l = suivant;
	continue;		/* Pour ne pas sauter un bloc en faisant deux fois ->suivant ! */
      }
    }

    bloc_l = bloc_l->suivant;
  }

  /* On libre d'ventuels axes secondaires */
  AXE_Liberer( bloc->dessin.X_secondaire, VRAI ); bloc->dessin.X_secondaire = NULL;
  AXE_Liberer( bloc->dessin.Y_secondaire, VRAI ); bloc->dessin.Y_secondaire = NULL;

  return BLOC_Liberer( bloc );
}

/* -------- Effacer & librer le contenu d'un bloc (en fonction de son type) -------- */

static int BLOC_Liberer(Bloc_Figure *bloc)
{
  switch ( bloc->type )
  {
  case BLOCFIGURE_TYPE_GRAPHES:
    {
      unsigned long i;

      for ( i = 0 ; i < bloc->donnees.graphes.nbr_graphes ; i++ ) {
	if ( VRAI == GRF_EstIsole( bloc->donnees.graphes.infos_graphe[ i ].graphes ) ) {
	  /* On efface les graphes isols [qui sinon ne seraient jamais librs !] */
	  GRF_eff_isole( bloc->donnees.graphes.infos_graphe[ i ].graphes );
	}
	/* Les autres sont simplement dlis */
	bloc->donnees.graphes.infos_graphe[ i ].graphes = NULL;
      }
    }
    free( bloc->donnees.graphes.infos_graphe );
    bloc->donnees.graphes.infos_graphe = NULL;
    break;

  case BLOCFIGURE_TYPE_TEXTE:
    free( bloc->donnees.texte ); bloc->donnees.texte = NULL;
    break;

  case BLOCFIGURE_TYPE_LEGENDE:
    bloc->donnees.a_legender = NULL;
    break;

  case BLOCFIGURE_TYPE_IMAGE:
    bloc->donnees.image = NULL;
    break;

  case BLOCFIGURE_TYPE_SURFACE:
    /* Rien de spcial  librer */
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type de bloc inconnu [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, bloc->type);
    return -2;
  }

  /* Libration russie */
  return 0;
}

/* -------- Changer les graphes associs  un bloc -------- */

int BLOC_AjouterGraphes(Bloc_Figure *bloc, struct graphe **graphes, unsigned long nbr_graphes)
{
  struct axes_graphe *nv_adresse;
  unsigned long i, nv_taille;

  /* Contrles prliminaires */
  if ( ( NULL == bloc ) || ( NULL == graphes ) || ( 0 == nbr_graphes ) ) return -1;
  if ( bloc->type != BLOCFIGURE_TYPE_GRAPHES ) {
    (void) fprintf(stderr, "%s, %s l.%u : Appel pour un bloc autre que de graphes !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* On augmente la mmoire associe aux blocs */
  nv_taille  = sizeof( struct axes_graphe ) * ( bloc->donnees.graphes.nbr_graphes + nbr_graphes );
  nv_adresse = ( struct axes_graphe * ) realloc( bloc->donnees.graphes.infos_graphe, nv_taille );
  if ( NULL == nv_adresse ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de rserver pour les nouveaux graphes !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }
  bloc->donnees.graphes.infos_graphe = nv_adresse;

  /* On installe les nouveaux graphes */
  for ( i = 0 ; i < nbr_graphes ; i++ ) {
    bloc->donnees.graphes.infos_graphe[ bloc->donnees.graphes.nbr_graphes + i ].graphes = graphes[ i ];
    bloc->donnees.graphes.infos_graphe[ bloc->donnees.graphes.nbr_graphes + i ].axes_secondaires = 0; /* Axes principaux par dfaut */
    bloc->donnees.graphes.infos_graphe[ bloc->donnees.graphes.nbr_graphes + i ].epaisseur =  2; /* paisseur : 2/10 mm */
    bloc->donnees.graphes.infos_graphe[ bloc->donnees.graphes.nbr_graphes + i ].ht_marque = 20; /* Hauteur de marque : 2 mm */
  }

  /* On met  jour le nombre de graphes dans le bloc */
  bloc->donnees.graphes.nbr_graphes += nbr_graphes;

  /* Si on met les tous premiers graphes : on en profite pour chercher les bornes */
  if ( nbr_graphes == bloc->donnees.graphes.nbr_graphes ) {
    BLOC_OptimiserBornes( bloc, BORNES_XY );
  }

  /* Termin */
  return 0;
}

/* -------- Position d'un point par rapport au bloc indiqu --------
    Renvoie le code de position (hors rgion en cas d'erreur)

    x et y sont supposs en coordonnes CRAN et le bloc initialis en fonction
 */

unsigned int BLOC_PositionPoint (Bloc_Figure *bloc, int x, int y, unsigned int marge)
{
  if ( NULL == bloc ) return TRC_EST_HORS_REGION;

  return TRC_PositionDansRegion2D( &(bloc->dessin), x, y, marge );
}

/* -------- Dessiner un bloc -------- */

void FIG_DessinerBloc(Fenetre *fenetre, Bloc_Figure *bloc, int x_zone, int y_zone, unsigned int larg_zone, unsigned int haut_zone)
{
  if ( NULL == bloc ) return;

  /* 1. Le cadre, en pointill, pour les dplacements */
/*   printf("Dimensions bloc : L=%u, H=%u en (%d, %d) [L=%g%%, H=%g%% en (%g%%, %g%%)]\n", */
/* 	 bloc->dessin.largeur, bloc->dessin.hauteur, bloc->dessin.x_BG,  bloc->dessin.y_BG, */
/* 	 bloc->dessin.larg_pc, bloc->dessin.haut_pc, bloc->dessin.x_BG_pc, bloc->dessin.y_BG_pc); */
  /* Attention, on veut le coin HG comme coordonnes de dpart ! */
  TRC_Couleur( VERT );
  TRC_Style( 3 );
  TRC_Z_Rectangle( fenetre, x_zone, y_zone, larg_zone, haut_zone,
		   bloc->dessin.x_BG, bloc->dessin.y_BG - bloc->dessin.hauteur + 1, bloc->dessin.largeur, bloc->dessin.hauteur );

  if ( ( 0 == bloc->dessin.largeur ) ||( 0 == bloc->dessin.hauteur ) ) {
    /* Cadre rduit  une ligne -> pas de dessin de toute faon ! */
    return;
  }

  /* En fonction du type de bloc */
  switch ( bloc->type )
  {
  case BLOCFIGURE_TYPE_GRAPHES:
    /* On dessine le cadre et les axes */
    TRC_DessineRegion2D( fenetre, &(bloc->dessin) );

    /* On dessine tous les graphes */
    if ( bloc->donnees.graphes.nbr_graphes > 0 ) {
      unsigned long i;
      struct axes_graphe *inf_graphe;

      for ( i = 0, inf_graphe = bloc->donnees.graphes.infos_graphe ; i < bloc->donnees.graphes.nbr_graphes ; i++, inf_graphe++ ) {
	/* Trac proprement dit */
/* 	printf(" Trac de %s dans (%d,%d)->(%d,%d) X:%f p/r, Y: %f p/r\n", inf_graphe->graphes->nom, bloc->dessin.x_BG, bloc->dessin.y_BG - bloc->dessin.hauteur, */
/* 	       bloc->dessin.x_BG + bloc->dessin.largeur, bloc->dessin.y_BG, echelle_x, echelle_y ); */

	TRC_GrapheDansRegion2D( fenetre, &(bloc->dessin), inf_graphe->axes_secondaires, inf_graphe->graphes, -1, -1, -1, AXE_X | AXE_Y, NULL, NULL );
      }
    }
    break;

  default:
    OPTION_A_FAIRE( "Bloc de type ..." );
    break;
  }
}

/* -------- Tracer un rticule sur un bloc -------- */

void BLOC_TracerReticule(Fenetre *fenetre, Bloc_Figure *bloc, unsigned int axes, int x, int y)
{
  if ( NULL == bloc) return;

  /* Rticule : en mode inversion, en noir, continu */
  TRC_Mode( MD_INVERSE );
  TRC_Couleur( NOIR );

  /* Rticule parallle  l'axe des X ? [donc horizontal] */
  if ( axes & AXE_X ) {
    TRC_LigneH( fenetre, bloc->dessin.x_BG, y, bloc->dessin.largeur );
  }

  /* Rticule parallle  l'axe des Y ? [donc vertical] */
  if ( axes & AXE_Y ) {
    TRC_LigneV( fenetre, x, bloc->dessin.y_BG - bloc->dessin.hauteur, bloc->dessin.hauteur );
  }
}
