/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998, Emmanuel CURIS

   Module d'affichage graphique des résultats

   Tracé des images "classiques"
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "echantillon.h" /* Pour avoir le contenu d'une composition */
#include "images.h"
#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int TRC_ImageComplete_Compo(Fenetre *fenetre, Image *image, int x, int y, unsigned int larg, unsigned int haut, double echelle);
static int TRC_ImageComplete_Maths(Fenetre *fenetre, Image *image, int x, int y, unsigned int larg, unsigned int haut, double echelle,
				   double z_min, double z_max, double echelle_Z, int couleur_base);

static int TRC_EchelleComposition(Fenetre *fenetre, Image *image, int x, int y, unsigned int larg, unsigned int haut);
static int TRC_EchelleReelle     (Fenetre *fenetre, DES_image *image, int x, int y, unsigned int larg, unsigned int haut);

static int TRC_CouleurCompo(struct matiere *composition, struct matiere **table, unsigned long nbr_compositions);

/* ~~~~~~~~ Realisation pratique ~~~~~ */

/* -------- Calculer l'échelle optimale pour faire tenir dans un repère orthonormé la zone indiquée
   Renvoie l'échelle en /pixels

larg_reelle, haut_reelle : la région du plan que l'on cherche à représenter
lg_zone, ht_zone         : les dimensions de la zone d'affichage à l'écran

   -------- */

double TRC_Echelle2D( double larg_reelle, double haut_reelle, unsigned int lg_zone, unsigned int ht_zone )
{
  double echelle_X, echelle_Y;
  double lg_predite, ht_predite;

  if ( ( 0 == lg_zone ) || ( 0 == ht_zone ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Dimensions de zone absurdes [L=%u, H=%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, lg_zone, ht_zone);
    return -1.0;
  }

  /* Échelles normales, pas forcément compatibles */
  echelle_X = larg_reelle / (double) lg_zone;
  echelle_Y = haut_reelle / (double) ht_zone;

  if ( ( 0.0 == echelle_X ) && ( 0.0 == echelle_Y ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Échelles absurdes [ L=%g / %u, H = %g / %u ]\n",
		   __FILE__, __FUNCTION__, __LINE__, larg_reelle, lg_zone, haut_reelle, ht_zone );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return 0.0;
  }

  /* Si l'une est absurde, l'autre est la seule utilisable */
  if ( 0.0 == echelle_X ) return echelle_Y;
  if ( 0.0 == echelle_Y ) return echelle_X;

  /* Si les deux sont utilisables, on utilise celle qui permet de faire tenir
     toute la région demandée  */
  lg_predite = echelle_Y * lg_zone;
  ht_predite = echelle_X * ht_zone;

/*   printf("%s : echelle X=%g -> %g prédite [%g] ; echelle Y=%g -> %g prédite [%g]\n", */
/* 	 __FUNCTION__, echelle_X, ht_predite, haut_reelle, echelle_Y, lg_predite, larg_reelle ); */

  if ( lg_predite < larg_reelle ) {
    if ( ht_predite < haut_reelle ) {
      (void) fprintf(stderr, "%s,%s l.%u : Aucune des deux échelles ne convient. Pourquoi ?\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }

    /* La largeur avec l'échelle des Y est plus petite que la largeur totale voulue
       ==> seule l'échelle X permet de représenter toute la région */
    return echelle_X;
  } else {
    if ( ht_predite < haut_reelle ) {
      /*  Seule l'échelle Y permet de représenter toute la région */
      return echelle_Y;
    }

#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Les deux échelles conviennent ??\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return echelle_X;
  }

  (void) fprintf(stderr, "%s,%s l.%u : On ne devrait jamais être là\n",
		 __FILE__, __FUNCTION__, __LINE__);
  return 0.0;
}

int TRC_PortionImageDansRectangle(Fenetre *fenetre, DES_image *des_image, int x, int y, unsigned int larg, unsigned int haut)
{
  double echelle /* en µm/pixel, si les X/Y sont en µ*/, echelle_Z;
  double lg_case, ht_case;
  double larg_image, haut_image;
  int points[ 10 ];
  Image *image;

  if ( NULL == des_image ) return -1;
  if ( NULL == des_image->image ) return -2;

#ifdef _DEVELOPPE_
  TRC_Couleur( ROUGE );
  TRC_Rectangle( fenetre, x, y, larg, haut );
#endif

  /* La taille est-elle suffisante */
  if ( ( larg < 10 ) || ( haut < 10 ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Région trop petite pour raisonnablement dessiner une image !\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return +4;
  }

  image = des_image->image;

  if ( ( des_image->x_max == des_image->x_min ) || ( des_image->y_max == des_image->y_min ) ) {
    TRC_Z_Message( fenetre, x, y, larg, haut, "Zoom trop important" );
    return +21;
  }

  /* Recherche des facteurs d'échelle optimaux */
  echelle_Z = image->val_max - image->val_min;
  if ( echelle_Z <= 0.0 ) {
    char *texte_erreur;

    texte_erreur = TXT_genere_chaine( "Image sans contraste [Z=%g]", image->val_min );
    TRC_Z_Message( fenetre, x, y, larg, haut, texte_erreur );
    free( texte_erreur );
    return +8;
  }
  echelle = TRC_Echelle2D( des_image->x_max - des_image->x_min, des_image->y_max - des_image->y_min, larg, haut);
  if ( echelle <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de trouver une échelle convenable\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -5;
  }

  /* Dimensions d'une case de l'image, en pixels */
  lg_case = image->l_reel / echelle;
  ht_case = image->h_reel / echelle;

  /* Dimension de l'image entière, compte-tenu de l'échelle */
  larg_image = lg_case * image->larg;
  haut_image = ht_case * image->haut;

  /* 1. Tracé de l'image */
  switch ( image->genre )
  {
  case IMG_GENRE_REELLE:
    if ( ( image->x_min >= des_image->x_min ) && ( image->y_min >= des_image->y_min ) &&
	 ( larg_image < (double) (larg + 1) ) && ( haut_image < (double) (haut + 1) ) ) {
      /* On compte un pixel de marge pour les erreurs d'arrondi double/entier */
#ifdef _DEVELOPPE_
      TRC_Couleur( BLEU );
      TRC_Rectangle( fenetre, x, y, larg_image, haut_image );
#endif

      /* On installe les couleurs */
      TRC_CreerDegrade( fenetre, des_image->couleur );

      /* On trace avec l'échelle indiquée, dans le rectangle indiqué */
      TRC_ImageComplete_Maths( fenetre, image, x, y, larg_image, haut_image, 1.0/echelle, image->val_min, image->val_max, echelle_Z, des_image->couleur );

      /* On remet la palette originale */
/*       (void) VDI_rappelle_palette(station); */
    } else {
      (void) fprintf(stderr, "%s [Min : X=%g/%g, Y=%g/%g ; Larg=%g/%u [D=%g] ; Haut=%g/%u [D=%g] ; echelle=%g]\n",
		     __FUNCTION__, image->x_min, des_image->x_min, image->y_min, des_image->y_min,
		     larg_image, larg, larg - larg_image, haut_image, haut, haut - haut_image, echelle);
      OPTION_A_FAIRE( "Tracé d'une image à valeurs réelles / partielle" );
    }
    break;

  case IMG_GENRE_COMPLEXE:
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Tracé d'une image à valeurs complexes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_GRAPHE:
    if ( ( image->x_min >= des_image->x_min ) && ( image->y_min >= des_image->y_min ) &&
	 ( larg_image < (double) (larg + 1) ) && ( haut_image < (double) (haut + 1) ) ) {
      /* On compte un pixel de marge pour les erreurs d'arrondi double/entier */
#ifdef _DEVELOPPE_
      TRC_Couleur( BLEU );
      TRC_Rectangle( fenetre, x, y, larg_image, haut_image );
#endif

      /* On installe les couleurs */
#ifdef ancien
      (void) VDI_sauver_palette(station);
      VDI_creer_degrade(station, des_image->couleur, None);
#endif

      /* On trace avec l'échelle indiquée, dans le rectangle indiqué */
      TRC_ImageComplete_Maths( fenetre, image, x, y, larg_image, haut_image, 1.0/echelle, image->val_min, image->val_max, echelle_Z, des_image->couleur );

#ifdef ancien
      /* On remet la palette originale */
      (void) VDI_rappelle_palette(station);
#endif
    } else {
      (void) fprintf(stderr, "%s [Min : X=%g/%g, Y=%g/%g ; Larg=%g/%u [D=%g] ; Haut=%g/%u [D=%g] ; echelle=%g]\n",
		     __FUNCTION__, image->x_min, des_image->x_min, image->y_min, des_image->y_min,
		     larg_image, larg, larg - larg_image, haut_image, haut, haut - haut_image, echelle);
      OPTION_A_FAIRE( "Tracé d'une image à valeurs graphes / partielle" );
    }
    break;

  case IMG_GENRE_DIAGRAMME:
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Tracé d'une image de type diagramme\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_PHOTO:
    (void) fprintf(stderr, "%s,%s l.%u : À FAIRE --- Tracé d'une image-photo\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case IMG_GENRE_COMPO:
    if ( ( image->x_min >= des_image->x_min ) && ( image->y_min >= des_image->y_min ) &&
	 ( larg_image < (double) (larg + 1) ) && ( haut_image < (double) (haut + 1) ) ) {
      TRC_ImageComplete_Compo( fenetre, image, x, y, larg_image, haut_image, 1.0/echelle );
    } else {
      OPTION_A_FAIRE( "Portion d'une image de type composition" );
    }
    break;
  }

  /* 2. Si demandé, tracé du cadre */
  if ( des_image->proprietes & IMG_AVEC_CADRE ) {
    /* Tracé du tour - en noir */
    TRC_Couleur( NOIR );
    TRC_Style( SOLID );
    TRC_Epaisseur( 1 );
    TRC_Rectangle( fenetre, x, y, larg, haut );
  }

  /* 3. Si demandé, tracé de la coupe [VERT] */
  if ( des_image->proprietes & IMG_AVEC_COUPE ) {
    if ( image->infos_affichage != NULL ) {
      if ( image->infos_affichage->coupe != NULL ) {
	int retour;

	TRC_Couleur( VERT );
	vst_alignment(_PRG_station->handle, BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_BAS, &retour, &retour);

	/* Le premier point */
	retour = TRC_AbsoluVersEcran( x, y, larg, haut, echelle, des_image->x_min, des_image->y_max,
				      image->infos_affichage->xp_1, image->infos_affichage->yp_1, points, points + 1);
	/* Le second point */
	retour = TRC_AbsoluVersEcran( x, y, larg, haut, echelle, des_image->x_min, des_image->y_max,
				      image->infos_affichage->xp_2, image->infos_affichage->yp_2, points + 2, points + 3);

	/* Les deux marques extrêmes */
	TRC_Marque( fenetre, points[ 0 ], points[ 1 ], PM_CROIX, 5 );
	TRC_Marque( fenetre, points[ 2 ], points[ 3 ], PM_CROIX, 5 );

	/* La trace de la coupe */
	TRC_Ligne( fenetre, points[ 0 ], points[ 1 ], points[ 2 ], points[ 3 ] );

	/* On marque quel point est qui */
	TRC_Mode( MD_TRANSPARENT );
	TRC_Texte( fenetre, points[ 0 ], points[ 1 ] + 3, BI_TEXTE_CENTRER_H, BI_TEXTE_CENTRER_V, 0, "1" );
	TRC_Texte( fenetre, points[ 2 ], points[ 3 ] + 3, BI_TEXTE_CENTRER_H, BI_TEXTE_CENTRER_V, 0, "2" );
	TRC_Mode( MD_REMPLACE );
      }
#ifdef _DEVELOPPE_
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : Pas d'infos d'affichage, mais coupe demandée !\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
    }
  }

  /* 4. Si demandé, tracé du centre du repère [VIOLET] */
  if ( des_image->proprietes & IMG_AVEC_CENTRE ) {
    if ( ( des_image->x_min <= 0.0 ) && ( des_image->x_max >= 0.0 ) &&
	 ( des_image->y_min <= 0.0 ) && ( des_image->y_max >= 0.0 ) ) {
      int x_centre, y_centre;
      /* Le centre est bien dans la région du plan dessinée */

      x_centre = x + ( 0.0 - des_image->x_min ) / echelle;
      y_centre = y + ( des_image->y_max - 0.0 ) / echelle;

      TRC_Couleur( VIOLET ); /* Pour qu'il ressorte bien */
      TRC_Cercle( fenetre, x_centre, y_centre, 2 );
    }
  }

  /* . Si demandé, tracé de  [] */
  if ( des_image->proprietes & IMG_AVEC_POINTS ) {
  }

  /* . Si demandé, tracé de  [] */
  if ( des_image->proprietes & IMG_AVEC_GRILLE ) {
  }

  /* . Si demandé, tracé de  [] */
  if ( des_image->proprietes & IMG_AVEC_GRADUATION ) {
  }

  /* . Si demandé, tracé de  [] */
  if ( des_image->proprietes & IMG_AVEC_AXES ) {
  }

  /* Dessin terminé, tout s'est bien passé */
  return 0;
}

/* -------- Trace d'une image dans un rectangle -----------\
 |                                                         |
 | station = la station VDI ou il faut dessiner            |
 | image   = l'image a afficher                            |
 | x, y    = coin superieur gauche du rectangle de dessin  |
 | larg, haut : dimensions du rectangle de dessin          |
 | mode    = convention pour les Y (écran, maths ?)        |
 \-------------------------------------------------------- */

void TRC_image(Fenetre *fenetre, DES_image *image, int x_HG, int y_HG, unsigned int larg, unsigned int haut, int mode)
{
  float echelle, l_case, h_case;
  float z_min, z_echelle;
  double *valeurs;
  unsigned int i, j, larg_image, haut_image; /* Dimensions de l'image totale, en pixels */
  int x_BG __attribute__((__unused__)), y_BG __attribute__((__unused__));
  int points[4];

  if (image == NULL) return;
  if (image->image == NULL) return;

  echelle = IMG_dim_trace(image->image, larg, haut,  &larg_image, &haut_image);
  l_case = image->image->l_reel * echelle;
  h_case = image->image->h_reel * echelle;

  if ( (h_case <= 0.0) || (l_case <= 0.0) ) {
    TRC_Z_Message( fenetre, x_HG, y_HG, larg, haut, "Image avec une dimension de case nulle" );
    return;
  }

  /* On en déduit les coordonnées du coin inférieur, gauche */
  x_BG = x_HG; y_BG = y_HG + haut - 1;

  z_echelle = image->image->val_max - image->image->val_min;
  if ( z_echelle <= 0.0 ) {
    TRC_Z_Message( fenetre, x_HG, y_HG, larg, haut, "Image sans contraste" );
    return;
  }
  z_min = image->image->val_min; /* Pour gagner du temps (?) et clarifier le code */
  valeurs = image->image->valeurs.R;

/*   (void) VDI_sauver_palette(station); */
/*   VDI_creer_degrade(station, image->couleur, None); */

  /* Affichage de l'image, pavé par pavé
     Idéalement, on devrait mettre un point de la couleur pour chaque (x, y) connu.
     On relie par l'équivalent des droites : des rectangles de couleur.

     On a donc le schéma suivant :  +-----+-----+
                                    |  X..|..X  |  X = valeur connue
                                    +-----+-----+  + = limite de pavé affiché
                                    |  X..|..X  |
                                    +-----+-----+
     ==> les valeurs connues sont au centre des cases
 */
  if (IMG_TRACE_NORMAL == mode) { /* Tracé avec les conventions informatiques */
    for ( i = 0 ; i < image->image->larg ; i++ ) {
      points[0] = x_HG + i * l_case;
      points[2] = x_HG + (i + 1) * l_case;

      points[ 1 ] = y_HG;
      for ( j = 0 ; j < image->image->haut ; j++, valeurs++ ) {
	int niveau;

	niveau = (int) (255.0 * (*valeurs - image->image->val_min) / z_echelle );
	TRC_Degrade( image->couleur, niveau );

	TRC_RectanglePlein( fenetre, points[ 0 ], points[ 1 ], l_case, h_case );
 	points[1] = y_HG + j * h_case;
 	points[3] = y_HG + (j + 1) * h_case;
      }
    }
  } else { /* Tracé avec les conventions mathématiques */
    int niveau;

    for (j = 0 ; j < image->image->haut - 1 ; j++) {
      points[3] = y_HG + haut_image - j * h_case;
      points[1] = points[3] - h_case;

      for (i = 0 ; i < image->image->larg ; i++) {
	points[0] = x_HG + i * l_case;
	points[2] = x_HG + (i + 1) * l_case;

	niveau = (int) rint( 255.0 * (*valeurs++ - z_min) / z_echelle );
	TRC_Degrade( image->couleur, niveau );

 	TRC_RectanglePlein( fenetre, points[ 0 ], points[ 1 ], l_case, h_case );
      }
    }

    /* Dernière ligne à part pour éviter des erreurs d'arrondi */
    points[3] = y_HG + h_case;
    points[1] = y_HG;

    for (i = 0 ; i < image->image->larg ; i++) {
      points[0] = x_HG + i * l_case;
      points[2] = x_HG + (i + 1) * l_case;

      niveau = (int) rint( 255.0 * (*valeurs++ - z_min) / z_echelle );
      TRC_Degrade( image->couleur, niveau );

      TRC_RectanglePlein( fenetre, points[ 0 ], points[ 1 ], l_case, h_case );
    }
    
#ifdef pas_de_debord
    /* Première case : un quart de rectangle seulement */
    points[0] = x;              points[1] = y + haut_image - h_case / 2.0;
    points[2] = x + l_case/2.0; points[3] = y + haut_image;

    niveau = (int) (255.0 * (valeurs[ 0 ] - z_min) / z_echelle );
    vsf_color(station->handle, niveau);
    v_bar(station->handle, points);

    /* Suite de la première ligne, sauf la dernière case */
    for (i = 1 ; i < image->image->larg - 1 ; i++) {
      points[0] = x + (i * l_case) - l_case / 2;
      points[2] = x + (i * l_case) + l_case / 2;

      niveau = (int) (255.0 * (valeurs[ i ] - z_min) / z_echelle );
      vsf_color(station->handle, niveau);
      v_bar(station->handle, points);
    }

    /* Dernière case de la première ligne */
    points[0] = x + ( (image->image->larg - 2) * l_case ) + l_case / 2;
    points[2] = x + larg_image;

    niveau = (int) (255.0 * (valeurs[ i ] - z_min) / z_echelle );
    vsf_color(station->handle, niveau);
    v_bar(station->handle, points);

    /* Les lignes suivantes, sauf la dernière ligne */
    for (j = 1, num_case = i + 1 ; j < image->image->haut - 1 ; j++) {
      /* Première case de la ligne */
      points[0] = x;              points[1] = y + haut_image - j * h_case - h_case / 2.0;
      points[2] = x + l_case/2.0; points[3] = y + haut_image - j * h_case + h_case / 2.0;

      niveau = (int) (255.0 * (valeurs[ num_case++ ] - z_min) / z_echelle );
      vsf_color(station->handle, niveau);
      v_bar(station->handle, points);

      /* Cases suivantes de la ligne, sauf la dernière */
      for (i = 1 ; i < image->image->larg - 1 ; i++) {
	points[0] = x + (i * l_case) - l_case / 2;
	points[2] = x + (i * l_case) + l_case / 2;

	niveau = (int) (255.0 * (valeurs[ num_case++ ] - z_min) / z_echelle );
	vsf_color(station->handle, niveau);
	v_bar(station->handle, points);
      }

      /* Dernière case de la ligne */
/*      points[0] = x + ( (image->image->larg - 2) * l_case ) + l_case / 2; */
      points[0] = x + larg_image - l_case / 2;
      points[2] = x + larg_image;

      niveau = (int) (255.0 * (valeurs[ num_case++ ] - z_min) / z_echelle );
      vsf_color(station->handle, niveau);
      v_bar(station->handle, points);
    }

    /* Première case de la dernière ligne */
    points[0] = x;              points[1] = y ;
    points[2] = x + l_case/2.0; points[3] = y /*+ haut_image - (image->image->haut - 1) * h_case*/ + h_case / 2;

    niveau = (int) (255.0 * (valeurs[ num_case++ ] - z_min) / z_echelle );
    vsf_color(station->handle, niveau);
    v_bar(station->handle, points);

    /* Suite de la première ligne, sauf la dernière case */
    for (i = 1 ; i < image->image->larg - 1 ; i++) {
      points[0] = x + (i * l_case) - l_case / 2;
      points[2] = x + (i * l_case) + l_case / 2;

      niveau = (int) (255.0 * (valeurs[ num_case++ ] - z_min) / z_echelle );
      vsf_color(station->handle, niveau);
      v_bar(station->handle, points);
    }

    /* Dernière case de la première ligne */
    points[0] = x + ( (image->image->larg - 2) * l_case ) + l_case / 2;
    points[2] = x + larg_image;

    niveau = (int) (255.0 * (valeurs[ num_case ] - z_min) / z_echelle );
    vsf_color(station->handle, niveau);
    v_bar(station->handle, points);
#endif
  }

/*   (void) VDI_rappelle_palette(station); */
}

/* -------- Affichage d'une image pour occuper de façon optimale le rectangle indiqué -------- */

int TRC_ImageDansRectangle(Fenetre *fenetre, Image *image, int couleur, int x, int y, unsigned int largeur, unsigned int hauteur)
{
  double echelle, echelle_Z;

  /* Contrôles préliminaires */
  if ( NULL == image ) return -1;
  if ( ( largeur < 5 ) || ( hauteur < 5 ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Région trop petite - affichage abandonné\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    return +1;
  }

  /* Quelle échelle est optimale ? */
  echelle = IMG_dim_trace( image, largeur, hauteur, NULL, NULL );
  if ( echelle <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_dim_trace()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  echelle_Z = image->val_max - image->val_min;
  if ( echelle_Z <= 0.0 ) {
    TRC_Z_Message( fenetre, x, y, largeur, hauteur, "Image sans contraste" );
    return +2;
  }

  /* On prépare les couleurs */
/*   (void) VDI_sauver_palette(station); */
/*   VDI_creer_degrade(station, couleur, None); */

  /* On trace avec l'échelle indiquée, dans le rectangle indiqué */
  TRC_ImageComplete_Maths( fenetre, image, x, y, largeur, hauteur, echelle, image->val_min, image->val_max, echelle_Z, couleur );

  /* On remet la palette originale */
/*   (void) VDI_rappelle_palette(station); */

  /* Dessin terminé, tout s'est bien passé */
  return 0;
}

/* -------- Affichage d'une image, de telle sorte que son centre soit au point indiqué,
   avec l'échelle indiquée

echelle : en pixels/µm, en supposant les X/Y en microns dans l'image
   -------- */

int TRC_ImagePositionCentre(Fenetre *fenetre, Image *image, int couleur, int x_centre, int y_centre, double echelle)
{
  int x_HG, y_HG;
  unsigned int largeur, hauteur;
  double echelle_Z;

  /* Contrôles préliminaires */
  if ( NULL == image ) return -1;
  if ( echelle <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Valeur de l'échelle absurde [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, echelle);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* Dimensions de l'image */
  largeur = (unsigned int) rint( ( image->larg * image->l_reel ) * echelle );
  hauteur = (unsigned int) rint( ( image->haut * image->h_reel ) * echelle );

  /* Coin supérieur gauche de l'image */
  x_HG = x_centre - largeur / 2;
  y_HG = y_centre - hauteur / 2;

  /* Échelle de couleurs */
  echelle_Z = image->val_max - image->val_min;
  if ( echelle_Z <= 0.0 ) {
    TRC_Z_Message( fenetre, x_HG, y_HG, largeur, hauteur, "Image sans contraste" );
    return +2;
  }

  switch ( image->genre )
  {
  case IMG_GENRE_REELLE:
  case IMG_GENRE_GRAPHE:
    /* On prépare les couleurs */
    TRC_CreerDegrade( fenetre, couleur );

    /* On trace avec l'échelle indiquée, dans le rectangle indiqué */
    TRC_ImageComplete_Maths( fenetre, image, x_HG, y_HG, largeur, hauteur, echelle, image->val_min, image->val_max, echelle_Z, couleur );

    /* On remet la palette originale */

    break;

  case IMG_GENRE_COMPO:
    TRC_ImageComplete_Compo( fenetre, image, x_HG, y_HG, largeur, hauteur, echelle );
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type d'image non géré --- %d\n",
		   __FILE__, __FUNCTION__, __LINE__, image->genre);
    return -1;
  }

  /* Dessin terminé, tout s'est bien passé */
  return 0;
}

/* -------- Tracé de l'image proprement dit, l'échelle étant supposée connue
   Image tracée en convention mathématiques (y croissants vers le haut)
   On suppose que l'image tient entièrement dans le rectangle, que les couleurs ont été définies,...

   Affichage de l'image, pavé par pavé
   Idéalement, on devrait mettre un point de la couleur pour chaque (x, y) connu.
   On relie par l'équivalent des droites : des rectangles de couleur.

   On a donc le schéma suivant :  +-----+-----+
                                  |  X..|..X  |  X = valeur connue
                                  +-----+-----+  + = limite de pavé affiché
                                  |  X..|..X  |
                                  +-----+-----+
   ==> les valeurs connues sont au centre des cases
   -------- */

/* -------- Cas d'une image de compositions -------- */

static int TRC_ImageComplete_Compo(Fenetre *fenetre, Image *image, int x, int y, unsigned int larg, unsigned int haut, double echelle)
{
  double l_case, h_case;
  struct matiere **compositions, **valeurs;
  unsigned int nbr_compositions;
  unsigned int i_ligne, i_colonne;
  int points[ 4 ] = { 0, 0, 0, 0 }, couleur;

  /* Le nombre de compositions est dans la borne haute */
  nbr_compositions = (unsigned int) ( image->val_max );
  if ( 0 == nbr_compositions ) {
    return 0; /* L'image n'est qu'un trou : toute blanche ! */
  }
  if ( nbr_compositions >= 16 ) {
    OPTION_A_FAIRE( "Mémoriser le % dans la matière pour avoir des niveaux de gris vraiment pertinents" );
    TRC_CreerDegrade( fenetre, BI_DEGRADE_GRIS );
  }

  /* On réserve pour la liste des compositions (trou exclus) */
  compositions = (struct matiere **) calloc( nbr_compositions, sizeof( struct matiere * ) );
  if ( NULL == compositions ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de réserver la table des compositions\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
  for ( i_ligne = 0 ; i_ligne < nbr_compositions ; i_ligne++ ) {
    compositions[ i_ligne ] = NULL;
  }

  /* Compte-tenu de l'échelle en X, dimensions d'une case */
  l_case = image->l_reel * echelle;
  h_case = image->h_reel * echelle;

  /* On trace par lignes : reprend l'agencement des données en mémoire... */
  points[ 3 ] = y + haut - 1;
  valeurs = image->valeurs.M;
  for ( i_ligne = 0 ; i_ligne < image->haut - 1 ; i_ligne++, valeurs++ ) {
    points[ 1 ] = points[ 3 ]; /* On repart de là où on était */
    points[ 3 ] = (int) rint( ( y + haut - 1 ) - (i_ligne + 1) * h_case );

    points[ 2 ] = x;
    for ( i_colonne = 0 ; i_colonne < image->larg - 1 ; i_colonne++, valeurs++ ) {
      points[ 0 ] = points[ 2 ];
      points[ 2 ] = (int) rint( x + (i_colonne + 1) * l_case );

      couleur = TRC_CouleurCompo( *valeurs, compositions, nbr_compositions );
      if ( nbr_compositions >= 16 ) {
	TRC_Degrade( BI_DEGRADE_GRIS, couleur - 1);
      } else {
	TRC_Couleur( couleur );
      }

      TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
    }

    /* Dernière colonne à part pour éviter des erreurs d'arrondi */
    points[ 0 ] = points[ 2 ];
    points[ 2 ] = x + larg - 1;

    couleur = TRC_CouleurCompo( *valeurs, compositions, nbr_compositions );
    if ( nbr_compositions >= 16 ) {
      TRC_Degrade( BI_DEGRADE_GRIS, couleur - 1);
    } else {
      TRC_Couleur( couleur );
    }

    TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
  }

  /* Dernière ligne à part pour éviter des erreurs d'arrondi */
  points[ 1 ] = points[ 3 ];
  points[ 3 ] = y;

  points[ 2 ] = x;
  for ( i_colonne = 0 ; i_colonne < image->larg - 1 ; i_colonne++, valeurs++) {
    points[ 0 ] = points[ 2 ];
    points[ 2 ] = (int) rint( x + (i_colonne + 1) * l_case );
    
    couleur = TRC_CouleurCompo( *valeurs, compositions, nbr_compositions );
    if ( nbr_compositions >= 16 ) {
      TRC_Degrade( BI_DEGRADE_GRIS, couleur - 1);
    } else {
      TRC_Couleur( couleur );
    }

    TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
  }

  /* Dernière colonne à part pour éviter des erreurs d'arrondi */
  points[ 0 ] = points[ 2 ];
  points[ 2 ] = x + larg - 1;

  couleur = TRC_CouleurCompo( *valeurs, compositions, nbr_compositions );
  if ( nbr_compositions >= 16 ) {
    TRC_Degrade( BI_DEGRADE_GRIS, couleur - 1);
  } else {
    TRC_Couleur( couleur );
  }
  TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );

  free( compositions ); compositions = NULL;
  return 0;
}

/* -------- Cas d'une image à valeurs réelles -------- */

static int TRC_ImageComplete_Maths(Fenetre *fenetre, Image *image, int x, int y, unsigned int larg, unsigned int haut, double echelle,
				   double z_min, double z_max, double echelle_Z, int couleur_base)
{
  int niveau;
  int points[ 4 ] = { 0, 0, 0, 0 };
  unsigned int i, j, largeur_case, hauteur_case;
  double l_case, h_case, z;
  double *valeurs = NULL;

  /* Compte-tenu de l'échelle en X, dimensions d'une case */
  l_case = image->l_reel * echelle; largeur_case = (unsigned int) rint( l_case );
  h_case = image->h_reel * echelle; hauteur_case = (unsigned int) rint( h_case );

  if ( ( largeur_case < 2 ) || ( hauteur_case < 2 ) ) {
    OPTION_A_FAIRE( "Points de moins d'un pixel !" );
    (void) fprintf(stderr, "%s, %s l.%u :  [largeur = %g arrondi à %u, hauteur = %g arrondi à %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, l_case, largeur_case, h_case, hauteur_case );
    return -10;
  }

  /* On trace par lignes : reprend l'agencement des données en mémoire... */
  points[ 3 ] = y + haut - 1;
  if ( IMG_GENRE_REELLE == image->genre ) {
    valeurs = image->valeurs.R;
  } else if ( IMG_GENRE_GRAPHE == image->genre ) {
    if ( NULL == image->infos_affichage ) {
      return -5;
    }
    if ( NULL == image->infos_affichage->moyennes ) {
      return -6;
    }
    valeurs = image->infos_affichage->moyennes;
  }
  if ( NULL == valeurs ) {
    /* Pas de table de valeurs dans l'image ! */
    AVERTISSEMENT( "L'image indiquée n'a pas de table de valeurs ! [%s]", image->nom );
    return -20;
  }
  for ( j = 0 ; j < image->haut - 1 ; j++, valeurs++ ) {
    points[ 1 ] = points[ 3 ]; /* On repart de là où on était */
    points[ 3 ] = (int) rint( ( y + haut - 1 ) - (j + 1) * h_case );

    points[ 2 ] = x;
    for ( i = 0 ; i < image->larg - 1 ; i++, valeurs++ ) {
      points[ 0 ] = points[ 2 ];
      points[ 2 ] = (int) rint( x + (i + 1) * l_case );

      z = *valeurs;
      if ( ( z >= z_min ) && ( z <= z_max ) ) {
	niveau = (int) rint( 255.0 * (z - z_min) / echelle_Z );
	TRC_Degrade( couleur_base, niveau );

	TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
      }
    }

    /* Dernière colonne à part pour éviter des erreurs d'arrondi */
    points[ 0 ] = points[ 2 ];
    points[ 2 ] = x + larg - 1;

    z = *valeurs;
    if ( ( z >= z_min ) && ( z <= z_max ) ) {
      niveau = (int) rint( 255.0 * (z - z_min) / echelle_Z );
      TRC_Degrade( couleur_base, niveau );
      
      TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
    }
  }

  /* Dernière ligne à part pour éviter des erreurs d'arrondi */
  points[ 1 ] = points[ 3 ];
  points[ 3 ] = y;

  points[ 2 ] = x;
  for ( i = 0 ; i < image->larg - 1 ; i++, valeurs++ ) {
    points[ 0 ] = points[ 2 ];	/* On repart de là où on était pour compenser les erreurs d'arrondi en entier */
    points[ 2 ] = (int) rint( x + (i + 1) * l_case );
    
    z = *valeurs;
    if ( ( z >= z_min ) && ( z <= z_max ) ) {
      niveau = (int) rint( 255.0 * (z - z_min) / echelle_Z );
      TRC_Degrade( couleur_base, niveau );
      
      TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
    }
  }

  /* Dernière colonne à part pour éviter des erreurs d'arrondi */
  points[ 0 ] = points[ 2 ];
  points[ 2 ] = x + larg - 1;

  z = *valeurs;
  if ( ( z >= z_min ) && ( z <= z_max ) ) {
    niveau = (int) rint( 255.0 * (z - z_min) / echelle_Z );
    TRC_Degrade( couleur_base, niveau );
      
    TRC_RectanglePlein( fenetre, points[ 0 ], points[ 3 ], points[ 2 ] - points[ 0 ] + 1, points[ 1 ] - points[ 3 ] + 1 );
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -~-~-~-~-~-~-~-~ Affichage de l'échelle des couleurs -~-~-~-~-~-~-~-~ */

int TRC_EchelleImage(Fenetre *fenetre, DES_image *des_image, int x, int y, unsigned int larg, unsigned int haut)
{
  int retour;

  if ( NULL == des_image->image ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'image !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

#ifdef _DEVELOPPE_
  TRC_Couleur( VERT_FONCE );
  TRC_Rectangle( fenetre, x, y, larg, haut );
#endif

  if ( des_image->image->val_min == des_image->image->val_max ) {
    TRC_Z_Message( fenetre, x, y, larg, haut, "Image sans contraste" );
    return +1;
  }

  switch ( des_image->image->genre )
  {
  case IMG_GENRE_REELLE:
  case IMG_GENRE_GRAPHE:	/* Comme reel... */
    retour = TRC_EchelleReelle( fenetre, des_image, x, y, larg, haut );
    break;

  case IMG_GENRE_COMPO:
    retour = TRC_EchelleComposition( fenetre, des_image->image, x, y, larg, haut );
    break;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type d'image inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, des_image->image->genre);
    retour = -48;
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le tracé de la légende [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  return retour;
}

/* -------- Échelle pour une image de type composition ------ */

/* Utilitaire : numéro de couleur associée (avec mise à jour de la table) */

static int TRC_CouleurCompo(struct matiere *composition, struct matiere **table, unsigned long nbr_compositions)
{
  unsigned long i;

  if ( NULL == composition ) return BLANC; /* Trous en blancs */
  if ( BLANC == composition->couleur ) {
    /* On cherche dans la table */
    for ( i = 0 ; i < nbr_compositions ; i++ ) {
      if ( table[ i ] == composition ) {
	/* Composition trouvée */
	return i + 1;
      }
      if ( NULL == table[ i ] ) break;/* Position vide */
    }
    if ( i < nbr_compositions ) { /* On mémorise, pour la prochaine */
      table[ i ] = composition;
    }

    composition->couleur = i + 1;
  }


  return composition->couleur;
}

static int TRC_EchelleComposition(Fenetre *fenetre, Image *image, int x, int y, unsigned int larg, unsigned int haut)
{
  unsigned int nbr_lignes, haut_ligne, nbr_colonnes, larg_colonne;
  unsigned int idx_composition, nbr_compositions, idx_ligne, idx_colonne = 0;
  unsigned long i;
  struct matiere **compositions, **valeurs;
  int attributs_texte[10];
  int haut_caractere;
  int x_case, y_case;

  /* Le nombre de compositions est dans la borne haute */
  nbr_compositions = (unsigned int) ( image->val_max );
  if ( nbr_compositions >= 16 ) {
    TRC_CreerDegrade( fenetre, BI_DEGRADE_GRIS );
  }

  /* Hauteur d'une ligne */
  /* On demande la hauteur et la largeur d'un caractère... */
  vqt_attributes( _PRG_station->handle, attributs_texte);
  haut_caractere = attributs_texte[7];

  haut_ligne = haut_caractere; /* Espace minimal entre deux textes : 3 pixels */

  /* Un pavé fait 8*8, +2 pixels d'espace = 10 ; 2 pixels autour... */
  nbr_lignes = (haut - 4) / haut_ligne;

  /* Il faudra donc combien de colonnes pour tout afficher ? */
  nbr_colonnes = 1 + nbr_compositions / nbr_lignes;

  /* Largeur d'une colonne */
  larg_colonne = larg / nbr_colonnes;
  if ( larg_colonne <= 10 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas assez de place pour tout afficher\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Position des cases [coin BG ; largeur = hauteur = haut_ligne] */
  x_case = x + 2;
  y_case = y + 2 + haut_ligne;

  /* Chaque case est entourée d'un bord noir */
  TRC_Couleur( BLANC );

  /* La légende pour les trous est toujours présente, c'est le blanc... */
  TRC_RectanglePlein( fenetre, x_case, y_case, haut_ligne, haut_ligne );
  TRC_Couleur( NOIR );
  TRC_Texte( fenetre, x_case + haut_ligne + 2, y_case, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, "Trou" );

  /* Chaque case est entourée d'un bord noir */
  TRC_Rectangle( fenetre, x_case, y_case, haut_ligne, haut_ligne );

  if ( 0 == nbr_compositions ) return 0; /* Il n'y a que des trous ! */

  /* On réserve pour la liste des compositions (trou exclus) */
  compositions = (struct matiere **) calloc( nbr_compositions, sizeof( struct matiere * ) );
  if ( NULL == compositions ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de réserver la table des compositions\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }
  for ( i = 0 ; i < nbr_compositions ; i++ ) {
    compositions[ i ] = NULL;
  }

  idx_composition = 0;
  idx_ligne = 1;
  /* On parcourt l'image pour trouver les compositions */
  for ( i = 0, valeurs = image->valeurs.M ; i < image->larg * (unsigned long) image->haut ; i++, valeurs++ ) {
    unsigned int i_compo;
    boolean traitee = FAUX;

    if ( NULL == *valeurs ) continue; /* Trous : déjà traités */

    /* La composition a-t-elle déjà été trouvée ? */
    for ( i_compo = 0 ; i_compo < idx_composition ; i_compo++ ) {
      if ( compositions[ i_compo ] == *valeurs ) {
	traitee = VRAI;
	break;
      }
    }

    /* Composition pas encore légendée */
    if ( FAUX == traitee ) {
      compositions[ idx_composition ] = *valeurs; /* On la mémorise pour après */
      if ( idx_ligne >= nbr_lignes ) {
	idx_colonne++;
	idx_ligne = 0;
      }

      /* On trace le carré de la bonne couleur */
      x_case = (x + 2) + larg_colonne * idx_colonne;
      y_case = (y + 2) + (haut_ligne + 3) * idx_ligne + haut_ligne; /* 3 pixels d'écart entre les pavés */

      if ( nbr_compositions >= 16 ) {
	TRC_Degrade( BI_DEGRADE_GRIS, idx_composition );
      } else {
	TRC_Couleur( idx_composition + 1 );
      }
      TRC_RectanglePlein( fenetre, x_case, y_case, haut_ligne, haut_ligne );

      /* Le cadre noir autour */
      TRC_Couleur( NOIR );
      TRC_Rectangle( fenetre, x_case, y_case, haut_ligne, haut_ligne );

      /* texte : nom de la composition */
      TRC_Texte( fenetre, x_case + haut_ligne + 3, y_case, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_HAUT, 0, (*valeurs)->nom );

      /* On passe à la composition suivante */
      idx_composition++;
      if ( idx_composition >= nbr_compositions ) /* On a fait toutes les compositions */
	break;

      /* Donc à la ligne suivante */
      idx_ligne++;
    }
  }

  /* Tout s'est bien passé */
  free( compositions ); compositions = NULL;
  return 0;
}

/* -------- Échelle pour une image à valeurs réelles -------- */

static int TRC_EchelleReelle(Fenetre *fenetre, DES_image *des_image, int x, int y, unsigned int larg, unsigned int haut)
{
  char legende[ 80 ], *txt_legende;
  int x_couleurs, y_couleurs, couleur;
  int points[ 10 ];
  double larg_case;
  unsigned int larg_couleurs, haut_couleurs;
  unsigned int i, nbr_teintes;
  struct aff_image *infos = NULL;

  /* Sur quelle partie dessine-t-on ? */
  x_couleurs = x;       y_couleurs = y;
  larg_couleurs = larg; haut_couleurs = (unsigned int) ( 0.666666 * (double) haut ); /* 2/3 : couleurs, 1/3 : axe, légende */

  /* Préparation de l'échelle des couleurs */
  infos = des_image->image->infos_affichage;
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Attention, image sans infos d'affichage\n",
		   __FILE__, __FUNCTION__, __LINE__);
    couleur = des_image->couleur;
    nbr_teintes = 255;
  } else {
    couleur = infos->couleur;
    if ( IMG_PALETTE == couleur ) {
      OPTION_A_FAIRE( "À FAIRE / palette libre" );
      nbr_teintes = infos->palette.nbr_teintes;
      return -58;
    } else {
      nbr_teintes = 255;
    }
  }

  /* Largeur d'une case de couleur */
  larg_case = larg_couleurs / (double) nbr_teintes;

  /* On installe les couleurs */
  TRC_CreerDegrade( fenetre, couleur );

  /* Tracé */
  points[ 1 ] = y_couleurs;

  points[ 2 ] = x_couleurs;
  for ( i = 0 ; i < nbr_teintes - 1 ; i++ ) {
    points[ 0 ] = points[ 2 ]; 	/* On repart de là où on en était pour compenser les problèmes d'arrondi */
    points[ 2 ] = x_couleurs + (int) rint( larg_case * ( i + 1 ) );

    TRC_Degrade( des_image->couleur, i );
    TRC_RectanglePlein( fenetre, points[ 0 ], points[ 1 ], points[ 2 ] - points[ 1 ] + 1, haut_couleurs );
  }

  /* La dernière à part pour ne pas déborder avec les erreurs d'arrondi ! */
  points[ 0 ] = x_couleurs + (int) ( larg_case * i );
  points[ 2 ] = x_couleurs + larg_couleurs - 1;
  TRC_Degrade( des_image->couleur, nbr_teintes );
  TRC_RectanglePlein( fenetre, points[ 0 ], points[ 1 ], larg_case, haut_couleurs );

  /* On remet la palette originale */
  (void) VDI_rappelle_palette( _PRG_station );

  /* Un rectangle autour : plus joli */
  TRC_Couleur( NOIR );
  TRC_Rectangle( fenetre, x_couleurs, y_couleurs, larg_couleurs, haut_couleurs );

  /* Le tiers bas : un axe, gradué */
  TRC_LigneH( fenetre, x_couleurs, y_couleurs + haut_couleurs + 10, larg_couleurs );
  TRC_gradue( fenetre, x_couleurs, points[ 1 ], points[ 2 ], points[ 1 ],
	      des_image->image->val_min, des_image->image->val_max, GRADUE_DECIMAL, 0, 10, VRAI );

  /* Valeurs extrêmes */
  (void) TXT_CNV_D_vTexte( des_image->image->val_min, -1.0, 2, 4, 4, legende, NULL );
  TRC_Texte( fenetre, x_couleurs, y + haut - 2, BI_TEXTE_FER_A_GAUCHE, BI_TEXTE_FER_EN_BAS, 0, legende );

  (void) TXT_CNV_D_vTexte( des_image->image->val_max, -1.0, 2, 4, 4, legende, NULL );
  TRC_Texte( fenetre, x_couleurs + larg_couleurs - 1, y + haut - 2, BI_TEXTE_FER_A_DROITE, BI_TEXTE_FER_EN_BAS, 0, legende );

  /* En-dessous : la légende, si elle existe */
  txt_legende = TXT_genere_chaine( "%s [%s]", IMG_QuelleLegende( des_image->image, AXE_Z ), IMG_QuelleUnite( des_image->image, AXE_Z ) );
  if ( txt_legende != NULL ) {
    TRC_Texte( fenetre, x + (larg/2), y + haut - 2, BI_TEXTE_CENTRER_H, BI_TEXTE_FER_EN_BAS, 0, txt_legende );
  }
  free( txt_legende ); txt_legende = NULL;

  /* Affichage terminé */
  return 0;
}

/* -------- Calcul des facteurs d'échelles et dimensions de l'image pour la faire tenir dans la fenêtre indiquée
      Renvoie le facteur d'echelle : pixels/µm - 0 en cas d'erreur

  lg_dessin, ht_dessin : dimensions de la région dans laquelle il faut dessiner
  larg_img, haut_img   : contient, à la fin, les dimensions de l'image en pixels
*/

double IMG_dim_trace(Image *image, unsigned int lg_dessin, unsigned int ht_dessin,
		     unsigned int *larg_img, unsigned int *haut_img)
{
  double l_image, h_image; /* Dimensions de l'image en microns */
  double echelle;
  unsigned int larg_image, haut_image;

  if ( image == NULL ) return 0.0;
  if ( (lg_dessin < 20) || (ht_dessin < 20) ) {
    (void) fprintf(stderr, "IMG_dim_trace : Zone de tracé trop petite [%d × %d]\n", lg_dessin, ht_dessin);
    return 0.0;
  }

  /* L'image est composée de image->larg points en largeur. Chaque point occupe une largeur image->l_reel
     Même chose en hauteur
  */
  l_image = image->l_reel * (double) (image->larg); /* Dimensions en microns */
  h_image = image->h_reel * (double) (image->haut); /* Il y a un-demi carré de chaque côté en plus,
						      car les valeurs sont au centre des cases => pas*larg et non pas*(larg-1) */

  if ( (l_image == 0.0) || (h_image == 0.0) ) {
    (void) fprintf(stderr, "IMG_dim_trace : Dimensions de l'image absurdes [%f × %f]\n", l_image, h_image);
    return 0.0;
  }

  if (h_image > l_image) { /* C'est la hauteur qui conditionne : doit tenir dans la fenêtre */
    echelle = ( (double) ht_dessin ) / h_image;
    larg_image = (unsigned int) (echelle * l_image); /* On contrôle l'autre dimension */
    if (larg_image > lg_dessin) { /* Déborde : la largeur limite */
      echelle = ( (double) lg_dessin ) / l_image;
      larg_image = (unsigned int) (echelle * l_image);
    }
    haut_image = (unsigned int) (echelle * h_image);
  } else { /* C'est la largeur qui conditionne */
    echelle = ( (double) lg_dessin ) / l_image;
    haut_image = (unsigned int) (echelle * h_image); /* On contrôle l'autre dimension */
    if (haut_image > ht_dessin ) { /* Déborde : la largeur limite */
      echelle = ( (double) ht_dessin ) / h_image;
      haut_image = (unsigned int) (echelle * h_image); /* On contrôle l'autre dimension */
    }
    larg_image = (unsigned int) (echelle * l_image);
  }

  if (larg_img != NULL) *larg_img = larg_image;
  if (haut_img != NULL) *haut_img = haut_image;

  return echelle;
}

/* -------- Conversion coordonnées image/écran -------- */

int TRC_AbsoluVersEcran(int x_HG, int y_HG, unsigned int larg, unsigned int haut, double echelle, double x_min, double y_max,
			double x_abs, double y_abs, int *x_ecran, int *y_ecran )
{
  double x_ecr, y_ecr;

#ifdef _DEVELOPPE_
  if ( ( NULL == x_ecran ) || ( NULL == y_ecran ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : NULL pour x_ecran/y_ecran\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -5;
  }
#endif
  if ( echelle <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Échelle absurde\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* On passe le point dans le repère local */
  x_abs -= x_min;
  y_abs = y_max - y_abs; /* Attention, axe des Y inversé à l'écran */

  /* On convertit en coordonnées écran : echelle en µm/pixels */
  x_abs /= echelle;
  y_abs /= echelle;

  /* On convertit en coordonnées entières */
  x_ecr = rint( x_abs );
  y_ecr = rint( y_abs );

  *x_ecran = x_HG + (int) x_ecr;
  *y_ecran = y_HG + (int) y_ecr;

  /* Contrôles : est-on dans la zone indiquée ? */
  if ( x_ecr < 0.0 ) return +1;
  if ( y_ecr < 0.0 ) return +2;
  if ( x_ecr > larg ) return +3;
  if ( y_ecr > haut ) return +4;

  /* On est à l'intérieur */
  return 0;
}

/* -------- Conversion coordonnées écran -> coordonnees réelles
   On suppose une orientation mathématique des axes => coin BG correspond à (x_min, y_min)
   echelle = en µ/pixel
   -------- */

int TRC_EcranVersAbsolu(int x_HG, int y_HG, unsigned int larg, unsigned int haut, double echelle, double x_min, double y_max,
			int x_ecran, int y_ecran, double *x_absolu, double *y_absolu)
{
  double x_abs, y_abs;
  int retour = 0;

#ifdef _DEVELOPPE_
  if ( ( NULL == x_absolu ) || ( NULL == y_absolu ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : NULL pour x_absolu/y_absolu\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }
#endif
  if ( echelle <= 0.0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Échelle absurde\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Écran -> repère absolu */
  x_abs = (double) (x_ecran - x_HG);
  y_abs = (double) (y_ecran - y_HG);

  /* Passage dans le repère absolu */
  x_abs *= echelle;
  y_abs *= echelle;

  /* Translation */
  x_abs += x_min;
  y_abs  = y_max - y_abs; /* Axe des Y inversé à l'écran */

  /* Mémorisation */
  *x_absolu = x_abs; *y_absolu = y_abs;

  /* Était-on dans la zone indiquée */
  if ( x_ecran < x_HG ) retour = +1;
  if ( x_ecran >= x_HG + (int) larg ) retour = +2;
  if ( y_ecran < y_HG ) retour = +3;
  if ( y_ecran >= y_HG + (int) haut ) retour = +4;

#ifdef _DEVELOPPE_
  if ( retour > 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : On est hors de la zone indiquée : (%d, %d) L=%u H=%u clic en %d %d\n",
		   __FILE__, __FUNCTION__, __LINE__, x_HG, y_HG, larg, haut, x_ecran, -y_ecran);
  }
#endif

  /* Terminé */
  return retour;
}

/* -------- Conversion coordonnées à l'écran -> coordonnées de l'image --------\
 |   Renvoie FAUX en cas de problèmes, VRAI si tout va bien                    |
 |   Les coordonnées image (dans le repère µmétrique absolu)                   |
 |     sont renvoyées dans x_image, y_image                                    |
 |                                                                             |
 | image                = l'image affichée                                     |
 | lg_dessin, ht_dessin = ses dimensions à l'écran                             |
 | mode                 = affichage avec les Y vers le bas ou le haut ?        |
 | x_ecran, y_ecran     = coordonnées à l'écran (dans le cadre de l'image)     |
 \---------------------------------------------------------------------------- */

boolean IMG_ecran_v_image(struct image *image, unsigned int lg_dessin, unsigned int ht_dessin, int mode,
			  int x_ecran, int y_ecran, double *x_image, double *y_image)
{
  double x_echelle, y_echelle;
  double lg_case, ht_case;
  double x, y;

  if (image == NULL) return FAUX;
  if ( (NULL == x_image) || (NULL == y_image) ) return FAUX;

  if ( (image->l_reel <= 0.0) || (image->h_reel <= 0.0) )
    return FAUX;

  x_echelle = ((double) lg_dessin) / ( image->l_reel * (double) image->larg);
  y_echelle = ((double) ht_dessin) / ( image->h_reel * (double) image->haut);

  lg_case = (double) lg_dessin / (double) image->larg; /* Dimensions d'une case, en pixels */
  ht_case = (double) ht_dessin / (double) image->haut;

  x = (double) x_ecran - lg_case / 2.0;
  if (IMG_TRACE_NORMAL == mode) {
    y = (double) y_ecran - ht_case / 2.0;
  } else {
    y_ecran = ht_dessin - y_ecran; /* On passe dans le repère Y vers le haut */
    y = (double) y_ecran - ht_case / 2.0;
  }

  *x_image = x / x_echelle + image->x_min;
  *y_image = y / y_echelle + image->y_min;

  return VRAI;
}

/* -------- Conversion coordonnées de l'image -> coordonnées à l'écran --------\
 |   Renvoie FAUX en cas de problèmes, VRAI si tout va bien                    |
 |   Les coordonnées écran (par rapport au coin HG du rectangle de l'image)    |
 |     sont renvoyées dans x_ecran, y_ecran                                    |
 |                                                                             |
 | image                = l'image affichée                                     |
 | lg_dessin, ht_dessin = ses dimensions à l'écran                             |
 | mode                 = affichage avec les Y vers le bas ou le haut ?        |
 | x_image, y_image     = coordonnées dans l'image (repère µmétrique)          |
 \---------------------------------------------------------------------------- */

boolean IMG_image_v_ecran(struct image *image, unsigned int lg_dessin, unsigned int ht_dessin, int mode,
			  double x_image, double y_image, int *x_ecran, int *y_ecran)
{
  double x_echelle, y_echelle;
  double lg_case, ht_case;

  if (image == NULL) return FAUX;
  if ( (NULL == x_ecran) || (NULL == y_ecran) ) return FAUX;

  if ( (image->l_reel <= 0.0) || (image->h_reel <= 0.0) )
    return FAUX;

  x_echelle = ((double) lg_dessin) / ( image->l_reel * (double) image->larg);
  y_echelle = ((double) ht_dessin) / ( image->h_reel * (double) image->haut);

  lg_case = (double) lg_dessin / (double) image->larg; /* Dimensions d'une case, en pixels */
  ht_case = (double) ht_dessin / (double) image->haut;

  x_image -= image->x_min; /* Position dans le repère avec (0,0) sur le coin BG de l'image */
  y_image -= image->y_min;

  *x_ecran = (int) rint(x_image * x_echelle);
  *x_ecran += (int) (lg_case / 2.0); /* Position compte-tenu de la taille des cases */
  if ( mode == IMG_TRACE_NORMAL ) {
    *y_ecran = (int) rint(y_image * y_echelle + ht_case / 2.0);
  } else {
    *y_ecran = (int) rint( (double) ht_dessin - y_image * y_echelle - ht_case / 2.0 );
  }

  return VRAI;
}
