/* Analyse d'une formule chimique
   (C) 2002, Emmanuel Curis
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ascii_utils.h"
#include "chimie_utils.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

typedef struct {
  char *symbole;
  char *formule;
} Compo_Nom_Groupe; /* Dictionnaire, groupes utilisateurs */

static Compo_Nom_Groupe *bibliotheque = NULL;
static unsigned long nombre_entrees = 0;

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static char *saute_charge(char *bloc);

static char *cherche_fin_bloc(char *bloc, int code);
static char *lit_nombre      (char *bloc, double *valeur);
static char *lit_indice      (char *bloc, double *indice);

static char *cherche_groupe  (Composition *mere, char *bloc);
static char *cherche_terme_suivant(char *bloc);
static char *trouver_virgule (char *bloc);

static int tri_entrees_bibliotheque(const void *e1, const void *e2);

static Composition *analyse_bloc (Composition *mere, char *chaine);
static Composition *decoupe_blocs(Composition *mere, char *chaine);

static Composition *nv_composition (Composition *mere);
static Composition *fin_composition(Composition *compo, double indice);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

Composition *nv_composition(Composition *mere)
{
  Composition *nv_compo;

  nv_compo = (Composition *) calloc( 1, sizeof(Composition) );
  if (nv_compo == NULL) {
    (void) fprintf(stderr, "%s,%s l.%u : mmoire insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  nv_compo->avant = mere;
  return nv_compo;
}

Composition *fin_composition(Composition *compo, double indice)
{
  Composition *avant;
  unsigned int i;

  if (compo == NULL) {
    (void) fprintf(stderr, "fin_composition : Erreur grave dans la structure : plus de composition en cours !\n");
    return NULL;
  }

  avant = compo->avant;
  if (avant == NULL) {
    (void) fprintf(stderr, "fin_composition : Erreur de syntaxe : trop de parenthses fermantes\n");
    return NULL;
  }

#ifdef _DEVELOPPE_
  printf("fin_composition : On travaille avec l'indice %f\n", indice);
#endif
  for (i = 0 ; i < 120 ; i++) {
    avant->atomes[i] += (float) ( indice * compo->atomes[i] );
  }
  free(compo); /* Plus besoin d'elle */
  return avant;
}

/* cherche la ) ou ] correspondante ... */

static char *cherche_fin_bloc(char *bloc, int code)
{
  char *fin_bloc;
  int niveau = 1; /* On appelle aprs ( ou [ => on est dans le niveau 1 */

  if (bloc == NULL) return NULL;

  if (code == '(') code = ')'; /* Caractre de fin  chercher */
  if (code == '[') code = ']'; 

  for (fin_bloc = bloc ; *fin_bloc != 0 ; fin_bloc++) {
    char c = *fin_bloc;

    if ( (c == '(') || (c == '[') ) {
      niveau++; /* On entre dans un nouveau niveau de parenthsage */
    }
    if ( (c == ')') || (c == ']') ) {
      niveau--;
      if (niveau == 0) {
	if (code != c) {
	  (void) fprintf(stderr, "cherche_fin_bloc: ATTENTION, ( ferm par ] ou rciproquement... [%s]\n", bloc);
	}
	return fin_bloc;
      }
    }
  }

/* On n'arrive ici que si l'on n'a pas trouv de parenthse fermante... */
  (void) fprintf(stderr, "cherche_fin_bloc : ERREUR : %c ouvert(e), mais non ferm(e)...\n", code);
  return NULL;
}

/* Lecture d'un indice aprs un symbole, une parenthse,...
   Renvoie la position dans la chane aprs cet indice (premier caractre aprs)

   Indice entier ou flottant, commence par un chiffre ou .
   Indice fractionnaire autoris, de la forme 3/8 par exemple.
   Indice fini par le premier caractre diffrent rencontr : ^,+,-,lettre,...
*/

static char *lit_indice(char *bloc, double *indice)
{
  char *fin_indice;
  double num = 1.0, den = 1.0;

  if ( indice == NULL) return NULL; 
  if ( bloc == NULL ) {
    *indice = 0.0;
    return NULL;
  }

  if (*bloc == 0) { /* Pas d'indice indiqu => coefficient 1 */
    *indice = 1.0;
#ifdef _DEVELOPPE_
    printf("      Pas d'indice lu\n");
#endif
    return bloc;
  }

  fin_indice = lit_nombre(bloc, &num); /* Le numrateur */
  if ('/' == *fin_indice) {
    fin_indice++; /* On saute le quotient */
    fin_indice = lit_nombre(fin_indice, &den);
    if (den == 0.0) {
      (void) fprintf(stderr, "lit_indice : Coefficient fractionnaire, dnominateur absent ou nul ! [%s]\n", bloc);
      den = 1.0; /* On ignore le dnominateur */
    }
  }
  if (num == 0.0) num = 1.0; /* Coefficient absent quivaut  1 */

  *indice = num / den;
  return fin_indice; /* Premier caractre aprs le coefficient stoechiomtrique */
}

static char *lit_nombre(char *bloc, double *valeur)
{
  char *fin_nombre;
  double nombre = 0.0;

  fin_nombre = bloc;
  while ( isdigit(*fin_nombre) ) {
    nombre *= 10.0;
    nombre += (*fin_nombre - '0');
    fin_nombre++;
  }
  if (*fin_nombre == '.') { /* Partie fractionnaire */
    double facteur = 1.0;

    fin_nombre++;
    while ( isdigit(*fin_nombre) ) {
      facteur /= 10.0;
      nombre += facteur * (*fin_nombre - '0');
      fin_nombre++;
    }
  }

  if (valeur != NULL) *valeur = nombre;
  return fin_nombre; /* Nombre lu */
}

/* Saute la charge, les liaisons, les caractres inconnus */

static char *saute_charge(char *bloc)
{
  char c;

  while (*bloc != 0) {
    c = *bloc;

    if ( isalpha(c) ) /* On commence un nouveau symbole */
      return bloc;

    switch (c)
    {
    case '+': /* Signe de charge */
    case '-': /* Charge ou liaison simple */
    case '=': /* Liaison double */
      break; /* Ignor purement et simplement */
 
    case '^': /* Exposant => charge suit */
      bloc++;
      bloc = lit_nombre(bloc, NULL); /* Charge ignore */
      break;

    case ')':
    case ']':
      (void) fprintf(stderr, "saute_charge : %c fermant(e) sans ouverture associe !\n", c);
      break;
    }
    bloc++;
  }

  return bloc;
}

/* Dcoupe une formule en blocs du type ...(...)...(...)...[...]... */

static Composition *decoupe_blocs(Composition *mere, char *bloc)
{
  char *deb_bloc, *fin_bloc;
  char *pos_par, *pos_virgule;

#ifdef _DEVELOPPE_
  printf("Bloc  dcouper : %s\n", bloc);
#endif

  deb_bloc = bloc;
  do {
    fin_bloc = deb_bloc + strlen( deb_bloc ); /* Premier caractre _aprs_ la fin du bloc */

    pos_par = strchr(bloc, '(');
    if (pos_par != NULL) /* Parenthse ouvrante -> bloc suivant */
      fin_bloc = pos_par;

    pos_par = strchr(bloc, '[');
    if ( (pos_par != NULL) && (pos_par < fin_bloc) )
      fin_bloc = pos_par; /* Crochet ouvrant, idem */

    pos_virgule = trouver_virgule( deb_bloc ); /* , = juxtaposition de formules */
    if ( pos_virgule != NULL ) {
      fin_bloc = pos_virgule; /* Virgule -> bloc suivant, mais faire une moyenne */
    }

    if ( *fin_bloc == 0 ) { /* Bloc sans (, [ ni , ==> bloc simple */
#ifdef _DEVELOPPE_
      printf("-> bloc isol\n");
#endif
      analyse_bloc(mere, deb_bloc); /* Analyse de la partie avant () */
      deb_bloc = fin_bloc;
    } else if ( *fin_bloc == ',' ) { /* Juxtaposition => il faut faire une moyenne ! */
      unsigned int nbr_termes = 2; /* Au moins deux termes, un de chaque ct de la , */
      char *deb_terme;
      Composition *moyenne;

      /* On compte les termes prsents */
      deb_terme = fin_bloc + 1; /* On se place aprs la 1re virgule */
      deb_terme = cherche_terme_suivant(deb_terme);
      while ( deb_terme != NULL ) {
	deb_terme = cherche_terme_suivant(deb_terme);
	nbr_termes++; /* Un terme de plus dans la moyenne */
      }
#ifdef _DEVELOPPE_
      printf("%u termes  moyenner...\n", nbr_termes);
#endif

      moyenne = nv_composition(mere);
      if (moyenne == NULL) { /* Impossible de moyenner => on a fini ce niveau d'analyse... */
	*fin_bloc = 0;
	(void) fprintf(stderr, "decoupe_blocs : Moyenne impossible... On ne garde que le premier [%s]\n", bloc);
	decoupe_blocs(mere, deb_bloc);
      } else {
	unsigned int i;

	deb_terme = cherche_terme_suivant(deb_bloc);
	if (deb_terme != NULL) *(deb_terme - 1) = 0; /* On remplace la , par une fin de bloc */
	for (i = 0 ; i < nbr_termes ; i++) {
	  decoupe_blocs(moyenne, deb_bloc); /* Analyse de ce terme */
	  deb_bloc = deb_terme;
	  deb_terme = cherche_terme_suivant(deb_bloc);
	  if (deb_terme != NULL) *(deb_terme - 1) = 0; /* On remplace la , par une fin de bloc */
	}

	/* Plus qu' faire la moyenne de tous ces blocs */
	fin_composition(moyenne, 1.0 / (double) (nbr_termes) );
      }
      deb_bloc = fin_bloc; /* De toute faon, xxx,yyy,zzz est forcment fini aprs cette analyse, donc on sort */
    } else if ( ( '(' == *fin_bloc ) || ( '[' == *fin_bloc ) ) { /* Bloc du type () ou [] */
      int code = *fin_bloc; /* Quel type de parenthse, () ou [] ? */
      double indice;
      Composition *parenthese;

      *fin_bloc = 0; /* On isole la partie avant () */
#ifdef _DEVELOPPE_
      printf(" Partie avant parenthses : %s\n", deb_bloc);
#endif
      analyse_bloc(mere, deb_bloc); /* Analyse de la partie avant () */

      /* On cherche la parenthse / le crochet fermant correspondant */
      deb_bloc = fin_bloc + 1; /* Premier caractre suivant la parenthse */
      fin_bloc = cherche_fin_bloc(deb_bloc, code);
      if (fin_bloc == NULL) {
	return NULL;
      }
      *fin_bloc = 0; /* On remplace la fermante par 0 pour avoir bien un bloc autonome */
      fin_bloc++; /* Premier caractre aprs la parenthse fermante */

      /* On analyse le bloc de la parenthse */
      parenthese = nv_composition(mere); /* Composition du bloc isol */
      if (parenthese == NULL) {
	return NULL;
      }
      decoupe_blocs(parenthese, deb_bloc);

      /* On cherche l'indice ventuel */
      fin_bloc = lit_indice(fin_bloc, &indice);
      fin_composition(parenthese, indice);

      /* On ignore ventuellement des trucs inutiles : charge par exemple */
      fin_bloc = saute_charge(fin_bloc);

      /* On est donc au dbut du bloc suivant */
      deb_bloc = fin_bloc;
    }
  } while ( *deb_bloc != 0 ); /* On n'est pas sur le dernier bloc */

  /* Si on est l, c'est que l'on a fini */
  return mere;
}

/* Cherche s'il existe des virgules au niveau "0" : distinction des cas Cu(Fe,Mg)2 et Cu(H2O),3NO2 */

static char *trouver_virgule (char *bloc)
{
  int niveau = 0;
  char *c = bloc;

  while ( *c != 0 ) {
    if ( ( '(' == *c ) || ( '[' == *c ) ) niveau++;
    if ( ( ']' == *c ) || ( ')' == *c ) ) {
      niveau--;
#ifdef _DEVELOPPE_
      if ( niveau < 0 ) {
	(void) fprintf(stderr, "trouver_virgule : plus de parenthses fermantes que d'ouvrantes [\"%s\"]\n", bloc);
      }
#endif
    }

    if ( ',' == *c ) {
      if ( 0 == niveau ) return c;
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "trouver_virgule : virgule dans des parenthses [\"%s\"]\n", bloc);
#endif
    }

    c++;
  }

  /* Si on est l : aucune virgule ne convenait */
  return NULL;
}

/* -------- Analyser un bloc pour y trouver les symboles et autres indices... -------- */

static Composition *analyse_bloc(Composition *mere, char *bloc)
{
  char *fin_symbole;

  if ( (mere == NULL) || (bloc == NULL) ) return NULL;

  if (*bloc == 0) {
    (void) fprintf(stderr, "analyse_bloc : Bloc vide !\n");
    return NULL;
  }

#ifdef _DEVELOPPE_
  printf(" bloc  analyser : %s\n", bloc);
#endif

  do {
    /* On cherche d'abord dans la table externe, utilisateur */
    fin_symbole = cherche_groupe(mere, bloc);
    if (fin_symbole == NULL) { /* Si on a chou, on cherche un symbole atomique */
      double indice;
      int Z;

      fin_symbole = bloc; /* Premier caractre du symbole */
      fin_symbole++; /* Premier caractre toujours suppos valable, minuscule */
      while ( (*fin_symbole >= 'a') && (*fin_symbole <= 'z') ) {
	fin_symbole++;
      }
#ifdef _DEVELOPPE_
      printf("   bloc : %s, sans symbole reste : %s [%d]\n", bloc, fin_symbole, *fin_symbole);
#endif

      Z = ATM_num_atomique( bloc );
      fin_symbole = lit_indice(fin_symbole, &indice);
      fin_symbole = saute_charge(fin_symbole);

      if ( (Z > 0) && (Z <= 120) ) {
	mere->atomes[Z - 1] += indice;
      } else {
	(void) fprintf(stderr, "Symbole inconnu : %s [indice : %f]\n", bloc, indice);
      }
    }
    bloc = fin_symbole; /* cherche_groupe se charge de lire l'indice et les charges
			   dans tous les cas, fin_symbole pointe donc sur le 1er caractre
			   du symbole suivant */

  } while (*bloc != 0);

  return mere;
}

/* ~~~~~~~~ Gestion des virgules = moyennes de termes ~~~~~~~~~~~~~~~~ */

/* -------- Renvoie la position juste aprs la virgule -------- */

static char *cherche_terme_suivant(char *bloc)
{
  char *c;
  if ( (bloc == NULL) || (*bloc == 0) ) return NULL;

#ifdef _DEVELOPPE_
  printf("Reste... %s\n", bloc);
#endif

  c = bloc;
  do {
    if ( (*c == '(' ) || (*c == '[') ) {
      c = cherche_fin_bloc(c + 1, *c); /* On passe  la parenthse fermante correspondante */
    }

    if ( *c == ',' ) { /* Terme suivant trouv */
      return c + 1;
    }
    c++;
  } while (*c != 0);

  return NULL; /* Pas de terme suivant trouv */
}

/* ~~~~~~~~ Gestion de la bibliothque de symboles de groupes ~~~~~~~~ */

/* ------ Groupes utilisateurs ------ */

static char *cherche_groupe(Composition *mere, char *bloc)
{
  size_t lg_bloc;
  unsigned long i;

  if ( (bloc == NULL) || (*bloc == 0) ) return NULL; /* Pas de groupe */
  if (bibliotheque == NULL) return NULL; /* Pas de dictionnaire connu */

  lg_bloc = strlen(bloc); /* Longueur du bloc = longueur maximale du symbole... */

  for (i = 0 ; i < nombre_entrees ; i++) {
    size_t lg_symbole;

    lg_symbole = strlen( bibliotheque[i].symbole );
    if (lg_symbole > lg_bloc) continue; /* Symbole plus long que la formule */

    if (strncmp( bibliotheque[i].symbole, bloc, lg_symbole) == 0) { /* On a trouv le symbole */
      Composition *groupe;
      double indice;

      bloc += lg_symbole; /* Premier caractre aprs le symbole reconnu */
#ifdef _DEVELOPPE_
      printf("Symbole reconnu : %s = %s\n", bibliotheque[i].symbole, bibliotheque[i].formule);
#endif

      groupe = nv_composition(mere);
      if (groupe != NULL) { /* On peut traiter le groupe */
	decoupe_blocs(groupe, bibliotheque[i].formule); /* Analyse de la formule du groupe considr */
      }
      bloc = lit_indice(bloc, &indice); /* Lecture de l'indice associ */
      if (groupe != NULL)
	fin_composition(groupe, indice);
      bloc = saute_charge(bloc);

      return bloc; /* Dbut du groupe suivant */
    }
  }

  return NULL; /* Pas de groupe trouv dans le dictionnaire */
}

static void libere_bibliotheque(void)
{
  unsigned long i;

  for (i = 0 ; i < nombre_entrees ; i++) {
    free(bibliotheque[i].symbole);
    free(bibliotheque[i].formule);
  }

  free(bibliotheque);
  bibliotheque = NULL;
  nombre_entrees = 0UL;
}

static int tri_entrees_bibliotheque(const void *e1, const void *e2)
{
  const Compo_Nom_Groupe *groupe1, *groupe2;

  groupe1 = (const Compo_Nom_Groupe *) e1;
  groupe2 = (const Compo_Nom_Groupe *) e2;
  if ( strlen( groupe1->symbole ) > strlen( groupe2->symbole ) ) /* Symbole plus long : plac avant */
    return -1;
  else
    return +1;
  /* L'ordre des lments de mme longueur n'a aucune importance... */
}

int Chimie_ChargeBibliotheque(const char *nom_fichier)
{
  FILE *biblio;
  unsigned long i = 0;
  char *txt;

  if (bibliotheque != NULL ) {
    libere_bibliotheque();
  }

  if (nom_fichier == NULL) return FAUX;
  biblio = fopen(nom_fichier, "r");
  if (biblio == NULL) {
    (void) fprintf(stderr, "Chimie_ChargeBibliotheque : Impossible d'ouvrir le fichier %s\n", nom_fichier);
    return FAUX;
  }

  /* Compter le nombre de lignes du fichier == nombre de groupes */
  nombre_entrees = compte_lignes(biblio, '#');
  if (nombre_entrees == 0) {
    (void) fprintf(stderr, "Chimie_ChargeBibliotheque : Fichier sans groupe indiqu !\n");
    (void) fclose(biblio);
    return FAUX;
  }

  bibliotheque = (Compo_Nom_Groupe *) calloc(nombre_entrees, sizeof(Compo_Nom_Groupe) );
  if (NULL == bibliotheque) {
    (void) fprintf(stderr, "Chimie_ChargeBibliotheque : Mmoire insuffisante pour charger la bibliothque\n");
    nombre_entrees = 0;
    (void) fclose(biblio);
    return FAUX;
  }

  rewind(biblio); /* On se remet au dbut */
  do {
    txt = lit_commande(biblio);
    if (txt == NULL) {
      break; /* Fin de fichier ? */
    }
    if (*txt == '#') { /* Commentaire */
      int c;

      free(txt);
      c = ignore_ligne(biblio);
      if (EOF == c) break;
      while ('#' == c) {
	c = ignore_ligne(biblio);
      }
    } else { /* C'est un groupe, on vient de lire le symbole */
      int c;

      bibliotheque[i].symbole = txt;
      bibliotheque[i].formule = lit_commande(biblio);
      i++;
      c = ignore_ligne(biblio);
      if (EOF == c) break;
      while ('#' == c) {
	c = ignore_ligne(biblio);
      }      
    }
  } while (i < nombre_entrees);

  (void) fclose( biblio );

  /* La table est lue, il faut maintenant la trier par longueur de symbole dcroissante */
  qsort( bibliotheque, nombre_entrees, sizeof(Compo_Nom_Groupe), tri_entrees_bibliotheque);

#ifdef _DEVELOPPE_
  /* Affichage de la table de groupes */
  for (i = 0; i < nombre_entrees ; i++) {
    printf("Symbole %s ==> formule %s\n", bibliotheque[i].symbole, bibliotheque[i].formule);
  }
#endif

  return VRAI; /* Chargement russi */
}

/* -------- Interface avec l'extrieur -------- */

Composition *Chimie_AnalyseFormule(const char *chaine)
{
  Composition *resultat;
  char *formule;
  int i;

  if ( NULL == chaine ) return NULL;
  formule = strdup( chaine ); /* Car on y mettra des 0,  la place des ()
			       En fait, il faudrait en profiter pour faire sauter les espaces... */
  if ( NULL == formule ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de prparer la chane de travail !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  resultat = nv_composition(NULL);
  if ( NULL == resultat ) {
    free( formule ); formule = NULL;
    return NULL;
  }

  decoupe_blocs( resultat, formule );

#ifdef _DEVELOPPE_
  /* Affichage du rsultat */
  for (i = 0 ; i < URANIUM ; i++) {
    if (resultat->atomes[i] > 0.0) {
      printf("  %s ==> %f\n", ATM_Z_vers_symbole(i + 1), resultat->atomes[i]);
    }
  }
#endif

  return resultat;
}

#ifdef FORM_CHIMIE_
int main(int n, char *args[]) /* ~~~~~~~~ Pour essais seulement ~~~~~~~~ */
{
  Composition *resultat = NULL;
  unsigned int i;

  if ( n < 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucun argument indiqu ! Donnez la formule  analyser en argument !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  Chimie_ChargeBibliotheque("tst.bib");

  printf( "\n--- Analyse de la formule %s ---\n", args[ 1 ] );

  resultat = Chimie_AnalyseFormule( args[ 1 ] );

  printf( "\n --- Rsultat de l'analyse :\n" );

  for ( i = 0 ; i < URANIUM ; i++ ) {
    if ( resultat->atomes[ i ] > 0.0 ) {
      printf( "  %s ==> %f\n", ATM_Z_vers_symbole( i + 1 ), resultat->atomes[ i ] );
    }
  }

  return 0;
}
#endif
