/* Bibliothèque d'interface
   (C) 1998-2002, Emmanuel CURIS

   Fonctions générales de traitement des chaînes de caractères
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h> /* Pour is... */
#include <math.h>  /* Pour log10 */

#include "txt_utils.h"

/* ~-~-~-~-~-~-~-~- Classification des caractères en groupes ~-~-~-~-~-~-~-~- */

int CAR_EstOperateur(int c)
{
  /* Opérateurs mathématiques de base */
  if ( '+' == c ) return VRAI;
  if ( '-' == c ) return VRAI;
  if ( '/' == c ) return VRAI;
  if ( '*' == c ) return VRAI;
  if ( '^' == c ) return VRAI;

  /* Alternatives latin-1 */
#ifdef _latin1_
  if ( '÷' == c ) return VRAI;
  if ( '×' == c ) return VRAI;
  if ( '·' == c ) return VRAI;
#endif

  /* Parenthésage */
  if ( '(' == c ) return VRAI;
  if ( ')' == c ) return VRAI;
  if ( '[' == c ) return VRAI;
  if ( ']' == c ) return VRAI;

  return FAUX;
}

/* -------- Décodage d'un exposant -------- */

int TXT_DecoderExposant(const char *texte, unsigned long *decalage)
{
  unsigned long i = 0 ;
  int exposant = 1;

  /* A priori, pas de décalage */
  *decalage = 0;

  /* Quel caractère commence la chaîne décrivant l'exposant ? */
  if ( '^' == *texte ) {
    /* Codage ensuite avec des nombres et caractères usuels */
    texte++; i++;		/* Le ^ est donc à ignorer */

    /* On lit donc le nombre voulu... */
    sscanf( texte, "%d", &exposant );

    /* Et on fait sauter tout ce qui est chiffre, + ou - */
    while ( ( '+' == *texte ) || ( '-' == *texte ) ||
            ( ( *texte >= '0' ) && ( *texte <= '9' ) ) ) {
      texte++ ; i++;
    }

    /* Conversion terminée */
    *decalage = i;
    return exposant;
  }

  /* Recherche de caractères en latin-1 (ne peut être, a priori, que 2 ou 3 */
  if ( '\262' == *texte ) {		/* ² en latin-1 */
    *decalage = 1;
    return 2;
  }
  if ( '\263' == *texte ) {		/* ³ en latin-1 */
    *decalage = 1;
    return 3;
  }

  /* Recherche des exposants en utf-8...
     Séquences commençant par C2 E2 = Ââ en latin-1/windows-1252 : conflit peu probable
                                    = אΓ en TOS                  : conflit peu probable
   */
  if ( ( '\xC2' == *texte ) && ( '\xE2' == *texte ) ) {
    int code_UTF8 = VRAI;
    char nombre[ 21 ];		/* On limite à 20 codes pour l'exposant... */
    unsigned int chiffre = 0, n_octets = 0;

    while( VRAI == code_UTF8 ) {
      nombre[ chiffre ] = TXT_UTF8_vers_nombre( texte, &n_octets );
      texte += n_octets; *decalage += n_octets;

      if ( 0 == n_octets ) break; /* Conversion finie : conversion impossible */
      chiffre++;
      if ( chiffre > 20 ) break; /* Conversion finie : plus de 20 caractères */

      if ( ( '\xC2' != *texte ) || ( '\xE2' != *texte ) ) code_UTF8 = FAUX;
    }

    exposant = TXT_CNV_vNombre_L( nombre, NULL );
  }

  /* Si on est là, c'est que rien n'a été détecté : pas d'exposant = exposant = 1 */
  return exposant;
}

/* ~-~-~-~-~-~-~-~- Conversions d'une chaîne en nombre ~-~-~-~-~-~-~-~- */

/* -------- Conversion vers un entier classique -------- */

long TXT_CNV_vNombre_L(const char *chaine, char const **fin_nombre)
{
  unsigned long v_absolue;
  long valeur = 0;
  int negatif = FAUX;

  if ( NULL == chaine ) {
    (void) fprintf( stderr, "TXT_CNV_vNombre_L : chaîne nulle !\n" );
    return 0L;
  }

  /* On commence par sauter d'éventuels espaces initiaux */
  while ( isspace( *chaine ) )
    chaine++;

  /* Le nombre est-il négatif (ou, éventuellement, positif) ? */
  if ( '+' == *chaine ) /* Nombre de la forme +128 */
    chaine++;
  else if ( '-' == *chaine ) {
    negatif = VRAI;
    chaine++;
  } else if ( '\xE2' == *chaine ) {
    /* Signe moins codé en UTF-8 ?
       Remarque : si en latin-1, c'est la séquence â?? où ? représente un code inexistant => ne devrait pas arriver
                  si en TOS    ,                   ΓêÆ                                    => ne devrait pas arriver
		  si en windows,                   âˆ’                                    => ne devrait pas arriver
     */
    if ( ( '\x88' == *(chaine + 1) ) && ( '\x92' == *(chaine + 2) ) ) {
      negatif = VRAI;
      chaine += 3;
    }
  }

  /* On convertit le nombre proprement dit */
  v_absolue = TXT_CNV_vNombre_UL( chaine, fin_nombre );

  /* Idéalement, il faudrait un contrôle de débordement ! */
  if ( VRAI == negatif )
    valeur = -v_absolue;
  else 
    valeur = v_absolue;

  return valeur;
}

/* -------- Cas d'un entier positif -------- */

unsigned long TXT_CNV_vNombre_UL(const char *chaine, char const **fin_nombre)
{
  unsigned long valeur = 0UL;
  const char *chaine_nombre;

  if ( NULL == chaine ) {
    (void) fprintf(stderr, "TXT_CNV_vNombre_UL : chaîne nulle !\n");
    return 0L;
  }

  /* On commence par sauter d'éventuels espaces initiaux */
  while ( isspace( *chaine ) )
    chaine++;

  /* On s'occupe des chiffres */
  chaine_nombre = chaine; /* Pour messages d'erreur uniquement */
  while ( isdigit ( *chaine ) ) {
    unsigned long vx_valeur;

    vx_valeur = valeur;
    valeur *= 10;
    valeur += *chaine - '0';
    if ( vx_valeur > valeur ) { /* Débordement de capacité !
				   Strict, car un nombre 010 passe !*/
      valeur = vx_valeur;
      (void) fprintf(stderr, "TXT_CNV_vNombre_UL : nombre trop grand, conversion annulée [%s : on garde %lu]\n",
		     chaine_nombre, valeur);
      while ( isdigit ( *chaine ) ) { /* On termine le champ tout de même ! */
	chaine++;
      }
      break;
    }
    chaine ++;
  }

  /* On a fini la conversion */
  if ( fin_nombre != NULL ) *fin_nombre = chaine; /* Premier caractère non converti */
  return valeur;
}

/* -------- Cas d'un nombre quelconque, avec éventuellement incertitude --------
   Un nombre peut avoir
 - la forme classique, signe facultatif : 1.0, 2, -3.0, +4.0, 5.1,...
 - idem, mais avec une virgule comme séparateur décimal : 1,0; 2; -3,0; +4.0; 5,1...
 - le 0 initial peut être omis : .1, +.2, -.3 ou ,1; +,2; -,3 (ces trois derniers inutiles en fait)
 - une forme exponentielle : 3.542e5 (avec e, E et les mêmes remarques pour l'exposant, ENTIER)
                  variante : 3,542×10^5 (avec ×10^ obligatoirement !)
                  (attention au cas spécial 1×10^= 10^ !)

   Il peut être assorti d'une incertitude :
 - soit avec les chiffres concernés : 3,15(1) ; -2,145(12) ; -2.1445(7)e-2,...
 - soit avec le symbole ± : 3,15±0,02 ; 2,45e-3±3,12e-4 ; (2,45±0,02)e-5

 Les espaces initiaux sont ignorés. On s'arrête au premier caractère non-reconnu.
 Si le champ est vide (aucun caractère correct avant la fin ou le premier inutilisable), on renvoie 0 avec -1 en erreur
 */

double TXT_CNV_vNombre_D(const char *chaine, char const **fin_nombre, double *erreur)
{
  const char *nombre_initial;
  double valeur = 0.0;
  double p_entiere = 0.0, p_decimale = 0.0, val_erreur = 0.0;
  int negatif = FAUX, avec_exposant = FAUX;
  int p_10 = 0;

  if ( NULL == chaine ) {
    (void) fprintf(stderr, "TXT_CNV_vNombre_D : chaîne nulle !\n");
    if ( NULL != erreur ) *erreur = TXT_DOUBLE_CHAINE_VIDE;
    return 0.0;
  }
  if ( NULL != erreur ) *erreur = 0.0; /* Nombre exact en général */

  /* On commence par sauter d'éventuels espaces initiaux */
  while ( isspace( *chaine ) )
    chaine++;
  if ( 0 == *chaine ) {		/* Chaîne vide : espaces uniquements... */
    if ( NULL != erreur ) *erreur = TXT_DOUBLE_CHAINE_VIDE;
    return 0.0;
  }

  nombre_initial = chaine; /* Pour les messages uniquement */
  /* Le nombre est-il négatif (ou, éventuellement, positif) ? */
  if ( '+' == *chaine ) /* Nombre de la forme +128 */
    chaine++;
  else if ( '-' == *chaine ) {
    negatif = VRAI;
    chaine++;
  } else if ( '\xE2' == *chaine ) {
    /* Signe moins codé en UTF-8 ?
       Remarque : si en latin-1, c'est la séquence â?? où ? représente un code inexistant => ne devrait pas arriver
                  si en TOS    ,                   ΓêÆ                                    => ne devrait pas arriver
		  si en windows,                   âˆ’                                    => ne devrait pas arriver
     */
    if ( ( '\x88' == *(chaine + 1) ) && ( '\x92' == *(chaine + 2) ) ) {
      negatif = VRAI;
      chaine += 3;
    }
  }

  if ( '(' == *chaine ) { /* Nombre de la forme [+|-](2,1±0,02)×10^-3, p. ex. */
    const char *nlle_fin;

    chaine++;
    valeur = TXT_CNV_vNombre_D( chaine, &nlle_fin, &val_erreur );
    if ( *nlle_fin != ')' ) {
      (void) fprintf(stderr, "TXT_CNV_vNombre_D : nombre à un format incorrect [%s] --- on garde %g (%f)\n",
		     nombre_initial, valeur, valeur);
      if ( fin_nombre != NULL ) *fin_nombre = chaine; /* Premier caractère non converti */
      return valeur;
    }
    chaine = nlle_fin + 1; /* On est après la parenthèse fermante */
  } else { /* Normalement, nombre classique */
    /* Conversion de la partie entière */
    while ( isdigit ( *chaine ) ) {
      p_entiere *= 10.0;
      p_entiere += (double) (*chaine - '0');
      chaine++;
    }

    /* Y a-t-il une partie décimale ? */
    if ( ( ',' == *chaine ) || ( '.' == *chaine ) ) {
      double exp = 1.0;
      chaine++;

      while ( isdigit ( *chaine ) ) {
	exp /= 10.0;
	p_10++;
	p_decimale += exp * (double) (*chaine - '0');
	chaine++;
      }
    }

    /* Y a-t-il une incertitude sur les derniers chiffres ? */
    if ( '(' == *chaine ) {
      double incertitude;
      const char *nlle_fin;

      chaine++;
      incertitude = TXT_CNV_vNombre_D( chaine, &nlle_fin, NULL );
      if ( *nlle_fin != ')' ) {
	(void) fprintf(stderr, "TXT_CNV_vNombre_D : incertitude à un format incorrect [%s] --- on garde %g (%f)\n",
		       nombre_initial, valeur, valeur);
	if ( erreur != NULL) *erreur = incertitude;
	if ( fin_nombre != NULL ) *fin_nombre = chaine; /* Premier caractère non converti */
	return valeur;
      }
      chaine = nlle_fin + 1;
      val_erreur = incertitude * pow( 10.0, (double) -p_10);
    }

    /* On assemble le tout */
    valeur = p_entiere + p_decimale;
  }
  if ( VRAI == negatif )
    valeur = -valeur;

  /* On a fini la « partie principale » du nombre : y a-t-il un exposant ? */
  if ( ( 'e' == *chaine ) || ( 'E' == *chaine ) ) {
    avec_exposant = VRAI;
    chaine++;
  } else if ( '^' == *chaine ) { /* Cas spécial : 10^xxx */
    if ( 10.0 == valeur ) {
      valeur = 1.0;		/* Puisque l'on fera * 10^exposant ensuite... */
      chaine++;
      avec_exposant = VRAI;
    }
  } else if ( ( '*'    == *chaine ) || ( 'x'    == *chaine ) || /* ASCII classique */
	      ( '\xD7' == *chaine ) || ( '\xB7' == *chaine ) ) { /* Latin-1/windows-1252 : × et · */
    if ( ( '1' == *(chaine + 1) ) && ( '0' == *(chaine + 2) ) ) {
      /* Repose sur l'évaluation paresseuse du C pour éviter un seg. fault si la chaîne s'arrête... */
      chaine += 3;
      avec_exposant = VRAI;
    }
  } else {			/* Le cas du codage UTF-8, au besoin */
    char code;
    unsigned int n_octets;

    code = TXT_UTF8_vers_operateur( chaine, &n_octets );
    if ( '*' == code ) {
      const char *c = chaine + n_octets;

      if ( ( '1' == *(c) ) && ( '0' == *(chaine + 1) ) ) {
	chaine += n_octets + 2;
	avec_exposant = VRAI;
      }
    }
  }

  /* Il y a bien un exposant */
  if ( VRAI == avec_exposant ) {
    const char *nlle_fin;
    long exposant;

    exposant = TXT_CNV_vNombre_L( chaine, &nlle_fin );
    chaine = nlle_fin;

    valeur *= pow(10.0, (double) exposant);
    val_erreur *= pow(10.0, (double) exposant);
  }

  /* Éventuellement, une incertitude globale */
  if ( ( '\xB1' == *chaine ) ||	/* ± en latin-1 / windows-1252, latin-9 */
       ( ( '\xC2' == *chaine ) && ( '\xB1' == *(chaine + 1 ) ) ) ) { /* ± en UTF-8 */
    const char *nlle_fin;

    chaine++;
    val_erreur = TXT_CNV_vNombre_D( chaine, &nlle_fin, NULL );
    chaine = nlle_fin;
  }

  /* On a fini */
  if ( erreur != NULL ) *erreur = val_erreur;
  if ( fin_nombre != NULL ) *fin_nombre = chaine; /* Premier caractère non converti */
  return valeur;
}

/* ~-~-~-~-~-~-~-~- Conversions d'un nombre en chaîne ~-~-~-~-~-~-~-~- */

/* -------- Conversion d'un chiffre à virgules, avec les conventions précisées -----------\
 |   Renvoie l'adresse de la chaîne contenant le résultat                                 |
 |                                                                                        |
 | valeur   : valeur à convertir
 | erreur   : erreur associée (< 0 : aucune, non indiquée ; sinon, même format...)        |
 | avant    : nombre maximal de chiffres avant la virgule. Si plus grand : notation 10^...|
 |            (0 : on s'arrange pour écrire tous les chiffres)                            |
 | apres    : nombre maximal de chiffres après la virgule                                 |
 |            (0 : précision maximale)
 | exposant : nombre maximal de chiffres après l'exposant                                 |
 | table    : si non-NULL, écrit le résultat dans la chaîne indiquée                      |
 | lg_resultat: si non-NULL, indique dedans le nombre de caractères dans la chaîne...     |
 \--------------------------------------------------------------------------------------- */

char *TXT_CNV_D_vTexte(double valeur, double erreur, unsigned int avant, unsigned int apres, unsigned int exposant,
		       char *resultat, unsigned int *lg_resultat)
{
  char *chaine, *debut_nombre;
  int val_exposant = 0, negatif = FAUX;

#ifdef AUTONOME
  printf( "%s : valeur = %g, erreur = %g, précision = %u:%u:%u, résultat : %p, longueur : %p\n", __FUNCTION__,
 	  valeur, erreur, avant, apres, exposant, (void *) resultat, (void *) lg_resultat);
#endif

  /* On contrôle les précisions */
  if ( 0 == avant ) {
    if ( NULL != resultat ) {
      (void) fprintf(stderr, "%s : attention, précision variable dans une chaîne externe [avant la virgule]\n", __FUNCTION__ );
      (void) fprintf(stderr, "                   aucun contrôle de débordement n'est possible !\n");
    }

    /* Précision = nombre de chiffres nécessaire pour tout afficher ! */
    if ( valeur != 0.0 ) {
      avant = (int) ( 1.0 + log10( fabs(valeur) ) );
    } else {
      avant = 1;
    }
  }
  if ( 0 == apres ) {
    if ( NULL != resultat ) {
      (void) fprintf(stderr, "%s : attention, précision variable dans une chaîne externe [après la virgule]\n", __FUNCTION__ );
      (void) fprintf(stderr, "                   aucun contrôle de débordement n'est possible !\n");
    }

    /* Précision = nombre de chiffres nécessaire pour tout afficher ! */
    apres = 20; /* Dépend en réalité de la machine, mais de toute façon, plus de 20 chiffres me semble sans intérêt ! */
  }

  /* Si nécessaire, on réserve la chaîne contenant le résultat */
  if ( NULL == resultat ) { /* On réserve la place pour la chaîne souhaitée */
    unsigned long lg_chaine;

    lg_chaine = avant + apres + exposant + 8; /* chiffres demandés, 0 final, virgule, *10^, - éventuels */
    if ( erreur > 0.0 ) {
      lg_chaine += avant + apres + exposant + 3; /* erreur et format (x±y)×10^exposant [mais si (x±10-3)×10² : sécurité...] */
#ifdef _utf8_
      lg_chaine += 2 * exposant + 1; /* Codage UTF-8 des exposants [2 ou 3/chiffre] et du ± [2 octets] */
#endif
    }
    lg_chaine++; /* +1 pour faciliter les arrondis */

    resultat = (char *) calloc( lg_chaine, 1 ); /* Chiffres, 0 final, virgule, *10^, - éventuels */
    if ( NULL == resultat ) {
      (void) fprintf( stderr, "%s,%s l.%u : impossible de réserver la chaîne résultat [%g ± %g] !\n",
		      __FILE__, __FUNCTION__, __LINE__, valeur, erreur);
      return NULL;
    }
  }

  /* On se place au début */
  chaine = resultat;

  /* Cas particulier : 0 */
  if ( 0.0 == valeur ) {
    *chaine++ = '0';
    *chaine = 0; /* 0 final... (sera effacé s'il y a une erreur) */
    if ( erreur > 0.0 ) { /* Il y a une erreur aussi */
#ifdef _utf8_
      *chaine++ = '\xC2'; *chaine++ = '\xB1'; /* Codage UTF-8 du ± */
#elif _latin1_
      *chaine++ = '\xD7';	/* Codage latin-1/windows-1252 du ± */
#elif _TOS_
      *chaine++ = '\xF1';	/* Codage TOS du ± */
#endif
      (void) TXT_CNV_D_vTexte( erreur, -1.0, avant, apres, exposant, chaine, lg_resultat );
#ifdef _utf8_
      if ( lg_resultat != NULL ) *lg_resultat += 3; /* 2 octets UTF-8 + 0 final */
#else
      if ( lg_resultat != NULL ) *lg_resultat += 2; /* 1 octet ± + 0 final */
#endif
    } else {
      if ( lg_resultat != NULL ) *lg_resultat = 1;
    }

    return resultat;
  }

  /* On se débarrasse du signe */
  if ( valeur < 0.0 ) {
    negatif = VRAI;
    valeur = -valeur;
  }

  /* Doit-on utiliser la notation scientifique ? */
  if ( ( valeur <  pow( 10.0, -(double) apres ) ) ||
       ( valeur >= pow( 10.0,  (double) avant ) ) ) { /* Notation scientifique */
#ifdef AUTONOME
    printf("%s : notation scientifique nécessaire... [%g = ", __FUNCTION__, valeur);
#endif

    val_exposant = (int) floor( log10(valeur) );
    valeur *= pow(10.0, - (double) val_exposant);
#ifdef AUTONOME
    printf("%g·10^%d\n", valeur, val_exposant);
#endif

    /* S'il y a une erreur, l'exposant est en facteur */
    if ( erreur > 0.0 ) {
      *chaine++ = '(';
      erreur *= pow( 10.0, - (double) val_exposant );
    }

    /* On donne tous les chiffres significatifs après la virgule ! */
    apres += avant - 1;
    avant = 1;
  }

  if ( VRAI == negatif ) {
    *chaine++ = '-';
  }
  debut_nombre = chaine;

  /* Recherche de la partie entière */
  if ( valeur < 1.0 ) { /* Nombre à virgule du type 0,... */
    *chaine++ = '0';
    *chaine = 0;

    /* On ne touche pas aux chiffres après la virgule, imposés par l'utilisateur */
  } else { /* Nombre plus petit que ce qui est permis : notation normale */
    unsigned int p_entiere;
    unsigned long nb_chiffres;

    /* Partie entière */
    p_entiere = (unsigned int) valeur;
    valeur -= (double) p_entiere;
    sprintf( chaine, "%u", p_entiere );
    nb_chiffres = TXT_nbr_car_UI( p_entiere );
    chaine += nb_chiffres;

#ifdef AUTONOME
    printf( "%s : partie entière = %u [%s / %s]\n", __FUNCTION__, p_entiere, resultat, chaine );
#endif

    /* On ne touche pas aux chiffres après la virgule, imposés par l'utilisateur */
  }

  /* La partie décimale, si nécessaire */
  if ( valeur > 0.0 ) {
    unsigned int i;

    *chaine++ = ',';
#ifdef AUTONOME
    printf( "%s : partie décimale %.20g... ", __FUNCTION__, valeur);
#endif
    for ( i = 0 ; i < apres ; i++ ) {
      unsigned int p_entiere;

      valeur *= 10.0;
      p_entiere = (unsigned int) valeur;
      valeur -= (double) p_entiere;
      if ( 10 == p_entiere ) { /* Peut arriver avec 0,3 = 0,2999999999999999889 => 0,2: sinon ! */
	(*(chaine - 1))++; /* On passe au chiffre suivant */
#ifdef AUTONOME
 	printf( "* Arrêt anticipé *" );
#endif
	break; /* Et on s'arrête */
      }
      *chaine++ = '0' + p_entiere;
#ifdef AUTONOME
      printf( "(%s)", resultat );
#endif
    }
    /* On arrondit au chiffre le plus proche [chiffre suivant >= 5 => chiffre supérieur] */
    valeur *= 10.0;
#ifdef AUTONOME
    printf( "/ch. suivant = %g, %u/", valeur, (unsigned int) valeur );
#endif
    if ( 5 <= (unsigned int) valeur ) {
      ( *(chaine - 1) )++;
      if ( ':' ==  *( (chaine - 1) ) ) { /* Il y avait un 9 avant */
	/* On remet un 9 et un en plus derrière => sera traité après */
	*(chaine - 1) = '9';
      }
    }

    *chaine = 0;
#ifdef AUTONOME
    printf(" faite [%s]\n", resultat);
#endif

    /* Si se termine par des 9, on voit s'il n'y a pas un problème d'arrondi... */
    if ( *(chaine - 1) == '9' ) {
      char *der_chiffre = chaine - 1;

      while ('9' == *der_chiffre) {
	der_chiffre--;
      }
      if ( ( (chaine - 1) - der_chiffre ) > 1 ) { /* Plus d'un seul 9 */
	char *pos_virgule = NULL;

	/* der_chiffre pointe sur le dernier chiffre autre que 9 : 12,45_8_999 */
	*(der_chiffre + 1) = '\0'; /* On fait sauter les 9 finaux */
	/* donc 12,45_8_ */
#ifdef AUTONOME
   	printf( "%s : 9 finaux après la virgule ôtés [%s]\n", __FUNCTION__, resultat);
#endif
	if ( ',' == *der_chiffre ) {
	  pos_virgule = der_chiffre;
/* 	  *(pos_virgule + 1) = '\0'; */
	  *der_chiffre-- = '\0'; /* On fait sauter ce qu'il y a après la virgule */
#ifdef AUTONOME
  	  printf("%s : , finale flottant ôtée [%s]\n", __FUNCTION__, resultat);
#endif
	}

	/* S'occuper des unités */
	if ( '9' == *der_chiffre ) { /* Nombre du genre 9,999 ou 19,999 ou 99,999... */
	  while ( '9' == *der_chiffre ) {
	    *der_chiffre = '0';
	    if ( debut_nombre == der_chiffre ) break;
	    der_chiffre--;
	  }
	  if ( debut_nombre == der_chiffre ) { /* Nombre du genre 9,99 ou 99,999 ou 999,999 => 10, 100, 1000
						  On a la place de la virgule pour décaler d'un cran */
	    while ( pos_virgule > der_chiffre ) {
	      *pos_virgule = *(pos_virgule - 1);
	      pos_virgule--;
	    }
	    debut_nombre[0] = '1';
	    if ( val_exposant != 0 ) { /* 999,999 -> 9,99×10^2 -> 1×10^3 */
	      val_exposant++;
	    }
	  } else { /* Nombre du genre 19,999 => 20 : on change le chiffre */
	    (*der_chiffre)++;
	  }
#ifdef AUTONOME
  	  printf("%s : 9 finaux avant la virgule ôtés [%s]\n", __FUNCTION__, resultat);
#endif
	} else { /* Nombre du genre 18,999 ou 18,15999 ou ... */
	  (*der_chiffre)++; /* 5 -> 6, 8 -> 9 avec les exemples précédents */
#ifdef AUTONOME
   	  printf("%s : dernier chiffre corrigé [%s]\n", __FUNCTION__, resultat);
#endif
	}
	chaine = der_chiffre + 1; /* Pour placer correctement la suite, si nécessaire ! */
	*chaine = 0;
      }
    } else if ( *(chaine - 1) == '0' ) { /* On ôte les 0 après la virgule, si nécessaire */
      while ( *(chaine - 1) == '0' ) {
	*(--chaine) = '\0';
      }
      if (*(chaine - 1) == ',') /* On ôte aussi la virgule */
	*(--chaine) = '\0';
    }
  }

  /* On s'occupe, si nécessaire, de l'erreur */
  if ( erreur > 0.0 ) {
    unsigned int lg_erreur;

    /* Le caractère ± */
#ifdef _utf8_
    *chaine++ = '\xC2'; *chaine++ = '\xB1'; /* Codage UTF-8 du ± */
#elif _latin1_
    *chaine++ = '\xD7';	/* Codage latin-1/windows-1252 du ± */
#elif _TOS_
    *chaine++ = '\xF1';	/* Codage TOS du ± */
#endif

    /* On écrit la valeur de l'erreur */
    (void) TXT_CNV_D_vTexte( erreur, -1.0, avant, apres, exposant, chaine, &lg_erreur );

    /* On se déplace d'autant */
    chaine += lg_erreur;
  }

  /* L'exposant, si nécessaire */
  if ( val_exposant != 0 ) {
    char *tmp;

    if ( erreur > 0.0 )
      *chaine++ = ')';
    sprintf( chaine, "×10^%d", val_exposant );
    chaine += 4;
    chaine += TXT_nbr_car_I( val_exposant );

    /* Contrôle : 1×10^n -> 10^n, forcément avec 1 comme 1er caractère, sinon (1±...)×10^ OK */
    tmp = resultat;
    if ( '-' == *tmp ) tmp++; /* Mais -1×10^n -> -10^n */

    if ( ( '1' == tmp[ 0 ] ) && ( ( '\xD7' == tmp[ 1 ] ) || /* × en latin-1/windows-1252 */
				  ( ( '\xC3' == tmp[ 1 ] ) && ( '\x97' == tmp[ 2 ] ) ) ) ) { /* × en UTF-8 */
      unsigned int i = 0, delta = 2;

      if ( '\xC3' == tmp[ 1 ] ) delta++; /* Code UTF-8 : occupe un octet de plus... */

      /* On décale tout pour faire sauter le 1× */
      do {
	tmp[ i ] = tmp[ i + delta ];
	i++;
      } while ( tmp[ i + 1 ] != 0 ) ;
      *chaine = 0;
      *(chaine - 1) = 0;
      chaine -= 2;
    }
  }

  if ( lg_resultat != NULL ) *lg_resultat = chaine - resultat;

#ifdef AUTONOME
  printf("%s : résultat (longueur : %u) = %s\n", __FUNCTION__, *lg_resultat, resultat);
#endif
  return resultat;
}

/* -------- Conversion d'un entier signé en chaîne -------- */

char *TXT_CNV_L_vTexte(long valeur, char *resultat, unsigned int *lg_resultat)
{
  unsigned long lg_nombre;

  lg_nombre = TXT_nbr_car_L( valeur ) + 1; /* N'oublions pas le 0 final ! */
  if ( NULL == resultat ) { /* On doit réserver la chaîne */
    resultat = (char *) malloc( lg_nombre );
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
  } else {
    if ( NULL == lg_resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : Longueur de chaîne passé inconnue\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
    if ( *lg_resultat < lg_nombre ) {
      (void) fprintf(stderr, "%s,%s l.%u : Chaîne passée trop courte [%u caractères, il en faudrait %lu]\n",
		     __FILE__, __FUNCTION__, __LINE__, *lg_resultat, lg_nombre);
      return NULL;
    }
  }

  /* Conversion proprement dite */
  (void) sprintf( resultat, "%ld", valeur );

  /* Terminé : on renvoie la bonne chaîne */
  return resultat;
}

/* -------- Conversion d'un entier non-signé en chaîne -------- */

char *TXT_CNV_UL_vTexte(unsigned long valeur, char *resultat, unsigned int *lg_resultat)
{
  unsigned long lg_nombre;

  lg_nombre = TXT_nbr_car_UL( valeur ) + 1; /* N'oublions pas le 0 final ! */
  if ( NULL == resultat ) { /* On doit réserver la chaîne */
    resultat = (char *) malloc( lg_nombre );
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
  } else {
    if ( NULL == lg_resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : Longueur de chaîne passé inconnue\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
    if ( *lg_resultat < lg_nombre ) {
      (void) fprintf(stderr, "%s,%s l.%u : Chaîne passée trop courte [%u caractères, il en faudrait %lu]\n",
		     __FILE__, __FUNCTION__, __LINE__, *lg_resultat, lg_nombre);
      return NULL;
    }
  }

  /* Conversion proprement dite */
  (void) sprintf( resultat, "%ld", valeur );

  /* Terminé : on renvoie la bonne chaîne */
  return resultat;
}
